indexing

	description: "Mixin routines for basic string/character manipulation."
	names: "string", "character", "text"
	patterns: "mixin"
	date: "$Date: 2006-06-23 10:34:45 -0400 (Fri, 23 Jun 2006) $"
	revision: "$Revision: 226 $"
	author: "Todd Plessel, US EPA/LM, plessel.todd@epa.gov"
	status: "unreviewed", "untested"

class

	BASIC_STRING

feature {NONE} -- Basic operations

	matches_word (word, words: STRING): BOOLEAN is
			-- Does `word' appear in the space-separated set of `words'?
		require
			words_exist: words /= Void
			words_count_at_least_three: words.count >= 3
			words_starts_with_space: words.item (1) = ' '
			words_ends_with_space: words.item (words.count) = ' '
			words_item_2_is_not_space: words.item (2) /= ' '
		local
			index: INTEGER
		do
			if word /= Void and then not word.is_empty then
				index := words.substring_index (word, 1)

				if index /= 0 and then
					words.item (index - 1) = ' ' and then
					words.item (index + word.count) = ' ' then
					Result := True
				end
			end
		end

	has_each_name (name_list, new_names: ARRAY [STRING]): BOOLEAN is
			-- Does `name_list' have each name in `new_names'?
		require
 			first_argument_not_empty:
				name_list /= Void and then
				name_list.count >= 1 and then
				name_list.item (1) /= Void and then
				name_list.item (1).count >= 1
  			second_argument_not_empty:
				new_names /= Void and then
				new_names.count >= 1 and then
				new_names.item (1) /= Void and then
				new_names.item (1).count >= 1
		local
			index: INTEGER
 		do
			from
    			index := 1
				Result := True
			invariant
				index_large_enough: index >= 1
				index_small_enough: index <= 1 + new_names.count
			variant
				index_increasing: 1 + new_names.count - index
			until
				Result = False or else index > new_names.count
			loop
				Result := name_list.has (new_names.item (index))
				index := index + 1
			end
		end

	skip_white_space (s: STRING; starting_index: INTEGER): INTEGER is
			-- Index of first non-white_space character in `s' or 0.
		require
			string_exists: s /= Void
			string_not_empty: not s.is_empty
			starting_index_within_bounds:
				starting_index >= 1 and starting_index <= s.count
		local
			index: INTEGER
		do
			from
				index := starting_index
			invariant
				index_large_enough: index >= starting_index
				index_small_enough: index <= 1 + s.count
			variant
				index_increasing: 1 + s.count - index
			until
				Result /= 0 or else index > s.count
			loop
				if not is_white_space (s.item (index)) then
					Result := index
				end

				index := index + 1
			end
		end

	delimited_trimmed_text (s: STRING; starting_index: INTEGER;
		delimiter: CHARACTER): PAIR [INTEGER, INTEGER] is
			-- First and last indices of white-space-trimmed,
			-- `delimiter'-terminated text in `s' after `starting_index' or
			-- [0, 0] if none.
		require
			string_exists: s /= Void
			string_not_empty: not s.is_empty
			starting_index_within_bounds:
				starting_index >= 1 and starting_index <= s.count
		do
			Result.put_first (skip_white_space (s, starting_index))

			if Result.first /= 0 then
				Result.put_second (s.index_of (delimiter, Result.first))

				from
				invariant
					result_second_large_enough: Result.second >= 0
					result_second_small_enough: Result.second <= Result.first
				variant
					result_second_decreasing: 1 + Result.second
				until
					Result.second = 0 or else
					not is_white_space (s.item (Result.second))
				loop
					Result.put_second (Result.second - 1)
				end

				if Result.second = 0 then
					Result.put_first (0)
				end
			end
		ensure
			result_valid:
				(Result.first = 0 and Result.second = 0) or
				(Result.first <= Result.second and
				Result.first >= starting_index and
				Result.second <= s.count and then
				(not is_white_space (s.item (Result.first)) and
				not is_white_space (s.item (Result.second)) and
				s.item (Result.first) /= delimiter and
				s.item (Result.second) /= delimiter))
		end

	end_word (s: STRING; starting_index: INTEGER): INTEGER is
			-- Index of last character of first space-delimited word in `s'
			-- after `starting_index' or 0 if none.
		require
			string_exists: s /= Void
			string_not_empty: not s.is_empty
			starting_index_within_bounds:
				starting_index >= 1 and starting_index <= s.count
		do
			Result := skip_white_space (s, starting_index)

			if Result /= 0 then

				from
				invariant
					result_large_enough: Result >= starting_index
					result_small_enough: Result <= 1 + s.count
				variant
					result_increasing: 1 + s.count - Result
				until
					Result > s.count or else is_white_space (s.item (Result))
				loop
					Result := Result + 1
				end

				Result := Result - 1
			end
		ensure
			result_valid:
				Result /= 0 implies
				Result >= starting_index and Result <= s.count and then
				not is_white_space (s.item (Result))
		end

	next_word (s: STRING; starting_index: INTEGER): INTEGER is
			-- Index of next space-delimited word in `s' after
			-- `starting_index' or 0 if none.
		require
			string_exists: s /= Void
			string_not_empty: not s.is_empty
			starting_index_within_bounds:
				starting_index >= 1 and starting_index <= s.count
		local
			index: INTEGER
		do
			from
				index := starting_index
			invariant
				index_large_enough: index >= starting_index
				index_small_enough: index <= 1 + s.count
			variant
				index_increasing: 1 + s.count - index
			until
				Result /= 0 or else index > s.count
			loop
				if is_white_space (s.item (index)) then
					Result := index
				end

				index := index + 1
			end

			if Result /= 0 then
				Result := skip_white_space (s, Result)
			end
		end

	nth_word (s: STRING; starting_index, word_number: INTEGER): INTEGER is
			-- Index of `word_number' space-delimited word in `s' after
			-- `starting_index' or 0 if none.
		require
			string_exists: s /= Void
			string_not_empty: not s.is_empty
			starting_index_within_bounds:
				starting_index >= 1 and starting_index <= s.count
			word_number_positive: word_number >= 1
		local
			the_word_count: INTEGER
		do
			from
				Result := skip_white_space (s, starting_index)
				the_word_count := 1
			invariant
				word_count_large_enough: the_word_count >= 0
				word_count_small_enough: the_word_count <= 1 + the_word_count
			variant
				word_count_increasing: 1 + word_number - the_word_count
			until
				the_word_count = word_number or else Result = 0
			loop
				Result := next_word (s,	Result)
				the_word_count := the_word_count + 1
			end
		end

	word_count (s: STRING; starting_index: INTEGER): INTEGER is
			-- How many white-space-delimited "words" appear in `s'
			-- beginning at index `starting_index'?
		require
			string_exists: s /= Void
			starting_index_within_bounds:
				starting_index >= 1 and starting_index <= s.count
		local
			index: INTEGER
		do
			from
				index := skip_white_space (s, starting_index)
			invariant
				index_large_enough: index >= 0
				index_small_enough: index <= 1 + s.count
			until
				index = 0
			loop
				Result := Result + 1
				index := next_word (s, index)
			end
		end

	all_printable (s: STRING): BOOLEAN is
			-- Is `s' composed only of printable characters?
		require
			string_exists: s /= Void
			string_not_empty: not s.is_empty
		local
			index: INTEGER
		do
			from
				index := 1
				Result := True
			invariant
				index_large_enough: index >= 1
				index_small_enough: index <= 1 + s.count
			variant
				index_increasing: 1 + s.count - index
			until
				not Result or else index > s.count
			loop
				Result := is_printable (s.item (index))
				index := index + 1
			end
		end

	is_printable (ch: CHARACTER): BOOLEAN is
			-- Is `ch' a printable character (i.e., not a control character)?
		do
			Result := isprint (ch) /= 0
		end

	is_alpha (ch: CHARACTER): BOOLEAN is
			-- Is `ch' in ['a'-'z', 'A'-'Z']?
		do
			Result := isalpha (ch) /= 0
		end

	is_alpha_numeric (ch: CHARACTER): BOOLEAN is
			-- Is `ch' in ['a'-'z', 'A'-'Z', 0-9]?
		do
			Result := isalnum (ch) /= 0
		end

	is_white_space (ch: CHARACTER): BOOLEAN is
			-- Is `ch' a blank, tab or return character?
		do
			Result := isspace (ch) /= 0
		end

	is_lower_case (ch: CHARACTER): BOOLEAN is
			-- Is `ch' lowercase?
		do
			Result := islower (ch) /= 0
		end

	is_upper_case (ch: CHARACTER): BOOLEAN is
			-- Is `ch' uppercase?
		do
			Result := isupper (ch) /= 0
		end

	is_digit (ch: CHARACTER): BOOLEAN is
			-- Is `ch' in [0-9]?
		do
			Result := isdigit (ch) /= 0
		end

	string_length (s: POINTER): INTEGER is
			-- Length of C string.
		require
			not_null: s /= default_pointer
		external
			"C (const char*): size_t | <string.h>"
		alias
			"strlen"
		end

feature {NONE} -- Implementation

	islower (ch: CHARACTER): INTEGER is
			-- Is `ch' a lowercase character ([a-z])?
		external
			"C [macro <ctype.h>]"
		end

	isupper (ch: CHARACTER): INTEGER is
			-- Is `ch' an uppercase character ([A-Z])?
		external
			"C [macro <ctype.h>]"
		end

	isprint (ch: CHARACTER): INTEGER is
			-- Is `ch' a printable character (i.e., not a control character)?
		external
			"C [macro <ctype.h>]"
		end

	isalpha (ch: CHARACTER): INTEGER is
			-- Is `ch' a printable character (i.e., not a control character)?
		external
			"C [macro <ctype.h>]"
		end

	isalnum (ch: CHARACTER): INTEGER is
			-- Is `ch' a printable character (i.e., not a control character)?
		external
			"C [macro <ctype.h>]"
		end

	isspace (ch: CHARACTER): INTEGER is
			-- Is `ch' a whitespace character (' ', '%T', '%N')?
		external
			"C [macro <ctype.h>]"
		end

	isdigit (ch: CHARACTER): INTEGER is
			-- Is `ch' a digit character [0-9]?
		external
			"C [macro <ctype.h>]"
		end

end -- class BASIC_STRING

