

indexing
	description: "Two-dimensional array with user-defined upper and lower bounds"
	patterns: "none"
	date: "$Date: 2006-12-06 13:14:08 -0500 (Wed, 06 Dec 2006) $"
	revision: "$Revision: 271 $"
	status: "unreviewed"




class 

	BOUNDED_ARRAY2 [G]
	
inherit
	
	ARRAY [G]
		rename
			make as array_make,
			item as array_item,
			put as array_put,
			force as array_force,
			resize as array_resize,
			clear_all as array_clear_all
		export
			{NONE}
				array_make, array_item, array_force,
				array_resize, array_clear_all
			{ARRAY2}
				array_put
			{ANY}
				copy, is_equal, area, to_c
		end

creation
	make, array_make, make_from_list, make_from_large_list

feature -- Initialization

	make (nb_row_lower, nb_row_upper, 
		  nb_column_lower, nb_column_upper: INTEGER) is
			-- Create a two dimensional array with valid row index 
			-- between [`nb_row_lower', `nb_row_upper'] and valid column index 
			-- between [`nb_column_lower', `nb_column_upper']
		require
			rows_ordered: nb_row_upper >= nb_row_lower
			columns_ordered: nb_column_upper >= nb_column_lower
			not_flat:  nb_row_upper - nb_row_lower >= 0
			not_thin: nb_column_upper - nb_column_lower >= 0
		do
			row_lower := nb_row_lower
			row_upper := nb_row_upper
			column_lower :=	 nb_column_lower
			column_upper := nb_column_upper
			
			array_make (1, (nb_row_upper - nb_row_lower + 1) *
						 (nb_column_upper - nb_column_lower + 1))
		ensure
			new_count: count = height * width
		end

	make_from_list (nu, nv: INTEGER; list: DS_LIST [G]) is
			-- Create a bounded array from `list', assuming u's 
			-- varying fastest
		require 
			nu_strictly_positive: nu >= 1
			nv_strictly_positive: nv >= 1
			list_not_void: list /= Void
			list_correct_size: list.count = nu * nv
		local
			i, j: INTEGER
		do
			make (0, nu - 1, 0, nv - 1)
			
			from
				list.start
			until
				list.after
			loop
				put (list.item_for_iteration, i, j)
				i := i + 1
				if i >= nu then
					i := 0
					j := j + 1
				end
				
				list.forth
			end
			
		end 

	make_from_large_list (nu, nv: INTEGER; list: LARGE_ARRAYED_LIST [G]) is
			-- Create a bounded array from `list', assuming u's 
			-- varying fastest
		require 
			nu_strictly_positive: nu >= 1
			nv_strictly_positive: nv >= 1
			list_not_void: list /= Void
			list_correct_size: list.count = nu * nv
		local
			i, j: INTEGER
		do
			make (0, nu - 1, 0, nv - 1)
			
			from
				list.start
			until
				list.after
			loop
				put (list.item_for_iteration, i, j)
				i := i + 1
				if i >= nu then
					i := 0
					j := j + 1
				end
				
				list.forth
			end
			
		end 

	initialize (v: G) is
			-- Make each entry have value `v'.
		local
			row, column: INTEGER
		do
			from
				row := 1
			until
				row > height
			loop
				from
					column := 1
				until
					column > width
				loop
					put (v, row, column)
					column := column + 1
				end
				row := row + 1
			end
		end

feature -- Access
	
	row_lower, row_upper, column_lower, column_upper: INTEGER
			-- The bounds of the rows and columns
	
	item (row, column: INTEGER): G is
			-- Entry at coordinates (`row', `column')
		require
			valid_row: (row_lower <= row) and (row <= row_upper)
			valid_column: (column_lower <= column) and (column <= column_upper)
		do
			Result := array_item ((row - row_lower) * width + column - column_lower + 1)
		end

feature -- Measurement

	height: INTEGER is
			-- Number of rows
		do
			Result := row_upper - row_lower + 1
		end

	width: INTEGER is
			-- Number of columns
		do
			Result := column_upper - column_lower + 1
		end

feature -- Comparison
feature -- Status report
feature -- Status setting
feature -- Cursor movement
feature -- Element change

	put (v: like item; row, column: INTEGER) is
			-- Assign item `v' at coordinates (`row', `column').
		require
			valid_row: (row_lower <= row) and (row <= row_upper)
			valid_column: (column_lower <= column) and (column <= column_upper)
		do
			array_put (v, (row - row_lower) * width + column - column_lower + 1)
		end

	force (v: like item; row, column: INTEGER) is
			-- Assign item `v' at coordinates (`row', `column').
			-- Resize if necessary.
		require
			row_large_enough: row_lower <= row
			column_large_enough: column_lower <= column
		do
			internal_resize (row_lower.min (row), row_upper.max (row),
							 column_lower.min (column), column_upper.max (column))
			put (v, row, column)
		end

feature -- Removal

	clear_all is
			-- Remove all items.
		do
			row_lower := 0
			row_upper := 0
			column_lower := 0
			column_upper := 0
			array_clear_all
		end

feature -- Resizing

	resize (nb_row_lower, nb_row_upper, 
		  nb_column_lower, nb_column_upper: INTEGER) is
			-- Rearrange array so that it can accommodate
			-- the new bounds without losing any previously
			-- entered items, nor changing their coordinates;
			-- do nothing if not possible.
		require
			rows_ordered: nb_row_upper >= nb_row_lower
			columns_ordered: nb_column_upper >= nb_column_lower
			not_flat:  nb_row_upper - nb_row_lower >= 0
			not_thin: nb_column_upper - nb_column_lower >= 0
		do
			internal_resize (nb_row_lower, nb_row_upper,
			nb_column_lower, nb_column_upper)
			
			row_lower := nb_row_lower
			row_upper := nb_row_upper
			column_lower :=	 nb_column_lower
			column_upper := nb_column_upper			
		end
	
feature -- Transformation
feature -- Conversion
feature -- Duplication
feature -- Basic operations
feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {NONE} -- Implementation
	
	internal_resize (nb_row_lower, nb_row_upper,
		nb_column_lower, nb_column_upper: INTEGER) is
			-- Rearrange array so that it can accommodate
			-- `nb_rows' rows and `nb_columns' columns,
			-- without losing any previously
			-- entered items, nor changing their coordinates;
			-- do nothing if not possible.
		require
			rows_ordered: nb_row_upper >= nb_row_lower
			columns_ordered: nb_column_upper >= nb_column_lower
			not_flat:  nb_row_upper - nb_row_lower >= 0
			not_thin: nb_column_upper - nb_column_lower >= 0
		local
			i, j: INTEGER
			new: like Current
		do
			create new.make (nb_row_lower, nb_row_upper,
							 nb_column_lower, nb_column_upper)
			from
				i := row_lower 
			invariant
				i_at_least_one: i >= row_lower
				i_small_enough: i <= row_lower + row_upper + 1
			variant
				i_increasing: row_lower + row_upper + 1 - i
			until
				i > row_upper
			loop
				from
					j := column_lower 
				invariant
					j_at_least_one: j >= column_lower
					j_small_enough: j <= column_lower + column_upper + 1
				variant
					j_increasing: column_lower + column_upper + 1  - j
				until
					j > column_upper
				loop
					new.put (item (i, j), i, j)
					j := j + 1
				end
				i := i + 1
			end
		end

invariant
	
	rows_ordered: row_upper >= row_lower
	columns_ordered: column_upper >= column_lower
	not_flat:  row_upper - row_lower >= 0
	not_thin: column_upper - column_lower >= 0

end -- BOUNDED_ARRAY2

