indexing

	description: "Sortable, resizable, trimmable lists implemented as arrays."
	names: "sort", "array", "list", "sequence"
	representation: "array"
	access: "index", "cursor", "membership"
	size: "fixed"
	contents: "generic"
	date: "$Date: 2006-08-10 11:12:59 -0400 (Thu, 10 Aug 2006) $"
	revision: "$Revision: 247 $"
	author: "Mark Bolstad, US EPA/LM, bolstad.mark@epa.gov"
	status: "reviewed", "tested"

class

	SORTABLE_ARRAYED_LIST [G -> COMPARABLE]

inherit

	DS_ARRAYED_LIST [G]
		rename
			sort as list_sort
		export
			{NONE} list_sort
		end

creation

	make

feature -- Initializing

	fill (value: G) is
			-- Fill the list with `value'
		local
			i: INTEGER
		do
			make (capacity)

			from
				i := 1
			invariant
				i_at_least_one: i >= 1
				i_small_enough: i <= 1 + capacity
			variant
				i_increasing: 1 + capacity - i
			until
				i > capacity
			loop
				force (value, i)
				i := i + 1
			end
		ensure
			default: true
		end

feature -- Resizing

	trim (new_count: INTEGER) is
			-- Reduce count to `new_count',
			-- effectively "removing" items outside of [1, `new_count'].
		require
			new_count_small_enough: new_count <= count
			new_count_large_enough: new_count >= 0
		do
			if new_count = 0 then
				wipe_out
			else
				count := new_count
				go_i_th (index.min (count + 1))
			end
		ensure
			count_set: count = new_count
			-- `item' (i) = old (`item' (i)) for i = 1 to `new_count'
		end

feature -- Transformation

	sort is
			-- Sort using Shellsort algorithm
			-- (insertion sort by diminishing increment).
			-- Sort is stable (unlike Quicksort) and performs
			-- not more than count ^ 1.5 comparisons (or even count ^ 1.25).
			-- Reference: Sedgewick, Robert, Algorithms, Second Edition,
			-- Addison-Wesley, 1988, page 108. ISBN 0-201-06673-4.
		local
			n, h, i, j, jh: INTEGER
			ajh, ith_value: G
			done: BOOLEAN
		do
			n := count

				-- Initialize h to value in sequence [..., 40, 13, 4, 1]
				-- that bounds n and results in about n ^ 1.25 comparisons.

			from
				h := 1
			until
				h > n
			loop
				h := 3 * h + 1
			end

			from
			until
				h = 1
			loop
				h := h // 3

				from
					i := h + 1
				until
					i > n
				loop
					ith_value := item (i)
					j := i
					jh := j - h
					ajh := item (jh)

					from
						done := False
					until
						done or else ajh = Void or else ajh <= ith_value
					loop
						replace (ajh, j)
						j := jh

						if j <= h then
							done := True
						else
							jh := j - h
							ajh := item (jh)
						end
					end

					replace (ith_value, j)
					i := i + 1
				end
			end
		end

	sort_with_auxillary (aux: DS_ARRAYED_LIST [ANY]) is
			-- Sort using Shellsort algorithm updating `aux' as we go
			-- (insertion sort by diminishing increment).
			-- Sort is stable (unlike Quicksort) and performs
			-- not more than count ^ 1.5 comparisons (or even count ^ 1.25).
			-- Reference: Sedgewick, Robert, Algorithms, Second Edition,
			-- Addison-Wesley, 1988, page 108. ISBN 0-201-06673-4.
		require
			same_size: count = aux.count
		local
			n, h, i, j, jh: INTEGER
			ajh, ith_value: G
			done: BOOLEAN
			bjh, aux_value: ANY
		do
			n := count

				-- Initialize h to value in sequence [..., 40, 13, 4, 1]
				-- that bounds n and results in about n ^ 1.25 comparisons.

			from
				h := 1
			until
				h > n
			loop
				h := 3 * h + 1
			end

			from
			until
				h = 1
			loop
				h := h // 3

				from
					i := h + 1
				until
					i > n
				loop
					ith_value := item (i)
					aux_value := aux.item (i)
					j := i
					jh := j - h
					ajh := item (jh)
					bjh := aux.item (jh)

					from
						done := False
					until
						done or else ajh = Void or else ajh <= ith_value
					loop
						replace (ajh, j)
						aux.replace (bjh, j)
						j := jh

						if j <= h then
							done := True
						else
							jh := j - h
							ajh := item (jh)
							bjh := aux.item (jh)
						end
					end

					replace (ith_value, j)
					aux.replace (aux_value, j)
					i := i + 1
				end
			end
		end

end -- class SORTABLE_ARRAYED_LIST
