

indexing
	description: "Colors represented as rgb triples plus alpha"
	names: "RGBA", "HSVA"
	representation: "fixed"
	access: "fixed"
	size: "fixed"
	contents: "DOUBLE"
	patterns: "none"



class

	RGBA_COLOR

inherit

	ANY
		undefine
			default_create, copy, out, is_equal
		end

	LINEAR_TUPLE_4D
		rename
			x as red, y as green, z as blue, w as alpha,
			make as linear_tuple_make, fill as linear_tuple_fill,
			zero as black, unit as white, is_zero as is_black,
			put as linear_tuple_put, scale as linear_tuple_scale,
			magnitude as luminance

		export
			{RGBA_COLOR}
			   infix "#", prefix "-",
			   prefix "+", linear_tuple_make, linear_tuple_fill,
			   two_norm, magnitude_squared, one_norm,
			   max_norm, put_x, put_y, put_z, normalize, all_finite,
			   linear_tuple_scale
		redefine
			default_create, summation
		end

creation
	default_create, make

feature -- Initialization

	default_create is
			-- Create color as white with alpha of 1 (i.e., visible)
		do
			make (1, 1, 1, 1)
		end

	make (initial_red, initial_green, initial_blue, initial_alpha: DOUBLE) is
			-- Initial cartesian component coordinates.
		require
			first_argument_in_range: initial_red >= 0.0
			second_argument_in_range: initial_green >= 0.0
			third_argument_in_range: initial_blue >= 0.0
			fourth_argument_in_range: initial_alpha >= 0.0 and
									  initial_alpha <= 1.0
		do
			linear_tuple_make (initial_red, initial_green,
						  initial_blue, initial_alpha)
		end

	make_from_hsva (hue, saturation, value, initial_alpha: DOUBLE) is
			-- Create the color from a HSV triple with alpha
		require
			hue_in_range: hue >= 0.0 and
						  hue <= 1.0
			saturation_in_range: saturation >= 0.0 and
								 saturation <= 1.0
			value_in_range: value >= 0.0 and
							value <= 1.0
			initial_alpha_in_range: initial_alpha >= 0.0 and
									initial_alpha <= 1.0
		local
			h, f, q, t, p: DOUBLE
			i: INTEGER
		do
			if hue = 1.0 then
				h := 0.0
			else
				h := hue * 6.0
			end

			i := real_64_truncated_to_integer (h)
			f := h - i
			p := value * (1 - saturation)
			q := value * (1 - saturation * f)
			t := value * (1 - saturation * (1 - f))
			inspect i
			when 0 then
				make (value, t, p, initial_alpha)
			when 1 then
				make (q, value, p, initial_alpha)
			when 2 then
				make (p, value, t, initial_alpha)
			when 3 then
				make (p, q, value, initial_alpha)
			when 4 then
				make (t, p, value, initial_alpha)
			when 5 then
				make (value, p, q, initial_alpha)
			end

		end

	fill (initial_value: DOUBLE) is
			-- Set each `item' to `initial_value'.
		require
			initial_value_in_range: initial_value >= 0.0 and initial_value <= 1.0
		do
			linear_tuple_fill (initial_value)
		end

feature -- Access
feature -- Measurement
feature -- Comparison
feature -- Status report
feature -- Status setting
feature -- Cursor movement
feature -- Element change

	put_red (new_red: DOUBLE) is
			-- Set the red component.
		require
			new_red_in_range: new_red >= 0.0
		do
			put_x (new_red)
		end

	put_green (new_green: DOUBLE) is
			-- Set the green component.
		require
			new_green_in_range: new_green >= 0.0
		do
			put_y (new_green)
		end

	put_blue (new_blue: DOUBLE) is
			-- Set the blue component.
		require
			new_blue_in_range: new_blue >= 0.0
		do
			put_z (new_blue)
		end

	put_alpha (new_alpha: DOUBLE) is
			-- Set the green component.
		require
			new_alpha_in_range: new_alpha >= 0.0
		do
			put_w (new_alpha)
		end

	put (value: DOUBLE; index: INTEGER) is
			-- Assign `value' to component at `index'.
		require
			index_in_range: index >= 1 and index <= 4
			value_in_range: value >= 0.0
			value_in_range: index = 4 implies value >= 0.0 and value <= 1.0
		do
			linear_tuple_put (value, index)
		end

feature -- Removal
feature -- Resizing
feature -- Transformation
feature -- Conversion
feature -- Duplication
feature -- Basic operations

	infix "+" (other: RGBA_COLOR): RGBA_COLOR is
			-- Add to colors; Clamping between 0 and 1
		require
			other_not_void: other /= Void
		do
			create Result.make (red + other.red,
								green + other.green,
								blue + other.blue,
								clamp_to_range (alpha + other.alpha, 0, 1))
		ensure
			Result_not_void: Result /= Void
			result_x_is_x_sum: about_equal (Result.red, red + other.red)
			result_y_is_y_sum: about_equal (Result.green, green + other.green)
			result_z_is_z_sum: about_equal (Result.blue, blue + other.blue)
			result_w_is_w_sum: about_equal (Result.alpha, clamp_to_range (alpha + other.alpha, 0, 1))
		end

	summation (other: like Current) is
			-- Sum `other' with Current
		local
			index: INTEGER
			left, right: DOUBLE
		do
			from
				index := 1
			invariant
				index_at_least_one: index >= 1
				index_small_enough: index <= 1 + count
			variant
				index_increasing: 1 + count - index
			until
				index > count - 1
			loop
				left := item (index)
				right := other.item (index)

				if left = -right then
						-- Avoid nan (e.g., -inf + inf).
					put (0.0, index)
				else
					put (left + right, index)
				end

				index := index + 1
			end

			put (clamp_to_range (alpha + other.alpha, 0, 1), count)
		end

	infix "*" (other: RGBA_COLOR): RGBA_COLOR is
			-- Add to colors; Clamping between 0 and 1
		require
			other_not_void: other /= Void
		do
			create Result.make (red * other.red,
								green * other.green,
								blue * other.blue,
								clamp_to_range (alpha * other.alpha, 0, 1))
		ensure
			Result_not_void: Result /= Void
			result_x_is_x_product: about_equal (Result.red, red * other.red)
			result_y_is_y_product: about_equal (Result.green, green * other.green)
			result_z_is_z_product: about_equal (Result.blue, blue * other.blue)
			result_w_is_w_product: about_equal (Result.alpha, clamp_to_range (alpha * other.alpha, 0, 1))
		end

	interpolate (other: RGBA_COLOR; parameter: DOUBLE): RGBA_COLOR is
			-- Interpolated color that lies `parameter' between
			--	`Current' and `other'.
		require
			other_not_void: other /= Void
			parameter_positive:	parameter >= 0
		local
			result_red, result_green, result_blue, result_alpha: DOUBLE
		do
			result_red := (other.red - red) * parameter + red
			result_green := (other.green - green) * parameter + green
			result_blue := (other.blue - blue) * parameter + blue
			result_alpha := (other.alpha - alpha) * parameter + alpha
			create Result.make (result_red, result_green, result_blue, result_alpha)
		ensure
			Result_not_void: Result /= Void
			alpha_in_range: Result.alpha >= 0 and Result.alpha <= 1
		end

	over (other: RGBA_COLOR): RGBA_COLOR is
			-- Composite two colors; `Current' "over" `other'
		require
			other_not_void: other /= Void
		local
			result_red, result_green, result_blue, result_alpha: DOUBLE
		do
			result_alpha := alpha + other.alpha - (alpha * other.alpha)

			result_red := alpha * red + other.alpha * other.red - (alpha * other.alpha * red)
			result_green := alpha * green + other.alpha * other.green - (alpha * other.alpha * green)
			result_blue := alpha * blue + other.alpha * other.blue - (alpha * other.alpha * blue)

			create Result.make (result_red / result_alpha, result_green / result_alpha,
								result_blue / result_alpha, result_alpha)
		ensure
			Result_not_void: Result /= Void
		end

	scale (scalar: DOUBLE): RGBA_COLOR is
			-- Scale `Current' by `scalar'
		require
			scalar_is_positive: scalar >= 0.0
		do
--			Result.make (clamp_to_range (red * scalar, 0.0, 1.0),
--						 clamp_to_range (green * scalar, 0.0, 1.0),
--						 clamp_to_range (blue * scalar, 0.0, 1.0),
--						 alpha)
			create Result.make (red * scalar, green * scalar, blue * scalar, alpha)
		end

feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {NONE} -- Implementation

	Opening_out_delimiter: CHARACTER is '('
			-- First character of result of `out'.

	Closing_out_delimiter: CHARACTER is ')'
			-- Second character of result of `out'.

invariant

	red_component_in_range: red >= 0.0
	green_component_in_range: green >= 0.0
	blue_component_in_range: blue >= 0.0
	alpha_component_in_range: alpha >= 0.0 and alpha <= 1.0

end -- RGB_COLOR

