

indexing
	description: "Class for storing pixel-based image data"
	patterns: "none"
	date: "$Date: 2006-12-08 13:32:17 -0500 (Fri, 08 Dec 2006) $"
	revision: "$Revision: 273 $"
	status: "unreviewed"




class

	IMAGE

inherit

	BASIC_NUMERIC

	NUMERIC_CONVERSION

creation
	make

feature -- Initialization

	make (nb_rows, nb_columns, nb_depth: INTEGER) is
			-- Create a two dimensional array which has `nb_rows'
			-- rows and `nb_columns' columns, with `nb_depth' pixel depth
			-- with lower bounds starting at 1.
		require
			not_flat: nb_rows > 0
			not_thin: nb_columns > 0
			not_shallow: nb_depth > 0
		do
			height := nb_rows
			width := nb_columns
			pixel_depth := nb_depth

			origin_is_upper_left := true
			image_maximum := 1
			normalized := false

			create tmp_color

			create image_data.make (1, pixel_depth * height * width)
			--image_data.fill
			create weights.make (1, height * width)
			--weights.fill
		ensure
			new_count: count * pixel_depth = image_data.count
		end

feature -- Access

	image_maximum: DOUBLE
		-- The largest value in the unnormalized image

	item (row, column: INTEGER): RGBA_COLOR is
			-- Entry at coordinates (`row', `column')
		require
			valid_row: (0 <= row) and (row < height)
			valid_column: (0 <= column) and (column < width)
		local
			i, a_index: INTEGER
		do
			from
				i := 1
				a_index := array_index (row, column)
			invariant
				i_at_least_one: i >= 1
				i_small_enough: i <= 1 + pixel_depth
			variant
				i_increasing: 1 + pixel_depth - i
			until
				i > pixel_depth
			loop
				tmp_color.put (image_data.item (a_index + (i - 1)), i)

				i := i + 1
			end

			Result := tmp_color
		end

	normalized: BOOLEAN
			-- Has the image data been normalized?

	normalize is
			-- `normalize' the image data such that the color values
			-- are between 0 and 1
		local
			i: INTEGER
		do
			normalized := true
			if image_maximum > 1 then
				from
					i := 1
				invariant
					i_at_least_one: i >= 1
					i_small_enough: i <= 1 + image_data.count
				variant
					i_increasing: 1 + image_data.count - i
				until
					i > image_data.count
				loop
					if pixel_depth = 4 then
						if i \\ 4 /= 0 then
							image_data.put (image_data.item (i) / real_64_truncated_to_real_32 (image_maximum), i)
						end
					else
						image_data.put (image_data.item (i) / real_64_truncated_to_real_32 (image_maximum), i)
					end
					i := i + 1
				end
			end
		end

feature -- Measurement

	height: INTEGER
			-- Number of rows

	width: INTEGER
			-- Number of columns

	pixel_depth: INTEGER
			-- The component depth of the pixels, i.e., Z is 1, RGB
			-- is 3, and RGBA is 4

	count, capacity: INTEGER_64 is
			-- Number of available indices
		do
			Result := image_data.count // pixel_depth
		ensure then
			consistent_with_bounds: Result = image_data.count // pixel_depth
		end

feature -- Comparison
feature -- Status report

	origin_is_upper_left: BOOLEAN
			-- The origin of the image is in the upper-left corner

feature -- Status setting

	set_origin (is_upper_left: BOOLEAN) is
			-- Set where the origin of the image is. Effects `put'
			-- and `item'
		do
			origin_is_upper_left := is_upper_left
		ensure
			origin_set: origin_is_upper_left = is_upper_left
		end

	checkerboard (size: INTEGER) is
			-- change the image to be a black and white checkerboard
			-- of `size'
		require
			size_positive: size >= 1
			size_less_than_width: size <= width
			size_less_than_height: size <= height
		local
			i, j, k, index: INTEGER
			x_mod, y_mod: DOUBLE
			val: DOUBLE
			white, green: RGBA_COLOR
		do
			weights.discard_items
			image_maximum := 1

			create white.make (0.5, 0.5, 0.5, 1)
			create green.make (0.5, 1, 0.5, 1)

			from
				i := 0
			invariant
				i_at_least_one: i >= 0
				i_small_enough: i <= width
			variant
				i_increasing: width - i
			until
				i >= width
			loop
				x_mod := (i \\ (2 * size)) / (2 * size)

				from
					j := 0
				invariant
					j_at_least_one: j >= 0
					j_small_enough: j <= height
				variant
					j_increasing: height - j
				until
					j >= height
				loop
					y_mod := (j \\ (2 * size)) / (2 * size)

					if x_mod < 0.5 then
						if y_mod < 0.5 then
							val := 0
						else
							val := 1
						end
					else
						if y_mod < 0.5 then
							val := 1
						else
							val := 0
						end
					end

					from
						k := 1
					invariant
						k_at_least_one: k >= 1
						k_small_enough: k <= 1 + pixel_depth
					variant
						k_increasing: 1 + pixel_depth - k
					until
						k >= pixel_depth
					loop
						index := array_index (j, i) + (k - 1)
						if val = 0 then
							image_data.put (real_64_to_real_32 (green.item (k)), index)
						else
							image_data.put (real_64_to_real_32 (white.item (k)), index)
						end

						k := k + 1
					end

					j := j + 1
				end

				i := i + 1
			end
		ensure
			default: true
		end

feature -- Cursor movement
feature -- Element change

	put (v: like item; weight: DOUBLE; row, column: INTEGER) is
			-- Assign item `v' with `weight' at coordinates (`row', `column').
		require
			valid_row: 0 <= row and row < height
			valid_column: 0 <= column and column < width
			v_not_void: v /= Void
		local
			i, a_index, index, weight_index: INTEGER
			val, v_item: DOUBLE
			empty: BOOLEAN
			weight_over_n1, weight_n_over_n1,  weight_n, weight_n1: DOUBLE
		do
			normalized := false
			weight_index := row * width + column + 1 -- Offset tp one-based

			weight_n := weights.item (weight_index)
			empty := about_equal (weight_n, 0)

			weight_n1 := weight_n + weight

			weights.put (real_64_to_real_32 (weight_n1), weight_index)

			-- Precompute some values
			a_index := array_index (row, column)
			weight_over_n1 := weight / weight_n1
			weight_n_over_n1 := weight_n / weight_n1

			if empty then
				from
					i := 1
					index := a_index
				invariant
					i_at_least_one: i >= 1
					i_small_enough: i <= 1 + pixel_depth
				variant
					i_increasing: 1 + pixel_depth - i
				until
					i > pixel_depth
				loop
					v_item := v.item (i)

					val := weight_over_n1 * v_item

					image_data.put (real_64_to_real_32 (val), index)

					if v_item > image_maximum then
						image_maximum := v_item
					end

					i := i + 1
					index := index + 1
				end
			else
				from
					i := 1
					index := a_index
				invariant
					i_at_least_one: i >= 1
					i_small_enough: i <= 1 + pixel_depth
				variant
					i_increasing: 1 + pixel_depth - i
				until
					i > pixel_depth
				loop
					v_item := v.item (i)

					val := image_data.item (index) * weight_n_over_n1 +
							weight_over_n1 * v_item

					image_data.put (real_64_to_real_32 (val), index)

					if v_item > image_maximum then
						image_maximum := v_item
					end

					i := i + 1
					index := index + 1
				end
			end
		end

	force (v: like item; weight: DOUBLE; row, column: INTEGER) is
			-- Assign item `v' at coordinates (`row', `column').
			-- Resize if necessary.
		require
			row_large_enough: 0 <= row
			column_large_enough: 0 <= column
			v_not_void: v /= Void
		do
			resize (row, column)
			put (v, weight, row, column)
		end

	reset is
			-- Reset the image to all black
		do
			image_data.discard_items
			weights.discard_items
		end

feature -- Removal

	wipe_out is
			-- Remove all items.
		do
			height := 0
			width := 0
			normalized := false
			image_maximum := 0
			image_data.discard_items
			weights.discard_items
		end

feature -- Resizing

	resize (nb_rows, nb_columns: INTEGER) is
			-- Rearrange array so that it can accommodate
			-- `nb_rows' rows and `nb_columns' columns,
			-- without losing any previously
			-- entered items, nor changing their coordinates;
			-- do nothing if not possible.
		require
			valid_row: nb_rows >= 1
			valid_column: nb_columns >= 1
		do
			if nb_rows > height then
				height := nb_rows
			end
			if nb_columns > width then
				width := nb_columns
			end

			image_data.conservative_resize (1, pixel_depth * height * width)
			weights.conservative_resize (1, height * width)
		end

feature -- Transformation
feature -- Conversion

	to_c: POINTER is
			-- Address of actual sequence of values,
			-- for passing to external (non-Eiffel) routines.
		local
			a: ANY
		do
			a := image_data.to_c
			Result := $a
		end

feature -- Duplication
feature -- Basic operations
feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {IMAGE_IO} -- Implementation

	image_data: ARRAY [REAL]
			-- Pixel values

	swap_index (i: INTEGER): INTEGER is
			-- Takes an index `i' from data that assumes row 1 is in
			-- the upper-left to an index with row 1 in the lower left
		require
			valid_i: i >= 1 and i <= image_data.count
		local
			new_i, row: INTEGER
		do
			row := (i - 1) // (width * pixel_depth)
			new_i := i - row * width * pixel_depth

			Result := pixel_depth * ((height - row - 1) * width) + new_i
			debug ("image")
				io.put_integer (i)
				io.put_string (" - ")
				io.put_integer (row)
				io.put_string (", ")
				io.put_integer (new_i)
				io.put_string (", ")
				io.put_integer (Result)
				io.put_string ("%N")
			end
		ensure
			Result_in_range: Result >= 1 and Result <= image_data.count
		end

feature {NONE} -- Implementation

	tmp_color: RGBA_COLOR
			-- Temporary used to avoid repeated memory allocation

	weights: ARRAY [REAL]
			-- Filter weights

	array_index (row, column: INTEGER): INTEGER is
			-- Function to compute offset into `image_data'
		do
			Result := pixel_depth * (row * width + column) + 1

			if origin_is_upper_left then
				Result := swap_index (Result)
			end
		ensure
			Result_in_range: Result >= 1 and Result <= image_data.count
		end

invariant
	tmp_color_not_void: tmp_color /= Void
	image_data_not_void: image_data /= Void
	weights_not_void: weights /= Void
	tmp_color_consistent: tmp_color.count = pixel_depth
	height_positive: height >= 0
	width_positive: width >= 0
	depth_positive: pixel_depth >= 0

end -- IMAGE

