

indexing
	description: "Manages a potentially large hunk of memory"
	patterns: "none"
	status: "unreviewed"
	Date: "$Date: 2007-01-25 13:00:51 -0500 (Thu, 25 Jan 2007) $"
	Revision: "$Revision: 280 $"


class
		
	LARGE_MEMORY_POINTER

inherit
	
	MEMORY
		redefine
			is_equal, copy, dispose, default_create
		end

	PLATFORM
		redefine
			is_equal, copy, default_create
		end

	BIT_SIZES
		undefine
			is_equal, copy, default_create
		end
	
	SINGLETON_ACCESSOR
		rename
			singleton as object_id
		undefine
			is_equal, copy, default_create
		redefine
			object_id
		end

create
	default_create, make_from_pointer, share_from_pointer
	
feature -- Initialization

	default_create is
			-- Create an intial object
			do
				count := 0
				object_id.next_id
				identifier := object_id.last_id
			end
		
	make_from_pointer (a_ptr: POINTER; n: INTEGER_64) is
			-- Copy `a_count' bytes from `a_ptr' into current.
		require
			a_ptr_not_null: a_ptr /= default_pointer
			n_non_negative: n >= 0
		do
			default_create
			
			if item /= default_pointer then
				if n > count then
					resize (n)
				end
			else
				item := allocate_external (n.max (1), True)
			end
			if item = default_pointer then
				(create {EXCEPTIONS}).raise ("No more memory")
			end
			memcpy_external (item, a_ptr, n)
			count := n.max (1)
		ensure
			item_set: item /= default_pointer
			count_set: count = n.max (1)
			is_shared_set: not is_shared
		end

	share_from_pointer (a_ptr: POINTER; n: INTEGER_64) is
			-- Use directly `a_ptr' with count `n' to hold current data.
		require
			a_ptr_not_null: a_ptr /= default_pointer
			n_non_negative: n >= 0
		do
			default_create

			item := a_ptr
			count := n
			is_shared := True
		ensure
			item_set: item /= default_pointer
			count_set: count = n
			is_shared_set: is_shared
		end
	
feature -- Access

	identifier: INTEGER
			-- A guaranteed unique value to identify this instance of
			-- an action
	
	item: POINTER
			-- Access to allocated memory.
			
	count: INTEGER_64
			-- Number of elements that Current can hold.

	is_shared: BOOLEAN
			-- Is `item' shared with another memory area?

	to_c: ANY is
			-- Generate an `ANY' wrapper to `item'
		do

			Result := item




		end
	
feature -- Comparison

	is_equal (other: like Current): BOOLEAN is
			-- Is `other' attached to an object considered equal to current object?
		do
			Result := count = other.count and then memcmp_external (item, other.item, count) = 0
		end

feature -- Duplication

	copy (other: like Current) is
			-- Update current object using fields of object attached
			-- to `other', so as to yield equal objects. If `is_shared'
			-- and current is not large enough to hold `other' create
			-- a new pointer area and `is_shared' is set to `False'.
		do
			if count >= other.count then
					-- No need to reallocate, it is safe to just copy.
				memcpy_external (item, other.item, other.count)
			else
					-- We need to reallocate memory here
				if is_shared then
						-- Before `item' was shared, so we simply allocate
						-- a new memory area from `other' and reset
						-- the `is_shared' flag.
					is_shared := False
					make_from_pointer (other.item, other.count)
				else
						-- Simply resize Current and copy data.
					resize (other.count)
					memcpy_external (item, other.item, other.count)
				end
			end
		ensure then
			sharing_status_not_preserved:
				(old is_shared and not is_shared) implies (other.count > old count)
		end
		
feature -- Access: Platform specific

	read_integer_8 (pos: INTEGER_64): INTEGER_8 is
			-- Read INTEGER_8 at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer8_bits // 8) <= count
		do
			read_data_external ($temp_integer_8, item, pos, integer8_bits // 8)
			Result := temp_integer_8
		end

	read_integer_16 (pos: INTEGER_64): INTEGER_16 is
			-- Read INTEGER_16 at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer16_bits // 8) <= count
		do
			read_data_external ($temp_integer_16, item, pos, integer16_bits // 8)
			Result := temp_integer_16
		end

	read_integer_32 (pos: INTEGER_64): INTEGER is
			-- Read INTEGER at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer32_bits // 8) <= count
		do
			read_data_external ($temp_integer_32, item, pos, integer32_bits // 8)
			Result := temp_integer_32
		end		

	read_integer_64 (pos: INTEGER_64): INTEGER_64 is
			-- Read INTEGER_64 at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer64_bits // 8) <= count
		do
			read_data_external ($temp_integer_64, item, pos, integer64_bits // 8)
			Result := temp_integer_64
		end

	read_boolean (pos: INTEGER_64): BOOLEAN is
			-- Read BOOLEAN at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + Boolean_bits // 8) <= count
		do
			read_data_external ($temp_boolean, item, pos, Boolean_bits // 8)
			Result := temp_boolean
		end

	read_character (pos: INTEGER_64): CHARACTER is
			-- Read CHARACTER at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + Character_bits // 8) <= count
		do
			read_data_external ($temp_character, item, pos, Character_bits // 8)
			Result := temp_character
		end		


	read_real (pos: INTEGER_64): REAL is




			-- Read REAL at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + real32_bits // 8) <= count
		do
			read_data_external ($temp_real, item, pos, real32_bits // 8)
			Result := temp_real
		end

	read_double (pos: INTEGER_64): DOUBLE is
			-- Read DOUBLE at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + real64_bits // 8) <= count
		do
			read_data_external ($temp_double, item, pos, real64_bits // 8)
			Result := temp_double
		end

	read_doubles (pos, num_items: INTEGER_64; vals: ARRAY [DOUBLE]) is
			-- Read DOUBLE at position `pos'.
		require
			pos_nonnegative: pos >= 0
			num_items_positive: num_items >0
			vals_not_void: vals /= Void
			vals_count_is_at_least_num_items: vals.count >= num_items
			valid_position: (pos + real64_bits // 8) <= count
		local
			a: ANY
		do
			a := vals.to_c
			read_data_external ($a, item, pos, num_items * real64_bits // 8)
		end

feature -- Element change: Platform specific

	put_integer_8 (i: INTEGER_8; pos: INTEGER_64) is
			-- Insert `i' at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer8_bits // 8) <= count
		do
			temp_integer_8 := i
			put_data_external($temp_integer_8, item, pos, integer8_bits // 8)
		ensure
			inserted: i = read_integer_8 (pos)
		end

	put_integer_16 (i: INTEGER_16; pos: INTEGER_64) is
			-- Insert `i' at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer16_bits // 8) <= count
		do
			temp_integer_16 := i
			put_data_external($temp_integer_16, item, pos, integer16_bits // 8)
		ensure
			inserted: i = read_integer_16 (pos)
		end

	put_integer_32 (i: INTEGER; pos: INTEGER_64) is
			-- Insert `i' at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer32_bits // 8) <= count
		do
			temp_integer_32 := i
			put_data_external($temp_integer_32, item, pos, integer32_bits // 8)
		ensure
			inserted: i = read_integer_32 (pos)
		end

	put_integer_64 (i: INTEGER_64; pos: INTEGER_64) is
			-- Insert `i' at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer64_bits // 8) <= count
		do
			temp_integer_64 := i
			put_data_external($temp_integer_64, item, pos, integer64_bits // 8)
		ensure
			inserted: i = read_integer_64 (pos)
		end

	put_boolean (b: BOOLEAN; pos: INTEGER_64) is
			-- Insert `b' at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + Boolean_bits // 8) <= count
		do
			temp_boolean := b
			put_data_external($temp_boolean, item, pos, Boolean_bits // 8)
		ensure
			inserted: b = read_boolean (pos)
		end	

	put_character (c: CHARACTER; pos: INTEGER_64) is
			-- Insert `' at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + Character_bits // 8) <= count
		do
			temp_character := c
			put_data_external($temp_character, item, pos, Character_bits // 8)
		ensure
			inserted: c = read_character (pos)
		end			


	put_real (r: REAL; pos: INTEGER_64) is




			-- Insert `r' at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + real32_bits // 8) <= count
		do
			temp_real := r
			put_data_external($temp_real, item, pos, real32_bits // 8)
		ensure
			inserted: r = read_real (pos)
		end

	put_double (d: DOUBLE; pos: INTEGER_64) is
			-- Insert `d' at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + real64_bits // 8) <= count
		do
			temp_double := d
			put_data_external($temp_double, item, pos, real64_bits // 8)
		ensure
			inserted: d = read_double (pos)
		end

feature -- Access: Little-endian format

	read_integer_8_le (pos: INTEGER_64): INTEGER_8 is
			-- Read INTEGER_8 at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer8_bits // 8) <= count
		do
			Result := read_integer_8 (pos)
		end

	read_integer_16_le (pos: INTEGER_64): INTEGER_16 is
			-- Read INTEGER_16 at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer16_bits // 8) <= count
		local
			l_high, l_low: INTEGER_16
		do

			l_low := (0x00FF).to_integer_16 & read_integer_8 (pos)




			l_high := read_integer_8 (pos + integer8_bits // 8)

			Result := (l_high.to_integer_16 |<< 8) | l_low




		end

	read_integer_32_le (pos: INTEGER_64): INTEGER is
			-- Read INTEGER at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer32_bits // 8) <= count
		local
			l_high, l_low: INTEGER
		do
			l_low := 0x0000FFFF & read_integer_16_le (pos)
			l_high := read_integer_16_le (pos + integer16_bits // 8)

			Result := (l_high.to_integer_32 |<< 16) | l_low




		end		

	read_integer_64_le (pos: INTEGER_64): INTEGER_64 is
			-- Read INTEGER_64 at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer64_bits // 8) <= count
		local
			l_high, l_low: INTEGER_64
		do
			l_low := 0x00000000FFFFFFFF & read_integer_32_le (pos)
			l_high := read_integer_32_le (pos + integer32_bits // 8)

			Result := (l_high.to_integer_64 |<< 32) | l_low




		end

feature -- Element change: Little-endian format

	put_integer_8_le (i: INTEGER_8; pos: INTEGER_64) is
			-- Insert `i' at position `pos' in big-endian format.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer8_bits // 8) <= count
		do
			put_integer_8 (i, pos)
		ensure
			inserted: i = read_integer_8_le (pos)
		end
		
	put_integer_16_le (i: INTEGER_16; pos: INTEGER_64) is
			-- Insert `i' at position `pos' in big-endian format.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer16_bits // 8) <= count
		do
			put_integer_8 ((i & 0x00FF).to_integer_8, pos)
			put_integer_8 ((((i & 0xFF00) |>> 8) & 0x00FF).to_integer_8, pos + integer8_bits // 8)
		ensure
			inserted: i = read_integer_16_le (pos)
		end

	put_integer_32_le (i: INTEGER; pos: INTEGER_64) is
			-- Insert `i' at position `pos' in big-endian format.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer32_bits // 8) <= count
		do
			put_integer_16_le ((i & 0x0000FFFF).to_integer_16, pos)
			put_integer_16_le ((((i & 0xFFFF0000) |>> 16) & 0x0000FFFF).to_integer_16, pos + integer16_bits // 8)
		ensure
			inserted: i = read_integer_32_le (pos)
		end

	put_integer_64_le (i: INTEGER_64; pos: INTEGER_64) is
			-- Insert `i' at position `pos' in big-endian format.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer64_bits // 8) <= count
		do
			put_integer_32_le ((i & 0x00000000FFFFFFFF).to_integer_32, pos)
			put_integer_32_le (
				(((i & 0xFFFFFFFF00000000) |>> 32) & 0x00000000FFFFFFFF).to_integer_32,
				pos+ integer32_bits // 8)
		ensure
			inserted: i = read_integer_64_le (pos)
		end

feature -- Access: Big-endian format

	read_integer_8_be (pos: INTEGER_64): INTEGER_8 is
			-- Read INTEGER_8 at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer8_bits // 8) <= count
		do
			Result := read_integer_8 (pos)
		end

	read_integer_16_be (pos: INTEGER_64): INTEGER_16 is
			-- Read INTEGER_16 at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer16_bits // 8) <= count
		local
			l_high, l_low: INTEGER_16
		do
			l_high := read_integer_8 (pos)

			l_low := (0x00FF).to_integer_16 & read_integer_8 (pos + integer8_bits // 8)
			Result := (l_high.to_integer_16 |<< 8) | l_low





		end

	read_integer_32_be (pos: INTEGER_64): INTEGER is
			-- Read INTEGER at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer32_bits // 8) <= count
		local
			l_high, l_low: INTEGER
		do
			l_high := read_integer_16_be (pos)
			l_low := 0x0000FFFF & read_integer_16_be (pos + integer16_bits // 8)

			Result := (l_high.to_integer_32 |<< 16) | l_low




		end		

	read_integer_64_be (pos: INTEGER_64): INTEGER_64 is
			-- Read INTEGER_64 at position `pos'.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer64_bits // 8) <= count
		local
			l_high, l_low: INTEGER_64
		do
			l_high := read_integer_32_be (pos)
			l_low := 0x00000000FFFFFFFF & read_integer_32_be (pos + integer32_bits // 8)

			Result := (l_high.to_integer_64 |<< 32) | l_low




		end

feature -- Element change: Big-endian format

	put_integer_8_be (i: INTEGER_8; pos: INTEGER_64) is
			-- Insert `i' at position `pos' in big-endian format.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer8_bits // 8) <= count
		do
			put_integer_8 (i, pos)
		ensure
			inserted: i = read_integer_8_be (pos)
		end
		
	put_integer_16_be (i: INTEGER_16; pos: INTEGER_64) is
			-- Insert `i' at position `pos' in big-endian format.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer16_bits // 8) <= count
		do
			put_integer_8 ((((i & 0xFF00) |>> 8) & 0x00FF).to_integer_8, pos)
			put_integer_8 ((i & 0x00FF).to_integer_8, pos + integer8_bits // 8)
		ensure
			inserted: i = read_integer_16_be (pos)
		end

	put_integer_32_be (i: INTEGER; pos: INTEGER_64) is
			-- Insert `i' at position `pos' in big-endian format.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer32_bits // 8) <= count
		do
			put_integer_16_be ((((i & 0xFFFF0000) |>> 16) & 0x0000FFFF).to_integer_16, pos)
			put_integer_16_be ((i & 0x0000FFFF).to_integer_16, pos + integer16_bits // 8)
		ensure
			inserted: i = read_integer_32_be (pos)
		end

	put_integer_64_be (i: INTEGER_64; pos: INTEGER_64) is
			-- Insert `i' at position `pos' in big-endian format.
		require
			pos_nonnegative: pos >= 0
			valid_position: (pos + integer64_bits // 8) <= count
		do
			put_integer_32_be (
				(((i & 0xFFFFFFFF00000000) |>> 32) & 0x00000000FFFFFFFF).to_integer_32, pos)
			put_integer_32_be ((i & 0x00000000FFFFFFFF).to_integer_32,
				pos + integer32_bits // 8)
		ensure
			inserted: i = read_integer_64_be (pos)
		end

feature -- Concatenation

	append (other: like Current) is
			-- Append `other' at the end of Current.
		require
			not_shared: not is_shared
		local
			new_count: INTEGER_64
		do
			new_count := count + other.count
			resize (new_count)
			memcpy_external (pointer_offset (item, count), other.item, other.count)
			count := new_count
		end

feature -- Resizing

	resize (n: INTEGER_64) is
			-- Reallocate `item' to hold `n' bytes.
		require
			not_shared: not is_shared
		do
			resize_external (1, $item, $count, n - count, True)
			if item = default_pointer then
				(create {EXCEPTIONS}).raise ("No more memory")
			end
		end

	allocate (n: INTEGER_64) is
			-- Allocate `item' with `n' bytes.
		require
			n_non_negative: n >= 0
		do
			if item /= default_pointer then
				if is_shared then
					item := allocate_external (n.max (1), True)
				else
					dispose
				end
			else
				item := allocate_external (n.max (1), True)
			end
			if item = default_pointer then
				(create {EXCEPTIONS}).raise ("No more memory")
			end
			count := n.max (1)
			is_shared := false
		ensure
			item_set: item /= default_pointer
			count_set: count = n.max (1)
			is_shared_set: not is_shared
		end

feature {LARGE_ARRAYED_LIST} -- Disposal

	dispose is
			-- Release memory pointed by `item'.
		local
			null: POINTER
		do
			debug ("memory")
				io.put_string ("Collecting id=")
				io.put_string (identifier.out)
				io.new_line
			end
			if not is_shared then
				free_external (item)
			end
			item := null
			is_shared := False
		ensure then
			shared_reset: not is_shared
		end

feature {NONE} -- Disposal

	temp_boolean: BOOLEAN
	temp_character: CHARACTER
	temp_double: DOUBLE
	temp_integer_8: INTEGER_8
	temp_integer_16: INTEGER_16
	temp_integer_32: INTEGER
	temp_integer_64: INTEGER_64

	temp_real: REAL




	
	allocate_external (num_bytes:INTEGER_64; zero: BOOLEAN): POINTER is
			-- Allocate memory for an array of `num_bytes' 
			-- bytes and returns a pointer to the
			-- allocated memory. If `zero', then the memory will be 
			-- zeroed.
		require
			num_bytes_positive: num_bytes >= 0
		external
			"C use <Memory.h>"
		alias
			"newMemory"
		end

	free_external (a_item: POINTER) is
		require
			a_item_not_void: a_item /= default_pointer
		external
			"C use <Memory.h>"
		alias
			"freeMemory"
		end

	resize_external (type_size: INTEGER; existing_data: POINTER;
					 existing_count: POINTER; delta_count: INTEGER_64;
					 zero: BOOLEAN) is
			-- Resize `existing_data' of `existing_count' 
			-- bytes by `type_size' * `delta_count' bytes and
			-- `zero' the extra memory. Negative `delta_count' implies
			-- `existing_data' will be shrunk.
		require
			type_size_positive: type_size >= 0
			existing_data_valid: existing_data /= default_pointer
			existing_count_valid: existing_count /= default_pointer
		external
			"C use <Memory.h>"
		alias
			"resizeMemory"
		end

	memcpy_external (a_dest, a_src: POINTER; a_n: INTEGER_64) is
		require
			a_n_greater_equal_zero: a_n >= 0
			a_dest_not_default: a_dest /= default_pointer
			a_src_not_default: a_src /= default_pointer
		external
			"C macro use <string.h>"
		alias
			"memcpy"
		end

	memcmp_external (a_dest, a_src: POINTER; a_n: INTEGER_64): INTEGER is
		require
			a_n_greater_equal_zero: a_n >= 0
			a_dest_not_default: a_dest /= default_pointer
			a_src_not_default: a_src /= default_pointer
		external
			"C macro use <string.h>"
		alias
			"memcmp"
		end

	read_data_external (a_dest, a_pointer: POINTER; a_pos: INTEGER_64;
						num_bytes: INTEGER_64) is
			-- Get `num_bytes' at the `a_pos'-th
			-- byte position starting at `a_pointer' and put in a_dest
		require
			a_pointer_not_default_pointer: a_pointer /= default_pointer
			a_dest_not_default_pointer: a_dest /= default_pointer
			a_pos_positive: a_pos >= 0
			num_bytes_positive: num_bytes >= 1
		do
			memcpy_external (a_dest, pointer_offset (a_pointer, a_pos), num_bytes)
		end

	put_data_external (a_src, a_dest: POINTER; a_pos: INTEGER_64;
					   num_bytes: INTEGER) is
			-- Put `num_bytes' from a_src at the `a_pos'-th
			-- byte position in `a_dest'
		require
			a_src_not_default_pointer: a_src /= default_pointer
			a_dest_not_default_pointer: a_dest /= default_pointer
			a_pos_positive: a_pos >= 0
			num_bytes_positive: num_bytes >= 1
		do
			memcpy_external (pointer_offset (a_dest, a_pos), a_src, num_bytes)			
		end

	pointer_offset (a_pointer: POINTER; bytes: INTEGER_64): POINTER is
			-- Offset `a_pointer' by `bytes'
		external
			"C macro use <Memory.h>"
		alias
			"pointer_offset"
		end
	
	object_id: OBJECT_IDENTIFIER is
			-- Create the id generator
		once
			create Result
		end

invariant
	valid_count: count >= 0
	
end -- class LARGE_MEMORY_POINTER
