
indexing
	description: "Abstract base class for linear algebra objects."
	names: "linear tuple"
	representation: "array"
	access: "fixed"
	size: "fixed"
	contents: "DOUBLE"
	patterns: "none"

deferred class

	LINEAR_TUPLE

inherit

	ANY
		-- Inherit explictly to satisfy SE 2.2
		undefine
			is_equal, out, copy
		end
	
	NUMERIC_COLLECTION
		export
			{ANY}
				is_nan
		undefine
			is_equal, out, copy
		end

	NUMERIC_CONVERSION
		export
			{NONE} all
		undefine
			is_equal, out, copy
		end
	
	DEBUG_OUTPUT
		undefine
			is_equal, out, copy
		end

feature -- Initialization

	make (the_count: INTEGER) is
			-- Make a zero vector with given number of components.
		require
			positive_length: the_count > 0
		deferred
		ensure
			count_is_length: count = the_count
			is_zero: is_zero
		end

	fill (initial_value: DOUBLE) is
			-- Set each `item' to `initial_value'.
		deferred
		ensure then
			items_set_to_initial_value: item (1) = initial_value
			-- Ditto for all the items.
		end

feature -- Access

	item (index: INTEGER): DOUBLE is
			-- Coordinate/component at `index'.
		require
			valid_index: index >= 1 and index <= count
		deferred
		end

	zero: like Current is
			-- Neutral element for "+" and "-".
		do
			Result := Current.deep_twin
			Result.fill (0.0)
		end

	one: like Current is
			-- Neutral element for "*".
		do
			Result := Current.deep_twin
			Result.fill (1.0)
		end

	unit: like Current is
			-- Unit-length (normalized) `Current'.
		require
			non_zero: magnitude /= 0.0
		do
			Result := Current.deep_twin
			Result.normalize
		end

feature -- Measurement

	one_norm: DOUBLE is
			-- Project down to a real by sum of absolute values of components.
		deferred
		end

	max_norm: DOUBLE is
			-- Project down to a real by selecting the largest absolute value
			-- component.
		deferred
		end

	two_norm, magnitude, to_real: DOUBLE is
			-- Euclidean magnitude (square-root of the sum of the squares of
			-- the components).
		do
			Result := square_root (magnitude_squared)
		end

	magnitude_squared: DOUBLE is
			-- Project down to a real by sum of squares of components.
		do
			Result := (Current # Current)
		end

feature -- Status report

	is_unit: BOOLEAN is
			-- Is this vector of unit length (within 1 +/- `Tolerance')?
		do
			Result := within_range (magnitude, 1.0 - Tolerance, 1.0 + Tolerance)
		end

feature -- Comparison

	infix "<" (other: like Current): BOOLEAN is
			-- Is to_real < other.to_real?
		do
			Result := to_real <	other.to_real
		end

	is_equal (other: like Current): BOOLEAN is
			-- Is each corresponding `item' within `Tolerance'?
		require else
			same_dimension: other.count = count
		deferred
		ensure then
			result_reasonable:
				(Result implies about_equal (item (1), other.item (1))) and
				(Result implies about_equal (item (count), other.item (count)))
		end

	is_zero: BOOLEAN is
			-- Are all coordinates/components exactly zero?
		deferred
		ensure then
			result_reasonable:
				Result implies item (1) = 0.0 and item (count) = 0.0
		end

	all_finite: BOOLEAN is
			-- Are all coordinates/components finite (not +/-inf/nan)?
		deferred
		ensure
			result_reasonable:
				Result implies (is_finite (item (1)) and
								is_finite (item (count)))
		end

feature -- Status report

	divisible (other: like Current): BOOLEAN is
			-- May current object be divided by `other'?
			-- Answer: no.
		do
		end

	exponentiable (other: NUMERIC): BOOLEAN is
			-- May current object be elevated to the power `other'?
		do
--- UNIMPLEMENTED: Result is False for now...
		end

feature -- Element change

	put (value: DOUBLE; index: INTEGER) is
			-- Assign `value' to component/coordinate at `index'.
		require
			non_nan_value: not is_nan (value)
			valid_index: index >= 1 and index <= count
		deferred
		ensure
			assigned: item (index) = value
		end

	normalize is
			-- Make unit length.
		require
			non_zero: magnitude /= 0.0
		deferred
		ensure
			normalized: is_unit
		end

feature -- Duplication

	copy (other: like Current) is
			-- Copy up to `count.min (other.count)' items from `other'.
		require else
			other_not_void: other /= Void
		deferred
		ensure then
			first_item_copied: item (1) = other.item (1)
			last_item_copied: item (count.min (other.count)) =
								other.item (count.min (other.count))			
			--- Ditto for items in-between.
		end

feature -- Basic operations

	prefix "+": like Current is
			-- Unary plus.
		do
			Result := Current.deep_twin
		ensure then
			result_equals_current: Result.is_equal (Current)
		end

	prefix "-": like Current is
			-- Unary minus.
		deferred
		ensure then
			result_equals_negative_current: (Result.sum_tuple (Current)).is_zero
		end

	summation (other: like Current) is
			-- Sum `other' with Current
		require
			same_dimension: count = other.count
		deferred
		end
	
	infix "#", dot (other: like Current): DOUBLE is
			-- Inner-product (dot-product) with `other'.
		require
			same_dimension: count = other.count
		deferred
		ensure
			result_non_nan: not is_nan (Result)
		end

	scale (scalar: DOUBLE): like Current is
			-- Scale by the real `scalar'.
		require
			non_nan_scalar: not is_nan (scalar)
		deferred
		ensure
			non_void_result: Result /= Void
			first_item_is_scaled: about_equal (Result.item (1), (item (1) * scalar))
			last_item_is_scaled: about_equal (Result.item (count), (item (count) * scalar))
			--- Ditto for items in-between.
		end

feature -- Conversion

	to_list: DOUBLE_ARRAYED_LIST is
			-- Create a row-major 1-d array from the elements of `Current'.
		do
			create Result.make (count)
			to_array_argument (Result)
		ensure
			Result_not_void: Result /= Void
			correct_count: Result.count = count
			elements_are_stored: check_list (Result)
		end

	to_array_argument (a: DOUBLE_ARRAYED_LIST) is
			-- Replace the items of `a' with the row-major elements of `Current'.
		require
			a_not_void: a /= Void
			a_correct_size: a.count >= count
		deferred
		end

	out: STRING is
			-- Yield a printable representation.
		deferred
		ensure then
			result_correct: Result /= Void and then Result.count >= 7
		end

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			Result := out
		end

feature {LINEAR_TUPLE} -- Implementation

	sum_tuple (other: LINEAR_TUPLE): LINEAR_TUPLE is
			   -- Sum with another tuple
		require
			same_dimension: count = other.count
		deferred
		end
	

	difference_tuple (other: LINEAR_TUPLE): LINEAR_TUPLE is
			   -- Difference with another tuple
		require
			same_dimension: count = other.count
		deferred
		end
	

feature {NONE} -- Implementation

	check_list (a: DOUBLE_ARRAYED_LIST): BOOLEAN is
			-- See if the elements in `a' are the same as `Current'
		local
			i64: INTEGER_64
			i: INTEGER
		do
			from
				i64 := 1
				i := 1
				Result := true
			invariant
				i64_at_least_one: i64 >= 1
				i64_small_enough: i64 <= 1 + count
				i_at_least_one: i >= 1
				i_small_enough: i <= 1 + count
			variant
				i64_increasing: 1 + count - i64
			until
				i64 >= count
			loop
				Result := Result and about_equal (a.item (i64), item (i))
				i64 := i64 + 1
				i := i + 1
			end
		end
			
	infix "/" (other: like Current): like Current is
			-- Division by `other'.
			-- (Precondition False.)
		do
		end

	infix "^" (other: NUMERIC): NUMERIC is
			-- Current object to the power `other'.
		do
--- UNIMPLEMENTED.
		end

	Opening_out_delimiter: CHARACTER is
			-- First character of result of `out'.
		deferred
		end

	Closing_out_delimiter: CHARACTER is
			-- Second character of result of `out'.
		deferred
		end

	is_nan_free: BOOLEAN is
			-- Are there no nan (Not a Number) items?
		deferred
		end

invariant

	no_nan_items: is_nan_free
	zero_is_correct: zero.is_zero
	is_zero_is_correct:
		is_zero implies
			(magnitude_squared = 0.0 and magnitude = 0.0 and two_norm = 0.0 and
			one_norm = 0.0 and max_norm = 0.0)
	magnitude_is_non_negative: magnitude >= 0.0
	magnitude_squared_is_non_negative: magnitude_squared >= 0.0
	two_norm_is_magnitude: two_norm = magnitude
	one_norm_is_non_negative: one_norm >= 0.0
	one_norm_is_large_enough: (( count = 1 implies
								 (one_norm + Tolerance >= absolute_value (item (1))))
							   or else
							   (one_norm + Tolerance >= (absolute_value (item (1)) +
														 absolute_value (item (count)))))
	
	max_norm_is_non_negative: max_norm >= 0.0
	max_norm_is_large_enough: max_norm >= absolute_value (item (1)) and
								max_norm >= absolute_value (item (count))
	max_norm_is_small_enough: max_norm <= one_norm
	magnitude_is_correct: about_equal (magnitude,
									   square_root (magnitude_squared))
	is_unit_correct:
		(is_nan_free and not is_zero ) implies (is_unit implies
			within_range (magnitude, 1.0 - Tolerance, 1.0 + Tolerance))
	pound_with_self_is_magnitude_squared:
		about_equal (Current # Current, magnitude_squared)
	dot_with_self_is_magnitude_squared:
		about_equal (dot (Current), magnitude_squared)
	unit_is_correct: (is_nan_free and not is_zero and magnitude /= 0.0) implies unit.is_unit

end -- deferred class LINEAR_TUPLE

