indexing

	description: "4 element linear algebra objects."
	names: "linear tuple"
	representation: "expanded", "DOUBLE"
	access: "fixed"
	size: "fixed"
	contents: "DOUBLE"

deferred class

	LINEAR_TUPLE_4D

inherit

	LINEAR_TUPLE_ND
			-- Inherit for implementation
		rename
			make as make_with_elements
		export
			{LINEAR_TUPLE_4D}
				make_with_elements
		redefine
			default_create, make_from_array, fill, make_from_c,
			item, unit, count, two_norm, magnitude, to_real,
			magnitude_squared, one_norm, max_norm, is_equal,
			is_zero, all_finite, put, normalize, prefix "+", prefix "-",
			infix "#", dot, scale, out, to_c, make_with_elements,
			to_array_argument, copy

		end

feature -- Initialization

	default_create is
			-- Create a zero vector
		do
			fill (0.0)
		end

	make (initial_x, initial_y, initial_z, initial_w: DOUBLE) is
			-- Initial cartesian component coordinates.
		require
			non_nan_first_argument: not is_nan (initial_x)
			non_nan_second_argument: not is_nan (initial_y)
			non_nan_third_argument: not is_nan (initial_z)
			non_nan_fourth_argument: not is_nan (initial_w)
		do
			x := initial_x
			y := initial_y
			z := initial_z
			w := initial_w
		ensure
			x_set: x = initial_x
			y_set: y = initial_y
			z_set: z = initial_z
			w_set: w = initial_w
		end

	fill (initial_value: DOUBLE) is
			-- Set each `item' to `initial_value'.
		do
			x := initial_value
			y := initial_value
			z := initial_value
			w := initial_value
		ensure then
			x_is_initial_value: x = initial_value
			y_is_initial_value: y = initial_value
			z_is_initial_value: z = initial_value
			w_is_initial_value: w = initial_value
		end

	make_from_array (values: ARRAY [DOUBLE]) is
			-- Make with these component values.
		do
			fill (0.0)

			if values.count >= 1 then
				x := values.item (1)
			end

			if values.count >= 2 then
				y := values.item (2)
			end

			if values.count >= 3 then
				z := values.item (3)
			end

			if values.count >= 4 then
				w := values.item (4)
			end
		end

	make_from_c (address: POINTER; the_count: INTEGER) is
			-- Re-construct from address of C storage of `the_count' DOUBLEs.
		local
			a: ARRAY [DOUBLE]
			p: EXTERNAL_POINTER [DOUBLE]
		do
			create a.make (1, the_count.min (count))
			copy_doubles (p.get_pointer (a), address, a.count)
			make_from_array (a)
		end

feature -- Access

	x: DOUBLE
			-- X-coordinate.

	y: DOUBLE
			-- Y-coordinate.

	z: DOUBLE
			-- Z-coordinate.

	w: DOUBLE
			-- W-coordinate.

	item (index: INTEGER): DOUBLE is
			-- Coordinate/component at `index'.
		do
			inspect
				index
			when 1 then
				Result := x
			when 2 then
				Result := y
			when 3 then
				Result := z
			else
				Result := w
			end
		ensure then
			index_one_implies_x: index = 1 implies Result = x
			index_two_implies_y: index = 2 implies Result = y
			index_three_implies_z: index = 3 implies Result = z
			index_four_implies_w: index = 4 implies Result = w
		end

	unit: like Current is
			-- Unit-length (normalized) `Current'.
		do
			Result := Current.deep_twin
			Result.make (x, y, z, w)
			Result.normalize
		end

feature -- Measurement

	count: INTEGER is 4
			-- The number of coordinates/components.

	two_norm, magnitude, to_real: DOUBLE is
			-- Euclidean magnitude (square-root of the sum of the squares of
			-- the components).
		do
			Result := square_root (magnitude_squared)
		end

	magnitude_squared: DOUBLE is
			-- Project down to a real by sum of squares.
		do
			Result := x * x + y * y + z * z + w * w
		end

	one_norm: DOUBLE is
			-- Project down to a real by sum of absolute values.
		do
			Result := absolute_value (x) + absolute_value (y) +
						absolute_value (z) + absolute_value (w)
		end

	max_norm: DOUBLE is
			-- Project down to a real by selecting the largest absolute value.
		do
			Result := (absolute_value (x).max (absolute_value (y))).max
						(absolute_value (z).max (absolute_value (w)))
		end

feature -- Comparison

	is_equal (other: like Current): BOOLEAN is
			-- Is each corresponding `item' within `Tolerance'?
		do
			Result := about_equal (x, other.x) and then
						about_equal (y, other.y) and then
						about_equal (z, other.z) and then
						about_equal (w, other.w)
		ensure then
			result_correct:
				Result = (about_equal (x, other.x) and
						about_equal (y, other.y) and
						about_equal (z, other.z) and
						about_equal (w, other.w))
		end

	is_zero: BOOLEAN is
			-- Are all coordinates/components exactly zero?
		do
			Result := x = 0.0 and then y = 0.0 and then z = 0.0 and then w = 0.0
		ensure then
			Result implies x = 0.0 and y = 0.0 and z = 0.0 and w = 0.0
		end

	all_finite: BOOLEAN is
			-- Are all coordinates/components finite (not +/-inf/nan)?
		do
			Result :=
				(is_finite (x) and then is_finite (y) and then is_finite (z)
				and then is_finite (w))
		ensure then
			result_correct:
				Result = (is_finite (x) and is_finite (y) and is_finite (z) and
							is_finite (w))
		end
feature -- Element change

	put_x (new_x: DOUBLE) is
			-- Set the x-coordinate.
		require
			non_nan_argument: not is_nan (new_x)
		do
			x := new_x
		ensure
			x_set: x = new_x
		end

	put_y (new_y: DOUBLE) is
			-- Set the y-coordinate.
		require
			non_nan_argument: not is_nan (new_y)
		do
			y := new_y
		ensure
			y_set: y = new_y
		end

	put_z (new_z: DOUBLE) is
			-- Set the z-coordinate.
		require
			non_nan_argument: not is_nan (new_z)
		do
			z := new_z
		ensure
			z_set: z = new_z
		end

	put_w (new_w: DOUBLE) is
			-- Set the w-coordinate.
		require
			non_nan_argument: not is_nan (new_w)
		do
			w := new_w
		ensure
			w_set: w = new_w
		end

	put (value: DOUBLE; index: INTEGER) is
			-- Assign_of `value' to component at `index'.
		do
			inspect
				index
			when 1 then
				x := value
			when 2 then
				y := value
			when 3 then
				z := value
			else
				w := value
			end
		ensure then
			index_one_implies_x: index = 1 implies x = value
			index_two_implies_y: index = 2 implies y = value
			index_three_implies_z: index = 3 implies z = value
			index_four_implies_w: index = 4 implies w = value
		end

	normalize is
			-- Make unit length.
		local
			scaled_x, scaled_y, scaled_z, scaled_w: DOUBLE
			magnitude_of_x, magnitude_of_y, magnitude_of_z: DOUBLE
			magnitude_of_w: DOUBLE
			scalar, recipricol_of_magnitude, largest_magnitude: DOUBLE
			square_root_of_one_half, square_root_of_one_third: DOUBLE
			square_root_of_one_fourth: DOUBLE
		do
			magnitude_of_x := absolute_value (x)
			magnitude_of_y := absolute_value (y)
			magnitude_of_z := absolute_value (z)
			magnitude_of_w := absolute_value (w)

			if is_infinity (magnitude_of_x) then

				if is_infinity (magnitude_of_y) then

					if is_infinity (magnitude_of_z) then

						if is_infinity (magnitude_of_w) then
								-- Make components equal magnitude.
							square_root_of_one_fourth := square_root (0.25)
							x := sign_of (x) * square_root_of_one_fourth
							y := sign_of (y) * square_root_of_one_fourth
							z := sign_of (z) * square_root_of_one_fourth
							w := sign_of (w) * square_root_of_one_fourth
						else
								-- x, y, z dominate so make them equal
								-- magnitude.
							square_root_of_one_third := square_root (1.0 / 3.0)
							x := sign_of (x) * square_root_of_one_third
							y := sign_of (y) * square_root_of_one_third
							z := sign_of (z) * square_root_of_one_third
							w := 0.0
						end

					elseif is_infinity (magnitude_of_w) then
							-- x, y, w dominate so make them equal magnitude.
						square_root_of_one_third := square_root (1.0 / 3.0)
						x := sign_of (x) * square_root_of_one_third
						y := sign_of (y) * square_root_of_one_third
						z := 0.0
						w := sign_of (w) * square_root_of_one_third
					else
							-- x and y dominate so make their components equal.
						square_root_of_one_half := square_root (0.5)
						x := sign_of (x) * square_root_of_one_half
						y := sign_of (y) * square_root_of_one_half
						z := 0.0
						w := 0.0
					end

				elseif is_infinity (magnitude_of_z) then

						if is_infinity (magnitude_of_w) then
								-- x, z, w dominate so make them equal
								-- magnitude.
							square_root_of_one_third := square_root (1.0 / 3.0)
							x := sign_of (x) * square_root_of_one_third
							y := 0.0
							z := sign_of (z) * square_root_of_one_third
							w := sign_of (w) * square_root_of_one_third
						else
								-- x, and z dominate so make them equal
								-- magnitude.
							square_root_of_one_half := square_root (0.5)
							x := sign_of (x) * square_root_of_one_half
							y := 0.0
							z := sign_of (z) * square_root_of_one_half
							w := 0.0
						end

				elseif is_infinity (magnitude_of_w) then
						-- x, and w dominate so make them equal magnitude.
					square_root_of_one_half := square_root (0.5)
					x := sign_of (x) * square_root_of_one_half
					y := 0.0
					z := 0.0
					w := sign_of (w) * square_root_of_one_half
				else
						-- x dominates.
					x := sign_of (x)
					y := 0.0
					z := 0.0
					w := 0.0
				end

			elseif is_infinity (magnitude_of_y) then

				if is_infinity (magnitude_of_z) then

					if is_infinity (magnitude_of_w) then
							-- y, z, w dominate so make them equal
							-- magnitude.
						square_root_of_one_third := square_root (1.0 / 3.0)
						x := 0.0
						y := sign_of (y) * square_root_of_one_third
						z := sign_of (z) * square_root_of_one_third
						w := sign_of (w) * square_root_of_one_third
					else
							-- y and z dominate so make their components equal.
						square_root_of_one_half := square_root (0.5)
						x := 0.0
						y := sign_of (y) * square_root_of_one_half
						z := sign_of (z) * square_root_of_one_half
						w := 0.0
					end

				elseif is_infinity (magnitude_of_w) then
						-- y and z dominate so make their components equal.
					square_root_of_one_half := square_root (0.5)
					x := 0.0
					y := sign_of (y) * square_root_of_one_half
					z := 0.0
					w := sign_of (w) * square_root_of_one_half
				else
						-- y dominates.
					x := 0.0
					y := sign_of (y)
					z := 0.0
					w := 0.0
				end

			elseif is_infinity (magnitude_of_z) then

				if is_infinity (magnitude_of_w) then
						-- z and w dominate so make their components equal.
					square_root_of_one_half := square_root (0.5)
					x := 0.0
					y := 0.0
					z := sign_of (y) * square_root_of_one_half
					w := sign_of (w) * square_root_of_one_half
				else
						-- z dominates.
					x := 0.0
					y := 0.0
					z := sign_of (z)
					w := 0.0
				end

			elseif is_infinity (magnitude_of_w) then
					-- w dominates.
				x := 0.0
				y := 0.0
				z := 0.0
				w := sign_of (w)
			else
					-- All components have non-infinite magnitude but may
					-- still be quite large so scale the components by the
					-- recipricol of the magnitude of the largest component
					-- to avoid overflow.
				largest_magnitude := magnitude_of_x.max (magnitude_of_y)
				largest_magnitude := largest_magnitude.max (magnitude_of_z)
				largest_magnitude := largest_magnitude.max (magnitude_of_w)
				scalar := 1.0 / largest_magnitude
					check
						scalar_is_positive: scalar > 0.0
					end
				scaled_x := x * scalar
				scaled_y := y * scalar
				scaled_z := z * scalar
				scaled_w := w * scalar
				recipricol_of_magnitude := 1.0 /
					square_root (scaled_x * scaled_x + scaled_y * scaled_y +
								scaled_z * scaled_z + scaled_w * scaled_w)
					check
						recipricol_of_magnitude_is_positive:
							recipricol_of_magnitude > 0.0
					end
				x := scaled_x * recipricol_of_magnitude
				y := scaled_y * recipricol_of_magnitude
				z := scaled_z * recipricol_of_magnitude
				w := scaled_w * recipricol_of_magnitude
					check
						scaling_successful:
							(square (x) + square (y) + square (z) + square (w))
							>= square (1.0 - Tolerance)
					end
			end
		end

feature -- Duplication

	copy (other: like Current) is
			-- Copy up to `count.min (other.count)' items from `other'.
		do
			x := other.x
			y := other.y
			z := other.z
			w := other.w
		end

feature -- Basic operations

	prefix "+": like Current is
			-- Positive of `Current'.
		do
			Result := Current.deep_twin
			Result.make (x, y, z, w)
		end

	prefix "-": like Current is
			-- Negative of `Current'.
		do
			Result := Current.deep_twin
			Result.make (-x, -y, -z, -w)
		ensure then
			Result_not_void: Result /= Void
			result_x_equals_negative_current_x: Result.x = -x
			result_y_equals_negative_current_y: Result.y = -y
			result_z_equals_negative_current_z: Result.z = -z
			result_w_equals_negative_current_w: Result.w = -w
		end

	infix "#", dot (other: like Current): DOUBLE is
			-- Inner-product with `other'.
		local
			the_product: DOUBLE
		do
			-- Avoid nan (e.g., 0.0 * inf).
			Result := product (x, other.x)

			-- Avoid nan (e.g., inf + -inf).
			the_product := product (y, other.y)
			Result := sum (Result, the_product)

			the_product := product (z, other.z)
			Result := sum (Result, the_product)

			the_product := product (w, other.w)
			Result := sum (Result, the_product)

		ensure then
			result_correct:
						   -- Note: addition is not associative with
						   -- respect to infinity, i.e. s = sum, p =
						   -- product, s (s (s (p, p), p), p) is not
						   -- equivilent to s (s (p, p), s (p, p))
				about_equal (Result,
							sum (sum (sum (product (x, other.x), product (y, other.y)),
								 product (z, other.z)), product (w, other.w)))
		end

	scale (scalar: DOUBLE): like Current is
			-- Scale by the real `scalar'.
		local
			result_x, result_y, result_z, result_w: DOUBLE
		do
			Result := Current.deep_twin

			result_x := x * scalar
			result_y := y * scalar
			result_z := z * scalar
			result_w := w * scalar
			
			Result.make (result_x, result_y, result_z, result_w)

		ensure then
			Result_not_void: Result /= Void
			result_x_correct: Result.x = filter_nan (x * scalar)
			result_y_correct: Result.y = filter_nan (y * scalar)
			result_z_correct: Result.z = filter_nan (z * scalar)
			result_w_correct: Result.w = filter_nan (w * scalar)
		end

feature -- Conversion

	out: STRING is
			-- Yield a printable representation.
		do
			create Result.make (80)
			Result.append_character (Opening_out_delimiter)
			Result.append_character (' ')
			Result.append_string (x.out)
			Result.append_character (' ')
			Result.append_string (y.out)
			Result.append_character (' ')
			Result.append_string (z.out)
			Result.append_character (' ')
			Result.append_string (w.out)
			Result.append_character (' ')
			Result.append_character (Closing_out_delimiter)
		ensure then
			result_long_enough: Result.count >= 11
		end

	to_array_argument (a: DOUBLE_ARRAYED_LIST) is
			-- Replace the items of `a' with the row-major elements of `Current'.
		local
			index: INTEGER_64
		do
			index := 1
			a.put (x, index)
			a.put (y, index + 1)
			a.put (z, index + 2)
			a.put (w, index + 3)
		ensure then
			x_stored: about_equal (x, a.item (1))
			y_stored: about_equal (y, a.item (2))
			z_stored: about_equal (z, a.item (3))
			w_stored: about_equal (w, a.item (4))
		end

	to_c: POINTER is
			-- Area of 2 row-major values for passing to C routines.
		local
			p: EXTERNAL_POINTER [DOUBLE]
		do
--			Result := p.get_pointer (to_array)
		end

feature {NONE}

	make_with_elements (the_count: INTEGER) is
		do
			default_create
		end

invariant

	count_is_four: count = 4
	magnitude_squared_is_sum_of_squares_of_components:
		about_equal (magnitude_squared, square (x) + square (y) + square (z) + square (w))
	one_norm_is_sum_of_absolute_values_of_components:
		about_equal (one_norm, absolute_value (x) + absolute_value (y) +
					absolute_value (z) + absolute_value (w))
	one_norm_is_sum_of_absolute_values_of_components:
		about_equal (one_norm, absolute_value (x) + absolute_value (y) +
					absolute_value (z) + absolute_value (w))
	max_norm_is_max_of_absolute_values_of_components:
		about_equal (max_norm, (absolute_value (x).max (absolute_value (y))).max
						(absolute_value (z).max (absolute_value (w))))

end -- class LINEAR_TUPLE_4D
