
indexing
	description: "N element linear algebra objects."
	names: "linear tuple"
	representation: "array"
	access: "fixed"
	size: "fixed"
	contents: "DOUBLE"
	patterns: "none"

deferred class

	LINEAR_TUPLE_ND

inherit

	LINEAR_TUPLE

feature -- Initialization

	make_from_c (address: POINTER; the_count: INTEGER) is
			-- Re-construct from address of C storage of `the_count' DOUBLEs.
		local
			a: ANY
		do
			make (the_count)
--			a := elements.area
--			copy_doubles ($a, address, the_count.min (count))
		end

	fill (initial_value: DOUBLE) is
			-- Set each `item' to `initial_value'.
		local
			index: INTEGER
		do
			from
				index := 1
			invariant
				index_at_least_one: index >= 1
				index_small_enough: index <= 1 + count
			variant
				index_increasing: 1 + count - index
			until
				index > count
			loop
				put (initial_value, index)
				index := index + 1
			end
		end

feature -- Access

	hash_code: INTEGER is
			-- Hash code value
		local
			index: INTEGER
		do
			from
				Result := item (1).hash_code
				index := 1
			invariant
				index_at_least_one: index >= 1
				index_small_enough: index <= 1 + count
			variant
				index_increasing: 1 + count - index
			until
				index > count
			loop
				Result := Result & item (index).hash_code
				index := index + 1
			end
		end

feature -- Measurement

	one_norm: DOUBLE is
			-- Project down to a real by sum of absolute values of components.
		local
			index: INTEGER
		do
			Result := absolute_value (item (1))

			from
				index := 2
			invariant
				index_at_least_one: index >= 2
				index_small_enough: index <= 1 + count
			variant
				index_increasing: 1 + count - index
			until
				Result > maximum_double or else index > count
			loop
				Result := Result + absolute_value (item (index))
				index := index + 1
			end
		end

	max_norm: DOUBLE is
			-- Project down to a real by selecting the largest absolute value
			-- component.
		local
			index: INTEGER
			item_magnitude: DOUBLE
		do
			Result := absolute_value (item (1))

			from
				index := 2
			invariant
				index_at_least_one: index >= 2
				index_small_enough: index <= 1 + count
			variant
				index_increasing: 1 + count - index
			until
				Result > maximum_double or else index > count
			loop
				item_magnitude := absolute_value (item (index))

				if item_magnitude > Result then
					Result := item_magnitude
				end

				index := index + 1
			end
		end

feature -- Comparison

	is_equal (other: like Current): BOOLEAN is
			-- Is each corresponding `item' within `Tolerance'?
		local
			index: INTEGER
		do
			from
				index := 1
				Result := True
			invariant
				index_at_least_one: index >= 1
				index_small_enough: index <= 1 + count
			variant
				index_increasing: 1 + count - index
			until
				Result = False or else index > count
			loop
				Result := about_equal (item (index), other.item (index))
				index := index + 1
			end
		end

	is_zero: BOOLEAN is
			-- Are all coordinates/components exactly zero?
		local
			index: INTEGER
		do
			from
				Result := True
				index := 1
			invariant
				index_at_least_one: index >= 1
				index_small_enough: index <= 1 + count
			variant
				index_increasing: 1 + count - index
			until
				Result = False or else index > count
			loop
				Result := item (index) = 0.0
				index := index + 1
			end
		end

	all_finite: BOOLEAN is
			-- Are all coordinates/components finite (not +/-inf/nan)?
		local
			index: INTEGER
		do
			from
				Result := True
				index := 1
			invariant
				index_at_least_one: index >= 1
				index_small_enough: index <= 1 + count
			variant
				index_increasing: 1 + count - index
			until
				Result = False or else index > count
			loop
				Result := is_finite (item (index))
				index := index + 1
			end
		end

feature -- Element change

	normalize is
			-- Make unit length.
		local
			index, index_of_first_largest_item, number_of_largest_items: INTEGER
			scalar, the_norm, indexed_item: DOUBLE
			magnitude_of_largest_item, magnitude_of_indexed_item: DOUBLE
				-- Magnitude of largest element found so far and
				-- of the currently indexed item.
			magnitude_of_largest_item_is_infinity: BOOLEAN
		do
				-- Find the magnitude, first index and count of the largest
				-- component:

			index_of_first_largest_item := 1
			number_of_largest_items := 1
			indexed_item := item (1)
			magnitude_of_largest_item := absolute_value (indexed_item)
			magnitude_of_largest_item_is_infinity :=
				is_infinity (magnitude_of_largest_item)

			from
				index := 2
			invariant
				index_at_least_one: index >= 2
				index_small_enough: index <= 1 + count
			variant
				index_increasing: 1 + count - index
			until
				index > count
			loop
				indexed_item := item (index)
				magnitude_of_indexed_item := absolute_value (indexed_item)

				if magnitude_of_indexed_item = magnitude_of_largest_item
						or else
						(magnitude_of_largest_item_is_infinity and then
						is_infinity (magnitude_of_indexed_item)) then
					number_of_largest_items := number_of_largest_items + 1
				elseif magnitude_of_indexed_item > magnitude_of_largest_item
						then
					magnitude_of_largest_item := magnitude_of_indexed_item
					magnitude_of_largest_item_is_infinity :=
						is_infinity (magnitude_of_largest_item)
					index_of_first_largest_item := index
					number_of_largest_items := 1
				end

				index := index + 1
			end

			if is_infinity (magnitude_of_largest_item) then
					-- The largest item(s) dominate the rest.

				if number_of_largest_items = 1 then
						-- There is one unique largest item so make that
						-- component one and all the rest zero.
					fill (0.0)
					put (1.0, index_of_first_largest_item)
				else
						-- There are `number_of_largest_items' components with
						-- equal (infinite) magnitude so make each of those
						-- components square_root (1 / number_of_largest_items)
						-- and the rest zero.

					from
						scalar := square_root (1.0 / number_of_largest_items)
						index := 1
					invariant
						index_at_least_one: index >= 1
						index_small_enough: index <= 1 + count
					variant
						index_increasing: 1 + count - index
					until
						index > count
					loop
						if is_infinity (absolute_value (item (index))) then
							put (scalar, index)
						else
							put (0.0, index)
						end

						index := index + 1
					end
				end
			else
					-- First scale the vector down by dividing each
					-- component by the magnitude of the largest item
					-- then normalize it.

				from
					scalar := 1.0 / magnitude_of_largest_item
					index := 1
				invariant
					index_at_least_one: index >= 1
					index_small_enough: index <= 1 + count
				variant
					index_increasing: 1 + count - index
				until
					index > count
				loop
					put (item (index) * scalar, index)
					the_norm := the_norm + square (item (index))
					index := index + 1
				end

					check
						the_norm_is_finite: not is_infinity (the_norm)
					end

				from
					scalar := 1.0 / square_root (the_norm)
					index := 1
				invariant
					index_at_least_one: index >= 1
					index_small_enough: index <= 1 + count
				variant
					index_increasing: 1 + count - index
				until
					index > count
				loop
					put (item (index) * scalar, index)
					index := index + 1
				end
			end
		end

feature -- Duplication

	copy (other: like Current) is
			-- Copy up to `count.min (other.count)' items from `other'.
		local
			index: INTEGER
		do
			from
				index := 1
			invariant
				index_large_enough: index >= 1
				index_small_enough: index <= 1 + count.min (other.count)
			variant
				index_increasing: 1 + count.min (other.count) - index
			until
				 index > count or else index > other.count
			loop
				put (other.item (index), index)
				index := index + 1
			end
		end

feature -- Basic operations

	prefix "-": like Current is
			-- Unary minus.
		local
			index: INTEGER
		do
			Result := Current.deep_twin

			from
				index := 1
			invariant
				index_at_least_one: index >= 1
				index_small_enough: index <= 1 + count
			variant
				index_increasing: 1 + count - index
			until
				index > count
			loop
				Result.put (-item (index), index)
				index := index + 1
			end
		end

	infix "#", dot (other: like Current): DOUBLE is
			-- Inner-product (dot-product) with `other'.
		local
			index: INTEGER
			left, right, the_product: DOUBLE
		do
			from
				index := 1
			invariant
				index_at_least_one: index >= 1
				index_small_enough: index <= 1 + count
			variant
				index_increasing: 1 + count - index
			until
				index > count
			loop
				left := item (index)

				if left /= 0.0 then
						-- Avoid nan (e.g., 0.0 * -inf).
					right := other.item (index)

					if right /= 0.0 then
							-- Avoid nan (e.g., inf * 0.0).
						the_product := left * right

						if Result = -the_product then
								-- Avoid nan (e.g., inf + -inf).
							Result := 0.0
						else
							Result := Result + the_product
						end
					end
				end

				index := index + 1
			end
		end

	summation (other: like Current) is
			-- Sum `other' with Current
		local
			index: INTEGER
			left, right: DOUBLE
		do
			from
				index := 1
			invariant
				index_at_least_one: index >= 1
				index_small_enough: index <= 1 + count
			variant
				index_increasing: 1 + count - index
			until
				index > count
			loop
				left := item (index)
				right := other.item (index)

				if left = -right then
						-- Avoid nan (e.g., -inf + inf).
					put (0.0, index)
				else
					put (left + right, index)
				end

				index := index + 1
			end
		end
	
	scale (scalar: DOUBLE): like Current is
			-- Scale by the real `scalar'.
		local
			index: INTEGER
			left: DOUBLE
		do
			Result := Current.deep_twin

			if scalar = 0.0 then
				Result.fill (0.0)
			else
				from
					index := 1
				invariant
					index_at_least_one: index >= 1
					index_small_enough: index <= 1 + count
				variant
					index_increasing: 1 + count - index
				until
					index > count
				loop
					left := item (index)

					if left = 0.0 then
							-- Avoid nan (e.g., 0.0 * inf).
						Result.put (0.0, index)
					else
						Result.put (left * scalar, index)
					end

					index := index + 1
				end
			end
		end

feature -- Conversion

	to_array_argument (a: DOUBLE_ARRAYED_LIST) is
			-- Replace the items of `a' with the row-major elements of `Current'.
		local
			i: INTEGER
		do
			from
				i := 1 
			invariant
				i_at_least_one: i >= 1
				i_small_enough: i <= 1 + count
			variant
				i_increasing: 1 + count - i
			until
				i >= count
			loop
				a.put (item (i), i)
				i := i + 1
			end
		end

	out: STRING is
			-- Yield a printable representation.
		local
			index: INTEGER
		do
			create Result.make (80)
			Result.append_character (Opening_out_delimiter)

			from
				index := 1
			invariant
				index_at_least_one: index >= 1
				index_small_enough: index <= 1 + count
			variant
				index_increasing: 1 + count - index
			until
				index > count
			loop
				Result.append_character (' ')
				Result.append (item (index).out)
				index := index + 1
			end

			Result.append_character (' ')
			Result.append_character (Closing_out_delimiter)
		end

feature {LINEAR_TUPLE} -- Implementation

	sum_tuple (other: LINEAR_TUPLE): LINEAR_TUPLE is
			   -- Sum with another tuple
		local
			index: INTEGER
			left, right: DOUBLE
		do
			Result := Current.deep_twin

			from
				index := 1
			invariant
				index_at_least_one: index >= 1
				index_small_enough: index <= 1 + count
			variant
				index_increasing: 1 + count - index
			until
				index > count
			loop
				left := item (index)
				right := other.item (index)

				if left = -right then
						-- Avoid nan (e.g., -inf + inf).
					Result.put (0.0, index)
				else
					Result.put (left + right, index)
				end

				index := index + 1
			end
		end
	

	difference_tuple (other: LINEAR_TUPLE): LINEAR_TUPLE is
			   -- Difference with another tuple
		local
			index: INTEGER
			left, right: DOUBLE
		do
			Result := Current.deep_twin

			from
				index := 1
			invariant
				index_at_least_one: index >= 1
				index_small_enough: index <= 1 + count
			variant
				index_increasing: 1 + count - index
			until
				index > count
			loop
				left := item (index)
				right := other.item (index)

				if left = right then
						-- Avoid nan (e.g., inf - inf).
					Result.put (0.0, index)
				else
					Result.put (left - right, index)
				end

				index := index + 1
			end
		end

feature {NONE} -- Implementation

	is_nan_free: BOOLEAN is
			-- Are there no nan (Not a Number) items?
		local
			index: INTEGER
		do
			from
				index := 1
				Result := True
			invariant
				index_at_least_one: index >= 1
				index_small_enough: index <= 1 + count
			variant
				index_increasing: 1 + count - index
			until
				Result = False or else index > count
			loop
				Result := not is_nan (item (index))
				index := index + 1
			end
		end

end -- LINEAR_TUPLE_ND

