indexing

	description: "M x N numeric matrices."
	names: "matrix", "numeric"
	representation: "none"
	access: "fixed", "index"
	size: "fixed"
	contents: "DOUBLE"
	date: "$Date: 2006-06-23 10:34:45 -0400 (Fri, 23 Jun 2006) $"
	revision: "$Revision: 226 $"
	author: "Mark Bolstad, US EPA/LM, bolstad.mark@epa.gov"
	status: "reviewed", "untested"

deferred class

	MATRIX

inherit

	NUMERIC
		redefine
			is_equal, out
		end

	NUMERIC_COLLECTION
		export
			{ANY}
				Tolerance, is_nan
		redefine
			is_equal, out
		end

feature -- Initialization

	fill (initial_value: DOUBLE) is
			-- Set each `item' to `initial_value'.
		deferred
		ensure then
			items_set_to_initial_value: about_equal (item (1, 1), initial_value)
			-- Ditto for all the items.
		end

	make_identity is
			-- Neutral element for "*".
		require
			square: row_count = column_count			
		deferred
		ensure
			current_is_identity: Current.is_equal (Current * Current)
		end

feature -- Access

	hash_code: INTEGER is
			-- Hash code value
		local
			column_index, row_index: INTEGER
		do
			Result := item (1, 1).hash_code
			from
				row_index := 1
			invariant
				row_index_at_least_one: row_index >= 1
				row_index_small_enough: row_index <= 1 + row_count
			variant
				row_index_increasing: 1 + row_count - row_index
			until
				row_index > row_count
			loop
				from
					column_index := 1
				invariant
					column_index_at_least_one: column_index >= 1
					column_index_small_enough: column_index <= 1 + column_count
				variant
					column_index_increasing: 1 + column_count - column_index
				until
					column_index > column_count
				loop
					Result := Result & item (row_index, column_index).hash_code
					column_index := column_index + 1
				end

				row_index := row_index + 1
			end
		end
	
	item (row, column: INTEGER): DOUBLE is
			-- Element at coordinates (`row', `column')
		require
			valid_row: row >= 1 and row <= row_count
			valid_column: column >= 1 and column <= column_count
		deferred
		end

	zero: like Current is
			-- Neutral element for "+" and "-".
		deferred
		ensure then
			result_exists: Result /= Void
		end

	one: like Current is
			-- Neutral element for "*"
		require else
			square: row_count = column_count
		deferred
		ensure then
			result_exists: Result /= Void
			square: Result.row_count = Result.column_count
			result_is_one: Current.is_equal (Current * Result)
		end

feature -- Measurement

	column_count: INTEGER is
			-- Number of columns.
		deferred
		end

	row_count: INTEGER is
			-- Number of rows.
		deferred
		end

	frobenius_norm, two_norm: DOUBLE is
			-- Square-root of the sum of the squares of the elements.
		deferred
		end

	one_norm: DOUBLE is
			-- Maximum absolute column sum of the elements.
		deferred
		end

	max_norm: DOUBLE is
			-- Maximum absolute row sum of the elements.
		deferred
		end

feature -- Comparison

	is_equal (other: like Current): BOOLEAN is
			-- Is each corresponding `item' within `Tolerance'?
		require else
			other_has_same_row_count: row_count = other.row_count
			other_has_same_column_count: column_count = other.column_count
		local
			column_index, row_index: INTEGER
		do
			Result := True

			from
				row_index := 1
			invariant
				row_index_at_least_one: row_index >= 1
				row_index_small_enough: row_index <= 1 + row_count
			variant
				row_index_increasing: 1 + row_count - row_index
			until
				not Result or else row_index > row_count
			loop
				from
					column_index := 1
				invariant
					column_index_at_least_one: column_index >= 1
					column_index_small_enough: column_index <= 1 + column_count
				variant
					column_index_increasing: 1 + column_count - column_index
				until
					not Result or else column_index > column_count
				loop
					Result := about_equal (item (row_index, column_index),
										   other.item (row_index, column_index))
					column_index := column_index + 1
				end

				row_index := row_index + 1
			end
		end

feature -- Status report

	divisible (other: like Current): BOOLEAN is
			-- May current object be divided by `other'?
			-- Answer: no.
		do
		end

	exponentiable (other: NUMERIC): BOOLEAN is
			-- May current object be elevated to the power `other'?
		do
			--- UNIMPLEMENTED: Result is False for now...
		end

	all_finite: BOOLEAN is
			-- Are all elements non-nan and not infinity or minus_infinity?
		deferred
		end

feature -- Element change

	put (value: like item; row, column: INTEGER) is
			-- Make `item' (`row', `column') = `value'.
		require
			valid_row: row >= 1 and row <= row_count
			valid_column: column >= 1 and column <= column_count
		deferred
		ensure
			assigned: about_equal (item (row, column), value)
		end

feature -- Transformation

		transpose is
			-- Transpose the current matrix.
		deferred
		ensure
			dimensions_switched: row_count = old column_count and
								 column_count = old row_count
		end

feature -- Basic operations

	prefix "+": like Current is
			-- Unary plus.
		deferred
		ensure then
			result_exists: Result /= Void
			result_equals_current: Result.is_equal (Current)
		end

	prefix "-": like Current is
			-- Unary minus.
		deferred
		ensure then
			result_exists: Result /= Void
		end

	infix "+" (other: like Current): like Current is
			-- Sum with `other' (commutative).
		require else
			other_exists: other /= Void
			dimensions_match: column_count = other.column_count and
							  row_count = other.row_count
		deferred
		ensure then
			result_exists: Result /= Void
		end

	infix "-" (other: like Current): like Current is
			-- Result of subtracting `other'.
		require else
			other_exists: other /= Void
			dimensions_match: column_count = other.column_count and
							  row_count = other.row_count
		deferred
		end

	infix "*" (other: like Current): like Current is
			-- Product by `other'.
		require else
			other_exists: other /= Void
			dimensions_correct: column_count = other.row_count
		deferred
		end

	infix "/" (other: like Current): like Current is
			-- Division by `other'.
			-- (Precondition False.)
		do
		end

	scale (scalar: DOUBLE): like Current is
			-- Result is the product of each element by `scalar'.
		deferred
		end

feature -- Conversion

	to_list: DOUBLE_ARRAYED_LIST is
			-- Create a row-major 1-d array from the elements of `Current'.
		do
			create Result.make (row_count * column_count)
			to_array_argument (Result)
		ensure
			Result_not_void: Result /= Void
			correct_count: Result.count = row_count * column_count
			elements_are_stored: check_list (Result)
		end

	to_array_argument (a: DOUBLE_ARRAYED_LIST) is
			-- Replace the items of `a' with the row-major elements of `Current'.
		require
			a_not_void: a /= Void
			a_correct_size: a.capacity >= row_count * column_count
		deferred
		end

	to_c: POINTER is
			-- Area of `count' doubles for passing to C routines.
			-- For passing the pointer to an instance of a reference
			-- NUMERIC_COMPARABLE object to a C routine:
			--	 Declare:
			--	   local the_area: ANY
			--	 Assign:
			--	   the_area := my_reference_numeric_comparable.to_c
			--	   the_count := my_reference_numeric_comparable.count
			--	 Call:
			--	   the_c_routine ($the_area, the_count)
			-- For achieving the same result for an instance of an expanded
			-- NUMERIC_COMPARABLE (containing only basic type (e.g., DOUBLE)
			-- attributes):
			--	 Call:
			--		the_c_routine ($my_expanded_numeric_comparable, the_count)
			-- For the above to work, expanded heirs must declare all of the
			-- appropriate expanded NUMERIC attributes at the start of the
			-- class, in order, so that the POINTER points at the contiguous
			-- block of writable basic (e.g., DOUBLE) attributes.
		do
			Result := to_list.storage.item
		end

	out: STRING is
			-- Yield a printable representation.
		local
			column_index, row_index: INTEGER
		do
			create Result.make (80)
			Result.append_character (Opening_out_delimiter)

			from
				row_index := 1
			invariant
				row_index_at_least_one: row_index >= 1
				row_index_small_enough: row_index <= 1 + row_count
			variant
				row_index_increasing: 1 + row_count - row_index
			until
				row_index > row_count
			loop
				from
					column_index := 1
				invariant
					column_index_at_least_one: column_index >= 1
					column_index_small_enough: column_index <= 1 + column_count
				variant
					column_index_increasing: 1 + column_count - column_index
				until
					column_index > column_count
				loop
					Result.append_character (' ')
					Result.append (item (row_index, column_index).out)
					column_index := column_index + 1
				end

				row_index := row_index + 1

				if row_index <= row_count then
					Result.append_character ('%N')
				end
			end

			Result.append_character (' ')
			Result.append_character (Closing_out_delimiter)
		ensure then
			result_correct: Result /= Void and then Result.count >= 5
		end

	print_using_io (ios: IO_INDENTED_MEDIUM) is
			-- Yield a printable representation.
		local
			column_index, row_index: INTEGER
		do
			ios.put_character (Opening_out_delimiter)

			from
				row_index := 1
			invariant
				row_index_at_least_one: row_index >= 1
				row_index_small_enough: row_index <= 1 + row_count
			variant
				row_index_increasing: 1 + row_count - row_index
			until
				row_index > row_count
			loop
				from
					column_index := 1
				invariant
					column_index_at_least_one: column_index >= 1
					column_index_small_enough: column_index <= 1 + column_count
				variant
					column_index_increasing: 1 + column_count - column_index
				until
					column_index > column_count
				loop
					if row_index > 1 and column_index = 1 then
						ios.put_character (' ')
					end
					ios.put_character (' ')
					ios.put_string (item (row_index, column_index).out)
					column_index := column_index + 1
				end

				row_index := row_index + 1

				if row_index <= row_count then
					ios.put_new_line
				end
			end

			ios.put_character (' ')
			ios.put_character (Closing_out_delimiter)
		end

feature {NONE} -- Implementation

	Opening_out_delimiter: CHARACTER is
			-- First character of result of `out'.
		do
			Result := '['
		end

	Closing_out_delimiter: CHARACTER is
			-- Second character of result of `out'.
		do
			Result := ']'
		end

	check_list (a: DOUBLE_ARRAYED_LIST): BOOLEAN is
			-- See if the elements in `a' are the same as `Current'
		local
			i64: INTEGER_64
			i, j: INTEGER
		do
			from
				i := 1
				i64 := 1
				Result := true
			invariant
				i_at_least_one: i >= 1
				i_small_enough: i <= 1 + row_count
			variant
				i_increasing: 1 + row_count - i
			until
				i > row_count
			loop
				from
					j := 1 
				invariant
					j_at_least_one: j >= 1
					j_small_enough: j <= 1 + column_count
				variant
					j_increasing: 1 + column_count - j
				until
					j > column_count
				loop
					Result := Result and about_equal (a.item (i64), item (i, j))
 
					j := j + 1
					i64 := i64 + 1
				end
				i := i + 1
			end
		end
			
invariant

	neutral_addition: all_finite implies Current.is_equal (Current + zero)
	self_subtraction: all_finite implies zero.is_equal (Current - Current)
	unary_plus_neutral: all_finite implies Current.is_equal (+Current)
	self_negation: all_finite implies zero.is_equal (-Current + Current)
	neutral_multiplication: all_finite implies Current.is_equal (Current * one)
--- The above assertions probably need to be modified to handle nan/inf.

	strictly_positive_column_count: column_count >= 1
	strictly_positive_row_count: row_count >= 1
	frobenius_norm_non_negative: all_finite implies not (frobenius_norm < 0.0)
	two_norm_is_frobenius_norm: about_equal (two_norm, frobenius_norm)
	one_norm_non_negative: all_finite implies not (one_norm < 0.0)
	max_norm_non_negative: all_finite implies not (max_norm < 0.0)

end -- deferred class MATRIX
