indexing

	description: "Geometric points in n-dimensional real cartesian space."
	names: "nd", "geometric", "point", "cartesian", "vertex"
	representation: "DOUBLE"
	access: "fixed"
	size: "fixed"
	contents: "DOUBLE"

deferred class

	POINT

inherit

	LINEAR_TUPLE
		export
			{POINT} one_norm, max_norm, two_norm, magnitude, to_real,
			magnitude_squared, unit, is_unit, normalize, infix "#", dot
		end

feature -- Basic operations

	infix "*" (other: DOUBLE): like Current is
			-- Product by `other'.
			-- (Not applicable in general.)
		do
			Result := scale (other)
		end

	infix "+" (other: VECTOR): like Current is
			-- Sum with `other'.
		require else
			other_non_void: other /= Void
			same_dimension: count = other.count
		deferred
		ensure then
			non_void_result: Result /= Void
			first_item_is_sum:
				Result.item (1) = filter_nan (item (1) + other.item (1))
			last_item_is_sum:
				Result.item (count) =
					filter_nan (item (count) + other.item (count))
			--- Ditto for items in-between.
		end

	infix "-" (other: like Current): VECTOR is
			-- Difference with `other'.
		require else
			other_non_void: other /= Void
			same_dimension: count = other.count
		deferred
		ensure then
			non_void_result: Result /= Void
			first_item_is_difference:
				Result.item (1) = filter_nan (item (1) - other.item (1))
			last_item_is_difference:
				Result.item (count) =
					filter_nan (item (count) - other.item (count))
			--- Ditto for items in-between.
		end

	distance_squared (other: like Current): DOUBLE is
			-- Square of Euclidean distance to `other'.
		do
			Result := (Current - other).magnitude_squared
		ensure
			non_negative_result: Result >= 0.0
		end

	distance (other: like Current): DOUBLE is
			-- Euclidean distance to `other'.
		do
			Result := (Current - other).magnitude
		ensure
			correct_result:
				about_equal (Result, square_root (distance_squared (other)))
		end

	interpolate (other: like Current; parameter: DOUBLE): like Current is
			-- Point that lies `parameter' distance along the vector
			-- from `Current' towards `other'.
		local
			one_minus_parameter, first, second, third: DOUBLE
			index: INTEGER
			is_unit_parameter: BOOLEAN
		do
			Result := Current.deep_twin

			if parameter /= 0.0 then
			one_minus_parameter := 1.0 - parameter
			is_unit_parameter := absolute_value (parameter) < 1.0

			from
				index := 1
			invariant
				index_at_least_one: index >= 1
				index_small_enough: index <= 1 + count
			variant
				index_increasing: 1 + count - index
			until
				index > count
			loop
				first := item (index)
				second := other.item (index)
				third :=  filter_nan (one_minus_parameter * first +
										parameter * second)

				if is_unit_parameter then
						-- Handle underflow by clamping:
					third := clamp_to_range (third,
											first.min (second),
											first.max (second))
				end

				Result.put (third, index)
				index := index + 1
				end
			end
		end

	midpoint (other: like Current): like Current is
			-- Mid-point along line between points `Current' and `other'.
		do
			Result := interpolate (other, 0.5)
		ensure
			result_item_1_large_enough:
				Result.item (1) >= item (1).min (other.item (1))
			result_item_1_small_enough:
				Result.item (1) <= item (1).max (other.item (1))
			-- Ditto for all items.
		end

feature {NONE} -- Implementation

	Opening_out_delimiter: CHARACTER is '('
			-- First character of result of `out'.

	Closing_out_delimiter: CHARACTER is ')'
			-- Second character of result of `out'.

end -- deferred class POINT
