indexing

	description: "Graphical transforms implemented as a 4 x 4 real matrix."
	names: "matrix", "transform"
	representation: "expanded", "DOUBLE"
	access: "fixed"
	size: "fixed"
	contents: "DOUBLE"
	date: "$Date: 2006-06-23 10:34:45 -0400 (Fri, 23 Jun 2006) $"
	revision: "$Revision: 226 $"
	author: "Mark Bolstad, US EPA/LM, bolstad.mark@epa.gov"
	status: "unreviewed", "untested"

class

	TRANSFORM

inherit

	MATRIX4
		rename
			scale as matrix_scale
		export
			{TRANSFORM} zero
		redefine
			default_create, one, zero, is_equal, inverse, matrix_scale,
			infix "+", infix "-", infix "*", infix "^",
			prefix "+", prefix "-", invert
		end
	
creation
	make, default_create, make_identity, make_with_elements
	
feature -- Initialization

	make, default_create is
			-- 
		do
			make_identity
		end

feature -- Access
	
	one: TRANSFORM is
			-- Neutral element for "*".
		once
			create Result.make_identity
		end

feature -- Measurement
feature -- Comparison

	is_equal (other: TRANSFORM): BOOLEAN is
			-- Is each corresponding `item' within `Tolerance'?
		do
			Result := (about_equal (a11, other.a11) and then
						about_equal (a12, other.a12) and then
						about_equal (a13, other.a13) and then
						about_equal (a14, other.a14) and then

						about_equal (a21, other.a21) and then
						about_equal (a22, other.a22) and then
						about_equal (a23, other.a23) and then
						about_equal (a24, other.a24) and then

						about_equal (a31, other.a31) and then
						about_equal (a32, other.a32) and then
						about_equal (a33, other.a33) and then
						about_equal (a34, other.a34) and then

						about_equal (a41, other.a41) and then
						about_equal (a42, other.a42) and then
						about_equal (a43, other.a43) and then
						about_equal (a44, other.a44))
		end

feature -- Status report
	
	is_identity: BOOLEAN is
			-- Is this transform the identity?
		do
			Result := is_equal (one)
		end 
	
	is_pure_rotation: BOOLEAN is
			-- Does `Current' consist of only rotations?
		local
			tmp: TRANSFORM
		do
			if all_finite then
				if about_equal (a14, 0) and about_equal (a41, 0) and 
					about_equal (a24, 0) and about_equal (a42, 0) and 
					about_equal (a34, 0) and about_equal (a43, 0) and 
					about_equal (a44, 1) then
					tmp := Current.twin
					tmp.transpose
					if one.is_equal (tmp * Current) then
						check
							determinant_is_one: about_equal (determinant, 1)
						end
						Result := true
					end
				end
			end
			
		end 

feature -- Status setting

	reset is
			-- Reset transfom
		do
			make_identity
		ensure
			Identity: is_identity
		end

feature -- Cursor movement
feature -- Element change
	
	rotate (angle: DOUBLE; axis: VECTOR3) is
			-- Apply the rotation, `angle' (in radians) to Current.
		require
			axis_normalized: axis.is_unit
		local
			xform: TRANSFORM
			sin_n, cos_n: DOUBLE
		do
			cos_n := cosine (angle)
			sin_n := sine (angle)
			create xform
			xform.create_rotation_matrix (angle, axis, sin_n, cos_n)

			xform := Current * xform
			Current.copy (xform)
			
			update_required := true
		end

	translate (x, y, z: DOUBLE) is
			-- Apply a translation to the current transformation.
		local
			xform: TRANSFORM
		do
			create xform
			xform.put (x, 1, 4)
			xform.put (y, 2, 4)
			xform.put (z, 3, 4)

			xform := Current * xform
			Current.copy (xform)
			
			update_required := true
		end

	scale (x, y, z: DOUBLE) is
			-- Apply a scale to the current transform.
		local
			xform: TRANSFORM
		do
			create xform
			xform.put (x, 1, 1)
			xform.put (y, 2, 2)
			xform.put (z, 3, 3)
			
			xform := Current * xform
			Current.copy (xform)
			
			update_required := true
		end
	
	skew (angle: DOUBLE; v1, v2: VECTOR3) is
		-- Apply a skew to the current transform. This operation 
		-- shifts all points along lines parallel to `v2'. Points 
		-- along `v1' are mapped onto another vector vv, where 
			-- `angle' is the angle between `v1' and vv.
		local
			right: VECTOR3
			r, s, m: TRANSFORM
			axis_angle, par, perp, ss: DOUBLE
		do
			v1.normalize
			v2.normalize
			axis_angle := arc_cosine (v1 # v2)
			
			if angle >= axis_angle or angle <= (axis_angle-pi) then
			else
				right := v1 * v2
				right.normalize
				
				create r.make_with_elements (right.x, right.y, right.z, 0,
									  v1.x,	   v1.y,	v1.z,	 0,
									  v1.x,	   v1.y,	v1.z,	 0,
									  0,	   0,		0,		 1)
				
				par := v1 # v2
				perp := sqrt (1 - par * par)
				ss := tangent (angle + arc_cosine (perp)) * perp - par
				create s
				s.put (ss, 3, 1)
				
				m := r.inverse * s * r
				
				r := m * Current
				Current.copy (r)
			end
			
			update_required := true
		end
	
feature -- Removal
feature -- Resizing
feature -- Transformation
	
	inverse: TRANSFORM is
			-- Attempt to invert `Current'.
			-- If successful (matrix was non-singular), `inverted' is True,
			-- otherwise `inverted' is False and the matrix is unchanged.
		do
			if update_required then
				cached_inverse := Current.twin
				cached_inverse.invert
				update_required := false
			end
			
			Result := cached_inverse
		end
	
	invert is
			-- Attempt to invert `Current'.
			-- If successful (matrix was non-singular), `inverted' is True,
			-- otherwise `inverted' is False and the matrix is unchanged.
		local
			the_determinant: DOUBLE
		do
			-- If not an affine matrix call the slow way
			if not (about_equal (a41, 0.0) and about_equal (a42, 0.0) and
					about_equal (a43, 0.0) and about_equal (a44, 1.0)) then
				Precursor
			else
				the_determinant := determinant
				
				if the_determinant /= 0.0 then
					
					the_determinant := 1.0 / the_determinant
					
					if the_determinant /= 0.0 then
						-- Determinant wasn't infinite.
						affine_adjugate (the_determinant)
						inverted := true
						update_required := true
						
						check
							inverted: all_finite and then not is_singular
						end
					
					end
				end
			end
		end
	
feature -- Conversion
feature -- Duplication
feature -- Basic operations
	
	concatonate_transform (xform: TRANSFORM) is
			-- Append `xform' to `Current'
		require 
			xform_not_void: xform /= Void
		local
			tmp: TRANSFORM
		do
			tmp := Current * xform 
			Current.copy (tmp)
			
			update_required := true
		end 

	infix "*" (o: TRANSFORM): TRANSFORM is
			-- Product by `o'.
		do
			create Result.make_with_elements (
						a11 * o.a11 + a12 * o.a21 + a13 * o.a31 + a14 * o.a41,
						a11 * o.a12 + a12 * o.a22 + a13 * o.a32 + a14 * o.a42,
						a11 * o.a13 + a12 * o.a23 + a13 * o.a33 + a14 * o.a43,
						a11 * o.a14 + a12 * o.a24 + a13 * o.a34 + a14 * o.a44,
						a21 * o.a11 + a22 * o.a21 + a23 * o.a31 + a24 * o.a41,
						a21 * o.a12 + a22 * o.a22 + a23 * o.a32 + a24 * o.a42,
						a21 * o.a13 + a22 * o.a23 + a23 * o.a33 + a24 * o.a43,
						a21 * o.a14 + a22 * o.a24 + a23 * o.a34 + a24 * o.a44,
						a31 * o.a11 + a32 * o.a21 + a33 * o.a31 + a34 * o.a41,
						a31 * o.a12 + a32 * o.a22 + a33 * o.a32 + a34 * o.a42,
						a31 * o.a13 + a32 * o.a23 + a33 * o.a33 + a34 * o.a43,
						a31 * o.a14 + a32 * o.a24 + a33 * o.a34 + a34 * o.a44,
						a41 * o.a11 + a42 * o.a21 + a43 * o.a31 + a44 * o.a41,
						a41 * o.a12 + a42 * o.a22 + a43 * o.a32 + a44 * o.a42,
						a41 * o.a13 + a42 * o.a23 + a43 * o.a33 + a44 * o.a43,
						a41 * o.a14 + a42 * o.a24 + a43 * o.a34 + a44 * o.a44)
		end

	infix "+" (other: TRANSFORM): TRANSFORM is
			-- Sum with `other' (commutative).
		do
			create Result.make_with_elements (
						filter_nan (a11 + other.a11),
						filter_nan (a12 + other.a12),
						filter_nan (a13 + other.a13),
						filter_nan (a14 + other.a14),
						filter_nan (a21 + other.a21),
						filter_nan (a22 + other.a22),
						filter_nan (a23 + other.a23),
						filter_nan (a24 + other.a24),
						filter_nan (a31 + other.a31),
						filter_nan (a32 + other.a32),
						filter_nan (a33 + other.a33),
						filter_nan (a34 + other.a34),
						filter_nan (a41 + other.a41),
						filter_nan (a42 + other.a42),
						filter_nan (a43 + other.a43),
						filter_nan (a44 + other.a44))
		ensure then
			result_a11_is_sum: about_equal (Result.a11,
											filter_nan (a11 + other.a11))
			result_a12_is_sum: about_equal (Result.a12, 
											filter_nan (a12 + other.a12))
			result_a13_is_sum: about_equal (Result.a13, 
											filter_nan (a13 + other.a13))
			result_a14_is_sum: about_equal (Result.a14, 
											filter_nan (a14 + other.a14))
			result_a21_is_sum: about_equal (Result.a21, 
											filter_nan (a21 + other.a21))
			result_a22_is_sum: about_equal (Result.a22, 
											filter_nan (a22 + other.a22))
			result_a23_is_sum: about_equal (Result.a23, 
											filter_nan (a23 + other.a23))
			result_a24_is_sum: about_equal (Result.a24, 
											filter_nan (a24 + other.a24))
			result_a31_is_sum: about_equal (Result.a31, 
											filter_nan (a31 + other.a31))
			result_a32_is_sum: about_equal (Result.a32, 
											filter_nan (a32 + other.a32))
			result_a33_is_sum: about_equal (Result.a33, 
											filter_nan (a33 + other.a33))
			result_a34_is_sum: about_equal (Result.a34, 
											filter_nan (a34 + other.a34))
			result_a41_is_sum: about_equal (Result.a41, 
											filter_nan (a41 + other.a41))
			result_a42_is_sum: about_equal (Result.a42, 
											filter_nan (a42 + other.a42))
			result_a43_is_sum: about_equal (Result.a43, 
											filter_nan (a43 + other.a43))
			result_a44_is_sum: about_equal (Result.a44, 
											filter_nan (a44 + other.a44))
		end

	infix "-" (other: TRANSFORM): TRANSFORM is
			-- Result of subtracting `other'.
		do
			create Result.make_with_elements (
						filter_nan (a11 - other.a11),
						filter_nan (a12 - other.a12),
						filter_nan (a13 - other.a13),
						filter_nan (a14 - other.a14),
						filter_nan (a21 - other.a21),
						filter_nan (a22 - other.a22),
						filter_nan (a23 - other.a23),
						filter_nan (a24 - other.a24),
						filter_nan (a31 - other.a31),
						filter_nan (a32 - other.a32),
						filter_nan (a33 - other.a33),
						filter_nan (a34 - other.a34),
						filter_nan (a41 - other.a41),
						filter_nan (a42 - other.a42),
						filter_nan (a43 - other.a43),
						filter_nan (a44 - other.a44))
		ensure then
			result_a11_is_difference: about_equal (Result.a11, 
												   filter_nan (a11 - other.a11))
			result_a12_is_difference: about_equal (Result.a12, 
												   filter_nan (a12 - other.a12))
			result_a13_is_difference: about_equal (Result.a13, 
												   filter_nan (a13 - other.a13))
			result_a14_is_difference: about_equal (Result.a14, 
												   filter_nan (a14 - other.a14))
			result_a21_is_difference: about_equal (Result.a21, 
												   filter_nan (a21 - other.a21))
			result_a22_is_difference: about_equal (Result.a22, 
												   filter_nan (a22 - other.a22))
			result_a23_is_difference: about_equal (Result.a23, 
												   filter_nan (a23 - other.a23))
			result_a24_is_difference: about_equal (Result.a24, 
												   filter_nan (a24 - other.a24))
			result_a31_is_difference: about_equal (Result.a31, 
												   filter_nan (a31 - other.a31))
			result_a32_is_difference: about_equal (Result.a32, 
												   filter_nan (a32 - other.a32))
			result_a33_is_difference: about_equal (Result.a33, 
												   filter_nan (a33 - other.a33))
			result_a34_is_difference: about_equal (Result.a34, 
												   filter_nan (a34 - other.a34))
			result_a41_is_difference: about_equal (Result.a41, 
												   filter_nan (a41 - other.a41))
			result_a42_is_difference: about_equal (Result.a42, 
												   filter_nan (a42 - other.a42))
			result_a43_is_difference: about_equal (Result.a43, 
												   filter_nan (a43 - other.a43))
			result_a44_is_difference: about_equal (Result.a44, 
												   filter_nan (a44 - other.a44))
		end

	infix "^" (exponent: NUMERIC): like Current is
			-- Current raised to the power `exponent'.
		do
				check
					implemented: False
				end
		end

	prefix "+": TRANSFORM is
			-- Unary plus.
		do
---			Result := Current
			create Result.make_with_elements (a11, a12, a13, a14,
										a21, a22, a23, a24,
										a31, a32, a33, a34,
										a41, a42, a43, a44)
		end

	prefix "-": TRANSFORM is
			-- Unary minus
		do
			create Result.make_with_elements (-a11, -a12, -a13, -a14,
										-a21, -a22, -a23, -a24,
										-a31, -a32, -a33, -a34,
										-a41, -a42, -a43, -a44)
		ensure then
			result_a11_is_negated: about_equal (result.a11, -a11)
			result_a12_is_negated: about_equal (result.a12, -a12)
			result_a13_is_negated: about_equal (result.a13, -a13)
			result_a14_is_negated: about_equal (result.a14, -a14)
			result_a21_is_negated: about_equal (result.a21, -a21)
			result_a22_is_negated: about_equal (result.a22, -a22)
			result_a23_is_negated: about_equal (result.a23, -a23)
			result_a24_is_negated: about_equal (result.a24, -a24)
			result_a31_is_negated: about_equal (result.a31, -a31)
			result_a32_is_negated: about_equal (result.a32, -a32)
			result_a33_is_negated: about_equal (result.a33, -a33)
			result_a34_is_negated: about_equal (result.a34, -a34)
			result_a41_is_negated: about_equal (result.a41, -a41)
			result_a42_is_negated: about_equal (result.a42, -a42)
			result_a43_is_negated: about_equal (result.a43, -a43)
			result_a44_is_negated: about_equal (result.a44, -a44)
		end

	matrix_scale (scalar: DOUBLE): TRANSFORM is
			-- Result of multiplying each element by `scalar'.
		do
			create Result.make_with_elements (
						a11 * scalar, a12 * scalar, a13 * scalar, a14 * scalar,
						a21 * scalar, a22 * scalar, a23 * scalar, a24 * scalar,
						a31 * scalar, a32 * scalar, a33 * scalar, a34 * scalar,
						a41 * scalar, a42 * scalar, a43 * scalar, a44 * scalar)
		ensure then
			result_a11_is_scaled: about_equal (result.a11, a11 * scalar)
			result_a12_is_scaled: about_equal (result.a12, a12 * scalar)
			result_a13_is_scaled: about_equal (result.a13, a13 * scalar)
			result_a14_is_scaled: about_equal (result.a14, a14 * scalar)
			result_a21_is_scaled: about_equal (result.a21, a21 * scalar)
			result_a22_is_scaled: about_equal (result.a22, a22 * scalar)
			result_a23_is_scaled: about_equal (result.a23, a23 * scalar)
			result_a24_is_scaled: about_equal (result.a24, a24 * scalar)
			result_a31_is_scaled: about_equal (result.a31, a31 * scalar)
			result_a32_is_scaled: about_equal (result.a32, a32 * scalar)
			result_a33_is_scaled: about_equal (result.a33, a33 * scalar)
			result_a34_is_scaled: about_equal (result.a34, a34 * scalar)
			result_a41_is_scaled: about_equal (result.a41, a41 * scalar)
			result_a42_is_scaled: about_equal (result.a42, a42 * scalar)
			result_a43_is_scaled: about_equal (result.a43, a43 * scalar)
			result_a44_is_scaled: about_equal (result.a44, a44 * scalar)
		end

	apply (pt: POINT3): POINT3 is
			-- Apply the current transformation.
		require
			pt_not_void: pt /= Void
		local
			vx, vy, vz, sx, sy, sz, sw: DOUBLE
		do
			vx := pt.x
			vy := pt.y
			vz := pt.z
			sx := a11 * vx + a12 * vy + a13 * vz + a14
			sy := a21 * vx + a22 * vy + a23 * vz + a24
			sz := a31 * vx + a32 * vy + a33 * vz + a34

				-- For perspective and uniform scaling
			
			sw := a41 * vx + a42 * vy + a43 * vz + a44

			if sw /= 1.0 and then sw /= 0.0 then
				sw := 1.0 / sw
				sx := sx * sw
				sy := sy * sw
				sz := sz * sw
			end

			create Result.make (sx, sy, sz)
		end

	apply_as_direction (v: VECTOR3): VECTOR3 is
			-- Apply the current transformation. Treat `v' as a direction.
		require
			v_not_void: v /= Void
		local
			vx, vy, vz, sx, sy, sz, sw: DOUBLE
		do
			vx := v.x
			vy := v.y
			vz := v.z
			sx := a11 * vx + a12 * vy + a13 * vz
			sy := a21 * vx + a22 * vy + a23 * vz
			sz := a31 * vx + a32 * vy + a33 * vz

				-- For perspective and uniform scaling
			
			sw := a41 * vx + a42 * vy + a43 * vz

			if sw /= 1.0 and then sw /= 0.0 then
				sw := 1.0 / sw
				sx := sx * sw
				sy := sy * sw
				sz := sz * sw
			end

			create Result.make (sx, sy, sz)
		end

	apply_nan (pt: POINT3): POINT3 is
			-- Apply the current transformation, removing NaNs if necessary.
		require
			pt_not_void: pt /= Void
		local
			vx, vy, vz, sx, sy, sz, sw: DOUBLE
		do
			vx := pt.x
			vy := pt.y
			vz := pt.z
			-- sx := a11 * vx + a12 * vy + a13 * vz + a14
			sx := sum (sum (product (a11, vx), product (a12, vy)),
						sum (product (a13, vz), a14))
			sy := sum (sum (product (a21, vx), product (a22, vy)),
						sum (product (a23, vz), a24))
			sz := sum (sum (product (a31, vx), product (a32, vy)),
						sum (product (a33, vz), a34))

				-- For perspective and uniform scaling

			sw := sum (sum (product (a41, vx), product (a42, vy)),
						sum (product (a43, vz), a44))

			if sw /= 1.0 and then sw /= 0.0 then
				sw := 1.0 / sw
				sx := product (sx, sw)
				sy := product (sy, sw)
				sz := product (sz, sw)
			end

			create Result.make (sx, sy, sz)
		end

	apply_as_direction_removing_nan (v: VECTOR3): VECTOR3 is
			-- Apply the current transformation. Treat `v' as a direction.
		require
			v_not_void: v /= Void
		local
			vx, vy, vz, sx, sy, sz, sw: DOUBLE
		do
			vx := v.x
			vy := v.y
			vz := v.z
			-- sx := a11 * vx + a12 * vy + a13 * vz 
			sx := sum (sum (product (a11, vx), product (a12, vy)),
						product (a13, vz))
			sy := sum (sum (product (a21, vx), product (a22, vy)),
						product (a23, vz))
			sz := sum (sum (product (a31, vx), product (a32, vy)),
						product (a33, vz))

				-- For perspective and uniform scaling

			sw := sum (sum (product (a41, vx), product (a42, vy)),
						product (a43, vz))

			if sw /= 1.0 and then sw /= 0.0 then
				sw := 1.0 / sw
				sx := product (sx, sw)
				sy := product (sy, sw)
				sz := product (sz, sw)
			end

			create Result.make (sx, sy, sz)
		end

feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {TRANSFORM} -- Implementation

	zero: TRANSFORM is
			-- Neutral element for "+" and "-".
		once
			create Result
			Result.fill (0.0)
		end

	create_rotation_matrix (angle: DOUBLE;
							axis: VECTOR3; 
							sin_t, cos_t: DOUBLE) is
			-- Apply the rotation to the current transform.
		local
			vx, vy, vz: DOUBLE
		do
			vx := axis.x
			vy := axis.y
			vz := axis.z

			a11 := vx * vx + (1.0 - vx * vx) * cos_t
			a12 := vx * vy * (1.0 - cos_t) - vz * sin_t
			a13 := vx * vz * (1.0 - cos_t) + vy * sin_t
			a14 := 0.0

			a21 := vx * vy * (1.0 - cos_t) + vz * sin_t
			a22 := vy * vy + (1.0 - vy * vy) * cos_t
			a23 := vy * vz * (1.0 - cos_t) - vx * sin_t
			a24 := 0.0

			a31 := vx * vz * (1.0 - cos_t) - vy * sin_t
			a32 := vy * vz * (1.0 - cos_t) + vx * sin_t
			a33 := vz * vz + (1.0 - vz * vz) * cos_t
			a34 := 0.0

			a41 := 0.0
			a42 := 0.0
			a43 := 0.0
			a44 := 1.0
			
		end

feature {NONE} -- Implementation
	
	affine_adjugate (scalar: DOUBLE) is
			-- Form the scaled transpose of the cofactor matrix.
		local
			ta11, ta12, ta13, ta14,
			ta21, ta22, ta23, ta24,
			ta31, ta32, ta33, ta34: DOUBLE
		do
			ta11 := a11
			ta12 := a12
			ta13 := a13
			ta14 := a14
			
			ta21 := a21
			ta22 := a22
			ta23 := a23
			ta24 := a24
			
			ta31 := a31
			ta32 := a32
			ta33 := a33
			ta34 := a34
			
				-- Compute the transpose of the cofactors:
			
			a11 := scalar *	 (ta22 * ta33 - ta32 * ta23)
			a12 := scalar * -(ta12 * ta33 - ta32 * ta13)
			a13 := scalar *	 (ta12 * ta23 - ta22 * ta13)

			a21 := scalar * -(ta21 * ta33 - ta31 * ta23)
			a22 := scalar *	 (ta11 * ta33 - ta31 * ta13)
			a23 := scalar * -(ta11 * ta23 - ta21 * ta13)

			a31 := scalar *	 (ta21 * ta32 - ta31 * ta22)
			a32 := scalar * -(ta11 * ta32 - ta31 * ta12)
			a33 := scalar *	 (ta11 * ta22 - ta21 * ta12)
			
			-- Calculate -C * inverse(A)
			a14 := -(ta14 * a11 + ta24 * a12 + ta34 * a13)
			a24 := -(ta14 * a21 + ta24 * a22 + ta34 * a23)
			a34 := -(ta14 * a31 + ta24 * a32 + ta34 * a33)
			
			-- Fill in the `affine' row
			a41 := 0.0
			a42 := 0.0
			a43 := 0.0
			a44 := 1.0
		end

invariant

---	non_singular: not is_singular

end -- class TRANSFORM

