indexing

	description: "Mixin constants and routines for basic numeric types."
	patterns: "mixin"
	date: "$Date: 2006-12-06 13:14:08 -0500 (Wed, 06 Dec 2006) $"
	revision: "$Revision: 271 $"
	status: "reviewed", "tested", "partial", "unimplemented: Maximum_integer"
	"unimplemented: Minimum_integer",
	"unimplemented: next_strictly_positive_integer"

expanded class

	BASIC_NUMERIC

feature {NONE} -- Access

	maximum_integer: INTEGER_64 is
			-- Largest representable 64-bit integer value.
		external
			"C macro use <NumericTypes.h>"
		alias
			"INTEGER_MAX"
		end

	minimum_integer: INTEGER_64 is
						-- Smallest (negative) representable 64-bit integer value.
		external
			"C macro use <NumericTypes.h>"
		alias
			"INTEGER_MIN"
				end

	maximum_double: DOUBLE is
			-- Largest representable 64-bit floating-point value.
		external
			"C macro use <NumericTypes.h>"
		alias
			"REAL_MAX"
		end

	smallest_double: DOUBLE is
			-- Smallest (positive) representable 64-bit floating-point value.
		external
			"C macro use <NumericTypes.h>"
		alias
			"REAL_MIN"
		end
	
	huge_value: DOUBLE is
			-- "Infinity-like" value.
		external
			"C macro use <math.h>"
		alias
			"HUGE_VAL"
		end

	machine_precision: DOUBLE is
			-- Smallest positive value such that x + Machine_precision /= x.
		external
			"C macro use <float.h>"
		alias
			"DBL_EPSILON"
		end

	is_little_endian: BOOLEAN is
			-- Is the platform least-significant-byte-first (LSB)?
		external
			"C macro use <NumericTypes.h>"
		alias
			"IS_LITTLE_ENDIAN"
		end

	is_native_xdr: BOOLEAN is
			-- Is the platform XDR compatible?
		external
			"C macro use <NumericTypes.h>"
		alias
			"IS_NATIVE_XDR"
		end

	pi: DOUBLE is
			-- Pi.
		external
			"C macro use <math.h>"
		alias
			"M_PI"
		end

	pi_over_two: DOUBLE is
			-- Pi / 2.
		external
			"C macro use <math.h>"
		alias
			"M_PI_2"
		end

	pi_over_four: DOUBLE is
			-- Pi / 4.
		external
			"C macro use <math.h>"
		alias
			"M_PI_4"
		end

	Tolerance: DOUBLE is 1.0e-6
			-- Basic recommended tolerance for numerical round-off error.

feature {NONE} -- Comparison

	is_nan (x: DOUBLE): BOOLEAN is
			-- Is `x' a nan (Not A Number)?
		external
			"C use <math.h>"
		alias
			"isnan"
		end

	is_infinity (x: DOUBLE): BOOLEAN is
			-- Is `x' greater than Maximum_double or does it have 0 recipricol?
		do
			Result := is_inf (x) /= 0
		end

	is_minus_infinity (x: DOUBLE): BOOLEAN is
			-- Is `x' less than -Maximum_double or does it have 0 recipricol?
		do
			if x /= 0.0 then
				Result := x < -maximum_double or else 1.0 / x = 0.0
			end
		end

	filter_nan (x: DOUBLE): DOUBLE is
			-- If `x' is a nan (Not A Number) result is 0.0 else result is `x'.
		do
			if not is_nan (x) then
				Result := x
			end
		end

	is_finite (x: DOUBLE): BOOLEAN is
			-- Is `x' not nan or +/-infinity?
		external
			"C use <math.h>"
		alias
			"finite"
		end

	about_equal (a, b: DOUBLE): BOOLEAN is
			-- Do `a' and `b' differ by not more than `Tolerance'?
		do
---			Result := (a - b).abs <= Tolerance
			Result := within_tolerance (a, b, Tolerance)
		end

	within_tolerance (a, b, epsilon: DOUBLE): BOOLEAN is
			-- Do `a' and `b' differ by not more than `epsilon'?
		require
			non_negative_epsilon: epsilon >= 0.0
		do
			if is_infinity (a) and is_infinity (b)	then
				Result := true		
			else	
				Result := absolute_value (a - b) <= epsilon
			end
--		external
--			"C use <NumericTypes.h>"
--		alias
--			"withinTolerance"
		end

	within_range (x, lower, upper: DOUBLE): BOOLEAN is
			-- Is `x' within [`lower', `upper']?
		require
			ordered: lower <= upper
		do
			Result := x >= lower and then x <= upper
		ensure
			result_correct: Result = (x >= lower and x <= upper)
		end

feature {NONE} -- Conversion

	next_strictly_positive_integer (x: INTEGER): INTEGER is
			-- `x' + 1 but skipping over negative and 0 values.
		external
			"C macro use <stddef.h>"
		alias
			"NEXT_STRICTLY_POSITIVE_INTEGER"
		ensure
			result_strictly_positive: Result > 0
		end

	truncate (x: DOUBLE): INTEGER is
			-- Truncate `x' to its integer part.
		external
			"C macro use <stddef.h>"
		alias
			" "
		end

	ceiling (x: DOUBLE): DOUBLE is
			-- Smallest integral value not less than argument
		external
			"C macro use <stddef.h>"
		alias
			"ceil"
		end

	radians (the_degrees: DOUBLE): DOUBLE is
			-- Convert degrees to radians.
		do
			Result := the_degrees * (pi / 180.0)
		ensure
			result_is_same_sign_or_zero:
				Result.sign = the_degrees.sign or else Result = 0.0
		end

	degrees (the_radians: DOUBLE): DOUBLE is
			-- Convert radians to degrees.
		do
			Result := the_radians * (180.0 / pi)
		ensure
			result_is_same_sign_or_zero:
				Result.sign = the_radians.sign or else Result = 0.0
		end

	clamp_to_range (x, lower, upper: DOUBLE): DOUBLE is
			-- Value within range or nearest endpoint.
		require
			non_nan_x: not is_nan (x)
			ordered: lower <= upper
		do
			if x < lower then
				Result := lower
			elseif x > upper then
				Result := upper
			else
				Result := x
			end
		ensure
			result_in_range: lower <= Result and Result <= upper
		end

	scalar_range (data: ARRAY [DOUBLE]): PAIR [DOUBLE, DOUBLE] is
			-- Minimum and maximum of scalar `data'.
		require
			argument_exists: data /= Void
		local
			index: INTEGER
			data_value, minimum, maximum: DOUBLE
		do
			from
				minimum := data.item (1)
				maximum := minimum
				index := 2
			invariant
				index_large_enough: index >= 2
				index_small_enough: index <= 1 + data.count
			variant
				1 + data.count - index
			until
				index > data.count
			loop
				data_value := data.item (index)

				if data_value < minimum then
					minimum := data_value
				elseif data_value > maximum then
					maximum := data_value
				end

				index := index + 1
			end

			Result.make (minimum, maximum)
		ensure
			result_ordered: Result.first <= Result.second
			result_small_enough: Result.first <= data.item (1) and
									Result.first <= data.item (data.count)
			result_large_enough: Result.second >= data.item (1) and
									Result.second >= data.item (data.count)
		end

	vector_range (data: ARRAY [DOUBLE]): PAIR [VECTOR3, VECTOR3] is
			-- Minimum and maximum of 3d vector `data'.
		require
			argument_exists: data /= Void
			argument_count_at_least_three: data.count >= 3
			argument_count_is_multiple_of_three: data.count \\ 3 = 0
		local
			index: INTEGER
			data_value: DOUBLE
			minimum1, maximum1, minimum2, maximum2, minimum3, maximum3: DOUBLE
			minimum_vector, maximum_vector: VECTOR3
		do
			from
				minimum1 := data.item (1)
				minimum2 := data.item (2)
				minimum3 := data.item (3)
				maximum1 := minimum1
				maximum2 := minimum2
				maximum3 := minimum3
				index := 4
			invariant
				index_large_enough: index >= 4
				index_small_enough: index <= 1 + data.count
			variant
				1 + data.count - index
			until
				index > data.count
			loop
				data_value := data.item (index)

				if data_value < minimum1 then
					minimum1 := data_value
				elseif data_value > maximum1 then
					maximum1 := data_value
				end

				index := index + 1
				data_value := data.item (index)

				if data_value < minimum2 then
					minimum2 := data_value
				elseif data_value > maximum2 then
					maximum2 := data_value
				end

				index := index + 1
				data_value := data.item (index)

				if data_value < minimum3 then
					minimum3 := data_value
				elseif data_value > maximum3 then
					maximum3 := data_value
				end

				index := index + 1
			end

			minimum_vector.make (minimum1, minimum2, minimum3)
			maximum_vector.make (maximum1, maximum2, maximum3)
			Result.make (minimum_vector, maximum_vector)
		ensure
			result_ordered: Result.first.x <= Result.second.x and
							Result.first.y <= Result.second.y and
							Result.first.z <= Result.second.z
			result_small_enough:
				Result.first.x <= data.item (1) and
				Result.first.x <= data.item (data.count - 2) and
				Result.first.y <= data.item (2) and
				Result.first.y <= data.item (data.count - 1) and
				Result.first.z <= data.item (3) and
				Result.first.z <= data.item (data.count)
			result_large_enough:
				Result.second.x >= data.item (1) and
				Result.second.x >= data.item (data.count - 2) and
				Result.second.y >= data.item (2) and
				Result.second.y >= data.item (data.count - 1) and
				Result.second.z >= data.item (3) and
				Result.second.z >= data.item (data.count)
		end

	ceiling2 (numerator, denominator: INTEGER): INTEGER is
			-- Smallest integer >= (numerator / denominator).
		require
			non_zero_denominator: denominator /= 0
		local
			negative: BOOLEAN
			positive_numerator, positive_denominator: INTEGER
			the_quotient, remainder: INTEGER
		do
			if numerator /= 0 then

				if numerator < 0 then
					negative := not negative
					positive_numerator := -numerator
				else
					positive_numerator := numerator
				end

				if denominator < 0 then
					negative := not negative
					positive_denominator := -denominator
				else
					positive_denominator := denominator
				end

				the_quotient := positive_numerator // positive_denominator
				remainder := positive_numerator \\ positive_denominator
				Result := the_quotient

				if remainder /= 0 then
					Result := Result + 1
				end

				if negative then
					Result := Result - 1
					Result := -Result
				end
			end
		end

	nop (arg: POINTER): POINTER is
			-- Return arg.
		external
			"C macro use <stddef.h>"
		alias
			" "
		end

feature {NONE} -- Basic operations

	sum (x, y: DOUBLE): DOUBLE is
			-- Sum of x and y (handling +/-inf).
		require
			first_argument_not_nan: not is_nan (x)
			second_argument_not_nan: not is_nan (y)
		do
			if x = -y then
				Result := 0.0
			else
				Result := x + y
			end
		ensure
			non_nan_result: not is_nan (Result)
		end

	difference (x, y: DOUBLE): DOUBLE is
			-- Difference of x and y (handling +/-inf).
		require
			first_argument_not_nan: not is_nan (x)
			second_argument_not_nan: not is_nan (y)
		do
			if x = y then
				Result := 0.0
			else
				Result := x - y
			end
		ensure
			non_nan_result: not is_nan (Result)
		end

	product (x, y: DOUBLE): DOUBLE is
			-- Product of x and y (handling +/-inf * 0).
		require
			first_argument_not_nan: not is_nan (x)
			second_argument_not_nan: not is_nan (y)
		do
			if x /= 0.0 and then y /= 0.0 then
				Result := x * y
			end
		ensure
			non_nan_result: not is_nan (Result)
		end

	quotient (x, y: DOUBLE): DOUBLE is
			-- Quotient of x and y (handling +/-inf).
		require
			first_argument_not_nan: not is_nan (x)
			second_argument_not_nan: not is_nan (y)
			second_argument_not_zero: y /= 0.0
		do
			if x = y then
				Result := 1.0
			elseif x = -y then
				Result := -1.0
			else
				Result := x / y
			end
		ensure
			non_nan_result: not is_nan (Result)
		end

	square_root (x: DOUBLE): DOUBLE is
			-- Square-root of `x'.
		require
			non_negative_argument: x >= 0.0
		external
			"C use <math.h>"
		alias
			"sqrt"
		end

	square (x: DOUBLE): DOUBLE is
			-- `x' * `x'.
		do
			Result := x * x
		end

	cube_root (x: DOUBLE): DOUBLE is
			-- Cube-root of `x'
		external
			"C use <math.h>"
		alias
			"cbrt"
		end

	cube (x: DOUBLE): DOUBLE is
			-- `x' * `x' * `x'.
		do
			Result := x * x * x
		end

	power_of_two (exponent: INTEGER): INTEGER is
			-- 2^exponent.
		require
			exponent_non_negative: exponent >= 0
			exponent_small_enough: exponent <= 31
		local
			iteration: INTEGER
		do
			from
				iteration := 1
				Result := 1
			until
				iteration > exponent
			loop
				Result := Result + Result
				iteration := iteration + 1
			end
		ensure
			result_is_strictly_positive: Result >= 1
			result_is_correct:
				(exponent = 0 and Result = 1) or
				(exponent > 0 and Result \\ 2 = 0)
				-- log2 (Result) = exponent.
		end

	sign_of (x: DOUBLE): DOUBLE is
			-- -1.0 if x < 0.0 else 1.0.
		do
			if x < 0.0 then
				Result := -1.0
			else
				Result := 1.0
			end
		end

	absolute_value (x: DOUBLE): DOUBLE is
			-- |`x'|.
		external
			"C use <math.h>"
		alias
			"fabs"
		end

	integer_absolute_value (x: INTEGER): INTEGER is
			-- |`x'|.
		do
			if x >= 0 then
				Result := x
			else
				Result := -x
			end

			if Result < 0 then
					-- Convert maximum magnitude negative integer to
					-- maximum positive integer.
				Result := -(Result + 1)
			end
		ensure
			non_negative: Result >= 0
		end

	hypotenuse (x, y: DOUBLE): DOUBLE is
			-- Length of hypotenuse of right triangle with sides `x' and `y'.
		require
			positive_x: x > 0.0
			positive_y: y > 0.0
		external
			"C use <math.h>"
		alias
			"hypot"
		end

	arc_tangent2 (y, x: DOUBLE): DOUBLE is
			-- Inverse tangent of `y' / `x'.
		require
			non_zero_x: x /= 0.0
		external
			"C use <math.h>"
		alias
			"atan2"
		end

	power (x, y: DOUBLE): DOUBLE is
			-- x^y
		require
			non_zero_y: y /= 0.0
		external
			"C use <math.h>"
		alias
			"pow"
		end

	sinh (x: DOUBLE): DOUBLE is
			-- Hyperbolic sine of `x'.
		external
			"C use <math.h>"
		end

	cosh (x: DOUBLE): DOUBLE is
			-- Hyperbolic cosine of `x'.
		external
			"C use <math.h>"
		end

	tanh (x: DOUBLE): DOUBLE is
			-- Hyperbolic tangent of `x'.
		external
			"C use <math.h>"
		end

	memory_copy (destination, source: POINTER; byte_count: INTEGER) is
			-- Copy `byte_count' bytes from `source' to `destination'.
		require
			destination_exists: destination /= destination.default
			source_exists: source /= source.default
			byte_count_strictly_positive: byte_count >= 1
		external
			"C use <string.h>"
		alias
			"memcpy"
		end

	uniform_random_value: DOUBLE is
			-- Yield a (48-bit) number in [0, 1] that is from a pseudo-random,
			-- uniformly distributed sequence.
		external
			"C use <stdlib.h>"
		alias
			"drand48"
		ensure
			result_at_least_zero: Result >= 0.0
			result_at_most_one: Result <= 1.0
		end

feature {NONE} -- Advanced operations
	
	binomial_coefficient (n, k: INTEGER): DOUBLE is
			-- Compute a binomial coefficient
		require
			k_positive: k > 0
			n_greater_than_k: n >= k
		do
			Result := 0.5 + exp (ln_of_factorial (k) - ln_of_factorial (n - k))
			Result := truncate (Result)
		end 

	ln_of_factorial (n: INTEGER): DOUBLE is
			-- Natural log of n!
		require
			n_positive: n >= 0
		do
			if n <= 1 then
				Result := 0
			else
				Result := ln_of_gamma (n + 1)
			end
		end
	
	ln_of_gamma (xx: DOUBLE): DOUBLE is
			-- Natural log of gamma function
		external
			"C use <gsl/gsl_sf_gamma.h>"
		alias
			"gsl_sf_lngamma"
		end
	
feature {NONE} -- Implementation

	is_inf (x: DOUBLE): INTEGER is
			-- Is `x' infinity
		external
			"C use <math.h>"
		alias
			"isinf"
		end

	nop_to_thwart_optimization (x: DOUBLE): DOUBLE is
			-- Returns `x' (intended to thwart optimizations that break is_nan).
		do
			Result := x
		end

	exp (x: DOUBLE): DOUBLE is
			-- Exponential of `v'.
		external
			"C use <math.h>"
		end

	log10 (x: DOUBLE): DOUBLE is
			-- Base10 logarithm of `v'.
		external
			"C use <math.h>"
		end

	cosine (v: DOUBLE): DOUBLE is
			-- Trigonometric cosine of radian `v' approximated
			-- in the range [-pi/4, +pi/4]
		external
			"C use <math.h>"
		alias
			"cos"
		end

	arc_cosine (v: DOUBLE): DOUBLE is
			-- Trigonometric arccosine of radian `v'
			-- in the range [0, pi]
		external
			"C use <math.h>"
		alias
			"acos"
		end

	sine (v: DOUBLE): DOUBLE is
			-- Trigonometric sine of radian `v' approximated
			-- in range [-pi/4, +pi/4]
		external
			"C use <math.h>"
		alias
			"sin"
		end

	arc_sine (v: DOUBLE): DOUBLE is
			-- Trigonometric arcsine of radian `v'
			-- in the range [-pi/2, +pi/2]
		external
			"C use <math.h>"
		alias
			"asin"
		end

	tangent (v: DOUBLE): DOUBLE is
			-- Trigonometric tangent of radian `v' approximated
			-- in range [-pi/4, +pi/4]
		external
			"C use <math.h>"
		alias
			"tan"
		end

	arc_tangent (v: DOUBLE): DOUBLE is
			-- Trigonometric arctangent of radian `v'
			-- in the range [-pi/2, +pi/2]
		external
			"C use <math.h>"
		alias
			"atan"
		end

	sqrt (v: DOUBLE): DOUBLE is
			-- Square root of `v'
		external
			"C use <math.h>"
		end

invariant

---	maximum_integer_large_enough: Maximum_integer > 2147483647
---	minimum_integer_small_enough: Minimum_integer < -2147483647 - 1
	maximum_double_large_enough: maximum_double > 1.0e+300
	minimum_double_large_enough: smallest_double > 0.0
	minimum_double_small_enough: smallest_double < 1.0e-300
	huge_value_large_enough:
		(is_native_xdr and huge_value > maximum_double) or else
		(not is_native_xdr and huge_value >= maximum_double)
	machine_precision_large_enough: machine_precision > 0.0
	machine_precision_small_enough: machine_precision <= 1.0e-6
	tolerance_large_enough: Tolerance >= machine_precision
	tolerance_small_enough: Tolerance < 1.0

end -- class BASIC_NUMERIC
