indexing

	description: "Mathematical quaternion types."
	patterns: "none"
	date: "$Date: 2006-12-06 13:14:08 -0500 (Wed, 06 Dec 2006) $"
	revision: "$Revision: 271 $"
	status: "unreviewed", "untested"

--- See /home/plessel/lib/internal/c++/src/libs/Math/Complex/Complex.[hC]
--- for suggested possible additional features.

class

	QUATERNION_REF

inherit

	VECTOR4
		rename
			infix "+" as vector4_infix_plus, 
			infix "-" as vector4_infix_minus, 
			infix "*" as vector4_infix_star, 
			prefix "-" as vector4_prefix_minus,
			prefix "+" as vector4_prefix_plus,
			scale as vector4_scale,
			is_equal as vector4_is_equal,
			zero as vector4_zero
		export
			{QUATERNION_REF}
			vector4_infix_plus, vector4_prefix_minus, vector4_prefix_plus,
			vector4_zero, vector4_infix_minus, 
			vector4_infix_star, infix "#", dot, one, put, vector4_scale,
			put_x, put_y, put_z, put_w, unit, is_unit, fill
			{ANY}
			within_range, is_finite
		redefine
			default_create
		end

creation

	default_create, make, make_from_axis, make_from_vectors,
	make_from_transformation

feature -- Initialization
	
	default_create is
			-- Initialization
		do
			Precursor {VECTOR4}
			create cached_axis
		end
		
	make_from_axis (the_axis: VECTOR3; the_angle: DOUBLE) is
			-- 
		require 
			axis_not_void: the_axis /= Void
			axis_is_finite: the_axis.all_finite
			angle_is_finite: is_finite (the_angle)
		local
			scale_factor: DOUBLE
		do
			the_axis.normalize
			
			scale_factor := sine (the_angle/2)
			x := the_axis.x * scale_factor
			y := the_axis.y * scale_factor
			z := the_axis.z * scale_factor
			w := cosine (the_angle/2)
			
			update_axis := true
			
		ensure
			axis_is_correct_for_positive_angle:
							the_angle > Tolerance implies axis.is_equal (the_axis)
			axis_is_correct_for_zero_angle:
							about_equal (the_angle, 0) implies 
								about_equal (axis.x, 0) and 
								about_equal (axis.y, 0) and 
								about_equal (axis.z, 1)
			axis_is_correct_for_negative_angle: 
							the_angle < -Tolerance implies axis.is_equal (-the_axis)
			-- Cosine computation is to force normalization of the angle
			angle_is_correct: about_equal (angle, arc_cosine (cosine (the_angle)))
		end 
	
	make_from_vectors (rotate_from, rotate_to: VECTOR3) is
			-- Create a quaterion that is the equivilent of rotating 
			-- `rotate_from' into `rotate_to'
		require
			rotate_from_is_finite: rotate_from.all_finite
			rotate_to_is_finite: rotate_to.all_finite
		local
			cost: DOUBLE
			tmp, tmp_axis: VECTOR3
		do
			rotate_from.normalize
			rotate_to.normalize
			cost := rotate_from # rotate_to
			
			if about_equal (cost, 1.0) then
				x := 0
				y := 0
				z := 0
				w := 1
				-- Find an axis to rotate about through the cross product
			elseif about_equal (cost, -1.0) then
				tmp.make (1, 0, 0)
				tmp := rotate_from * tmp
				if about_equal (tmp.magnitude, 0.0) then
					tmp.make (0, 1, 0)
					tmp := rotate_from * tmp
				end
				
				tmp.normalize
				make (tmp.x, tmp.y, tmp.z, 0)
			else
				tmp_axis := rotate_from * rotate_to
				tmp_axis.normalize
				
				tmp_axis := tmp_axis.scale (sqrt (0.5 * (1.0 - cost)))
				
				make (tmp_axis.x, tmp_axis.y, tmp_axis.z, sqrt (0.5 * (1.0 + cost)))
			end
			
			update_axis := true
		ensure
			rotation_correct: rotate_to.is_equal (rotation_matrix.apply_as_direction (rotate_from))
		end
	
	make_from_transformation (xform: TRANSFORM) is
			-- Create a quaternion from a rotation matrix
		require 
			is_pure_rotation_matrix: xform.is_pure_rotation
		local
			i, j, k: INTEGER
		do
			if xform.a11 > xform.a22 then
				if xform.a11 > xform.a33 then
					i := 1
				else
					i := 3
				end
			else
				if xform.a22 > xform.a33 then
					i := 2
				else
					i := 3
				end
			end
			
			if xform.a11 + xform.a22 + xform.a33 > xform.item (i, i) then
				w := sqrt (xform.a11 + xform.a22 + xform.a33 + xform.a44)/2
				
				x := (xform.a32 - xform.a23)/ (4 * w)
				y := (xform.a13 - xform.a31)/ (4 * w)
				z := (xform.a21 - xform.a12)/ (4 * w)
			else
				-- i, j, k are mutually exclusive in [1..3]
				j := i \\ 3 + 1
				k := (i + 1) \\ 3 + 1
				
				put (sqrt (xform.item (i, i) - xform.item (j, j) - 
						   xform.item (k, k) + xform.a44)/2, i)
				
				put ((xform.item (j, i) + xform.item (i, j)) / (4 * item (i)), j)
				put ((xform.item (k, i) + xform.item (i, k)) / (4 * item (i)), k)
				put ((xform.item (k, j) - xform.item (j, k)) / (4 * item (i)), 4)
			end

			update_axis := true
		ensure 
			quaterion_is_correct: xform.is_equal (rotation_matrix)
		end 

feature -- Access
	
	axis: VECTOR3 is
			-- The rotation axis.
		do
			if update_axis then
				compute_cached_axis
			end
			
			Result := cached_axis
		ensure 
			result_is_normalized: Result.is_unit
		end 
	
	angle: DOUBLE is
			-- The rotation angle about `axis' (in radians).
		do
			if update_axis then
				compute_cached_axis
			end
			
			Result := cached_angle
		ensure 
			angle_is_finite: is_finite (Result)
			angle_in_range: angle >= 0 and angle <= pi + Tolerance
		end 

	rotation_matrix: TRANSFORM is
			-- Build a rotation matrix from the quaternion
		do
			create Result.make_with_elements (1.0 - (2.0 * (y * y + z * z)),
											  2.0 * (x * y - z * w),
											  2.0 * (z * x + y * w),
											  0.0,
											  
											  2.0 * (x * y + z * w),
											  1.0 - (2.0 * (z * z + x * x)),
											  2.0 * (y * z - x * w),
											  0.0,
											  
											  2.0 * (z * x - y * w),
											  2.0 * (y * z + x * w),
											  1.0 - (2.0 * (y * y + x * x)),
											  0.0,
											  
											  0.0,
											  0.0,
											  0.0,
											  1.0)
		end
	
	zero: like Current is
			-- Neutral element for `+' and `-'
		do
			create Result
		end 

feature -- Measurement
feature -- Comparison
	
	is_equal (other: like Current): BOOLEAN is
			-- Is each corresponding `item' within `Tolerance'?
		do
			Result := about_equal (x, other.x) and then
			about_equal (y, other.y) and then
			about_equal (z, other.z) and then
			about_equal (w, other.w)
		ensure
			result_correct:
						   Result = (about_equal (x, other.x) and
									 about_equal (y, other.y) and
									 about_equal (z, other.z) and
									 about_equal (w, other.w))
		end

feature -- Status report
feature -- Status setting
feature -- Cursor movement
feature -- Element change
feature -- Removal
feature -- Resizing
feature -- Transformation
feature -- Conversion
feature -- Duplication
feature -- Basic operations
	
	infix "+" (other: like Current): like Current is
			-- Addition of two quaternions
		require 
			other_not_void: other /= Void
		do
			create Result.make (x + other.x,
								y + other.y,
								z + other.z,
								w + other.w)
			
			update_axis := true
		ensure 
			Result_not_void: Result /= Void
		end 

	infix "-" (other: like Current): like Current is
			-- Subtraction of two quaternions
		require 
			other_not_void: other /= Void
		do
			create Result.make (x - other.x,
								y - other.y,
								z - other.z,
								w - other.w)
			
			update_axis := true
		ensure 
			Result_not_void: Result /= Void
		end 

	infix "*" (other: like Current): like Current is
			-- Given two rotations, figure out the equivalent single rotation
		do
			create Result.make (w*other.x + x*other.w + y*other.z - z*other.y,
								w*other.y + y*other.w + z*other.x - x*other.z,
								w*other.z + z*other.w + x*other.y - y*other.x,
								w*other.w - x*other.x - y*other.y - z*other.z)
			
			Result.normalize			 
			update_axis := true
			
		ensure
			is_correct: Result.rotation_matrix.is_equal (Current.rotation_matrix * other.rotation_matrix)
		end
	
	prefix "+": like Current is
			-- Positve of `Current'.
		do
			create Result.make (+x, +y, +z, +w)
		end
	
	prefix "-": like Current is
			-- Negative of `Current'.
		do
			create Result.make (-x, -y, -z, -w)
		ensure
			result_x_equals_negative_current_x: Result.x = -x
			result_y_equals_negative_current_y: Result.y = -y
			result_z_equals_negative_current_z: Result.z = -z
			result_w_equals_negative_current_w: Result.w = -w
		end
	
	scale (scalar: DOUBLE): like Current is
			-- Scale the quaternion by `scalar'
		require
			non_nan_scale: not is_nan (scalar)
		local
			result_x, result_y, result_z, result_w: DOUBLE
		do
			if scalar /= 0.0 then
				
				if x /= 0.0 then
					result_x := x * scalar
				end
				
				if y /= 0.0 then
					result_y := y * scalar
				end
				
				if z /= 0.0 then
					result_z := z * scalar
				end
				
				if w /= 0.0 then
					result_w := w * scalar
				end
				
				create Result.make (result_x, result_y, result_z, result_w)
				update_axis := true
			end
		ensure then
			result_x_correct: Result.x = filter_nan (x * scalar)
			result_y_correct: Result.y = filter_nan (y * scalar)
			result_z_correct: Result.z = filter_nan (z * scalar)
			result_w_correct: Result.w = filter_nan (w * scalar)
		end

	inverse: like Current is
			-- invert `Current'.
		do
			Result := Current.twin
			Result.invert
		ensure
			result_is_inverse: (Result + Current).rotation_matrix.is_identity
		end
	
	invert is
			-- invert `Current'.
		local
			inverse_norm: DOUBLE
		do
			inverse_norm := 1.0 / magnitude_squared
			make (-x * inverse_norm, -y * inverse_norm,
				  -z * inverse_norm, w * inverse_norm)
			update_axis := true
		end
	
	interpolation (q: like Current; t: DOUBLE): like Current is
			-- Spherical linear interpolation: as t goes from 0 to 1, 
			-- the result goes from `Current' to `q'.
		require 
			q_not_void: q /= Void
			t_in_range: within_range (t, 0.0, 1.0)
		local
			omega, cosom, sinom: DOUBLE
			scalerot0, scalerot1: DOUBLE
			tq: like Current
		do
			-- Calculate the cosine of the quaternions
			cosom := x * q.x + y * q.y + z * q.z + w * q.w
			
			-- Adjust the signs
			if cosom < 0.0 then
				cosom := -cosom
				tq := -q
			else
				tq := q
			end
			
			-- Calculate the interpolating coefficients
			if 1.0 - cosom > Tolerance then
				omega := arc_cosine (cosom)
				sinom := sine (omega)
				scalerot0 := sine ((1 - t) * omega) / sinom
				scalerot1 := sine (t * omega) / sinom
			else
				-- The quaterions are nearly equal, use linear interpolation
				scalerot0 := 1 - t
				scalerot1 := t
			end
			
			create Result.make (scalerot0 * x + scalerot1 * tq.x,
								scalerot0 * y + scalerot1 * tq.y,
								scalerot0 * z + scalerot1 * tq.z,
								scalerot0 * w + scalerot1 * tq.w)
		end 

feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {NONE} -- Implementation
	
	cached_angle: DOUBLE
			-- The rotation angle about `axis' (in radians).
	
	cached_axis: VECTOR3
			-- The rotation axis
	
	update_axis: BOOLEAN
			-- Flag used to indicate if the axis needs to be computed
	
	compute_cached_axis is
			-- The rotation axis
		do
			cached_axis.make (x, y, z)
			if cached_axis.magnitude < Tolerance then
				cached_axis.make (0, 0, 1)
				cached_angle := 0
			else
				cached_axis.normalize
				cached_angle := 2 * arc_cosine (w)
			end
			
			update_axis := false
		end
	
end -- class QUATERNION_REF


