indexing

	description: "3D cartesian axis-aligned bounding boxes."
	date: "$Date: 2006-12-10 07:10:42 -0500 (Sun, 10 Dec 2006) $"
	revision: "$Revision: 274 $"
	status: "reviewed", "tested"

class

	AXIS_ALIGNED_BOX

inherit

	INTERSECTABLE
		redefine
			is_equal, out, default_create, intersect
		end

	BASIC_NUMERIC
		redefine
			default_create, is_equal, out
		end

	DEBUG_OUTPUT
		redefine
			default_create, is_equal, out
		end

creation
	default_create, make

feature -- Initialization

	default_create is
			-- An empty box
		do
			Precursor {INTERSECTABLE}
			create lower
			create upper
		end

	make (first_corner, second_corner: POINT3) is
			-- Arbitrary-ordered vertices to be bounded.
		do
			create lower.make (first_corner.x.min (second_corner.x),
						first_corner.y.min (second_corner.y),
						first_corner.z.min (second_corner.z))
			create upper.make (first_corner.x.max (second_corner.x),
						first_corner.y.max (second_corner.y),
						first_corner.z.max (second_corner.z))
		end

feature -- Access

	lower: POINT3
			-- Lower corner.

	upper: POINT3
			-- Upper corner.

	intersection: ARRAY [DOUBLE] is
			-- Result of 'intersect' operation: ordered parametric coordinates
			-- along the ray of first and second point of intersection.
			-- In degenerate cases (e.g., ray hits a corner or originates from
			-- inside the box), Result.count = 1, otherwise Result.count = 2.
			-- (Use RAY.project to convert a parametric coordinate into POINT3.)
		do
			Result := cached_intersection
		ensure
			result_cached: Result = cached_intersection
		end

feature -- Measurement

	volume: DOUBLE is
			-- Volume of box.
		do
			Result := product (product (difference (upper.x, lower.x),
										difference (upper.y, lower.y)),
								difference (upper.z, lower.z))
		ensure
			Result_non_negative: Result >= 0.0
		end

	is_degenerate: BOOLEAN is
			-- Is lower = upper?
		do
			Result := lower.is_equal (upper)
		ensure
			result_correct:
				Result = lower.is_equal (upper)
		end

feature -- Comparison

	is_equal (other: like Current): BOOLEAN is
			-- Is `other' equal to `Current' (within `Tolerance')?
		do
			Result := lower.is_equal (other.lower) and then
						upper.is_equal (other.upper)
		end

	infix "<=" (other: AXIS_ALIGNED_BOX): BOOLEAN is
			-- Is `Current' a subset (sub-volume) of `other'?
		do
			Result := other.lower.x <= lower.x and then
						other.lower.y <= lower.y and then
						other.lower.z <= lower.z and then
						upper.x <= other.upper.x and then
						upper.y <= other.upper.y and then
						upper.z <= other.upper.z
		end

	infix "<" (other: AXIS_ALIGNED_BOX): BOOLEAN is
			-- Is `Current' a proper subset (sub-volume) of `other'?
		do
			Result := other.lower.x < lower.x and then
						other.lower.y < lower.y and then
						other.lower.z < lower.z and then
						upper.x < other.upper.x and then
						upper.y < other.upper.y and then
						upper.z < other.upper.z
		end

	infix ">=" (other: AXIS_ALIGNED_BOX): BOOLEAN is
			-- Is `Current' a superset (super-volume) of `other'?
		do
			Result := other <= Current
		end

	infix ">" (other: AXIS_ALIGNED_BOX): BOOLEAN is
			-- Is `Current' a proper superset (super-volume) of `other'?
		do
			Result := other < Current
		end

	overlaps (other: AXIS_ALIGNED_BOX): BOOLEAN is
			-- Does `Current' intersect `other'?
		do
			Result := not
						(lower.x > other.upper.x or else
						upper.x < other.lower.x or else
						lower.y > other.upper.y or else
						upper.y < other.lower.y)
		end

	contains (other: AXIS_ALIGNED_BOX): BOOLEAN is
			-- Does `Current' completely contain `other'?
		do
			Result := has (other.lower) and then has (other.upper)
		end

	all_finite: BOOLEAN is
			-- Are all coordinates finite (not +/-inf/nan)?
		do
			Result := (lower.all_finite and then upper.all_finite)
		ensure
			result_correct:
				Result = (lower.all_finite and upper.all_finite)
		end

feature -- Status report

	has (this_point: POINT3): BOOLEAN is
			-- Is `this_point' on or inside?
		do
			Result := within_range (this_point.x, lower.x, upper.x) and then
						within_range (this_point.y, lower.y, upper.y) and then
						within_range (this_point.z, lower.z, upper.z)
		end

	has_2d (x, y: DOUBLE): BOOLEAN is
			-- Is this_point on or inside the box (only considering x and y)?
		do
			Result := within_range (x, lower.x, upper.x) and then
						within_range (y, lower.y, upper.y)
		end

	centroid: POINT3 is
			-- Where is the centroid of the box?
		do
			create Result.make ((upper.x + lower.x) * 0.5,
						(upper.y + lower.y) * 0.5,
						(upper.z + lower.z) * 0.5)
		end

feature -- Element change

	set_lower (new_lower: POINT3) is
			-- Set `lower'.
		require
			x_less_than_upper: new_lower.x <= upper.x
			y_less_than_upper: new_lower.y <= upper.y
			z_less_than_upper: new_lower.z <= upper.z
		do
			lower := new_lower
		ensure
			lower_set: lower = new_lower
		end

	set_upper (new_upper: POINT3) is
			-- Set `upper'.
		require
			x_greater_than_lower: lower.x <= new_upper.x
			y_greater_than_lower: lower.y <= new_upper.y
			z_greater_than_lower: lower.z <= new_upper.z
		do
			upper := new_upper
		ensure
			upper_set: upper = new_upper
		end

	transform (xform: TRANSFORM) is
			-- Transform the box by `xform'
		require
			xform_not_void: xform /= Void
		local
			i: INTEGER
			pt: POINT3
			corners: ARRAY [POINT3]
		do
			create corners.make (1, 8)
			i := 1
			create pt.make (lower.x, lower.y, lower.z)
			corners.put (xform.apply (pt), i)
			i := i + 1
			create pt.make (lower.x, lower.y, upper.z)
			corners.put (xform.apply (pt), i)
			i := i + 1
			create pt.make (lower.x, upper.y, lower.z)
			corners.put (xform.apply (pt), i)
			i := i + 1
			create pt.make (lower.x, upper.y, upper.z)
			corners.put (xform.apply (pt), i)
			i := i + 1
			create pt.make (upper.x, lower.y, lower.z)
			corners.put (xform.apply (pt), i)
			i := i + 1
			create pt.make (upper.x, lower.y, upper.z)
			corners.put (xform.apply (pt), i)
			i := i + 1
			create pt.make (upper.x, upper.y, lower.z)
			corners.put (xform.apply (pt), i)
			i := i + 1
			create pt.make (upper.x, upper.y, upper.z)
			corners.put (xform.apply (pt), i)
			i := i + 1

			lower := corners.item (1).twin
			upper := corners.item (1).twin
			from
				i := 2
			invariant
				i_at_least_one: i >= 2
				i_small_enough: i <= 2 + 8
			variant
				i_increasing: 2 + 8 - i
			until
				i > 8
			loop
				pt := corners.item (i)
				-- Since we're looking down the Z-axis, max in Z is
				-- closer, but need to keep the ordering to avoid
				-- violating the invariant
				lower.make (lower.x.min (pt.x),
							lower.y.min (pt.y),
							lower.z.min (pt.z))
				upper.make (upper.x.max (pt.x),
							upper.y.max (pt.y),
							upper.z.max (pt.z))
				i := i + 1
			end
		end

feature -- Resizing

	grow (point: POINT3) is
			-- Grow the box, if necessary, to accomodate the new point.
		do
			lower.make (lower.x.min (point.x),
						lower.y.min (point.y),
						lower.z.min (point.z))
			upper.make (upper.x.max (point.x),
						upper.y.max (point.y),
						upper.z.max (point.z))
		ensure
			argument_contained: has (point)
		end

feature -- Conversion

	out: STRING is
			-- Printable representation of `Current'.
		do
			create Result.make (160)
			Result.append_character ('[')
			Result.append_character (' ')
			Result.append (lower.out)
			Result.append_character (' ')
			Result.append (upper.out)
			Result.append_character (' ')
			Result.append_character (']')
		end

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			Result := out
		end

feature -- Basic operations

	intersect (ray: RAY) is
			-- Generate the intersection of `ray' against this volume.
			--| Haines, Eric, "Ray/Box Intersection" in
			--| Glassner, "An Introduction to Ray Tracing", Academic Press,
			--| 1989, pp 65-67.
		local
			t_near, t_far: DOUBLE
					-- Parametric coordinates along the ray of the near and far
					-- points of intersection with the box.

			t1, t2: DOUBLE
					-- Intersection distances to each pair of bounding planes.

			direction, origin: DOUBLE
					-- Ray's coordinate direction and origin.

			x_intersected, y_intersected, z_intersected: BOOLEAN
					-- Did the coordinate intersect?
		do
			intersection_count := intersection_count + 1

			intersected := False
			t_near := -huge_value
			t_far := huge_value
			direction := ray.direction.x
			origin := ray.origin.x

			if direction = 0.0 then
				-- Ray direction is parallel to the bounding x-planes so
				-- it intersects only if it originates at a point with
				-- x coordinates that lie between the bounding x-planes.
				x_intersected := within_range (origin, lower.x, upper.x)
			else
				-- Ray is not parallel to the bounding x-planes so
				-- calculate near and far intersection distances of planes.
				direction := 1.0 / direction
				t1 := product (difference (lower.x, origin), direction)
				t2 := product (difference (upper.x, origin), direction)

				if t1 > t2 then
					t_near := t2
					t_far := t1
				else
					t_near := t1
					t_far := t2
				end

				check
					t_near_not_more_than_t_far: t_near <= t_far
				end
				x_intersected := t_far >= 0.0
			end

			if x_intersected then
				direction := ray.direction.y
				origin := ray.origin.y

				if direction = 0.0 then
					-- Ray direction is parallel to the bounding y-planes so
					-- it intersects only if it originates at a point with
					-- y coordinates that lie between the bounding y-planes.
					y_intersected := within_range (origin, lower.y, upper.y)
				else
					-- Ray is not parallel to the bounding y-planes so
					-- calculate near and far intersection distances of
					-- planes.
					direction := 1.0 / direction
					t1 := product (difference (lower.y, origin), direction)
					t2 := product (difference (upper.y, origin), direction)

					if t1 > t2 then

						if t2 > t_near then
							t_near := t2
						end

						if t1 < t_far then
							t_far := t1
						end
					else

						if t1 > t_near then
							t_near := t1
						end

						if t2 < t_far then
							t_far := t2
						end
					end

					y_intersected := t_near <= t_far and then t_far >= 0.0
				end

				if y_intersected then
					direction := ray.direction.z
					origin := ray.origin.z

					if direction = 0.0 then
						-- Ray direction is parallel to the bounding
						-- z-planes so it intersects only if it originates
						-- at a point with z coordinates that lie between
						-- the bounding z-planes.
						z_intersected := within_range (origin, lower.z, upper.z)
					else
						-- Ray is not parallel to the bounding z-planes so
						-- calculate near and far intersection distances of
						-- planes.
						direction := 1.0 / direction
						t1 := product (difference (lower.z, origin), direction)
						t2 := product (difference (upper.z, origin), direction)

						if t1 > t2 then

							if t2 > t_near then
								t_near := t2
							end

							if t1 < t_far then
								t_far := t1
							end
						else

							if t1 > t_near then
								t_near := t1
							end

							if t2 < t_far then
								t_far := t2
							end
						end

						z_intersected := t_near <= t_far and then t_far >= 0.0
					end

					if z_intersected then
						if t_near < 0.0 or else about_equal (t_near, t_far) then

							if one_intersection = Void then
								create one_intersection.make (1, 1)
							end

							cached_intersection := one_intersection
							cached_intersection.put (t_far, 1)
						else

							if two_intersections = Void then
								create two_intersections.make (1, 2)
							end

							cached_intersection := two_intersections
							cached_intersection.put (t_near, 1)
							cached_intersection.put (t_far, 2)
						end

						intersection_hit := intersection_hit + 1
						intersected := True
					end
				end
			end
		end

feature {NONE} -- Implementation

	cached_intersection: ARRAY [DOUBLE]
			-- Refers to either `one_intersection' or `two_intersections'.

	one_intersection: ARRAY [DOUBLE]
			-- Cached array of one parametric coordinate.

	two_intersections: ARRAY [DOUBLE]
			-- Cached array of two parametric coordinates.

invariant

	lower_not_void: lower /= Void
	upper_not_void: upper /= Void
	x_ordered: lower.x <= upper.x
	y_ordered: lower.y <= upper.y
	z_ordered: lower.z <= upper.z
	irreflexive_comparison: not (Current < Current)
	volume_non_negative: volume >= 0.0
	volume_exactly_zero_if_degenerate:
		(lower.x = upper.x or lower.y = upper.y or lower.z = upper.z) implies
		volume = 0.0
	has_bounding_points: has (lower) and has (upper)
	intersection_non_negative:
		intersected implies
			intersection /= Void and
			(intersection.count = 1 or intersection.count = 2) and
			intersection.item (1) >= 0.0 and
			(intersection.count = 2 implies intersection.item (2) >= 0.0)
	cached_intersection_assigned_correctly:
		cached_intersection /= Void implies
			cached_intersection = one_intersection or
			cached_intersection = two_intersections

end -- class AXIS_ALIGNED_BOX

