indexing

	description: "Geometers - basic knowledge about 3d vector geometry."
	names: "3d", "cartesian", "geometry", "vector"
	patterns: "mixin"
	representation: "DOUBLE"
	access: "fixed"
	size: "fixed"
	contents: "POINT3", "VECTOR3"
	date: "$Date: 2006-06-23 10:34:45 -0400 (Fri, 23 Jun 2006) $"
	revision: "$Revision: 226 $"
	author: "Todd Plessel, US EPA/LM, plessel.todd@epa.gov"
	status: "unreviewed", "tested", "partial", "unimplemented: colinear2"

class

	GEOMETER

inherit

	BASIC_NUMERIC

feature {NONE} -- Basic operations

	colinear (point1, point2, point3: POINT3): BOOLEAN is
			-- Do the three points lie on the same straight line?
			--| 1. Form three vectors: point1->point2 and point1->point3 and
			--|    point2->point3.
			--| 2. If any vector is zero, then two or all three
			--| points are coincident which is degenerate and considered
			--| colinear.
			--| 3. Otherwise, normalize the vectors and compute the
			--| dot-products: point1->point2 * point1->point3 and
			--|               point1->point2 * point2->point3
			--| 4. If both dot-products are approximately +/-1 then
			--| the points are colinear since the angle between the vectors is
			--| arccosine (+/-1) = 0 (or 180).
			--| Swokowski, Earl, "Calculus With Analytic Geometry, Second
			--| Edition", Prindle, Weber & Schmidt, Boston, MA, 1979, page 670.
		local
			vector1, vector2, vector3: VECTOR3
				-- Vectors point1->point2, point1->point3, point2->point3.

			dot_product_1: DOUBLE
				-- Dot-product of vector1 and vector2.

			dot_product_2: DOUBLE
				-- Dot-product of vector1 and vector3.
		do
			Result := True
			vector1 := point2 - point1

			if vector1.magnitude_squared /= 0.0 then
				vector2 := point3 - point1

				if vector2.magnitude_squared /= 0.0 and then
					not point1.is_equal (point3) then
					vector1.normalize
					vector2.normalize
					dot_product_1 := vector1.dot (vector2)
					Result := within_range (absolute_value (dot_product_1),
											1.0 - Tolerance, 1.0 + Tolerance)

					if Result then
							-- Compute third vector and dot product.
						vector3 := point3 - point2

						if vector3.magnitude_squared /= 0.0 then
							vector3.normalize
							dot_product_2 := vector1.dot (vector3)
							Result := within_range
										(absolute_value (dot_product_2),
										1.0 - Tolerance, 1.0 + Tolerance)
						end
					end
				end
			end
		end

	distance_from_point_to_line (point, line_point_1, line_point_2: POINT3):
		DOUBLE is
			-- Shortest absolute distance between `point' and the infinite
			-- undirected line containing the two points `line_point_1' and
			-- line_point_2'.
			-- If line length is infinite or point-line-point distance is
			-- infinite then Result may be huge_value.
			--| Swokowski, Earl, "Calculus With Analytic Geometry, Second
			--| Edition", Prindle, Weber & Schmidt, Boston, MA, 1979, page 696.
		require
			non_degenerate_line: not line_point_1.is_equal (line_point_2)
		local
			point_vector: VECTOR3
				-- Vector from `line_point_1' to `point'.

			line_vector: VECTOR3
				-- Vector from `line_point_1' to `line_point_2'.

			line_length: DOUBLE
				-- Magnitude of `line_vector'.
		do
			line_vector := line_point_2 - line_point_1
			line_length := line_vector.magnitude
				check
					line_length_non_zero: line_length /= 0.0
				end

			line_length := 1.0 / line_length

			if line_length = 0.0 then
					-- E.g., 1 / inf = 0 so avoid possible nan (0 * inf).
					-- Result will be 0 iff colinear, otherwise Result = inf.
				if not colinear (point, line_point_1, line_point_2) then
					Result := huge_value
				end
			else
				point_vector := point - line_point_1
				Result := line_length *
							line_vector.cross (point_vector).magnitude
			end
		ensure
			result_non_negative: Result >= 0.0
		end

	distance_from_point_to_plane (point, plane_point_1, plane_point_2,
			plane_point_3: POINT3): DOUBLE is
			-- Shortest absolute distance between `point' and the infinite
			-- plane given by the three plane points.
			--| Swokowski, Earl, "Calculus With Analytic Geometry, Second
			--| Edition", Prindle, Weber & Schmidt, Boston, MA, 1979, page 688.
		require
			plane_points_unique:
				not plane_point_1.is_equal (plane_point_2) and
				not plane_point_1.is_equal (plane_point_3) and
				not plane_point_2.is_equal (plane_point_3)
			plane_points_non_colinear:
				not colinear (plane_point_1, plane_point_2, plane_point_3)
		local
			point_vector: VECTOR3
				-- Vector from `plane_point_1' to `point'.

			plane_vector_1, plane_vector_2, plane_normal: VECTOR3
				-- Normal vector of plane.
		do
			point_vector := point - plane_point_1

			if not point_vector.is_zero and then
				not point.is_equal (plane_point_2) and then
				not point.is_equal (plane_point_3) then
				plane_vector_1 := plane_point_2 - plane_point_1
				plane_vector_2 := plane_point_3 - plane_point_1
				plane_normal := plane_vector_1.cross (plane_vector_2)
					check
						plane_vector_1_non_zero: not plane_vector_1.is_zero
						plane_vector_2_non_zero: not plane_vector_2.is_zero
						plane_normal_non_zero:
							plane_normal.magnitude_squared /= 0.0
					end
				Result := (point_vector.dot (plane_normal.unit)).abs
			end
		ensure
			result_non_negative: Result >= 0.0
		end

	projection_of_point_onto_line (point, line_point_1, line_point_2: POINT3):
		DOUBLE is
			-- Signed distance from `line_point_1' to location where `point'
			-- projects onto directed line from `line_point_1' to line_point_2'.
			--| Swokowski, Earl, "Calculus With Analytic Geometry, Second
			--| Edition", Prindle, Weber & Schmidt, Boston, MA, 1979, page 673.
		require
			non_degenerate_line: not line_point_1.is_equal (line_point_2)
		local
			point_vector: VECTOR3
				-- Vector from `line_point_1' to `point'.

			line_vector: VECTOR3
				-- Vector from `line_point_1' to `line_point_2'.
		do
			line_vector := line_point_2 - line_point_1
				check
					line_length_non_zero: line_vector.magnitude /= 0.0
				end
			line_vector.normalize
			point_vector := point - line_point_1
			Result := point_vector.dot (line_vector)
		end

	determinant3 (vector1, vector2, vector3: VECTOR3): DOUBLE is
			-- Determinant of 3 x 3 matrix with given row vectors.
			--| Computed by Laplace expansion of the first row.
			--| Swokowski, Earl, "Calculus With Analytic Geometry, Second
			--| Edition", Prindle, Weber & Schmidt, Boston, MA, 1979, page 690.
		local
			left, right, term: DOUBLE
				-- To hold value of intermediate expressions to avoid nans,
				-- e.g., from -inf + inf and 0 * inf.
		do
			if not (vector1.is_equal (vector2) or else
				vector1.is_equal (vector3) or else
				vector2.is_equal (vector3)) then
					-- Check for degenerate matrix first to ensure that
					-- result is 0 even when +/-inf is involved.

				if vector1.x /= 0.0 then

					if vector2.y /= 0.0 and then vector3.z /= 0.0 then
						left := vector2.y * vector3.z
					end

					if vector2.z /= 0.0 and then vector3.y /= 0.0 then
						right := vector2.z * vector3.y
					end

					if left /= right then
						Result := vector1.x * (left - right)
					end
				end

				if vector1.y /= 0.0 then

					if vector2.x /= 0.0 and then vector3.z /= 0.0 then
						left := vector2.x * vector3.z
					else
						left := 0.0
					end

					if vector2.z /= 0.0 and then vector3.x /= 0.0 then
						right := vector2.z * vector3.x
					else
						right := 0.0
					end

					if left /= right then
						term := -vector1.y * (left - right)

						if Result = -term then
							Result := 0.0
						else
							Result := Result + term
						end
					end
				end

				if vector1.z /= 0.0 then

					if vector2.x /= 0.0 and then vector3.y /= 0.0 then
						left := vector2.x * vector3.y
					else
						left := 0.0
					end

					if vector2.y /= 0.0 and then vector3.x /= 0.0 then
						right := vector2.y * vector3.x
					else
						right := 0.0
					end

					if left /= right then
						term := vector1.z * (left - right)

						if Result = -term then
							Result := 0.0
						else
							Result := Result + term
						end
					end
				end
			end
		end

	clip_line (p1, p2: POINT3; box: AXIS_ALIGNED_BOX):
		TRIPLE [POINT3, POINT3, BOOLEAN] is
			-- Clip line segment `p1->p2' to `box'.
			-- If Result.third = True then
			--   an intersection occured and
			--   Result.first and Result.second are the possibly moved endpoints
			-- otherwise Result.third = False
			--   and no intersection occured and
			-- Result.first = Result.second = zero.
			--| Uses Liang-Barsky Line Clipping Algorithm.
			--| See Pokorny, Cornel K., and Curtis F. Gerald,
      		--| _Computer Graphics: The Princples Behind the Art and Science_,
      		--| Franklin, Beedle & Associates, Irvine CA, 1989, pp 134-142.
		require
			non_degenerate_line: not p1.is_equal (p2)
		local
			box_has_p1, box_has_p2: BOOLEAN
				-- Does box contain p1 or p2?

			dx, dy, dz: DOUBLE
				-- Components of vector p1->p2.

			x, y, z, s, xmin, xmax, ymin, ymax, zmin, zmax: DOUBLE
				-- Clamped results and box bounds.

			t: TRIPLE [DOUBLE, DOUBLE, BOOLEAN]
				-- Result of `clip_coordinate': component of line endpoint
				-- clipped to each boundary pair.
		do
				-- Clip to: left, right, bottom, top, back, front:

			Result.put_first (p1)
			Result.put_second (p2)
			box_has_p1 := box.has (p1)
			box_has_p2 := box.has (p2)

			if box_has_p1 and then box_has_p2 then
					-- Handle this case simply to avoid algorithm failure
					-- in case infinities arise.
				Result.put_third (True)
			else
				t.put_second (1.0)

				dx := difference (p2.x, p1.x)
				t := clip_coordinate (-dx, difference (p1.x, box.lower.x),
										t.first, t.second)

				if t.third then
					t := clip_coordinate (dx, difference (box.upper.x, p1.x),
											t.first, t.second)

					if t.third then
						dy := difference (p2.y, p1.y)
						t := clip_coordinate (-dy,
												difference (p1.y, box.lower.y),
												t.first, t.second)

						if t.third then
							t := clip_coordinate (dy,
												difference (box.upper.y, p1.y),
													t.first, t.second)

							if t.third then
								dz := difference
									(Result.second.z, Result.first.z)
								t := clip_coordinate (-dz,
												difference (p1.z, box.lower.z),
														t.first, t.second)

								if t.third then
									t := clip_coordinate (dz,
												difference (box.upper.z, p1.z),
															t.first, t.second)

									if t.third then
											-- At least some of the line is
											-- within the box so we must
											-- calculate the possibly new
											-- endpoints.
											-- The usual formulation can fail
											-- to yield coordinates within the
											-- box so clamping is used to ensure
											-- that adjusted clipped points at
											-- least lie within the box.

										xmin := box.lower.x
										xmax := box.upper.x
										ymin := box.lower.y
										ymax := box.upper.y
										zmin := box.lower.z
										zmax := box.upper.z

										if t.second < 1.0 then
												-- Compute new end point.
											s := t.second
											x := clamp_clipped_coordinate
												(s, dx, p2.x, p1.x, xmin, xmax)
											y := clamp_clipped_coordinate
												(s, dy, p2.y, p1.y, ymin, ymax)
											z := clamp_clipped_coordinate
												(s, dz, p2.z, p1.z, zmin, zmax)
											Result.second.put_x (x)
											Result.second.put_y (y)
											Result.second.put_z (z)
										end

										if t.first > 0.0 then
												-- Compute new start point.
											s := t.first
											x := clamp_clipped_coordinate
												(s, dx, p1.x, p1.x, xmin, xmax)
											y := clamp_clipped_coordinate
												(s, dy, p1.y, p1.y, ymin, ymax)
											z := clamp_clipped_coordinate
												(s, dz, p1.z, p1.z, zmin, zmax)
											Result.first.put_x (x)
											Result.first.put_y (y)
											Result.first.put_z (z)
										end

										if box_has_p1 or else box_has_p2 then
												-- One of the original points is
												-- contained so we must yield a
												-- clipped result. This may
												-- require clamping the other
												-- point if the algorithm failed
												-- due to infinities.

											if box.has (Result.first) then

												if not box.has (Result.second)
													then
														-- Clamp second.
													x := clamp_to_range
														(Result.second.x,
														xmin, xmax)
													y := clamp_to_range
														(Result.second.y,
														ymin, ymax)
													z := clamp_to_range
														(Result.second.z,
														zmin, zmax)
													Result.second.put_x (x)
													Result.second.put_y (y)
													Result.second.put_z (z)
												end
											elseif box.has (Result.second) then
													-- Clamp first.
												x := clamp_to_range
													(Result.first.x, xmin, xmax)
												y := clamp_to_range
													(Result.first.y, ymin, ymax)
												z := clamp_to_range
													(Result.first.z, zmin, zmax)
												Result.first.put_x (x)
												Result.first.put_y (y)
												Result.first.put_z (z)
											end
										end

										Result.put_third
											(box.has (Result.first) and then
											box.has (Result.second))
									end
								end
							end
						end
					end
				end
			end

			if not Result.third then
				Result.first.make (0.0, 0.0, 0.0)
				Result.second.make (0.0, 0.0, 0.0)
			end
		ensure
			result_implies_contained:
				(Result.third and
					box.has (Result.first) and box.has (Result.second)) or else
				(not Result.third and
					Result.first.is_zero and Result.second.is_zero)
		end

feature {NONE} -- Implementation

	clip_coordinate (p, q, t1, t2: DOUBLE): TRIPLE [DOUBLE, DOUBLE, BOOLEAN] is
			-- Compute parameterized coordinates of clipped coordinate where
			-- `p' and `q' are the boundary points and `t1', and `t2' are the
			-- endpoints of the line to clip.
			-- If the segment t1->t2 intersects a boundary then Result.third
			-- is True and Result.first and Result.second are the new possibly
			-- moved clipped line endpoints otherwise Result.third is False and
			-- Result.first and Result.second are zero.
			--| Part of Liang-Barsky Line Clipping Algorithm.
		local
			r: DOUBLE
					-- Intersection coordinate.
		do
			Result.put_first (t1)
			Result.put_second (t2)
			Result.put_third (True)
				-- First assume that line is not completely clipped.

			if p < 0.0 then
					-- Line from outside to inside of that boundary.
				r := quotient (q, p)

				if r > t2 then
						-- Intersection past segment end point.
					Result.put_third (False)
				elseif r > t1 then
						-- Intersection is past start point.
					Result.put_first (r)
				end

			elseif p > 0.0 then
					-- Line from inside to outside of that boundary.
				r := quotient (q, p)

				if r < t1 then
						-- Intersection is before start point.
					Result.put_third (False)
				elseif r < t2 then
						-- Intersection is before end point.
					Result.put_second (r)
				end

			elseif q < 0.0 then
					-- Line is parallel to that boundary.
					check
						p_is_zero: p = 0.0
					end
				Result.put_third (False)
			end

			if Result.third = False then
				Result.put_first (0.0)
				Result.put_second (0.0)
			end
		end

	clamp_clipped_coordinate (scale, delta, origin0, origin, minimum, maximum:
		DOUBLE): DOUBLE is
			-- Compute clipped coordinate from origin scaled by delta and
			-- clamps it to given range.
		do
			if scale = 0.0 or else scale = 1.0 or else
				not within_range (delta, -maximum_double, maximum_double) then
				Result := origin0
			else
				Result := sum (origin, product (scale, delta))
			end

			Result := clamp_to_range (Result, minimum, maximum)
		ensure
			result_within_bounds: within_range (Result, minimum, maximum)
		end

	colinear2 (point1, point2, point3: POINT3): BOOLEAN is
			-- Do the three points lie on the same straight line?
			--| Compute the direction numbers (the vectors point1->point2 and
			--| point2->point3) and verify that they are proportional.
			--| Protter, Murray and Charles Morrey, Jr.,
			--| "Analytic Geometry, Second Edition", Addison-Wesley, Reading,
			--| MA, 1975, pp 209, 214, 215.
		local
			x_numerator, x_denominator, x_ratio: DOUBLE
			y_numerator, y_denominator, y_ratio: DOUBLE
			z_numerator, z_denominator, z_ratio: DOUBLE
			x_matched, y_matched: BOOLEAN
		do
			Result := True
			x_denominator := difference (point3.x, point2.x)

			if about_equal (x_denominator, 0.0) then
				Result := about_equal (point3.x, point1.x)
				x_matched := True
			else
				x_numerator := difference (point2.x, point1.x)
				x_ratio := quotient (x_numerator, x_denominator)
			end

			if Result then
				y_denominator := difference (point3.y, point2.y)

				if about_equal (y_denominator, 0.0) then
					Result := about_equal (point3.y, point1.y)
					y_matched := True
				else
					y_numerator := difference (point2.y, point1.y)
					y_ratio := quotient (y_numerator, y_denominator)
					Result := x_matched or else about_equal (x_ratio, y_ratio)
				end

				if Result then
					z_denominator := difference (point3.z, point2.z)

					if about_equal (z_denominator, 0.0) then
						Result := about_equal (point3.z, point1.z)
					else
						z_numerator := difference (point2.z, point1.z)
						z_ratio := quotient (z_numerator, z_denominator)
						Result := (x_matched or else
									about_equal (x_ratio, z_ratio)) and then
									(y_matched or else
									about_equal (y_ratio, z_ratio))
					end
				end
			end
		end

end -- class GEOMETER

