indexing

	description: "Positioned infinite directed lines in 3D cartesian space."
	date: "$Date: 2006-12-06 13:14:08 -0500 (Wed, 06 Dec 2006) $"
	revision: "$Revision: 271 $"
	status: "reviewed", "tested"

class

	RAY

inherit

	DEBUG_OUTPUT
		redefine
			is_equal, out, default_create
		end
	
creation
	default_create, make
	
feature -- Initialization

	default_create is
			-- Initialize with `origin' (0,0,0) and `direction' <1,0,0>.
		do
			if origin = Void and direction = Void then
				create origin.make (0.0, 0.0, 0.0)
				create direction.make (0.0, 0.0, 1.0)
			else
				origin.make (0.0, 0.0, 0.0)
				direction.make (0.0, 0.0, 1.0)
			end
		ensure then
			origin_zero:origin.is_zero
			direction_along_z_axis:
				direction.x = 0.0 and direction.y = 0.0 and direction.z = 1.0
		end

	make (initial_origin: POINT3; initial_direction: VECTOR3) is
			-- Initialize with given origin and direction.
		require
			initial_direction_is_unit_vector: initial_direction.is_unit
		do
			origin := initial_origin
			direction := initial_direction
		ensure
			origin_set: origin = initial_origin
			direction_set: direction = initial_direction
		end

feature -- Access

	origin: POINT3
			-- The position of the tail/root of the ray.

	direction: VECTOR3
			-- The unit vector that the ray is oriented along.

feature -- Comparison

	is_equal (other: RAY): BOOLEAN is
			-- Are both origin and direction within `Tolerance' of other's?
		do
			Result := (origin.is_equal (other.origin) and then
						direction.is_equal (other.direction))
		ensure then
			result_correct:
				Result = (origin.is_equal (other.origin) and
						direction.is_equal (other.direction))
		end

	all_finite: BOOLEAN is
			-- Are all coordinates/components finite (not +/-inf/nan)?
		do
			Result := (origin.all_finite and then direction.all_finite)
		ensure then
			result_correct:
				Result = (origin.all_finite and direction.all_finite)
		end

feature -- Element change

	set_origin (new_origin: POINT3) is
			-- Set `origin' to `new_origin'.
		do
			origin := new_origin
		ensure
			origin_set: origin = new_origin
		end

	set_direction (new_direction: VECTOR3) is
			-- Set `direction' to `new_direction'.
		require
			argument_is_unit_length: new_direction.is_unit
		do
			direction := new_direction
		ensure
			direction_set: direction = new_direction
		end

	set_towards_point (towards_point: POINT3) is
			-- Re-orient `direction' to point towards `towards_point'.
		require
			argument_not_origin: towards_point /= origin
		do
			direction := (towards_point - origin).unit
		end

feature -- Conversion

	out: STRING is
			-- Printable representation of `Current'.
		do
			create Result.make (160)
			Result.append_character ('(')
			Result.append_character (' ')
			Result.append (origin.out)
			Result.append_character (' ')
			Result.append (direction.out)
			Result.append_character (' ')
			Result.append_character (')')
		end

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			Result := out
		end

feature -- Basic operations

	apply (xform: TRANSFORM) is
			-- Apply a transform.
		do
			origin := xform.apply (origin)
			direction := xform.apply_as_direction (direction)
				check
					direction_is_normalizable: direction.magnitude /= 0.0
				end
		end

	project (distance: DOUBLE): POINT3 is
			-- Point along the ray `distance' from the `origin'.
		do
			Result := origin + direction.scale (distance)
		end

invariant

	origin_not_void: origin /= Void
	direction_not_void: direction /= Void
	
end -- class RAY

