indexing

	description: "3D cartesian geometric line segment cells."
	names: "3d", "cartesian", "geometry", "cell", "line", "segment"
	representation: "DOUBLE"
	access: "fixed"
	size: "fixed"
	contents: "POINT3"
	date: "$Date: 2006-06-23 10:34:45 -0400 (Fri, 23 Jun 2006) $"
	revision: "$Revision: 226 $"
	author: "Todd Plessel, US EPA/LM, plessel.todd@epa.gov"
	status: "unreviewed", "tested", "partial", "unimplemented: intersect"

class

	LINE

inherit

	MESH_CELL
		redefine
			default_create, is_simplex, intersect
		end

creation

	make, default_create, make_from_array

feature -- Initialization

	default_create is
			-- Make degenerate line with two zero vertices.
		local
			zero: POINT_CELL
			zero_vertex: POINT3
		do
			Precursor
			
			intersected := False
			distance := 0.0
			contoured := False
			the_parametric_coordinate := 0.0
			
			create zero_vertex
			
			if vertices = Void then
				create vertices.make (1, 2)
			end
			
			vertices.put (zero_vertex.twin, 1)
			vertices.put (zero_vertex.twin, 2)

			if the_contours = Void then
				create zero.default_create
				create the_contours.make (1, 1)
				the_contours.put (zero, 1)
			else
				the_contours.item (1).default_create
			end

			if scalar_derivatives = Void then
				create scalar_derivatives.make (1, 3)
			else
				scalar_derivatives.clear_all
			end

			if vector_derivatives = Void then
				create vector_derivatives.make (1, 9)
			else
				vector_derivatives.clear_all
			end

			if the_simplex = Void then
				create the_simplex.make (1, 1)
				the_simplex.put (Current, 1)
			end

			if one_intersection = Void then
				create one_intersection.make (1, 1)
			end

			if two_intersections = Void then
				create two_intersections.make (1, 2)
			end

			cached_intersection := two_intersections
			cached_intersection.put (0.0, 1)
			cached_intersection.put (1.0, 2)
			one_intersection.put (0.0, 1)
		ensure then
			first_vertex_is_zero: item (1).is_zero
			second_vertex_is_zero: item (2).is_zero
			the_parametric_coordinate_is_zero: the_parametric_coordinate = 0.0
			intersected_false: intersected = False
			distance_is_zero: distance = 0.0
			contoured_false: contoured = False
		end

	make (initial_first_endpoint, initial_second_endpoint: POINT3) is
			-- Make line with these initial endpoint verticies.
		do
			default_create
			put (initial_first_endpoint, 1)
			put (initial_second_endpoint, 2)
		ensure
			first_vertex_set: item (1).is_equal (initial_first_endpoint)
			second_vertex_set: item (2).is_equal (initial_second_endpoint)
			the_parametric_coordinate_is_zero: the_parametric_coordinate = 0.0
			intersected_false: intersected = False
			distance_is_zero: distance = 0.0
			contoured_false: contoured = False
		end

	make_from_array (the_vertices: ARRAY [POINT3]) is
		    -- Make cell with given points.
	    do
		make (the_vertices.item (1), the_vertices.item (2))
	    end

feature -- Access

	item (vertex_number: INTEGER): POINT3 is
			-- Access a vertex.
		do
			Result := vertices.item (vertex_number)
		end

	weight (vertex_number: INTEGER): DOUBLE is
			-- Interpolation weight at given vertex for current parametric
			-- coordinate.
		do
			if vertex_number = 1 then
				Result := 1.0 - the_parametric_coordinate
			else
				Result := the_parametric_coordinate
			end
		end

	parametric_coordinate (coordinate_number: INTEGER): DOUBLE is
			-- Parametric coordinate for interpolation along the line.
		do
			Result := the_parametric_coordinate
		end

	parametric_position: POINT3 is
			-- XYZ coordinates of position at current parametric coordinate
			-- along the line.
		do
			if the_parametric_coordinate = 0.0 then
				Result := item (1)
			elseif the_parametric_coordinate = 1.0 then
				Result := item (2)
			else
				Result := item (1).interpolate (item (2),
												the_parametric_coordinate)
			end
		ensure then
			zero_yields_first_endpoint:
				parametric_coordinate (1) = 0.0 implies
					Result.is_equal (item (1))
			one_yields_second_endpoint:
				parametric_coordinate (1) = 1.0 implies
					Result.is_equal (item (2))
		end

	edge (edge_number: INTEGER): LINE is
			-- Edge of a cell.
			-- `Result' is `Current'.
		do
			Result := Current
		ensure then
			result_is_current: Result = Current
		end

	face (face_number: INTEGER): MESH_CELL is
			-- Pre-condition is false.
		do
		end

	nearest_point: POINT3 is
			-- Point nearest to `parametric_position'.
		do
			if the_parametric_coordinate <= 0.5 then
				Result := item (1)
			else
				Result := item (2)
			end
		end

	nearest_edge: LINE is
			-- `Result' is `Current'.
		do
			Result := Current
		ensure then
			result_is_current: Result = Current
		end

	nearest_face: MESH_CELL is
			-- Pre-condition is false.
		do
		end

	simplices: ARRAY [LINE] is
			-- `Result' is one `Current' and is cached.
		do
			Result := the_simplex
		ensure then
			result_is_current: Result = the_simplex
		end

	contours: ARRAY [POINT_CELL] is
			-- Cell(s) generated by last successful `contour' operation.
			-- `Result' is a cached point.
		do
			Result := the_contours
		end

feature -- Measurement

	dimension: INTEGER is 1
			-- Spatial dimensionality:
			-- `Result' is 1.

	vertex_count: INTEGER is 2
			-- Number of vertices in cell.
			-- `Result' is 2.

	edge_count: INTEGER is 1
			-- Number of edges in cell.
			-- `Result' is 1.

	face_count: INTEGER is 0
			-- Number of faces in cell.
			-- `Result' is 0.

	centroid: POINT3 is
			-- Location of cell center.
		do
			Result := item (1).midpoint (item (2))
		end

	perimeter: DOUBLE is
			-- Sum of edge lengths.
		do
			Result := item (1).distance (item (2))
		end

	area: DOUBLE is 0.0
			-- Area of cell faces.
			-- `Result' is 0.

	volume: DOUBLE is 0.0
			-- Volume of cell.
			-- `Result' is 0.

	extent: AXIS_ALIGNED_BOX is
			-- 3d cartesian axis-aligned minimum bounding-box enclosing cell.
		do
			Result.make (item (1), item (2))
		end

	intersection: ARRAY [DOUBLE] is
			-- Result of 'intersect' operation: if `intersected' then
			-- Result.item (1) is the parametric coordinate along the ray of
			-- the (first) point of intersection and if the ray points along the
			-- line then Result.count = 2 and Result.item (1) is one of the line
			-- endpoints and Result.item (2) is the other endpoint.
			-- (Use RAY.project to convert a parametric coordinate into POINT3.)
		do
			Result := cached_intersection
		ensure then
			result_cached: Result = cached_intersection
		end

feature -- Status report

	is_simplex: BOOLEAN is True
			-- Is the cell a simplex for its dimension?
			-- `Result' is True.

	is_degenerate: BOOLEAN is
			-- Is the line just two coincident points?
		do
			Result := item (1).is_equal (item (2))
		end

feature -- Element change

	put, replace (new_vertex: POINT3; vertex_number: INTEGER) is
			-- Make `new_vertex' `item (vertex_number)'.
		do
			vertices.put (new_vertex, vertex_number)
			intersected := False
			distance := 0.0
			the_parametric_coordinate := 0.0
			contoured := False
		end

	extend (new_vertex: POINT3) is
			-- Pre-condition is `False'.
		do
		end

	set_parametric_coordinate (value: DOUBLE; coordinate_number: INTEGER) is
			-- Set the given parametric coordinate for subsequent interpolation.
		do
			the_parametric_coordinate := value
		end

feature -- Remove

	remove (vertex_number: INTEGER) is
			-- Pre-condition is `False'.
		do
		end

feature -- Basic operations

	contour (interpolant_value: DOUBLE; vertex_data: ARRAY [DOUBLE]) is
			-- Generate contour point that approximates the
			-- interpolated position of `value' along the line assuming
			-- the given scalar values at each endpoint.
			-- If the `interpolant_value' is along the line,
			-- `contoured' is `True' and the result is available as `contours'.
			-- Otherwise `contoured' is `False'.
		local
			denominator: DOUBLE
				-- Of normalization expression.

			parameter: DOUBLE
				-- Parametric coordinate of `position' along the line.

			position: POINT3
				-- Position of `interpolant_value' along the line.

			first_vertex_value: DOUBLE
				-- Scalar value at first vertex.
		do
			contoured := False
			first_vertex_value := vertex_data.item (1)
			denominator := difference (vertex_data.item (2), first_vertex_value)

			if denominator = 0.0 then
				contoured := about_equal (interpolant_value, first_vertex_value)

				if contoured then
						-- Choose the midpoint when contouring equal values.
					the_contours.item (1).put (centroid, 1)
				end
			else
				parameter := quotient (difference (interpolant_value,
													first_vertex_value),
										denominator)
				contoured := within_range (parameter, 0.0, 1.0)

				if contoured then
					position := item (1).interpolate (item (2), parameter)
					the_contours.item (1).put (position, 1)
				end
			end
		end

	interpolate_scalar (vertex_data: ARRAY [DOUBLE]): DOUBLE is
			-- Interpolate scalar `vertex_data' to `parametric_position'.
		do
			Result := sum (product (vertex_data.item (1),
									1.0 - the_parametric_coordinate),
							product (vertex_data.item (2),
									the_parametric_coordinate))
		end

	interpolate_vector (vertex_data: ARRAY [DOUBLE]): VECTOR3 is
			-- Interpolate 3d vector `vertex_data' to `parametric_position'.
 		local
			one_minus_parametric_coordinate: DOUBLE
		do
			one_minus_parametric_coordinate := 1.0 - the_parametric_coordinate
			Result.put_x (sum (product (vertex_data.item (1),
										one_minus_parametric_coordinate),
								product (vertex_data.item (4),
										the_parametric_coordinate)))
			Result.put_y (sum (product (vertex_data.item (2),
										one_minus_parametric_coordinate),
								product (vertex_data.item (5),
										the_parametric_coordinate)))
			Result.put_z (sum (product (vertex_data.item (3),
										one_minus_parametric_coordinate),
								product (vertex_data.item (6),
										the_parametric_coordinate)))
		end

	gradient (vertex_data: ARRAY [DOUBLE]): ARRAY [DOUBLE] is
			-- Compute the derivatives in the x-y-z coordinate directions
			-- of scalar or 3D vector vertex data.
			-- The result for scalar vertex data is <ds/dx,ds/dy,ds/dz> and
			-- the result for 3d vector data (v) is:
			-- <dvx/dx,dvx/dy,dvx/dz,dvy/dx,dvy/dy,dvy/dz,dvz/dx,dvz/dy,dvz/dz>.
			-- Result values will be zero for degenerate cell dimensions.
			-- `Result' is cached.
			--| See Schroeder, Will, et al. "The Visualization Toolkit",
			--| Prentice Hall, Upper Saddle River, New Jersey, 1996, pp 236-238.
		local
			vertex_data_is_scalar: BOOLEAN
				-- Is the vertex data scalar?

			unit_line_vector: VECTOR3
				-- Unit vector from 1st line endpoint to 2nd line endpoint.

			line_length: DOUBLE
				-- The distance between the line endpoints.

			component_vector: VECTOR3
				-- Difference in data (component) values divided by the
				-- length of the line times the unit_line_vector.
		do
			vertex_data_is_scalar := vertex_data.count = 2
			unit_line_vector := item (2) - item (1)
			line_length := unit_line_vector.magnitude

			if line_length = 0.0 then
					-- Line is degenerate so result is all zeros.
				scalar_derivatives.clear_all
				vector_derivatives.clear_all
			else
				unit_line_vector.normalize
				line_length := 1.0 / line_length

				if vertex_data_is_scalar then
					component_vector := scaled_data_difference_vector
											(line_length,
											vertex_data.item (2),
											vertex_data.item (1),
											unit_line_vector)
					scalar_derivatives.put (component_vector.x, 1)
					scalar_derivatives.put (component_vector.y, 2)
					scalar_derivatives.put (component_vector.z, 3)
				else
					component_vector := scaled_data_difference_vector
											(line_length,
											vertex_data.item (4),
											vertex_data.item (1),
											unit_line_vector)
					vector_derivatives.put (component_vector.x, 1)
					vector_derivatives.put (component_vector.y, 2)
					vector_derivatives.put (component_vector.z, 3)
					component_vector := scaled_data_difference_vector
											(line_length,
											vertex_data.item (5),
											vertex_data.item (2),
											unit_line_vector)
					vector_derivatives.put (component_vector.x, 4)
					vector_derivatives.put (component_vector.y, 5)
					vector_derivatives.put (component_vector.z, 6)
					component_vector := scaled_data_difference_vector
											(line_length,
											vertex_data.item (6),
											vertex_data.item (3),
											unit_line_vector)
					vector_derivatives.put (component_vector.x, 7)
					vector_derivatives.put (component_vector.y, 8)
					vector_derivatives.put (component_vector.z, 9)
				end
			end

			if vertex_data_is_scalar then
					Result := scalar_derivatives
			else
					Result := vector_derivatives
			end
		end

	evaluate_position (position: POINT3) is
			-- Evaluate the position with respect to the cell geometry.
			-- If `position' is within the cell then
			--   (1) `intersected' will be True and
			--   (2) `parametric_coordinate (1)' will be set to interpolate to
			--       `position' and
			--   (3) `distance' will be zero and
			--   (4) `intersection.count' = 1
			-- Else `position' is outside the cell and then
			--   (1) `intersected' will be False and
			--   (2) `parametric_coordinate (1)' will be set to the point on
			--       the cell that is closest (by Euclidean distance) to
			--       `position' and
			--   (3) `distance' will report the distance from
			--       `parametric_position' to `position'.
			--| `position' is on the line iff the projection distance of
			--| `position' onto the line is within the length of the line and
			--| the three points are colinear.
			--| Protter, Murray, et al, "Analytic Geometry, Second Edition",
			--| Addison Wesley, Reading, MA, 1975, page 209.
			--| Swokowski, Earl, "Calculus With Analytic Geometry, Second
			--| Edition", Prindle, Weber & Schmidt, Boston, MA, 1979, page 673
			--| and 696.
		local
			parameter: DOUBLE
				-- Parametric coordinate of `position' along line.
		do
			intersected := False
			distance := huge_value
			the_parametric_coordinate := 0.0
			cached_intersection := one_intersection
			cached_intersection.put (0.0, 1)

			if position.is_equal (item (1)) then
				distance := 0.0
				intersected := True
			elseif position.is_equal (item (2)) then
				the_parametric_coordinate := 1.0
				distance := 0.0
				intersected := True
			elseif is_degenerate then
					-- Compute point-to-point distance.
				distance := position.distance (item (1))
			else
					-- Project position onto line.
					-- If the projection distance is within the length of the
					-- line and the three points are colinear then an
					-- intersection occured.
				parameter := projection_of_point_onto_line (position,
															item (1), item (2))
				parameter := quotient (parameter, perimeter)
				the_parametric_coordinate := clamp_to_range (parameter, 0.0,1.0)
				intersected := parametric_position.is_equal (position)

				if intersected then
					distance := 0.0
				else
						-- Compute point-to-point distance.
					distance := position.distance (parametric_position)
				end
			end
		end

	intersect (ray: RAY) is
			-- Calculate point of intersection with ray.
			-- If intersection occurs then
			--   (1) `intersected' will be True and
			--   (2) `distance' will be zero and
			--   (3) `parametric_coordinate' (and thus `parametric_position')
			--       will be set to the (first/nearest) point of intersection
			--       and
			--   (4) `intersection.item (1)' will be the parametric coordinate
			--       along `ray' of the point of intersection and
			--   (5) if the ray intersects the line twice (i.e., by being
			--       colinear and either pointing towards the line or
			--       originating from within the line) then
			--         (a) `intersection.count' = 2 and
			--         (b) `intersection.item (2)' will be the second
			--             parametric coordinate of intersection along `ray'.
			-- Else no intersection occurs and then
			--   (1) `intersected' will be False and
			--   (2) `distance' will be huge_value if `ray' points away from
			--       the line else `distance' will be equal to the shortest
			--       skew line distance.
			--   (3) `parametric_coordinate' will be set to the coordinate of
			--       the point of nearest approach of the ray to the line.
			--| See Goldman, Ronald, "Intersection of Two Lines in Three-space",
			--| Glassner, Andrew, Editor, "Graphics Gems Volume 1",
			--| Academic Press, Cambridge, MA, 1990, page 304.
		local
			point: POINT_CELL
				-- Used in case of degenerate line.

			line_endpoint_1, line_endpoint_2: POINT3
				-- Endpoints of this line segment.

			ray_endpoint: POINT3
				-- Position of tail/root of ray.

			unit_line_vector: VECTOR3
				-- Normalized vector pointing from line_endpoint_1
				-- towards line_endpoint_2.

			line_to_ray_vector: VECTOR3
				-- Vector from line_endpoint_1 to ray_endpoint.

			denominator: DOUBLE
				-- Square of the magnitude of above cross-product.
				-- If zero then the line and the ray are parallel.

			line_vector_cross_ray: VECTOR3
				-- Cross-product of unit_line_vector and ray.

			numerator: DOUBLE
				-- Determinant of the 3 x 3 matrix used to compute the
				-- parametric_coordinate_along_line.

			parametric_coordinate_along_line: DOUBLE
				-- Parametric coordinate along line of point of
				-- intersection, or closest approach.

			parametric_coordinate_along_ray: DOUBLE
				-- Parametric coordinate along ray of point of
				-- intersection, or closest approach.

			distance_along_ray_to_line_endpoint_1: DOUBLE
				-- Parametric coordinate along ray of first line endpoint.

			distance_along_ray_to_line_endpoint_2: DOUBLE
				-- Parametric coordinate along ray of second line endpoint.

			compute_skew_distance: BOOLEAN
				-- Does ray miss line but point towards at least part of it?
		do

			--- This routine is not implemented correctly.
			check incorrect_implementation: False end
			
			intersection_count := intersection_count + 1
			
			intersected := False
			distance := huge_value
			line_endpoint_1 := item (1)
			line_endpoint_2 := item (2)
			ray_endpoint := ray.origin

			if is_degenerate then
					-- Use POINT_CELL implementation.
				create point.make (item (1))
				point.intersect (ray)
				distance := point.distance
				the_parametric_coordinate := 0.0

				if point.intersected then
					cached_intersection.put (point.intersection.item (1), 1)
				else
					cached_intersection.put (0.0, 1)
				end

				intersected := point.intersected
			else
				unit_line_vector := (line_endpoint_2 - line_endpoint_1).unit
				line_vector_cross_ray := unit_line_vector.cross (ray.direction)
				denominator := line_vector_cross_ray.magnitude_squared

---				if about_equal (denominator, 0.0) then
				if denominator = 0.0 then
						-- The line and the ray are parallel.
						-- Check if they are colinear and the ray points towards
						-- or within the line segment.

					if colinear (line_endpoint_1, line_endpoint_2, ray_endpoint)
							then
						distance_along_ray_to_line_endpoint_1 :=
							ray.direction.dot (line_endpoint_1 - ray_endpoint)
						distance_along_ray_to_line_endpoint_2 :=
							ray.direction.dot (line_endpoint_2 - ray_endpoint)

						if sign_of (distance_along_ray_to_line_endpoint_1) /=
							sign_of (distance_along_ray_to_line_endpoint_2) then
								-- Different signs indicate that the ray lies
								-- within the line so the first point of
								-- intersection is the ray endpoint and the
								-- second point of intersection is the positive
								-- line endpoint.

							if distance_along_ray_to_line_endpoint_1 >= 0.0 then
									-- Ray points towards first line endpoint.

								if distance_along_ray_to_line_endpoint_1 <=
									Tolerance then
										-- One intersection: ray origin is at
										-- first line endpoint.
									cached_intersection.put (0.0, 1)
									the_parametric_coordinate := 0.0
								else
										-- Two intersections: ray origin and
										-- first line endpoint.
									cached_intersection := two_intersections
									cached_intersection.put (0.0, 1)
									cached_intersection.put
									(distance_along_ray_to_line_endpoint_1, 2)
									line_to_ray_vector :=
										ray_endpoint - line_endpoint_1
									the_parametric_coordinate :=
										unit_line_vector.dot (line_to_ray_vector)
								end
							else
									-- Ray points towards second line endpoint.
									check
										negative_first_endpoint_distance:
										distance_along_ray_to_line_endpoint_1
										< 0.0
										non_negative_second_endpoint_distance:
										distance_along_ray_to_line_endpoint_2
										>= 0.0
									end

								if distance_along_ray_to_line_endpoint_2 <=
									Tolerance then
										-- One intersection: ray origin is at
										-- second line endpoint.
									cached_intersection := one_intersection
									cached_intersection.put (0.0, 1)
									the_parametric_coordinate := 1.0
								else
										-- Two intersections: ray origin and
										-- second line endpoint.
									cached_intersection := two_intersections
									cached_intersection.put (0.0, 1)
									cached_intersection.put
									(distance_along_ray_to_line_endpoint_1, 2)
									line_to_ray_vector :=
										ray_endpoint - line_endpoint_1
									the_parametric_coordinate :=
										unit_line_vector.dot (line_to_ray_vector)
								end
							end

							distance := 0.0
							intersected := True
						elseif distance_along_ray_to_line_endpoint_1 >= 0.0
								and then
								distance_along_ray_to_line_endpoint_2 >= 0.0 then
								-- Both signs positive indicates that the ray
								-- points towards both line endpoints.
								-- The first point of intersection is the
								-- nearest line endpoint and the second point of
								-- intersection is the farthest line endpoint.
							cached_intersection := two_intersections

							if distance_along_ray_to_line_endpoint_1 <
									distance_along_ray_to_line_endpoint_2 then
									-- The first line endpoint is closer.
								the_parametric_coordinate := 0.0
								cached_intersection.put
									(distance_along_ray_to_line_endpoint_1, 1)
								cached_intersection.put
									(distance_along_ray_to_line_endpoint_2, 2)
							else
									-- The second line endpoint is closer.
								the_parametric_coordinate := 1.0
								cached_intersection.put
									(distance_along_ray_to_line_endpoint_2, 1)
								cached_intersection.put
									(distance_along_ray_to_line_endpoint_1, 2)
							end

							distance := 0.0
							intersected := True
						end
					end
				else
						-- The lines are non-parallel and either intersect or
						-- are skewed.
					if ray_endpoint.is_equal (line_endpoint_1) then
							-- Ray originates at first line endpoint so this is
							-- an intersection.
						cached_intersection := one_intersection
						cached_intersection.put (0.0, 1)
						the_parametric_coordinate := 0.0
						distance := 0.0
						intersected := True
					elseif ray_endpoint.is_equal (line_endpoint_2) then
							-- Ray originates at second line endpoint so this is
							-- an intersection.
						cached_intersection := one_intersection
						cached_intersection.put (0.0, 1)
						the_parametric_coordinate := 1.0
						distance := 0.0
						intersected := True
					else
							-- Compute intersection point or skew distance.
						line_to_ray_vector := ray_endpoint - line_endpoint_1
						numerator := determinant3 (line_to_ray_vector,
													ray.direction,
													line_vector_cross_ray)
						parametric_coordinate_along_line :=
							quotient (numerator, denominator)

						if within_range (parametric_coordinate_along_line, 0,1)
								then
								-- The ray intersects the line once at this
								-- parametric coordinate along the line vector.
								-- Must also compute the parametric coordinate
								-- along the ray of this point of intersection
								-- to be sure it is non-negative.
							the_parametric_coordinate :=
								parametric_coordinate_along_line

							numerator := determinant3 (line_to_ray_vector,
														unit_line_vector,
														line_vector_cross_ray)
							parametric_coordinate_along_ray :=
								quotient (numerator, denominator)

							if parametric_coordinate_along_ray >= 0.0 then
								cached_intersection := one_intersection
								cached_intersection.put
									(parametric_coordinate_along_ray, 1)
								distance := 0.0
								intersected := True
							end
						else
								-- Ray likely misses line, but must still set
								-- parametric coordinate nearest point of
								-- closest approach and
								-- compute the distance of the miss -
								-- which will be huge_value if the ray points
								-- away from the line (as determined by having
								-- both dot products negative).
							the_parametric_coordinate :=
								clamp_to_range
									(parametric_coordinate_along_line, 0.0, 1.0)

							distance_along_ray_to_line_endpoint_1 :=
								ray.direction.dot
									(line_endpoint_1 - ray_endpoint)

							if distance_along_ray_to_line_endpoint_1 >= 0.0 then
								compute_skew_distance := True
							else
								distance_along_ray_to_line_endpoint_2 :=
									ray.direction.dot
										(line_endpoint_2 - ray_endpoint)
								compute_skew_distance :=
									distance_along_ray_to_line_endpoint_2 >= 0.0
							end

							if compute_skew_distance then
									-- Compute distance between closest points.
								numerator :=
									determinant3 (line_to_ray_vector,
													unit_line_vector,
													line_vector_cross_ray)
								parametric_coordinate_along_ray :=
									quotient (numerator, denominator)

								distance := parametric_position.distance
											(ray.project
											(parametric_coordinate_along_ray))

								if distance <= Tolerance then
										-- Ray intersects line at infinity!
										-- Be sure it is not behind the ray.

									if parametric_coordinate_along_ray >= 0.0
											then
										cached_intersection := one_intersection
										cached_intersection.put
											(parametric_coordinate_along_ray, 1)
										distance := 0.0
										intersected := True
									else
										distance := huge_value
									end
								end
							end
						end
					end
				end
			end
			
			if intersected then
				intersection_hit := intersection_hit + 1
			end
		ensure then
			intersected_implies_intersection_is_consistent:
				intersected implies
					not ray.origin.all_finite or
					not is_finite (intersection.item (1)) or
					not is_finite (perimeter) or
					ray.project (intersection.item (1)).is_equal
								(parametric_position)
			not_intersected_implies_distance_is_strictly_positive:
				not intersected implies distance > 0.0
		end

feature {NONE} -- Implementation

	scaled_data_difference_vector (scalar, value1, value2: DOUBLE;
		unit_vector: VECTOR3): VECTOR3 is
			-- Scaled data value difference vector:
			-- Result = unit_vector * (scalar * (value1 - value2)).
		require
			scalar_non_nan: not is_nan (scalar)
			value1_non_nan: not is_nan (value1)
			value2_non_nan: not is_nan (value2)
			unit_vector_is_unit: unit_vector.is_unit
		local
			scaled_data_difference: DOUBLE
				-- scalar * (value1 - value2)
		do
			scaled_data_difference :=
				product (scalar, difference (value1, value2))
			Result := unit_vector.scale (scaled_data_difference)
		end

	vertices: ARRAY [POINT3]
			-- The two endpoints of the line.

	the_simplex: ARRAY [LINE]
			-- The cached result of simplicies.

	the_contours: ARRAY [POINT_CELL]
			-- The result of a successful `contour' operation.

	the_parametric_coordinate: DOUBLE
			-- The parametric_coordinate along the line.

	scalar_derivatives: ARRAY [DOUBLE]
			-- ds/dx, ds/dy, ds/dz of the scalar value.

	vector_derivatives: ARRAY [DOUBLE]
			-- dvx/dx, dvx/dy, dvx/dz, dvy/dx, dvy/dy, dvy/dz, dvz/dx, dvz/dy,
			-- dvz/dz of the vector value.

	cached_intersection: ARRAY [DOUBLE]
			-- Refers to either `one_intersection' or `two_intersections'.

	one_intersection: ARRAY [DOUBLE]
			-- Cached array of one parametric coordinate.

	two_intersections: ARRAY [DOUBLE]
			-- Cached array of two parametric coordinates.

invariant

	dimension_is_one: dimension = 1
	edge_count_is_one: edge_count = 1
	area_is_zero: area = 0.0
	volume_is_zero: volume = 0.0
	the_simplex_exists: the_simplex /= Void
	simplices_is_current: the_simplex.item (1) = Current
	contoured_implies_point_contour:
		contoured implies contours.item (1).vertex_count = 1
	is_simplex: is_simplex
	degenerate_implies_coincident:
		is_degenerate implies item (1).is_equal (item (2))
	current_cell_is_one: current_cell = 1
	scalar_derivatives_exists: scalar_derivatives /= Void
	scalar_derivatives_count_is_3: scalar_derivatives.count = 3
	vector_derivatives_exists: vector_derivatives /= Void
	vector_derivatives_count_is_9: vector_derivatives.count = 9
	intersection_exists: intersected implies intersection /= Void
	intersection_count_is_one_or_two: (intersected implies
									   (intersection.count = 1 or intersection.count = 2))
	cached_intersection_assigned:
		cached_intersection /= Void and
			(cached_intersection = one_intersection or
			cached_intersection = two_intersections)

end -- class LINE

