indexing

	description: "3D cartesian geometric cells."
	date: "$Date: 2006-12-06 13:14:08 -0500 (Wed, 06 Dec 2006) $"
	revision: "$Revision: 271 $"
	status: "reviewed", "tested"

deferred class

	MESH_CELL

inherit

	INTERSECTABLE
		redefine
			default_create
		end
	
	GEOMETER
		export
			{ANY}
				about_equal, is_nan
		redefine
			default_create
		end

feature -- Initialization
	
	default_create is
			-- Initialize
		do
			Precursor {INTERSECTABLE} 
			Precursor {GEOMETER} 
		end
	
	make_from_array (vertices: ARRAY [POINT3]) is
			-- Make cell with given points.
		require
			vertices_exist: vertices /= Void
			at_least_one_vertex: vertices.count >= 1
--- FIX: add `minimum_vertex_count' and check against it...
---			not_too_many_vertices:
---				(not extendible implies vertices.count = vertex_count) and
---				extendible implies vertices.count >= vertex_count
--- Can't check against anything if there is no object yet!
		deferred
		ensure
		end

feature -- Access

	intersection: ARRAY [DOUBLE] is
			-- Result of 'intersect' operation: ordered parametric coordinates
			-- along the ray of first, second, third, ... unique points of
			-- intersection.
			-- (Use RAY.project to convert a parametric coordinate into POINT3.)
		require
			intersected: intersected
		deferred
		ensure
			result_exists: Result /= Void and then Result.count >= 1
			first_value_non_negative: Result.item (1) >= 0.0
			last_value_non_negative: Result.item (Result.count) >= 0.0
			-- Ditto for each intermediate value.
		end

	item (vertex_number: INTEGER): POINT3 is
			-- Access a vertex (as a POINT3).
		require
			vertex_number_at_least_one: vertex_number >= 1
			vertex_number_not_more_than_vertex_count:
				vertex_number <= vertex_count
		deferred
		end

	edge (edge_number: INTEGER): LINE is
			-- Edge of a cell.
		require
			dimension_positive: dimension > 0
			edge_number_at_least_one: edge_number >= 1
			edge_number_not_more_than_edge_count: edge_number <= edge_count
		deferred
		ensure
			result_exists: Result /= Void
			result_is_1d: Result.dimension = 1
			result_is_simplex: Result.is_simplex
		end

	face (face_number: INTEGER): MESH_CELL is
			-- Face of a cell.
		require
			dimension_at_least_2: dimension > 1
			face_number_at_least_one: face_number >= 1
			face_number_not_more_than_face_count: face_number <= face_count
		deferred
		ensure
			result_exists: Result /= Void
			result_is_2d: Result.dimension = 2
			result_edge_count_at_least_three: Result.edge_count >= 3
			result_edge_count_not_more_than_current_edge_count:
				Result.edge_count <= edge_count
		end

	cell (cell_number: INTEGER): MESH_CELL is
			-- Cell of a composite or Current if not composite.
			-- (Default `Result' is `Current`.)
		require
			cell_number_at_least_one: cell_number >= 1
			cell_number_not_more_than_cell_count: cell_number <= cell_count
		do
			Result := Current
		ensure
			result_exists: Result /= Void
		end

	nearest_point: POINT3 is
			-- Point nearest to `parametric_position'.
		deferred
		end

	nearest_edge: LINE is
			-- Edge nearest to `parametric_position'.
		require
			dimension_at_least_one: dimension >= 1
		deferred
		ensure
			result_exists: Result /= Void
			result_is_one_dimensional: Result.dimension = 1
		end

	nearest_face: MESH_CELL is
			-- Face nearest to `parametric_position'.
		require
			dimension_at_least_three: dimension > 2
		deferred
		ensure
			result_exists: Result /= Void
			result_is_two_dimensional: Result.dimension = 2
		end

	weight (vertex_number: INTEGER): DOUBLE is
			-- Interpolation weight at given vertex for current parametric
			-- coordinate(s).
		require
			vertex_number_at_least_one: vertex_number >= 1
			vertex_number_not_more_than_vertex_count:
				vertex_number <= vertex_count
		deferred
		ensure
			result_non_negative: Result >= 0.0
			result_not_more_than_one: Result <= 1.0
		end

	parametric_coordinate (coordinate_number: INTEGER): DOUBLE is
			-- Parametric coordinate for interpolation.
			-- Each `intersect' operation sets these coordinates as follows:
			-- If the ray intersects the cell then
			--	 (1) `intersected' will be True and
			--	 (2) the parametric coordinate(s) will be set to interpolate to
			--		 the first/nearest point of intersection and
			--	 (3) `distance' will be zero.
			--	 (4) `intersection' will contain ordered parametric coordinates
			--		 along the ray of first, second, third, ... unique points of
			--		 intersection.
			-- Else the ray does not intersect the cell and then
			--	 (1) `intersected' will be False and
			--	 (2) the parametric coordinate(s) will be set to the point on
			--		 the cell that is closest (by Euclidean distance) to the
			--		 ray and
			--	 (3) `distance' will report the distance from
			--		 `parametric_position' to the ray.
		require
			dimension_positive: dimension > 0
			coordinate_number_at_least_one: coordinate_number >= 1
			coordinate_number_not_more_than_dimension:
				coordinate_number <= dimension
		deferred
		ensure
			result_non_negative: Result >= 0.0
			result_not_more_than_one: Result <= 1.0
		end

	parametric_position: POINT3 is
			-- XYZ coordinates of position at current parametric coordinates.
		deferred
		end

	simplices: ARRAY [MESH_CELL] is
			-- Simplices of decomposed cell.
			-- Result may be one or more cells.
			-- Result is cached.
			-- Each dimension + 1 vertices defines a simplex.
			-- Each simplex depends on the dimension:
			--	 dimension = 0 implies point (a single point).
			--	 dimension = 1 implies line(s).
			--	 dimension = 2 implies triangles(s).
			--	 dimension = 3 implies tetrahedron(s).
		deferred
		ensure
			result_exists: Result /= Void
			-- no_fewer_vertices_than_current:
			--	 sum (Result.item (i).vertex_count) >= vertex_count
			current_if_simplex: is_simplex implies Result.item (1) = Current
			results_are_simplex:
				Result.item (1).is_simplex and
				Result.item (Result.count).is_simplex
			results_have_equal_vertex_count:
				Result.item (1).vertex_count =
				Result.item (Result.count).vertex_count
			results_have_correct_vertex_count:
				Result.item (1).vertex_count = dimension + 1 
		end

	contours: ARRAY [MESH_CELL] is
			-- Cell(s) generated by last successful `contour' operation.
			-- Result is cached.
		require
			contoured: contoured
		deferred
		end

feature -- Measurement

	dimension: INTEGER is
			-- Spatial dimensionality:
			-- 0 = point, 1 = line, 2 = surface, 3 = volume.
		deferred
		end

	vertex_count: INTEGER is
			-- Number of vertices in cell.
		deferred
		end

	edge_count: INTEGER is
			-- Number of edges in cell.
		deferred
		end

	face_count: INTEGER is
			-- Number of faces in cell.
		deferred
		end

	cell_count: INTEGER is
			-- Number of cells in composite (Result > 1 iff composite).
			-- (Default `Result' is 1.)
		do
			Result := 1
		end

	centroid: POINT3 is
			-- Location of cell center.
		deferred
		end

	perimeter: DOUBLE is
			-- Sum of edge lengths.
		deferred
		end

	area: DOUBLE is
			-- Area of cell faces.
		deferred
		end

	volume: DOUBLE is
			-- Volume of cell.
		deferred
		end

	extent: AXIS_ALIGNED_BOX is
	
			-- 3d cartesian axis-aligned minimum bounding-box enclosing cell.
		deferred
		end

feature -- Status report

	is_simplex: BOOLEAN is
			-- Is the cell a simplex for its dimension?
			-- Simplices are the minimal cell forms for each dimension:
			-- dimension = 0 implies point
			-- dimension = 1 implies line
			-- dimension = 2 implies triangle
			-- dimension = 3 implies tetrahedron
			-- (Default `Result' is `False'.)
		do
		end

	is_degenerate: BOOLEAN is
			-- Does the cell have length, area or volume that is less than
			-- normal because of coincident vertices?
		deferred
		end

	extendible: BOOLEAN is
			-- May vertices be added?
			-- (Default `Result' is `False'.)
		do
		end

	is_composite: BOOLEAN is
			-- Is the cell a collection of cells?
			-- (Default `Result' is `False'.)
		do
		end

	distance: DOUBLE
			-- After each `evaluate_position (position)' operation,
			-- `distance' will be the absolute distance between
			-- `parametric_position' and `position' or zero if `intersected'.
			-- After each `intersect' operation,
			-- `distance' will be the absolute distance between
			-- `parametric_position' and the ray's nearest approach or
			-- zero if `intersected'.

	current_cell: INTEGER is
			-- Currently selected cell number of composite.
			-- (Default `Result' is 1.)
		do
			Result := 1
		end

	contoured: BOOLEAN
			-- Did the last `contour' operation generate any results?
			-- (If so, obtain them as `contours'.)

feature -- Status setting

	select_cell (cell_number: INTEGER) is
			-- Select the current cell of a composite.
			-- (Default is do nothing.)
		require
			cell_number_at_least_one: cell_number >= 1
			cell_number_not_more_than_cell_count: cell_number <= cell_count
		do
		end

feature -- Element change

	put (new_vertex: POINT3; vertex_number: INTEGER) is
			-- Make `new_vertex' the `vertex_number'th vertex
			-- (by insertion or replacement).
		require
			vertex_number_at_least_one: vertex_number >= 1
			vertex_number_not_more_than_vertex_count:
				vertex_number <= vertex_count
		deferred
		ensure
			set_ith_vertex_to_new_vertex:
				item (vertex_number).is_equal (new_vertex)
			intersected_false: intersected = False
			distance_is_zero: distance = 0.0
			contoured_false: contoured = False
			first_parametric_coordinate_is_zero:
				dimension > 0 implies parametric_coordinate (1) = 0.0
			last_parametric_coordinate_is_zero:
				dimension > 0 implies parametric_coordinate (dimension) = 0.0
			-- Ditto for each parametric coordinate.
		end

	replace (new_vertex: POINT3; vertex_number: INTEGER) is
			-- Make `new_vertex' the `vertex_number'th vertex (by replacement).
		require
			vertex_number_at_least_one: vertex_number >= 1
			vertex_number_not_more_than_vertex_count:
				vertex_number <= vertex_count
		deferred
		ensure
			set_numbered_vertex_to_new_vertex:
				item (vertex_number).is_equal (new_vertex)
			count_unchanged: vertex_count = old vertex_count
			intersected_false: intersected = False
			distance_is_zero: distance = 0.0
			contoured_false: contoured = False
			first_parametric_coordinate_is_zero:
				dimension > 0 implies parametric_coordinate (1) = 0.0
			last_parametric_coordinate_is_zero:
				dimension > 0 implies parametric_coordinate (dimension) = 0.0
			-- Ditto for each parametric coordinate.
		end

	extend (new_vertex: POINT3) is
			-- Make `new_vertex' the last vertex (by extension).
		require
			extendible: extendible
		deferred
		ensure
			count_incremented: vertex_count = old vertex_count + 1
			intersected_false: intersected = False
			distance_is_zero: distance = 0.0
			contoured_false: contoured = False
			first_parametric_coordinate_is_zero:
				dimension > 0 implies parametric_coordinate (1) = 0.0
			last_parametric_coordinate_is_zero:
				dimension > 0 implies parametric_coordinate (dimension) = 0.0
			-- Ditto for each parametric coordinate.
		end

	set_parametric_coordinate (value: DOUBLE; coordinate_number: INTEGER) is
			-- Set the given parametric coordinate for subsequent interpolation.
		require
			value_non_negative: value >= 0.0
			value_not_more_than_one: value <= 1.0
			dimension_positive: dimension > 0
			coordinate_number_at_least_one: coordinate_number >= 1
			coordinate_number_not_more_than_dimension:
				coordinate_number <= dimension
		deferred
		ensure
			parametric_coordinate_set:
				parametric_coordinate (coordinate_number) = value
		end

feature -- Removal

	remove (vertex_number: INTEGER) is
			-- Remove numbered vertex.
		require
			extendible: extendible
			vertex_number_at_least_one: vertex_number >= 1
			vertex_number_not_more_than_vertex_count:
				vertex_number <= vertex_count
			vertex_count_large_enough: vertex_count >= dimension + 2
		deferred
		ensure
			count_decremented: vertex_count = old vertex_count - 1
			intersected_false: intersected = False
			distance_is_zero: distance = 0.0
			contoured_false: contoured = False
			first_parametric_coordinate_is_zero:
				dimension > 0 implies parametric_coordinate (1) = 0.0
			last_parametric_coordinate_is_zero:
				dimension > 0 implies parametric_coordinate (dimension) = 0.0
			-- Ditto for each parametric coordinate.
		end

feature -- Basic operations

	contour (interpolant_value: DOUBLE; vertex_data: ARRAY [DOUBLE]) is
			-- Generate contour cell(s) whose vertices approximate the
			-- interpolated position(s) of `value' within the cell assuming
			-- the given scalar values at each cell vertex.
			-- The result will be points for point and line cells,
			-- lines for face cells and faces for volume cells.
			-- If the interpolant_value is within cell, `contoured' is True and
			-- the result is available as `contours'.
			-- Otherwise `contoured' is False.
		require
			interpolant_value_non_nan: not is_nan (interpolant_value)
			vertex_data_exists: vertex_data /= Void
			one_vertex_value_per_vertex: vertex_data.count = vertex_count
		deferred
		end

	interpolate_scalar (vertex_data: ARRAY [DOUBLE]): DOUBLE is
			-- Interpolate scalar `vertex_data' to `parametric_position'.
		require
			argument_exists: vertex_data /= Void
			vertex_data_count_is_vertex_count: vertex_data.count = vertex_count
		deferred
		ensure
			result_is_within_range: scalar_within_range (Result, vertex_data)
		end

	interpolate_vector (vertex_data: ARRAY [DOUBLE]): VECTOR3 is
			-- Interpolate 3d vector `vertex_data' to `parametric_position'.
		require
			argument_exists: vertex_data /= Void
			vertex_data_count_is_vertex_count_times_3:
				vertex_data.count = vertex_count * 3
		deferred
		ensure
			result_is_within_range: vector_within_range (Result, vertex_data)
		end

	gradient (vertex_data: ARRAY [DOUBLE]): ARRAY [DOUBLE] is
			-- Compute the derivatives in the x-y-z coordinate directions
			-- of scalar or 3D vector vertex data.
			-- The result for scalar vertex data is <ds/dx,ds/dy,ds/dz> and
			-- the result for 3d vector data is:
			-- <dvx/dx,dvx/dy,dvx/dz,dvy/dx,dvy/dy,dvy/dz,dvz/dx,dvz/dy,dvz/dz>.
			-- Result values will be zero for degenerate cell dimensions.
		require
			argument_exists: vertex_data /= Void
			vertex_data_are_either_scalar_or_3d_vector:
				vertex_data.count = vertex_count or else
				vertex_data.count = vertex_count * 3
		deferred
		ensure
			result_exists: Result /= Void
			vertex_data_are_scalar_implies_three_values_in_result:
				vertex_data.count = vertex_count implies Result.count = 3
			vertex_data_are_3d_vector_implies_nine_values_in_result:
				vertex_data.count = vertex_count * 3 implies Result.count = 9
		end

	evaluate_position (position: POINT3) is
			-- Evaluate the position with respect to the cell geometry.
			-- If `position' is within or on the cell then
			--	 (1) `intersected' will be True and
			--	 (2) the parametric coordinate(s) will be set to interpolate to
			--		 `position' and
			--	 (3) `distance' will be zero.
			-- Else `position' is outside the cell and then
			--	 (1) `intersected' will be False and
			--	 (2) the parametric coordinate(s) will be set to the point on
			--		 the cell that is closest (by Euclidean distance) to
			--		 `position' and
			--	 (3) `distance' will report the distance from
			--		 `parametric_position' to `position'.
		deferred
		ensure
			intersected_implies_intersection_count_is_one:
				intersected implies intersection.count = 1
			intersected_implies_intersection_is_zero:
				intersected implies intersection.item (1) = 0.0
			intersected_implies_parametric_position_is_position:
				intersected implies parametric_position.is_equal (position)
			not_intersected_implies_distance_is_parametric_position_to_position:
				not intersected implies
					about_equal (distance,
								parametric_position.distance (position))
		end

feature {NONE} -- Implementation

	scalar_within_range (scalar: DOUBLE; data: ARRAY [DOUBLE]): BOOLEAN is
			-- Is `scalar' within the range of scalar `data'?
		require
			data_exists: data /= Void
		local
			data_range: PAIR [DOUBLE, DOUBLE]
		do
			data_range := scalar_range (data)
			Result := within_range (scalar,
									data_range.first - Tolerance,
									data_range.second + Tolerance)
		end

	vector_within_range (vector: VECTOR3; data: ARRAY [DOUBLE]): BOOLEAN is
			-- Is `vector' within the range of vector `data'?
		require
			data_exists: data /= Void
			data_count_at_least_three: data.count >= 3
			data_count_is_multiple_of_three: data.count \\ 3 = 0
		local
			data_range: PAIR [VECTOR3, VECTOR3]
		do
			data_range := vector_range (data)
			Result :=
				within_range (vector.x,
								data_range.first.x - Tolerance,
								data_range.second.x + Tolerance)
				and then
				within_range (vector.y,
								data_range.first.y - Tolerance,
								data_range.second.y + Tolerance)
				and then
				within_range (vector.z,
								data_range.first.z - Tolerance,
								data_range.second.z + Tolerance)
		end

invariant

	dimension_non_negative: dimension >= 0
	dimension_not_more_than_three: dimension <= 3
	vertex_count_large_enough: vertex_count >= dimension + 1
	vertex_count_one_if_point: dimension = 0 implies vertex_count = 1
	edge_count_large_enough: edge_count >= dimension
	edge_count_zero_if_point: dimension = 0 implies edge_count = 0
	face_count_zero_if_sub_planar: dimension <= 1 implies face_count = 0
	face_count_large_enough: dimension > 1 implies face_count >= 1
	cell_count_large_enough: cell_count >= 1
	weight_is_one_if_point: dimension = 0 implies weight (1) = 1.0
---	extent_has_centroid: extent.has (centroid)
	perimeter_non_negative: perimeter >= 0.0
	perimeter_is_zero_if_point: dimension = 0 implies perimeter = 0.0
	area_non_negative: area >= 0.0
	area_is_zero_if_sub_planar: dimension < 2 implies area = 0.0
	volume_non_negative: volume >= 0.0
	volume_is_zero_if_sub_volume: dimension < 3 implies volume = 0.0
	simplex_implies_minimal_vertex_count: is_simplex implies
											vertex_count = dimension + 1
	simplex_implies_not_extendible: is_simplex implies not extendible
	simplex_implies_not_composite: is_simplex implies not is_composite
	simplex_implies_simplices_is_current:
								is_simplex implies simplices.item (1) = Current
	not_composite_implies_current_cell_is_one: not is_composite implies
												current_cell = 1
	composite_implies_extendible: is_composite implies extendible
	composite_implies_not_simplex: is_composite implies not is_simplex
	extendible_implies_not_simplex: extendible implies not is_simplex
	current_cell_at_least_one: current_cell >= 1
	current_cell_not_more_than_cell_count: current_cell <= cell_count
	intersected_implies_zero_distance: intersected implies distance = 0.0
	non_negative_distance: distance >= 0.0
	contoured_implies_contours: contoured implies contours /= Void

end -- class MESH_CELL

