indexing

	description: "3D cartesian geometric point cells."
	names: "3d", "cartesian", "geometry", "cell", "point"
	representation: "DOUBLE"
	access: "fixed"
	size: "fixed"
	contents: "POINT3"
	date: "$Date: 2006-06-23 10:34:45 -0400 (Fri, 23 Jun 2006) $"
	revision: "$Revision: 226 $"
	author: "Todd Plessel, US EPA/LM, plessel.todd@epa.gov"
	status: "unreviewed", "tested"

class

	POINT_CELL

inherit

	MESH_CELL
		redefine
			default_create, is_simplex, intersect
		end

creation

	make, default_create, make_from_array

feature -- Initialization

	default_create is
			-- Make cell with zero vertex.
		do
			Precursor
			
			create vertex.make (0.0, 0.0, 0.0)
			intersected := False
			distance := 0.0
			contoured := False

			if scalar_derivatives = Void then
				create scalar_derivatives.make (1, 3)
			else
				scalar_derivatives.clear_all
			end

			if vector_derivatives = Void then
				create vector_derivatives.make (1, 9)
			else
				vector_derivatives.clear_all
			end

			if the_simplex = Void then
				create the_simplex.make (1, 1)
				the_simplex.put (Current, 1)
			end

			if intersection = Void then
				create intersection.make (1, 1)
			end

			intersection.put (0.0, 1)

		ensure then
			item_is_zero: item (1).is_zero
			intersected_false: intersected = False
			distance_is_zero: distance = 0.0
			contoured_false: contoured = False
		end

	make (initial_vertex: POINT3) is
			-- Make cell with this initial vertex.
		do
			default_create
			vertex := initial_vertex
		ensure
			vertex_set: item (1).is_equal (initial_vertex)
			intersected_false: intersected = False
			distance_is_zero: distance = 0.0
			contoured_false: contoured = False
		end

	make_from_array (vertices: ARRAY [POINT3]) is
		    -- Make cell with given points.
	    do
		make (vertices.item (1))
	    end

feature -- Access

	item (vertex_number: INTEGER): POINT3 is
			-- Access a vertex.
		do
			Result := vertex
		end

	weight (vertex_number: INTEGER): DOUBLE is
			-- Interpolation weight at given vertex for current parametric
			-- coordinate(s).
			-- (Always 1.)
		do
			Result := 1.0
		end

	parametric_coordinate (coordinate_number: INTEGER): DOUBLE is
			-- Pre-condition is `False'.
		do
		end

	parametric_position: POINT3 is
			-- `Result' is item (1).
		do
			Result := item (1)
		ensure then
			result_is_item_one: Result.is_equal (item (1))
		end
 
	edge (edge_number: INTEGER): LINE is
			-- Pre-condition is false.
		do
		end

	face (face_number: INTEGER): MESH_CELL is
			-- Pre-condition is false.
		do
		end

	nearest_point: POINT3 is
			-- Point nearest to `parametric_position'.
			-- `Result' is `Current.item (1)'
		do
			Result := Current.item (1)
		end

	nearest_edge: LINE is
			-- Pre-condition is false.
		do
		end

	nearest_face: MESH_CELL is
			-- Pre-condition is false.
		do
		end

	simplices: ARRAY [MESH_CELL] is
			-- `Result' is array of one `Current' and is cached.
		do
			Result := the_simplex
		ensure then
			result_is_current: Result.item (1) = Current
		end

	contours: ARRAY [MESH_CELL] is
			-- Cell(s) generated by last successful `contour' operation.
			-- `Result' is array of one `Current' and is cached.
		do
			Result := the_simplex
		ensure then
			result_item_one_is_current: Result.item (1) = Current
		end

	intersection: ARRAY [DOUBLE]
			-- Result of 'intersect' operation: if `intersected' then
			-- Result.item (1) is the parametric coordinate along the ray of
			-- the point of intersection.
			-- (Use RAY.project to convert a parametric coordinate into POINT3.)

feature -- Measurement

	dimension: INTEGER is 0
			-- Spatial dimensionality:
			-- `Result' is 0.

	vertex_count: INTEGER is 1
			-- Number of vertices in cell.
			-- `Result' is 1.

	edge_count: INTEGER is 0
			-- Number of edges in cell.
			-- `Result' is 0.

	face_count: INTEGER is 0
			-- Number of faces in cell.
			-- `Result' is 0.

	centroid: POINT3 is
			-- Location of cell center.
			-- `Result' is `item (1)'.
		do
			Result := item (1)
		end

	perimeter: DOUBLE is 0.0
			-- Sum of edge lengths.
			-- `Result' is 0.

	area: DOUBLE is 0.0
			-- Area of cell faces.
			-- `Result' is 0.

	volume: DOUBLE is 0.0
			-- Volume of cell.
			-- `Result' is 0.

	extent: AXIS_ALIGNED_BOX is
			-- 3d cartesian axis-aligned minimum bounding-box enclosing cell.
			-- `Result' is degenerate.
		do
			create Result.make (item (1), item (1))
		end

feature -- Status report

	is_simplex: BOOLEAN is True
			-- Is the cell a simplex for its dimension?
			-- `Result' is True.

	is_degenerate: BOOLEAN is True
			-- Does the cell have length area or volume that is less than
			-- normal because of coincident vertices?
			-- `Result' is `True'.

feature -- Element change

	put, replace (new_vertex: POINT3; vertex_number: INTEGER) is
			-- Make `new_vertex' `item (1)'.
		do
			vertex := new_vertex
			intersected := False
			distance := 0.0
			contoured := False
		end

	extend (new_vertex: POINT3) is
			-- Pre-condition is `False'.
		do
		end

	set_parametric_coordinate (value: DOUBLE; coordinate_number: INTEGER) is
			-- Pre-condition is `False'.
		do
		end

feature -- Remove

	remove (vertex_number: INTEGER) is
			-- Pre-condition is `False'.
		do
		end

feature -- Basic operations

	contour (interpolant_value: DOUBLE; vertex_data: ARRAY [DOUBLE]) is
			-- If values match, `contoured' is True and
			-- the result is available in `contours' (as Current).
			-- Otherwise `contoured' is False.
		do
			contoured := about_equal (interpolant_value, vertex_data.item (1))
		end

	interpolate_scalar (vertex_data: ARRAY [DOUBLE]): DOUBLE is
			-- Interpolate scalar `vertex_data' to `parametric_position'.
			-- Since points are degenerate the result is the argument.
		do
			Result := vertex_data.item (1)
		ensure then
			result_is_argument_item_1:
				about_equal (Result, vertex_data.item (1))
		end

	interpolate_vector (vertex_data: ARRAY [DOUBLE]): VECTOR3 is
			-- Interpolate 3d vector `vertex_data' to `parametric_position'.
			-- Since points are degenerate the result is the argument.
		do
			Result.make (vertex_data.item (1),
						vertex_data.item (2),
						vertex_data.item (3))
		ensure then
			result_is_argument_item_1:
				about_equal (Result.x, vertex_data.item (1)) and
				about_equal (Result.y, vertex_data.item (2)) and
				about_equal (Result.z, vertex_data.item (3))
		end

	gradient (vertex_data: ARRAY [DOUBLE]): ARRAY [DOUBLE] is
			-- Compute the derivatives in the x-y-z coordinate directions
			-- of scalar or 3D vector vertex data.
			-- The result for scalar vertex data is <ds/dx,ds/dy,ds/dz> and
			-- the result for 3d vector data is:
			-- <dvx/dx,dvx/dy,dvx/dz,dvy/dx,dvy/dy,dvy/dz,dvz/dx,dvz/dy,dvz/dz>.
			-- For point, the (cached) result will contain all zeros.
		do
			if vertex_data.count = 1 then
				Result := scalar_derivatives
			else
				Result := vector_derivatives
			end
		end

	evaluate_position (position: POINT3) is
			-- If `position' equals item (1) then
			--   (1) `intersected' will be True and
			--   (2) `distance' will be zero and
			--   (3) `intersection.item (1)' will be zero.
			-- Else
			--   (1) `intersected' will be False and
			--   (2) `distance' will be the distance from `parametric_position'
			--       to `position'.
		do
			intersected := position.is_equal (vertex)

			if intersected then
				distance := 0.0
				intersection.put (0.0, 1)
			else
				distance := vertex.distance (position)
			end
		end

	intersect (ray: RAY) is
			-- Determine if `ray' intersects.
			-- If intersection occurs then
			--   (1) `intersected' will be True and
			--   (2) `distance' will be zero and
			--   (3) `intersection.item (1)' will be the distance between
			--       `ray.origin' and `item (1)'
			-- Else intersection does not occur and then
			--   (1) `intersected' will be False and
			--   (2) `distance' will be huge_value if `ray' points away from
			--       `item (1)' else `distance' will be equal to the shortest
			--       distance between the infinite ray line and `item (1)'.
		local
			ray_arrow_point: POINT3
		do
			intersection_count := intersection_count + 1
			
			intersected := False
			distance := huge_value

			if ray.origin.is_equal (vertex) then
					-- Ray originates at vertex so it intersects.
				intersection.put (0.0, 1)
				distance := 0.0
				intersected := True
			elseif ray.direction.dot (vertex - ray.origin) > 0.0 then
					-- Ray points towards point so it might intersect.
				ray_arrow_point := ray.origin + ray.direction

				if not ray.origin.is_equal (ray_arrow_point) then
						-- Must check for "degenerate" ray such as
						-- <(inf, 0, 0) <1, 0, 0>> that yields
						-- points that project only onto ray.origin.
					distance :=
						distance_from_point_to_line (vertex, ray.origin,
													ray_arrow_point)

					if distance <= Tolerance then
							-- Close enough to be considered a hit.
						intersection.put (ray.origin.distance (vertex), 1)
						distance := 0.0
						intersected := True
					end
				end
			end
			
			if intersected then
				intersection_hit := intersection_hit + 1
			end
		ensure then
			intersected_implies_intersection_is_consistent:
				intersected implies
					not ray.origin.all_finite or
					not is_finite (intersection.item (1)) or
					ray.project (intersection.item (1)).is_equal
								(parametric_position)
			intersected_implies_intersection_is_point_distance:
				intersected implies
					not is_finite (intersection.item (1)) or
					about_equal (intersection.item (1),
								ray.origin.distance (item (1)))
			not_intersected_implies_distance_is_strictly_positive:
				not intersected implies distance > 0.0
		end

feature {NONE} -- Implementation

	vertex: POINT3
			-- The vertex of the point.

	the_simplex: ARRAY [POINT_CELL]
			-- The cached result of simplicies.

	scalar_derivatives: ARRAY [DOUBLE]
			-- ds_dx, ds_dy, ds_dz of the scalar value. Always all zero.

	vector_derivatives: ARRAY [DOUBLE]
			-- dvx_dx, dvx_dy, dvx_dz, dvy_dx, dvy_dy, dvy_dz, dvz_dx, dvz_dy,
			-- dvz_dz of the vector value. Always all zero.

invariant

	dimension_is_zero: dimension = 0
	weight_is_one: weight (1) = 1.0
	perimeter_is_zero: perimeter = 0.0
	area_is_zero: area = 0.0
	volume_is_zero: volume = 0.0
	parametric_position_is_item: parametric_position.is_equal (item (1))
	simplices_is_current: simplices.item (1) = Current
	contoured_implies_contour_is_current:
		contoured implies contours.item (1) = Current
	centroid_is_item: centroid.is_equal (item (1))
	extent_lower_is_item: extent.lower.is_equal (item (1))
	extent_upper_is_item: extent.upper.is_equal (item (1))
	is_simplex: is_simplex
	is_degenerate: is_degenerate
	current_cell_is_one: current_cell = 1
	the_simplex_exists: the_simplex /= Void
	the_simplex_is_current: the_simplex.item (1) = Current
	scalar_derivatives_exists: scalar_derivatives /= Void
	vector_derivatives_exists: vector_derivatives /= Void
	three_scalar_derivatives: scalar_derivatives.count = 3
	nine_vector_derivatives: vector_derivatives.count = 9
	scalar_derivatives_all_zero: scalar_derivatives.item (1) = 0.0 and
									scalar_derivatives.item (2) = 0.0 and
									scalar_derivatives.item (3) = 0.0
	vector_derivatives_all_zero: vector_derivatives.item (1) = 0.0 and
									vector_derivatives.item (2) = 0.0 and
									vector_derivatives.item (3) = 0.0
									vector_derivatives.item (4) = 0.0
									vector_derivatives.item (5) = 0.0
									vector_derivatives.item (6) = 0.0
									vector_derivatives.item (7) = 0.0
									vector_derivatives.item (8) = 0.0
									vector_derivatives.item (9) = 0.0
	intersection_exists: intersection /= Void
	intersection_count_is_one: intersection.count = 1

end -- class POINT_CELL

