indexing

	description: "3D cartesian polylines."
	names: "3d", "cartesian", "polyline", "line", "curve", "polygon"
	representation: "DOUBLE"
	access: "fixed"
	size: "fixed"
	contents: "POINT3"
	date: "$Date: 2006-06-23 10:34:45 -0400 (Fri, 23 Jun 2006) $"
	revision: "$Revision: 226 $"
	author: "Todd Plessel, US EPA/LM, plessel.todd@epa.gov"
	status: "unreviewed", "tested", "partial", "unimplemented: intersect"

class

	POLYLINE

inherit

	MESH_CELL
		redefine
			default_create, extendible, intersect
		end

creation

	default_create, make, make_simple, make_from_array

feature -- Initialization

	default_create is
			-- Make degenerate polyline with two zero vertices.
		local
			zero_vertex: POINT3
		do
			Precursor
			
			intersected := False
			distance := 0.0
			contoured := False
			the_parametric_coordinate := 0.0
			
			create zero_vertex
			
			if vertices = Void then
				create vertices.make (1, 2)
			else
				vertices.make (1, 2)
			end
			vertices.put (zero_vertex, 1)
			vertices.put (zero_vertex, 2)

			create the_nearest_point

			if the_contour_points = Void then
				create the_contour_points.make (1, 1)
			else
				the_contour_points.clear_all
			end

			if the_edge = Void then
				create the_edge
			else
				the_edge.default_create
			end

			if the_nearest_edge = Void then
				create the_nearest_edge
			else
				the_nearest_edge.default_create
			end

			the_simplices := Void

			if cached_intersection = Void then
				create cached_intersection.make (1, 2)
			else
				cached_intersection.make (1, 2)
			end

			cached_intersection.put (0.0, 1)
			cached_intersection.put (1.0, 2)
		ensure then
			first_vertex_is_zero: item (1).is_zero
			second_vertex_is_zero: item (2).is_zero
			the_parametric_coordinate_is_zero: the_parametric_coordinate = 0.0
			intersected_false: intersected = False
			distance_is_zero: distance = 0.0
			contoured_false: contoured = False
			no_cached_simplices: the_simplices = Void
		end

	make_simple (initial_first_vertex, initial_second_vertex: POINT3) is
			-- Make polyline with these two initial verticies.
		do
			default_create
			put (initial_first_vertex, 1)
			put (initial_second_vertex, 2)
			create the_nearest_point.make (initial_first_vertex.x,
									initial_first_vertex.y,
									initial_first_vertex.z)			
			the_edge.make (initial_first_vertex, initial_second_vertex)
			the_nearest_edge.make (initial_first_vertex, initial_second_vertex)
		ensure
			first_vertex_set: item (1).is_equal (initial_first_vertex)
			second_vertex_set: item (2).is_equal (initial_second_vertex)
			the_parametric_coordinate_is_zero: the_parametric_coordinate = 0.0
			intersected_false: intersected = False
			distance_is_zero: distance = 0.0
			contoured_false: contoured = False
		end

	make, make_from_array (vertex_list: ARRAY [POINT3]) is
			-- Create polyline with these vertices.
---		require else
---			argument_has_at_least_two_vertices: vertex_list.count >= 2
--- FIX: re-enambe assertion after:
--- (1) fixing mal.xdr and (2) fixing checking in client.
--- For now, just strengthen the implementation.
		do
			if vertex_list.count = 1 then
					-- Degenerate one-segment polyline.
				make_simple (vertex_list.item (1), vertex_list.item (1))
			else
					-- Non-degenerate polyline.
				make_simple (vertex_list.item (1), vertex_list.item (2))

				if vertex_list.count > 2 then
						-- Polyline with more than one segment.
					vertices := vertex_list.twin
				end
			end
		ensure then
---			second_vertex_set: item (2).is_equal (vertex_list.item (2))
			the_parametric_coordinate_is_zero: the_parametric_coordinate = 0.0
			intersected_false: intersected = False
			distance_is_zero: distance = 0.0
			contoured_false: contoured = False
		end

feature -- Access

	item (vertex_number: INTEGER): POINT3 is
			-- Access a vertex.
		do
			Result := vertices.item (vertex_number)
		end

	weight (vertex_number: INTEGER): DOUBLE is
			-- Interpolation weight at given vertex for current parametric
			-- coordinate. Zero if `parametric_position' does not lie within
			-- a line segment containing this vertex number.
		local
			first_bracketed_vertex_index: INTEGER
				-- First vertex number that brackets the `parametric_position'.
		do
			first_bracketed_vertex_index := lower_parametric_vertex_index

			if vertex_number = first_bracketed_vertex_index then
				Result := 1.0 - the_local_parametric_coordinate
			elseif vertex_number = first_bracketed_vertex_index + 1 then
				Result := the_local_parametric_coordinate
			end
		end

	parametric_coordinate (coordinate_number: INTEGER): DOUBLE is
			-- Parametric coordinate for interpolation along the polyline.
		do
			Result := the_parametric_coordinate
		end

	parametric_position: POINT3 is
			-- XYZ coordinates of position at current parametric coordinate
			-- along the polyline.
		do
			Result := nearest_edge.parametric_position
		end

	edge (edge_number: INTEGER): LINE is
			-- Edge of a cell.
			-- `Result' is cached.
		do
			the_edge.put (item (edge_number), 1)
			the_edge.put (item (edge_number + 1), 2)
			Result := the_edge
		ensure then
			result_cached: Result = the_edge
		end

	face (face_number: INTEGER): MESH_CELL is
			-- Pre-condition is false.
		do
		end

	nearest_point: POINT3 is
			-- Point nearest to `parametric_position'.
		local
			first_bracketed_vertex_index: INTEGER
				-- First vertex number that brackets the `parametric_position'.
		do
			first_bracketed_vertex_index := lower_parametric_vertex_index

			if the_local_parametric_coordinate <= 0.5 then
				Result := item (first_bracketed_vertex_index)
			else
				Result := item (first_bracketed_vertex_index + 1)
			end
		end

	nearest_edge: LINE is
			-- Edge nearest to `parametric_position'.
			-- `Result' is cached.
		local
			first_bracketed_vertex_index: INTEGER
				-- First vertex number that brackets the `parametric_position'.
		do
			first_bracketed_vertex_index := lower_parametric_vertex_index
			the_nearest_edge.put (item (first_bracketed_vertex_index), 1)
			the_nearest_edge.put (item (first_bracketed_vertex_index + 1), 2)
			the_nearest_edge.set_parametric_coordinate
				(the_local_parametric_coordinate, 1)
			Result := the_nearest_edge
		ensure then
			result_cached: Result = the_nearest_edge
		end

	nearest_face: MESH_CELL is
			-- Pre-condition is false.
		do
		end

	simplices: ARRAY [MESH_CELL] is
			-- `Result' is cached (possibly recomputed) sequence of line cells.
		local
			vertex_index: INTEGER
					-- Index on successive vertices of polyline.

			first_vertex: POINT3
					-- First vertex of current segment.

			second_vertex: POINT3
					-- Second vertex of current segment.

			line: LINE
					-- Current line segment.

			lines: ARRAY [LINE]
					-- Temporary result (for exception-safety).
		do
			if the_simplices = Void then
					-- Must (re)-compute simplices (regenerating the cache).
				create lines.make (1, edge_count)

				from
					vertex_index := 1
				invariant
					vertex_index_large_enough: vertex_index >= 1
					vertex_index_small_enough: vertex_index <= 1 + vertex_count
				variant
					1 + vertex_count - vertex_index
				until
					vertex_index = vertex_count
				loop
					first_vertex := item (vertex_index)
					second_vertex := item (vertex_index + 1)
					create line.make (first_vertex, second_vertex)
					lines.put (line, vertex_index)
					vertex_index := vertex_index + 1
				end

				the_simplices := lines
			end

			Result := the_simplices
		ensure then
			result_is_cached: Result = the_simplices
			result_count_is_edge_count: Result.count = edge_count
		end

	contours: ARRAY [MESH_CELL] is
			-- Cell(s) generated by last successful `contour' operation.
			-- `Result' is a cached sequence of point cells.
		do
			Result := the_contour_points
		end

feature -- Measurement

	dimension: INTEGER is 1
			-- Spatial dimensionality:
			-- `Result' is 1.

	vertex_count: INTEGER is
			-- Number of vertices in polyline.
			-- `Result' is at least 2.
		do
			Result := vertices.count
		end

	edge_count: INTEGER is
			-- Number of edges in polyline.
			-- `Result' is 1 less than `vertex_count'.
		do
			Result := vertex_count - 1
		end

	face_count: INTEGER is 0
			-- Number of faces in polyline.
			-- `Result' is 0.

	centroid: POINT3 is
			-- Location of polyline center.
			--| Average of vertices.
		local
			vertex_index: INTEGER
			portion, x, y, z, x_min, x_max, y_min, y_max, z_min, z_max: DOUBLE
			vertex: POINT3
		do
			from
				portion := 1.0 / vertex_count
				vertex := item (1)
				x_min := vertex.x
				y_min := vertex.y
				z_min := vertex.z
				x_max := x_min
				y_max := y_min
				z_max := z_min
				Result.make (x_min * portion, y_min * portion, z_min * portion)
				vertex_index := 2
			invariant
				vertex_index_large_enough: vertex_index >= 2
				vertex_index_small_enough: vertex_index <= 1 + vertex_count
			variant
				1 + vertex_count - vertex_index
			until
				vertex_index > vertex_count
			loop
				vertex := item (vertex_index)
				x := vertex.x
				y := vertex.y
				z := vertex.z

				if x < x_min then
					x_min := x
				elseif x > x_max then
					x_max := x
				end

				if y < y_min then
					y_min := y
				elseif y > y_max then
					y_max := y
				end

				if z < z_min then
					z_min := z
				elseif z > z_max then
					z_max := z
				end

				Result.make (sum (Result.x, x * portion),
							sum (Result.y, y * portion),
							sum (Result.z, z * portion))
				vertex_index := vertex_index + 1
			end

			Result.make (clamp_to_range (Result.x, x_min, x_max),
							clamp_to_range (Result.y, y_min, y_max),
							clamp_to_range (Result.z, z_min, z_max))
		end

	perimeter: DOUBLE is
			-- Sum of edge lengths.
		local
			vertex_index: INTEGER
		do
			from
				vertex_index := 1
			invariant
				vertex_index_large_enough: vertex_index >= 1
				vertex_index_small_enough: vertex_index <= 1 + vertex_count
				result_non_negative: Result >= 0.0
			variant
				1 + vertex_count - vertex_index
			until
				vertex_index = vertex_count
			loop
				Result := Result +
					item (vertex_index).distance (item (vertex_index + 1))
				vertex_index := vertex_index + 1
			end
		end

	area: DOUBLE is 0.0
			-- Area of cell faces.
			-- `Result' is 0.

	volume: DOUBLE is 0.0
			-- Volume of cell.
			-- `Result' is 0.

	extent: AXIS_ALIGNED_BOX is
			-- 3d cartesian axis-aligned minimum bounding-box enclosing cell.
		local
			vertex_index: INTEGER
		do
			from
				create Result.make (item (1), item (2))
				vertex_index := 3
			invariant
				vertex_index_large_enough: vertex_index >= 3
				vertex_index_small_enough: vertex_index <= 1 + vertex_count
			variant
				1 + vertex_count - vertex_index
			until
				vertex_index > vertex_count
			loop
				Result.grow (item (vertex_index))
				vertex_index := vertex_index + 1
			end
		end

	intersection: ARRAY [DOUBLE] is
			-- Result of 'intersect' operation: if `intersected' then
			-- Result.item (1) is the parametric coordinate along the ray of
			-- the first/nearest point of intersection and if the ray intersects
			-- more than once then Result.count is the number of intersections
			-- and Result.item (i) is the parametric coordinate along the ray of
			-- each successive intersection.
			-- (Use RAY.project to convert a parametric coordinate into POINT3.)
		do
			Result := cached_intersection
		ensure then
			result_cached: Result = cached_intersection
		end

feature -- Status report

	is_degenerate: BOOLEAN is
			-- Is the polyline just two coincident points?
		do
			Result := vertex_count = 2 and then item (1).is_equal (item (2))
		end

	extendible: BOOLEAN is True
			-- May vertices be added?
			-- `Result' is `True'.

feature -- Element change

	put, replace (new_vertex: POINT3; vertex_number: INTEGER) is
			-- Make `new_vertex' `item (vertex_number)'.
		do
			vertices.put (new_vertex, vertex_number)
			the_parametric_coordinate := 0.0
			intersected := False
			distance := 0.0
			contoured := False
			the_simplices := Void
			the_nearest_point.make (item (1).x, item (1).y, item (1).z)
		ensure then
			simplices_invalidated: the_simplices = Void
		end

	extend (new_vertex: POINT3) is
			-- Append `new_vertex'.
		local
			old_count: INTEGER
		do
			old_count := vertices.count

			vertices.force (new_vertex, old_count + 1)

			the_parametric_coordinate := 0.0
			intersected := False
			distance := 0.0
			contoured := False
			the_simplices := Void
			the_nearest_point.make (item (1).x, item (1).y, item (1).z)
		ensure then
			simplices_invalidated: the_simplices = Void
		end

	set_parametric_coordinate (value: DOUBLE; coordinate_number: INTEGER) is
			-- Set the given parametric coordinate for subsequent interpolation.
		do
			the_parametric_coordinate := value
		end

feature -- Removal

	remove (vertex_number: INTEGER) is
			-- Remove numbered vertex.
		local
			i: INTEGER
			vertices_copy: ARRAY [POINT3]
		do
			from
				i := 1 
			invariant
				i_at_least_one: i >= 1
				i_small_enough: i <= 1 + vertex_number - 1
			variant
				i_increasing: 1 + vertex_number - 1 - i
			until
				i >= vertex_number - 1
			loop
				vertices_copy.put (vertices.item (i), i)
				i := i + 1
			end
			
			from
				i := vertex_number + 1 
			invariant
				i_at_least_one: i >= vertex_number + 1
				i_small_enough: i <= vertex_number + 1 + vertex_count
			variant
				i_increasing: vertex_number + 1 + vertex_count - i
			until
				i >= vertex_count
			loop
				vertices_copy.put (vertices.item (i), i - 1)
				i := i + 1
			end

			vertices := vertices_copy
			vertices_copy := Void

			the_parametric_coordinate := 0.0
			intersected := False
			distance := 0.0
			contoured := False
			the_simplices := Void
			the_nearest_point.make (item (1).x, item (1).y, item (1).z)
		end

feature -- Basic operations

	contour (interpolant_value: DOUBLE; vertex_data: ARRAY [DOUBLE]) is
			-- Generate contour point(s) whose vertices approximate the
			-- interpolated position(s) of `value' along the polyline assuming
			-- the given scalar values at each vertex.
			-- If the interpolant_value is along the line, `contoured' is `True'
			-- and the result is available as `contours'.
			-- Otherwise `contoured' is `False'.
		local
			vertex_index: INTEGER
				-- Index on successive vertices of polyline.

			first_vertex: POINT3
				-- First vertex of current segment.

			second_vertex: POINT3
				-- Second vertex of current segment.

			first_vertex_value: DOUBLE
				-- Scalar value at first vertex of current segment.

			second_vertex_value: DOUBLE
				-- Scalar value at second vertex of current segment.

			found: BOOLEAN
				-- Have we found a contour point on the current segment?

			contour_count: INTEGER
				-- Number of contour points generated.

			contour_point: POINT_CELL
				-- The contour point of current segment (if exists).

			denominator: DOUBLE
				-- Of normalization expression.

			parameter: DOUBLE
				-- Parametric coordinate of `position' along the line.

			position: POINT3
				-- Position of `interpolant_value' along the line.

			temp_result: DS_ARRAYED_LIST [POINT_CELL]
				-- Temporary result to copy to `contour_points'.
		do
			from
				contoured := False
				create temp_result.make (0)
				vertex_index := 1
			invariant
				vertex_index_large_enough: vertex_index >= 1
				vertex_index_small_enough: vertex_index <= 1 + vertex_count
			variant
				1 + vertex_count - vertex_index
			until
				vertex_index = vertex_count
			loop
				found := False
				first_vertex_value := vertex_data.item (vertex_index)
				second_vertex_value := vertex_data.item (vertex_index + 1)
				denominator := difference (second_vertex_value,
											first_vertex_value)

				if denominator = 0.0 then
						-- Current segment is degenerate.

					found := about_equal (interpolant_value, first_vertex_value)

					if found then
							-- Choose midpoint of segment as position.
						first_vertex := item (vertex_index)
						second_vertex := item (vertex_index + 1)
						position := first_vertex.midpoint (second_vertex)
					end
				else
					parameter := quotient (difference (interpolant_value,
														first_vertex_value),
											denominator)

					found := within_range (parameter, 0.0, 1.0)

					if found then
							-- Interpolant lies within this segment.
						first_vertex := item (vertex_index)
						second_vertex := item (vertex_index + 1)
						position := first_vertex.interpolate (second_vertex,
																parameter)
					end
				end

				if found then
						-- Create and append a new contour point.
					create contour_point.make (position)
					temp_result.force_last (contour_point)
					contour_count := contour_count + 1
				end

				vertex_index := vertex_index + 1
			end

				-- Copy temp_result to contour_points.
			from
				vertex_index := 1
				the_contour_points.make (1, temp_result.count)
				temp_result.start
			invariant
				index_large_enough: vertex_index >= 1
				index_small_enough: vertex_index <= 1 + temp_result.count
			variant
				index_increasing: 1 + temp_result.count - vertex_index
			until
				vertex_index > temp_result.count
			loop
				the_contour_points.put (temp_result.item_for_iteration, vertex_index)
				vertex_index := vertex_index + 1
				temp_result.forth
			end

			contoured := the_contour_points.count > 0
		end

	interpolate_scalar (vertex_data: ARRAY [DOUBLE]): DOUBLE is
			-- Interpolate scalar `vertex_data' to `parametric_position'.
		local
			first_bracketed_vertex_index: INTEGER
				-- First vertex number that brackets the `parametric_position'.
			bracketed_edge: LINE
				-- Line bracketing parametric_position.
			edge_data: ARRAY [DOUBLE]
				-- Subset of vertex_data that is for bracketed_edge.
		do
			first_bracketed_vertex_index := lower_parametric_vertex_index
			create edge_data.make (1, 2)
			edge_data.put (vertex_data.item (first_bracketed_vertex_index), 1)
			edge_data.put (vertex_data.item (first_bracketed_vertex_index + 1), 2)
			create bracketed_edge.make (item (first_bracketed_vertex_index),
										item (first_bracketed_vertex_index + 1))
			bracketed_edge.set_parametric_coordinate
				(the_local_parametric_coordinate, 1)
			Result := bracketed_edge.interpolate_scalar (edge_data)
		end

	interpolate_vector (vertex_data: ARRAY [DOUBLE]): VECTOR3 is
			-- Interpolate vector `vertex_data' to `parametric_position'.
		local
			first_bracketed_vertex_index: INTEGER
				-- First vertex number that brackets the `parametric_position'.
			bracketed_edge: LINE
				-- Line bracketing parametric_position.
			edge_data: ARRAY [DOUBLE]
				-- Subset of vertex_data that is for bracketed_edge.
		do
			first_bracketed_vertex_index := lower_parametric_vertex_index
			create edge_data.make (1, 2 * 3)
			edge_data.put (vertex_data.item (first_bracketed_vertex_index), 1)
			edge_data.put (vertex_data.item (first_bracketed_vertex_index + 1), 2)
			edge_data.put (vertex_data.item (first_bracketed_vertex_index + 2), 3)
			edge_data.put (vertex_data.item (first_bracketed_vertex_index + 3), 4)
			edge_data.put (vertex_data.item (first_bracketed_vertex_index + 4), 5)
			edge_data.put (vertex_data.item (first_bracketed_vertex_index + 5), 6)
			create bracketed_edge.make (item (first_bracketed_vertex_index),
										item (first_bracketed_vertex_index + 1))
			bracketed_edge.set_parametric_coordinate
				(the_local_parametric_coordinate, 1)
			Result := bracketed_edge.interpolate_vector (edge_data)
		end

	gradient (vertex_data: ARRAY [DOUBLE]): ARRAY [DOUBLE] is
			-- Compute the derivatives in the x-y-z coordinate directions
			-- of scalar or 3D vector vertex data.
			-- The result for scalar vertex data is (for each line segment):
			-- <ds/dx,ds/dy,ds/dz> and the result for 3d vector data (v) is
			-- (for each line segment):
			-- <dvx/dx,dvx/dy,dvx/dz,dvy/dx,dvy/dy,dvy/dz,dvz/dx,dvz/dy,dvz/dz>.
			-- Result values will be zero for degenerate cell dimensions.
		local
			component_count: INTEGER
					-- Number of components in gradient.
					-- 3 if vertex_data is scalar or 9 if vertex_data is vector.

			edge_index: INTEGER
					-- Index on successive edges (line segments) of polyline.

			offset: INTEGER
				-- Offset into vertex_data of the current edge.

			vertex_data_values_per_edge: INTEGER
				-- Number of values per vertex per edge.

			edge_vertex_data, e: ARRAY [DOUBLE]
			-- Subset of vertex_data for each successive edge.

			i, j: INTEGER
		do
			if vertex_data.count = vertex_count then
				component_count := 3
				vertex_data_values_per_edge := 2
			else
				component_count := 9
				vertex_data_values_per_edge := 6
			end

			create Result.make (1, edge_count * component_count)
			create edge_vertex_data.make (1, vertex_data_values_per_edge)

			from
				edge_index := 1
				offset := 1
			invariant
				edge_index_large_enough: edge_index >= 1
				edge_index_small_enough: edge_index <= 1 + edge_count
			variant
				1 + edge_count - edge_index
			until
				edge_index > edge_count
			loop
				from
					i := offset
					j := 1
				invariant
					i_at_least_one: i >= offset
					i_small_enough: i <= offset + vertex_data_values_per_edge
				variant
					i_increasing: offset + vertex_data_values_per_edge - i
				until
					i >= vertex_data_values_per_edge
				loop
					edge_vertex_data.put (vertex_data.item (i), j)
					j := j + 1
					i := i + 1
				end

				from
					i := 1
					j := offset
					e := edge (edge_index).gradient (edge_vertex_data)
				invariant
					i_at_least_one: i >= 1
					i_small_enough: i <= 1 + component_count
				variant
					i_increasing: 1 + component_count - i
				until
					i >= component_count
				loop
					Result.put (e.item (i), j)
					j := j + 1
					i := i + 1
				end
				offset := offset + component_count
				edge_index := edge_index + 1
			end
		end

	evaluate_position (position: POINT3) is
			-- Evaluate the position with respect to the cell geometry.
			-- If `position' is within the cell then
			--	 (1) `intersected' will be True and
			--	 (2) `parametric_coordinate (1)' will be set to interpolate to
			--		 `position' and
			--	 (3) `distance' will be zero and
			--	 (4) `intersection.count' = 1
			-- Else `position' is outside the cell and then
			--	 (1) `intersected' will be False and
			--	 (2) `parametric_coordinate (1)' will be set to the point on
			--		 the cell that is closest (by Euclidean distance) to
			--		 `position' and
			--	 (3) `distance' will report the distance from
			--		 `parametric_position' to `position'.
		local
			vertex_index: INTEGER
				-- Index on each vertex of polyline (in order).

			closest_line: LINE
				-- (First) closest line cell to `position'.

			line: LINE
				-- Current line cell being checked (in order).

			swap_line: LINE
				-- Temporary reference for swapping lines.

			line_perimeter: DOUBLE
				-- Perimeter of line.

			total_perimeter: DOUBLE
				-- The total length of the polyline.

			partial_perimeter: DOUBLE
				-- The length of the polyline up to `position'.
		do
			intersected := False
			distance := huge_value
			the_parametric_coordinate := 0.0
			cached_intersection.make (1, 1)
			cached_intersection.put (0.0, 1)

				-- For each edge, accumulate the perimeter and check for the
				-- edge that is closest to position and use this to set the
				-- parametric coordinate.

			create closest_line.make (item (1), item (2))
			closest_line.evaluate_position (position)
			total_perimeter := closest_line.perimeter
			partial_perimeter := product (total_perimeter,
										closest_line.parametric_coordinate (1))
			create line.default_create

			from
				vertex_index := 2
			invariant
				vertex_index_large_enough: vertex_index >= 2
				vertex_index_small_enough: vertex_index <= 1 + vertex_count
				total_perimeter_non_negative: total_perimeter >= 0.0
				partial_perimeter_non_negative: partial_perimeter >= 0.0
				partial_perimeter_not_more_than_total:
					partial_perimeter <= total_perimeter
			variant
				1 + vertex_count - vertex_index
			until
				vertex_index = vertex_count
			loop
				line.make (item (vertex_index), item (vertex_index + 1))
				line_perimeter := line.perimeter

				if not closest_line.intersected then
					line.evaluate_position (position)

					if line.distance < closest_line.distance then
						partial_perimeter := total_perimeter +
								product (line.parametric_coordinate (1),
										line_perimeter)
						swap_line := closest_line
						closest_line := line
						line := swap_line
					end
				end

				total_perimeter := total_perimeter + line_perimeter
				vertex_index := vertex_index + 1
			end

			if total_perimeter /= 0.0 then
				the_parametric_coordinate :=
					quotient (partial_perimeter, total_perimeter)
			else
				the_parametric_coordinate := 0.0
			end

			the_local_parametric_coordinate :=
				closest_line.parametric_coordinate (1)
			distance := closest_line.distance
			intersected := closest_line.intersected
		end

	intersect (ray: RAY) is
			-- Calculate point of intersection with ray.
			-- If intersection occurs then
			--	 (1) `intersected' will be True and
			--	 (2) `distance' will be zero and
			--	 (3) `parametric_coordinate (1)' (and `parametric_position')
			--		 will be the first/nearest point of intersection and
			--	 (4) `intersection.item (1)' will be the parametric coordinate
			--		 along `ray' of the first/nearest point of intersection and
			--	 (5) if the ray intersects the cell more than once then
			--		   (a) `intersection.count' is number of intersections and
			--		   (b) `intersection.item (i)' will be the parametric
			--			   coordinate of each intersection along `ray'.
			-- Else no intersection occurs and then
			--	 (1) `intersected' will be False and
			--	 (2) `distance' will be huge_value if `ray' points away from
			--		 the cell else `distance' will be equal to the shortest
			--		 skew line distance.
			--	 (3) `parametric_coordinate (1)' will be the coordinate of the
			--		 point of nearest approach of the ray to the line.
		local
			vertex_index: INTEGER
				-- Index on each vertex of polyline (in order).

			closest_line: LINE
				-- (First) closest line cell to ray.

			line: LINE
				-- Current line cell being checked (in order).

			swap_line: LINE
				-- Temporary reference for swapping lines.

			total_perimeter: DOUBLE
				-- The total length of the polyline.

			partial_perimeter: DOUBLE
				-- The length of the polyline up to `position'.

			intersection_counter: INTEGER
				-- Number of intersections recorded.

			intersections: ARRAY [DOUBLE]
			-- Temporary result to copy to cached_intersection.

			i: INTEGER
		do

			--- This routine uses LINE.interpolate which is not implemented correctly.
			check incorrect_implementation: False end

			intersection_count := intersection_count + 1
			
			intersected := False
			distance := 0.0
			the_parametric_coordinate := 0.0
			create intersections.make (1, 0)

				-- For each edge, accumulate the perimeter and check for the
				-- edge that is closest to position and use this to set the
				-- parametric coordinate. Meanwhile, accumulate intersections.

			create closest_line.make (item (1), item (2))
			closest_line.intersect (ray)

			if closest_line.intersected then
					-- Start with the first intersections.
				intersection_counter := closest_line.intersection.count
				intersections := closest_line.intersection.twin
			end

			total_perimeter := closest_line.perimeter
			partial_perimeter := product (total_perimeter,
										closest_line.parametric_coordinate (1))
			create line.default_create

			from
				vertex_index := 2
			invariant
				vertex_index_large_enough: vertex_index >= 2
				vertex_index_small_enough: vertex_index <= 1 + vertex_count
				total_perimeter_non_negative: total_perimeter >= 0.0
				partial_perimeter_non_negative: partial_perimeter >= 0.0
				partial_perimeter_not_more_than_total:
					partial_perimeter <= total_perimeter
			variant
				1 + vertex_count - vertex_index
			until
				vertex_index = vertex_count
			loop
				line.make (item (vertex_index), item (vertex_index + 1))
				line.intersect (ray)

				if line.intersected then
						-- Add the next intersection(s).
					intersections.resize (1, intersection_counter + line.intersection.count)

					from
						i := 1 
					invariant
						i_at_least_one: i >= 1
						i_small_enough: i <= 1 + line.intersection.count
					variant
						i_increasing: 1 + line.intersection.count - i
					until
						i >= line.intersection.count
					loop
						intersections.put (line.intersection.item (i), intersection_counter + i)
						i := i + 1
					end
					intersection_counter :=
						intersection_counter + line.intersection.count
				end

				if (line.distance < closest_line.distance) or else
					(line.distance = closest_line.distance and then
					line.parametric_position.distance_squared (ray.origin) <
					closest_line.parametric_position.distance_squared
						(ray.origin) ) then
						-- Line is first/nearest to ray so update closest.
						-- "First/nearest" means, if two line segments are
						-- equidistant to the ray, choose the one that will
						-- encounter/approach the ray first. This will be the
						-- one that is nearest to the ray position.
					swap_line := closest_line
					closest_line := line
					line := swap_line
					partial_perimeter := total_perimeter +
								product (closest_line.parametric_coordinate (1),
										closest_line.perimeter)
				end

				total_perimeter := total_perimeter + line.perimeter
				vertex_index := vertex_index + 1
			end

			if total_perimeter /= 0.0 then
				the_parametric_coordinate :=
					quotient (partial_perimeter, total_perimeter)
			else
				the_parametric_coordinate := 0.0
			end

			cached_intersection := intersections.twin

			the_local_parametric_coordinate :=
				closest_line.parametric_coordinate (1)
			distance := closest_line.distance
			intersected := closest_line.intersected
		
			if intersected then
				intersection_hit := intersection_hit + 1
			end
		ensure then
			intersected_implies_intersection_is_consistent:
				intersected implies
					not ray.origin.all_finite or
					not is_finite (intersection.item (1)) or
					ray.project (intersection.item (1)).is_equal
								(parametric_position)
			not_intersected_implies_distance_is_strictly_positive:
				not intersected implies distance > 0.0
		end

feature {NONE} -- Implementation

	lower_parametric_vertex_index: INTEGER is
			-- Compute the number of the first vertex in the pair of vertices
			-- that bracket the parametric_position.
		local
			vertex_index: INTEGER
			current_segment_length: DOUBLE
				-- Length of current line segment
				-- [vertex_index, vertex_index + 1]
			partial_perimeter: DOUBLE
				-- Length of polyline up to vertex_index.
			parametric_perimeter: DOUBLE
				-- Length of polyline up to parametric coordinate
				-- where parametric_coordinate (1) = 1.0 implies
				-- (total) perimeter.
		do
			if parametric_coordinate (1) = 0.0 then
				the_local_parametric_coordinate := 0.0
				Result := 1
			elseif parametric_coordinate (1) = 1.0 then
				the_local_parametric_coordinate := 1.0
				Result := vertex_count - 1
			else
					-- Must compute entire perimeter then loop over vertices
					-- and compute partial_perimeter until a bracketing pair
					-- of vertices is found.
				from
					vertex_index := 1
					parametric_perimeter := perimeter * parametric_coordinate (1)
				invariant
					vertex_index_large_enough: vertex_index >= 1
					vertex_index_small_enough: vertex_index <= 1 + vertex_count
					current_segment_length_non_negative:
						current_segment_length >= 0.0
					parametric_perimeter_non_negative:
						parametric_perimeter >= 0.0
					parametric_perimeter_not_more_than_perimeter:
						parametric_perimeter <= perimeter
					-- parametric_perimeter_constant
					partial_perimeter_non_negative: partial_perimeter >= 0.0
					partial_parametric_perimeter_not_more_than_perimeter:
						partial_perimeter <= perimeter
				variant
					1 + vertex_count - vertex_index
				until
					partial_perimeter >= parametric_perimeter or else
						vertex_index = vertex_count
				loop
					current_segment_length :=
						item (vertex_index).distance (item (vertex_index + 1))
					partial_perimeter := partial_perimeter +
						current_segment_length
					vertex_index := vertex_index + 1
				end

					check
						partial_perimeter_large_enough:
							partial_perimeter >= parametric_perimeter
						partial_perimeter_small_enough:
							partial_perimeter <= perimeter
						current_segment_length_large_enough:
							current_segment_length >= 0.0
						current_segment_length_small_enough:
							current_segment_length <= partial_perimeter
					end

				Result := vertex_index - 1
						-- Also compute the_local_parametric_coordinate.

				if current_segment_length = 0.0 then
					the_local_parametric_coordinate := 0.0
				else
						-- Normalized coordinate of `parametric_position'
						-- along bracketing segment.

					the_local_parametric_coordinate :=
						quotient (difference (parametric_perimeter,
												difference (partial_perimeter,
														current_segment_length)),
								current_segment_length)

						check
							reasonable_local_parametric_coordinate:
							within_range (the_local_parametric_coordinate,
											0.0, 1.0)
						end
				end
			end
		ensure
			result_large_enough: Result >= 1
			result_small_enough: Result < vertex_count
		end

	vertices: ARRAY [POINT3]
		-- Representation of points in polyline.

	the_simplices: ARRAY [LINE]
			-- The result of `simplicies'.

	the_contour_points: ARRAY [POINT_CELL]
			-- The result of a successful `contour' operation.

	the_parametric_coordinate: DOUBLE
			-- The parametric coordinate along the line.

	the_edge: LINE
			-- The cached result of `edge'.

	the_nearest_point: POINT3
			-- The cached result of `nearest_point'.

	the_nearest_edge: LINE
			-- The cached result of `nearest_edge'.

	the_local_parametric_coordinate: DOUBLE
			-- The percentage from `lower_parametric_vertex_index' towards
			-- `lower_parametric_vertex_index' + 1 that the
			-- `parametric_position' lies (or zero).

	cached_intersection: ARRAY [DOUBLE]
			-- Result of `intersection': count is >= 1.

invariant

	dimension_is_one: dimension = 1
	vertex_count_at_least_two: vertex_count >= 2
	edge_count_is_vertex_count_minus_one: edge_count = vertex_count - 1
	area_is_zero: area = 0.0
	volume_is_zero: volume = 0.0
	contoured_implies_point_contour:
		contoured implies contours.item (1).vertex_count = 1
	not_is_simplex: not is_simplex
	degenerate_implies_two_coincident_vertices:
		is_degenerate implies vertex_count = 2 and item (1).is_equal (item (2))
	current_cell_is_one: current_cell = 1
	the_edge_exists: the_edge /= Void
	the_nearest_edge_exists: the_nearest_edge /= Void
	the_local_parametric_coordinate_large_enough:
		the_local_parametric_coordinate >= 0.0
	the_local_parametric_coordinate_small_enough:
		the_local_parametric_coordinate <= 1.0

end -- class POLYLINE

