

indexing
	description: "Abstract interface to graphics routines. Based on OpenGL."
	patterns: "none"
	date: "$Date: 2007-01-09 16:27:59 -0500 (Tue, 09 Jan 2007) $"
	revision: "$Revision: 278 $"
	status: "unreviewed"


deferred class 
	RENDERER
	
inherit
	
	RENDER_CONSTANTS
	
	SINGLETON_ACCESSOR
		redefine
			singleton
		end

feature -- Initialization
	
	initialize is
			-- Initialize a renderer
		do
			show_extents := false
			singleton.next_id
			identifier := singleton.last_id
		end

	initialize_extensions is
			-- Initialize any extension outside of the core renderer
		require
			valid_context: context_enabled
		deferred
		ensure
			extensions_initialized: extensions_initialized
		end 
	
feature -- Access
	
	global_lights: SCENE_LIGHTS
			-- Defines all of the  lights that are used for the entire 
			-- scene. 
	
	viewing_camera: SCENE_CAMERA
			-- The camera 
	
	use_display_lists: BOOLEAN
			-- Use display lists if possible
	
	identifier: INTEGER
			-- A guaranteed unique value to identify this instance of 
			-- the renderer
	
	show_extents: BOOLEAN
			-- Should we show the extents of an object
	
	draw_style: INTEGER is
			-- What is the current drawing style?
		deferred
		end

	extensions_initialized: BOOLEAN
			-- Were any extensions initialized?
	
feature -- Element change
	
	set_show_extents (draw_extents: BOOLEAN) is
			-- Set `show_extents' to `draw_extents'
		do
			show_extents := draw_extents
		ensure
			show_extents_set: show_extents = draw_extents
		end
	
	set_global_lights (new_lights: SCENE_LIGHTS) is
			-- Set `global_lights' to `new_lights'
		require
			new_lights_not_void: new_lights /= Void
		do
			global_lights := new_lights
		ensure
			global_lights_set: global_lights = new_lights
		end
	
	set_viewing_camera (new_camera: SCENE_CAMERA) is
			-- Set `viewing_camera' to `new_camera'
		require 
			new_camera_not_void: new_camera /= Void
		do
			viewing_camera := new_camera
		ensure 
			viewing_camera_set: viewing_camera = new_camera
		end 
	
	set_display_list_usage (true_or_false: BOOLEAN) is
			-- Update `use_display_lists' to `true_or_false'
		do
			use_display_lists := true_or_false
		ensure 
			use_display_lists_set: use_display_lists = true_or_false
		end 

feature -- Basic operations
	
	draw_lights is
			-- Draw all the lights
		do
			global_lights.apply (Current)
		end
	
	load_viewing_transform is
			-- Tromp the current matrix stack with `viewing_camera'
		require 
			viewing_camera_not_void: viewing_camera /= Void
		do
			load_matrix (viewing_camera.world_to_camera)
		end 

feature -- Quadric Routines
	
	normal_style (how: INTEGER) is
			-- Define whether or not to use normals
		require
			how_in_range: how = rd_none or how = rd_flat or how = rd_smooth
		deferred
		end
	
	set_draw_style (how: INTEGER) is
			-- Define whether or not to use normals
		require
			how_in_range: how = rd_line or how = rd_point or how = rd_fill
		deferred
		ensure
			draw_style_set: draw_style = how
		end
	
	orientation (how: INTEGER) is
			-- Define whether or not to use normals
		require
			how_in_range: how = rd_inside or how = rd_outside
		deferred
		end
	
	use_textures (on_or_off: BOOLEAN) is
			-- Use Textures?
		deferred
		end

feature -- NURBS Primitives
	
	new_nurbs_renderer: NURBS_RENDERER is
			-- Create a NURBS renderer
		require
			valid_context: context_enabled
		deferred
		ensure
			Result_not_void: Result /= Void
		end
	
feature -- High Level Primitives
	
	cone (radius, height, sweep_angle,
		  u_minimum, u_maximum, v_minimum, v_maximum: DOUBLE) is
		require
			valid_context: context_enabled
			correct_order_u: u_minimum < u_maximum
			correct_order_v: v_minimum < v_maximum
		deferred
		end
	
	cube (lower, upper: POINT3) is
		require
			valid_context: context_enabled
			lower_not_void: lower /= Void
			upper_not_void: upper /= Void
		deferred
		end
	
	cylinder (base_radius, top_radius, z_minimum, z_maximum, sweep_angle,
			u_minimum, u_maximum, v_minimum, v_maximum: DOUBLE) is
		require
			valid_context: context_enabled
			correct_order_u: u_minimum < u_maximum
			correct_order_v: v_minimum < v_maximum
		deferred
		end
	
	disk (radius, height, sweep_angle: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	sphere (radius, z_minimum, z_maximum, sweep_angle,
			u_minimum, u_maximum, v_minimum, v_maximum: DOUBLE) is
		require
			valid_context: context_enabled
			correct_order_z: z_minimum < z_maximum
			correct_order_u: u_minimum < u_maximum
			correct_order_v: v_minimum < v_maximum
		deferred
		end
	
	torus (inner_radius, outer_radius, phi_min, phi_max, sweep_angle,
			u_minimum, u_maximum, v_minimum, v_maximum: DOUBLE) is
		require
			valid_context: context_enabled
			correct_order_u: u_minimum < u_maximum
			correct_order_v: v_minimum < v_maximum
		deferred
		end
	
feature -- Shading Language

	has_shading_language: BOOLEAN is
			-- Does this implementation support a shading language?
		require
			valid_context: context_enabled
		deferred
		end 

	create_vertex_shader: INTEGER is
			-- Create a vertex shader
		require
			valid_context: context_enabled
			has_shading_language: has_shading_language
		deferred
		end 

	create_fragment_shader: INTEGER is
			-- Create a fragment shader
		require
			valid_context: context_enabled
			has_shading_language: has_shading_language
		deferred
		end 

	set_shader_source ( shader_id: INTEGER; buffer: STRING) is
			-- Set the source of a shader
		require
			valid_context: context_enabled
			buffer_not_empty: buffer /= Void implies buffer.count > 0
			has_shading_language: has_shading_language
		deferred
		end 

	compile_shader (shader_id: INTEGER) is
			-- Compile the shader corresponding to `shader_id'
		require
			valid_context: context_enabled
			has_shading_language: has_shading_language
		deferred
		end 

	shader_compiled (shader_id: INTEGER): BOOLEAN is
			-- Has `shader_id' been compiled
		require
			valid_context: context_enabled
			has_shading_language: has_shading_language
		deferred
		end 

	shader_status (shader_id: INTEGER): STRING is
			-- What is the status of `shader_id'?
		require
			valid_context: context_enabled
			has_shading_language: has_shading_language
		deferred
		end 

	create_shader_program: INTEGER is
			-- Create a shader program
		require
			valid_context: context_enabled
			has_shading_language: has_shading_language
		deferred
		end 

	attach_shader_object (program, object: INTEGER) is
			-- Attach a shader to the program
		require
			valid_context: context_enabled
			has_shading_language: has_shading_language
		deferred
		end 

	link_shader_program (shader_id: INTEGER) is
			-- Link the program defined by `shader_id'?
		require
			valid_context: context_enabled
			has_shading_language: has_shading_language
		deferred
		end 
	
	use_shader_program (shader_id: INTEGER) is
			-- Load the program defined by `shader_id'?
		require
			valid_context: context_enabled
			has_shading_language: has_shading_language
		deferred
		end 
	
	set_uniform_parameters (program_id: INTEGER; parameter: STRING;
						   values: REAL_ARRAYED_LIST) is
			-- Set the value for the uniform `parameter' in 
			-- `program_id'
		require
			valid_context: context_enabled
			has_shading_language: has_shading_language
		deferred
		end 

	get_uniform_variable_names (program_id: INTEGER): DS_ARRAYED_LIST [STRING] is
			-- What uniform variables are used by `program_id'?
		require
			valid_context: context_enabled
			has_shading_language: has_shading_language
		deferred
		end 
	
feature -- Primitives 
	
	render_begin (mode: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	render_end is
		require
			valid_context: context_enabled
		deferred
		end
	
	rect (x1, y1, x2, y2: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	set_vertex (pt: POINT) is
		require
			valid_context: context_enabled
			pt_not_void: pt /= Void
		do
			if pt.count = 2 then
				vertex2 (pt.item (1), pt.item (2))
			elseif pt.count = 3 then
				vertex3 (pt.item (1), pt.item (2), pt.item (3))
			elseif pt.count = 4 then
				vertex4 (pt.item (1), pt.item (2), pt.item (3), pt.item (4))
			end
		end
	
	vertex2 (x, y: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	vertex3 (x, y, z: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	vertex4 (x, y, z, w: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	vertex_pointer (p: POINTER) is
		require
			valid_context: context_enabled
			p_not_void: p /= default_pointer
		deferred
		end
	
feature -- Coordinate Transforms 

	rotate (angle, x, y, z: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	scale (x, y, z: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	translate (x, y, z: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	multiply_matrix (m: MATRIX4) is
			-- Interface to glMultMatrix
		require 
			m_not_void: m /= Void
			valid_context: context_enabled
		deferred
		end -- multiply_matrix

	frustum (left, right, bottom, top, near, far: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	ortho (left, right, bottom, top, near, far: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	load_identity is
		require
			valid_context: context_enabled
		deferred
		end
	
	load_matrix (m: MATRIX4) is
		require
			valid_context: context_enabled
			m_not_void: m /= Void
		deferred
		end
	
	matrix_mode (mode: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	pop_matrix is
		require
			valid_context: context_enabled
		deferred
		end
	
	push_matrix is
		require
			valid_context: context_enabled
		deferred
		end
	
	viewport (left, right, bottom, top: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
feature -- Coloring and Lighting

	rgb_color (c: RGB_COLOR) is
		require
			valid_context: context_enabled
		do
			color_3 (c.red, c.green, c.blue)
		end
	
	rgba_color (c: RGBA_COLOR) is
		require
			valid_context: context_enabled
		do
			color_4 (c.red, c.green, c.blue, c.alpha)
		end
	
	color_3 (red, green, blue: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	color_4 (red, green, blue, alpha: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	color_material (face, mode: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	color_index (index: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	normal (nx, ny, nz: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	material (light_number, pname: INTEGER; parameter: ARRAY [DOUBLE]) is
		require
			parameter_not_void: parameter /= Void
			parameter_count_positive: parameter.count > 0
			valid_context: context_enabled
		deferred
		end
	
	set_light_position (light_number: INTEGER; position: POINT4) is
		require
			position_not_void: position /= Void
			light_number_is_positive: light_number >= 0
			light_number_less_than_maximum_allowed: light_number < 
													maximum_number_of_lights
			valid_context: context_enabled
		deferred
		end
	
	set_light_direction (light_number: INTEGER; direction: VECTOR4) is
		require
			direction_not_void: direction /= Void
			light_number_is_positive: light_number >= 0
			light_number_less_than_maximum_allowed: light_number < 
												   maximum_number_of_lights
			valid_context: context_enabled
		deferred
		end
	
	set_diffuse_light_color (light_number: INTEGER; new_color: RGBA_COLOR) is
		require
			light_number_is_positive: light_number >= 0
			light_number_less_than_maximum_allowed: light_number < 
												   maximum_number_of_lights
			valid_context: context_enabled
		deferred
		end
	
	set_ambient_light_color (light_number: INTEGER; new_color: RGBA_COLOR) is
		require
			light_number_is_positive: light_number >= 0
			light_number_less_than_maximum_allowed: light_number < 
												   maximum_number_of_lights
			valid_context: context_enabled
		deferred
		end
	
	set_specular_light_color (light_number: INTEGER; new_color: RGBA_COLOR) is
		require
			light_number_is_positive: light_number >= 0
			light_number_less_than_maximum_allowed: light_number < 
												   maximum_number_of_lights
			valid_context: context_enabled
		deferred
		end
	
	set_spotlight_exponent (light_number: INTEGER; exponent: DOUBLE) is
		require
			light_number_is_positive: light_number >= 0
			light_number_less_than_maximum_allowed: light_number < 
												   maximum_number_of_lights
			valid_context: context_enabled
		deferred
		end
	
	set_spotlight_cutoff (light_number: INTEGER; cutoff: DOUBLE) is
		require
			light_number_is_positive: light_number >= 0
			light_number_less_than_maximum_allowed: light_number < 
												   maximum_number_of_lights
			valid_context: context_enabled
		deferred
		end
	
	set_constant_attenuation (light_number: INTEGER; factor: DOUBLE) is
		require
			light_number_is_positive: light_number >= 0
			light_number_less_than_maximum_allowed: (light_number < 
													 maximum_number_of_lights)
			factor_is_positive: factor >= 0.0
			valid_context: context_enabled
		deferred
		end
	
	set_linear_attenuation (light_number: INTEGER; factor: DOUBLE) is
		require
			light_number_is_positive: light_number >= 0
			light_number_less_than_maximum_allowed: (light_number < 
													 maximum_number_of_lights)
			factor_is_positive: factor >= 0.0
			valid_context: context_enabled
		deferred
		end
	
	set_quadratic_attenuation (light_number: INTEGER; factor: DOUBLE) is
		require
			light_number_is_positive: light_number >= 0
			light_number_less_than_maximum_allowed: (light_number < 
													 maximum_number_of_lights)
			factor_is_positive: factor >= 0.0
			valid_context: context_enabled
		deferred
		end
	
	light (light_number, pname: INTEGER; parameter: VECTOR) is
		require
			light_number_is_positive: light_number >= 0
			light_number_less_than_maximum_allowed: light_number < 
												   maximum_number_of_lights
			parameter_not_void: parameter /= Void
			parameter_count_valid: parameter.count = 3 or parameter.count = 4
			valid_context: context_enabled
		deferred
		end
	
	assign_light_index: INTEGER is
			-- Generate an index for use with `light'
		require
			valid_context: context_enabled
		deferred
		ensure
			Result_is_valid: Result >= 0 and Result <= maximum_number_of_lights 
							 or Result = -1
		end
	
	enable_light (light_index: INTEGER) is
			-- Turn on the light at `light_index'
		require 
			light_index_is_positive: light_index >= 0
			light_index_less_than_maximum_allowed: light_index < 
												   maximum_number_of_lights
		deferred
		end 

	disable_light (light_index: INTEGER) is
			-- Turn off the light at `light_index'
		require 
			light_index_is_positive: light_index >= 0
			light_index_less_than_maximum_allowed: light_index < 
												   maximum_number_of_lights
		deferred
		end
	
	maximum_number_of_lights: INTEGER is
			-- The maximum number of active lights allowed by the renderer
		require
			valid_context: context_enabled
		deferred
		ensure
			Result_is_positive: Result >= 1
		end
	
	shade_model (mode: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	enable_two_sided_lighting is
		require
			valid_context: context_enabled
		deferred
		end
	
	enable_one_sided_lighting is
		require
			valid_context: context_enabled
		deferred
		end
	
	set_scene_ambient_color (new_color: RGBA_COLOR) is
		require
			valid_context: context_enabled
		deferred
		end
	
	front_face (mode: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	color_pointer (p: POINTER) is
		require
			valid_context: context_enabled
			p_not_void: p /= default_pointer
		deferred
		end
	
feature -- Clipping
	
	clip_plane (which_plane: INTEGER; equation: EWG_DOUBLE_ARRAY) is
			-- Interface to the glClipPlane
		require 
			positive_plane: which_plane >= 0
			equation_not_void: equation /= Void
			equation_count_is_4: equation.count = 4
			valid_context: context_enabled
		deferred
		end -- clip_plane
	
feature -- Rasterization 

	raster_position_2 (x, y: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	raster_position_3 (x, y, z: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	raster_position_4 (x, y, z, w: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	point_size (size: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	line_width (width: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	polygon_mode (face, mode: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
feature -- Pixel Operations 

	read_buffer (mode: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	pixel_store (pname, param: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	pixel_transfer (pname:INTEGER; param: REAL_32) is
		require
			valid_context: context_enabled
		deferred
		end
	
	draw_pixels (width, height, mode: INTEGER; pixels: POINTER) is
		require
			valid_context: context_enabled
		deferred
		end
	
feature -- Texture Mapping

	texture_coord_1 (s: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	texture_coord_2 (s, t: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	texture_coord_3 (s, t, r: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	texture_coord_4 (s, t, r, q: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	texture_env (target, pname, param: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	texture_gen (coord, pname: INTEGER; param: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	texture_parameter (target, pname, param: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
feature -- Fog
feature -- Frame Buffer Operations

	scissor (x, y, width, height: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	stencil_func (func, ref, mask: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	stencil_mask (mask: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	stencil_op (fail, zfail, zpass: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	clear (mask: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	clear_color (red, green, blue, alpha: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	clear_depth (depth: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	clear_index (index: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
	draw_buffer (mode: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	draw_arrays (mode, offset, count: INTEGER) is
		require
			valid_context: context_enabled
			offset_is_positive: offset >= 0
			count_is_positive: count >= 0
		deferred
		end
	
	draw_elements (mode, offset, count: INTEGER; indices: POINTER) is
		require
			valid_context: context_enabled
			offset_is_positive: offset >= 0
			count_is_positive: count >= 0
			indices_not_void: indices /= default_pointer
		deferred
		end
	
feature -- Evaluators

	map_1d (target: INTEGER; u1, u2: DOUBLE; stride, order: INTEGER; points: POINTER) is
		require
			valid_context: context_enabled
			points_not_void: points /= default_pointer
		deferred
		end
	
	map_grid_1d (un: INTEGER; u1, u2: DOUBLE) is
		require
			valid_context: context_enabled
			un_positive: un > 0
		deferred
		end
	
	map_2d (target: INTEGER; u1, u2: DOUBLE; ustride, uorder: INTEGER;
	v1, v2: DOUBLE; vstride, vorder: INTEGER; points: POINTER) is
		require
			valid_context: context_enabled
			points_not_void: points /= default_pointer
		deferred
		end

	map_grid_2d (un: INTEGER; u1, u2: DOUBLE; vn: INTEGER; v1, v2: DOUBLE) is
		require
			valid_context: context_enabled
			un_positive: un > 0
			vn_positive: vn > 0
		deferred
		end
	
	eval_mesh_1d (mode, i1, i2: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
		
	eval_mesh_2d (mode, i1, i2, j1, j2: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
		
feature -- Selection and Feedback

	render_mode (mode: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
feature -- Display Lists

	generate_lists (count: INTEGER): INTEGER is
			-- Generate a contiguous set of empty display lists
		require
			valid_context: context_enabled
			count_is_positive: count > 0
		deferred
		end
	
	delete_lists (count, base: INTEGER) is
			-- Delete a contiguous set of empty display lists
		require
			valid_context: context_enabled
			count_is_positive: count > 0
			base_is_non_negative: base >= 0
		deferred
		end
	
	list_base (offset: INTEGER) is
			-- Set the display list base for call_lists
		require
			valid_context: context_enabled
		deferred
		end
	
	new_list (list, mode: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	end_list is
		require
			valid_context: context_enabled
		deferred
		end
	
	call_list (list: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	call_lists (n, type: INTEGER; lists: POINTER) is
		require
			valid_context: context_enabled
		deferred
		end
	
feature -- Modes and Execution

	disable (mode: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	enable (mode: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	enable_client_state (mode: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	disable_client_state (mode: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
	finish is
		require
			valid_context: context_enabled
		deferred
		end
	
	flush is
		require
			valid_context: context_enabled
		deferred
		end
	
	load_font (font_ptr, offset, count, display_list_base: INTEGER) is
		require
			valid_context: context_enabled
		deferred
		end
	
feature -- State Queries
	
	context_enabled: BOOLEAN is
			-- Is there a valid context for rendering
		deferred
		end
	
	push_attribute (mask: INTEGER) is
			-- Push the attribute stack
		deferred
		end
	
	pop_attribute is
			-- Pop the attribute stack
		deferred
		end
	
feature -- Implementation
	
	integer_or (i, j: INTEGER): INTEGER is
		external
			"C inline"
		alias "[ 
				($i | $j ) 
			   ]"
		end
		
	perspective (fovy, aspect, znear, zfar: DOUBLE) is
		require
			valid_context: context_enabled
		deferred
		end
	
feature {NONE} -- Implementation
	
	singleton: OBJECT_IDENTIFIER is
			-- Create the id generator
		once
			create Result
		end
	
end -- RENDERER

