

indexing
	description: "A table that maps locations to buckets"
	patterns: "none"
	status: "unreviewed"
	date: "$Date: 2006-09-04 15:55:08 -0400 (Mon, 04 Sep 2006) $"
	revision: "$Revision: 264 $"



class

	BUCKET_TABLE

inherit

	NUMERIC_CONVERSION

creation
	make

feature -- Initialization

	make (x_count, y_count,
		  new_x_resolution, new_y_resolution,
		  new_width, new_height: INTEGER) is
			--
		require
			x_count_positive: x_count > 0
			y_count_positive: y_count > 0
			new_x_resolution: new_x_resolution > 0
			new_y_resolution: new_y_resolution > 0
			new_width: new_width > 0
			new_height: new_height > 0
		local
			p: BUCKET_TABLE_ITEM
			i: INTEGER
		do
			bucket_count_in_x := x_count
			bucket_count_in_y := y_count

			width := new_width
			height := new_height

			bucket_x_resolution := new_x_resolution
			inv_bucket_x_resolution := 1.0 / bucket_x_resolution
			bucket_y_resolution := new_y_resolution
			inv_bucket_y_resolution := 1.0 / bucket_y_resolution

			create buckets.make (1, bucket_count_in_x * bucket_count_in_y)

			from
				i := buckets.lower
			invariant
				i_at_least_one: i >= buckets.lower
				i_small_enough: i <= buckets.lower + buckets.upper
			variant
				i_increasing: buckets.lower + buckets.upper - i
			until
				i > buckets.upper
			loop
				create p.make (false, Void, Void)
				buckets.put (p, i)

				i := i + 1
			end
		ensure
			bucket_count_in_x_set: bucket_count_in_x = x_count
			bucket_count_in_y_set: bucket_count_in_y = y_count
			width_set: width = new_width
			height_set: height = new_height
		end

feature -- Access

	bucket_count_in_x, bucket_count_in_y: INTEGER
			-- The number of buckets in x and y respectively

	bucket_x_resolution, bucket_y_resolution: INTEGER
			-- The resolution of an image bucket

	width, height: INTEGER
			-- The domain over which the buckets are valid

	last_bucket: INTEGER
			-- The index of the last bucket accessed

	item_at_point (x, y: DOUBLE): BUCKET_TABLE_ITEM is
			-- The `item' at point (`x', `y')
		require
			x_in_range: x >= 0 and x <= width
			y_in_range: y >= 0 and y <= height
		do
			Result := buckets.item (to_bucket_index (x, y))
		end

	item (row, col: INTEGER): BUCKET_TABLE_ITEM is
			-- The `item' at location (`row', `col')
		require
			col_in_range: col >= 0 and col < bucket_count_in_x
			row_in_range: row >= 0 and row < bucket_count_in_y
		do
			Result := buckets.item (bucket_index (row, col))
		end

feature -- Measurement

	count: INTEGER is
			-- The total number of buckets
		do
			Result := bucket_count_in_x * bucket_count_in_y
		end

feature -- Comparison
feature -- Status report

	bucket_index (row, col: INTEGER): INTEGER is
			-- The bucket_number for location (`row', `col')
		require
			col_in_range: col >= 0 and col <= bucket_count_in_x
			row_in_range: row >= 0 and row <= bucket_count_in_y
		do
			Result := row * bucket_count_in_x + col + 1
			last_bucket := Result
		ensure
			Result_positive: Result >= 1
		end

	to_bucket_index (x, y: DOUBLE): INTEGER is
			-- Convert `x', `y' to an index into the buckets
		require
			x_positive: x >= 0 and x <= width
			y_positive: y >= 0 and y <= height
		local
			row, col: INTEGER
		do
			col := real_64_truncated_to_integer ((x.floor * inv_bucket_x_resolution).floor)
			row := real_64_truncated_to_integer ((y.floor * inv_bucket_y_resolution).floor)
			Result := bucket_index (row, col)
			last_bucket := Result
		ensure
			Result_positive: Result > 0 and Result <= buckets.count
		end

	bound (row, col: INTEGER): AXIS_ALIGNED_BOX is
			-- What's the bounds of the bucket at location (`row', `col')
		require
			col_in_range: col >= 0 and col <= bucket_count_in_x
			row_in_range: row >= 0 and row <= bucket_count_in_y
		local
			lower, upper: POINT3
		do
			create lower.make (col * bucket_x_resolution,
							   row * bucket_y_resolution,
							   0)
			create upper.make ((col + 1) * bucket_x_resolution,
							   (row + 1) * bucket_y_resolution,
							   0)
			create Result.make (lower, upper)
		ensure
			Result_not_void: Result /= Void
			x_in_range: Result.lower.x >= 0 and
						Result.upper.x <= (width * inv_bucket_x_resolution).ceiling * bucket_x_resolution + 1
			y_in_range: Result.lower.y >= 0 and
						Result.upper.y <= (height * inv_bucket_y_resolution).ceiling * bucket_y_resolution + 1

		end

	bound_at_point (x, y: DOUBLE): AXIS_ALIGNED_BOX is
			-- What's the bounds of the bucket at  point (`x', `y')
		require
			x_in_range: x >= 0 and x <= width
			y_in_range: y >= 0 and y <= height
		local
			row, col: INTEGER
		do
			col := real_64_truncated_to_integer ((x.floor * inv_bucket_x_resolution).floor)
			row := real_64_truncated_to_integer ((y.floor * inv_bucket_y_resolution).floor)
			Result := bound (row, col)
		end

	bucket_at_point_processed (x, y: DOUBLE): BOOLEAN is
			-- Has the bucket at point (`x', `y') been processed
		require
			x_in_range: x >= 0 and x <= width
			y_in_range: y >= 0 and y <= height
		do
			Result := buckets.item (to_bucket_index (x, y)).processed
		end

	processed (row, col: INTEGER): BOOLEAN is
			-- Has the bucket at point (`row', `col') been processed
		require
			col_in_range: col >= 1 and col <= bucket_count_in_x
			row_in_range: row >= 1 and row <= bucket_count_in_y
		do
			Result := buckets.item (bucket_index (row, col)).processed
		end

feature -- Status setting

	mark_as_processed (row, col: INTEGER) is
			-- Mark bucket (`row', `col') as processed
		require
			col_in_range: col >= 0 and col < bucket_count_in_x
			row_in_range: row >= 0 and row < bucket_count_in_y
		do
			buckets.item (bucket_index (row, col)).make (true, Void, Void)
		end

	mark_bucket_at_point_as_processed (x, y: DOUBLE) is
			-- The `item' at point (`x', `y')
		require
			x_in_range: x >= 0 and x <= width
			y_in_range: y >= 0 and y <= height
		do
			buckets.item (to_bucket_index (x, y)).make (true, Void, Void)
		end

feature -- Cursor movement
feature -- Element change

	put (val: BUCKET_ENTITY; x, y: DOUBLE) is
			-- The `item' at point (`x', `y')
		require
			val_not_void: val /= Void
			x_in_range: x >= 0 and x <= width
			y_in_range: y >= 0 and y <= height
		local
			list: SORTABLE_ARRAYED_LIST [BUCKET_ENTITY]
			p: BUCKET_TABLE_ITEM
		do
			p := buckets.item (to_bucket_index (x, y))
			list := p.entity_list
			if list = Void then
				create list.make (1)
				p.replace_second (list)
			end

			list.force_last (val)
		ensure
			item_at_x_y_not_void: item_at_point (x, y) /= Void
		end

	forward_micropolygon (mp: MICROPOLYGON; x, y: DOUBLE) is
			-- Forward a micropolygon to the bucket at point (`x', `y')
		require
			mp_not_void: mp /= Void
		local
			list: DS_ARRAYED_LIST [MICROPOLYGON]
			p: BUCKET_TABLE_ITEM
			xx, yy: DOUBLE
		do
			xx := x
			yy := y
			if x < 0 then
				xx := 0
			elseif x >= width then
				xx := width - 1
			else
				xx := x
			end

			if y < 0 then
				yy := 0
			elseif y >= height then
				yy := height - 1
			else
				yy := y
			end
			p := buckets.item (to_bucket_index (xx, yy))
			
			list := p.mp_list
			
			if list = Void then
				create list.make (1)
				p.replace_mp_list (list)
			end

			list.force_last (mp)

				check
					no_duplicate_micropolygons: p.no_duplicate_micropolygons
				end
		ensure
			mp_list_at_x_y_not_void: (x >= 0 and y >= 0 and x <= width and y <= height) implies
									 buckets.item (to_bucket_index (x, y)).mp_list /= Void 
		end

	put_in_bucket (val: BUCKET_ENTITY; row, col: INTEGER) is
			-- Put `val' into the bucket at (`row', `col')
		require
			val_not_void: val /= Void
			col_in_range: col >= 0 and col < bucket_count_in_x
			row_in_range: row >= 0 and row < bucket_count_in_y
		local
			list: SORTABLE_ARRAYED_LIST [BUCKET_ENTITY]
			p: BUCKET_TABLE_ITEM
		do
			p := buckets.item (bucket_index (row, col))
			list := p.entity_list
			if list = Void then
				create list.make (1)
				p.replace_second (list)
			end

			list.force_last (val)
		ensure
			item_not_void: item (row, col) /= Void
		end

	set_resolution (new_x_resolution, new_y_resolution: INTEGER) is
			-- Set `bucket_x_resolution' to `new_x_resolution' and
			-- `bucket_y_resolution' to `new_y_resolution'
		require
			new_x_resolution: new_x_resolution > 0
			new_y_resolution: new_y_resolution > 0
		do
			bucket_x_resolution := new_x_resolution
			inv_bucket_x_resolution := 1.0 / bucket_x_resolution
			bucket_y_resolution := new_y_resolution
			inv_bucket_y_resolution := 1.0 / bucket_y_resolution
		ensure
			bucket_x_resolution_set: bucket_x_resolution = new_x_resolution
			bucket_y_resolution_set: bucket_y_resolution = new_y_resolution
		end

feature -- Removal
feature -- Resizing
feature -- Transformation
feature -- Conversion
feature -- Duplication
feature -- Basic operations
feature -- Miscellaneous

	dump_buckets (dump_entities: BOOLEAN) is
			-- Print the status of the buckets
		local
			i, j: INTEGER
			fi: FORMAT_INTEGER
		do
			from
				i := 0
				create fi.make (5)
			invariant
				i_at_least_one: i >= 0
				i_small_enough: i <= bucket_y_resolution
			variant
				i_increasing: bucket_y_resolution - i
			until
				i >= bucket_y_resolution
			loop
				from
					j := 0
				invariant
					j_at_least_one: j >= 0
					j_small_enough: j <= bucket_x_resolution
				variant
					j_increasing: bucket_x_resolution - j
				until
					j >= bucket_x_resolution
				loop
					if dump_entities then
						io.put_string (fi.formatted (buckets.item (bucket_index (i, j)).entity_count))
					else
						io.put_string (fi.formatted (buckets.item (bucket_index (i, j)).mp_count))
					end
					io.put_string (", ")
					j := j + 1
				end
				io.new_line
				i := i + 1
			end
		end

feature -- Obsolete
feature -- Inapplicable
feature {NONE} -- Implementation

	buckets: ARRAY [BUCKET_TABLE_ITEM]
			-- The storage item
	
	inv_bucket_x_resolution, inv_bucket_y_resolution: REAL_64
			-- Memoized inverse

invariant

	bucket_not_void: buckets /= Void
	bucket_count_in_x_positive: bucket_count_in_x > 0
	bucket_count_in_y_positive: bucket_count_in_y > 0
	count_correct: (count = buckets.count and
					buckets.count = bucket_count_in_x * bucket_count_in_y)

end -- BUCKET_TABLE

