

indexing
	description: "A grid of 4-sided polygons, each smaller than a pixel"
	patterns: "none"
	status: "unreviewed"
	date: "$Date: 2007-03-29 13:43:21 -0400 (Thu, 29 Mar 2007) $"
	revision: "$Revision: 286 $"




class

	MICROPOLYGON_GRID

inherit

	SINGLETON_ACCESSOR
		rename
			singleton as shading_data_pool
		redefine
			shading_data_pool
		select
			shading_data_pool
		end

	SINGLETON_ACCESSOR
		rename
			singleton as object_id
		redefine
			object_id
		end

	BASIC_NUMERIC
		export
			{NONE} all
		end

creation
	make

feature -- Initialization

	make (new_width, new_height: INTEGER) is
			-- Create a new grid
		local
			i: INTEGER
			sd: SHADING_DATA
		do
			object_id.next_id
			identifier := object_id.last_id

			width := new_width
			height := new_height

			shaded := false
			released := false

			create grid.make (1, width * height)

			from
				i := 1
			invariant
				i_at_least_one: i >= 1
				i_small_enough: i <= 1 + grid.count
			variant
				i_increasing: 1 + grid.count - i
			until
				i > grid.count
			loop
				sd := shading_data_pool.item
				sd.add_reference (Current)
				grid.put (sd, i)
				i := i + 1
			end
		end

feature -- Access

	width: INTEGER
	height: INTEGER
			-- The width and height of the grid

	identifier: INTEGER
			-- A guaranteed unique value to identify this instance of
			-- an action
	
	shaded: BOOLEAN
			-- Has this grid been shaded

	released: BOOLEAN
			-- Has the memory for the grid points been released?

	parent_element: SCENE_DRAWABLE
			-- The element that was diced to this grid

	two_sided: BOOLEAN is
			-- Is this grid `two_sided'?
		do
			Result := parent_element.geometry.two_sided
		end

	item (i, j: INTEGER): SHADING_DATA is
			-- Get the value at location (`i', `j' )
		require
			i_in_range: i > 0 and i <= width
			j_in_range: j > 0 and j <= height
		do
			Result := grid.item (grid_index (i, j))
		ensure
			Result_not_void: Result /= Void
		end

	indexed_item (i: INTEGER): SHADING_DATA is
			-- Get the value at index `i'
		require
			i_in_range: i > 0 and i <= width * height
		do
			Result := grid.item (i)
		ensure
			Result_not_void: Result /= Void
		end

	vertex_count: INTEGER is
			-- The number of points in the grid
		do
			Result := width * height
		ensure
			Result_positive: Result > 0
		end

	count: INTEGER is
			-- The number of polygons in the grid
		do
			Result := (width - 1) * (height - 1)
		ensure
			Result_positive: Result > 0
		end

	extents: AXIS_ALIGNED_BOX is
			-- The bounds of the grid
		local
			pt1, pt2, pt3, pt4: POINT3
			lower, upper: POINT3
		do
			pt1 := item (1, 1).point
			pt2 := item (width, 1).point
			pt3 := item (1, height).point
			pt4 := item (width, height).point

			create lower.make (pt1.x.min (pt3.x), pt1.y.min (pt2.y), pt1.z.min (pt2.z.min (pt3.z.min (pt4.z))))
			create upper.make (pt1.x.max (pt3.x), pt1.y.max (pt2.y), pt1.z.max (pt2.z.max (pt3.z.max (pt4.z))))
			create Result.make (lower, upper)
		end

feature -- Measurement
feature -- Comparison
feature -- Status report

	is_reset: BOOLEAN is
			-- Has the grid been `reset'
		local
			i: INTEGER
		do
			from
				i := 1
				Result := true
			invariant
				i_at_least_one: i >= 1
				i_small_enough: i <= 1 + grid.count
			variant
				i_increasing: 1 + grid.count - i
			until
				i > grid.count or not Result
			loop
				Result := grid.item (i) = Void
				i := i + 1
			end
		end

	is_backfacing (camera: SCENE_CAMERA): BOOLEAN is
			-- Are all the elements of the grid backfacing?
		require
			camera_not_void: camera /= Void
		do
			Result := (item (1, 1).geometric_normal.transform (camera.world_to_camera).dot_with_vector (camera.projection_direction) < 0 and then
					   item (width, 1).geometric_normal.transform (camera.world_to_camera).dot_with_vector (camera.projection_direction) < 0 and then
					   item (1, height).geometric_normal.transform (camera.world_to_camera).dot_with_vector (camera.projection_direction) < 0 and then
					   item (width, height).geometric_normal.transform (camera.world_to_camera).dot_with_vector (camera.projection_direction) < 0)
		end

	is_trimmed: BOOLEAN is
			-- Have all the elements of the grid been trimmed away?
		require
			default: true
		do
			Result := false
		ensure
			default: true
		end

	all_diced: BOOLEAN is
			-- Are all the neigbors of the parent element diced
		local
			i, j, the_count: INTEGER
		do
			from
				i := parent_element.geometry.left
				Result := true
			invariant
				i_at_least_one: i >= parent_element.geometry.left
				i_small_enough: i <= parent_element.geometry.left + parent_element.geometry.below + 1
			variant
				i_increasing: parent_element.geometry.left + parent_element.geometry.below + 1 - i
			until
				not Result or i > parent_element.geometry.below
			loop
				if parent_element.geometry.has_neighbor (i) then
					from
						j := 1
						the_count := parent_element.geometry.neighbor (i).count
					invariant
						j_at_least_one: j >= 1
						j_small_enough: j <= 1 + the_count
					variant
						j_increasing: 1 + the_count - j
					until
						not Result or j > the_count
					loop
						Result := parent_element.geometry.neighbor (i).item (j).diced

						j := j + 1
					end
				end

				i := i + 1
			end
		end

	has_left_patch: BOOLEAN is
			-- Has a left patch been generated for this grid?
		do
			Result := left_patch /= Void and then left_patch.count > 0
		end

	has_right_patch: BOOLEAN is
			-- Has a right patch been generated for this grid?
		do
			Result := right_patch /= Void and then right_patch.count > 0
		end

	has_above_patch: BOOLEAN is
			-- Has a above patch been generated for this grid?
		do
			Result := above_patch /= Void and then above_patch.count > 0
		end

	has_below_patch: BOOLEAN is
			-- Has a below patch been generated for this grid?
		do
			Result := below_patch /= Void and then below_patch.count > 0
		end

	has_patch (patch: PATCH_OBJECT): BOOLEAN is
			-- Is `patch' attached to this grid?
		require
			patch_not_void: patch /= Void
		do
			Result := (has_left_patch and then left_patch.has (patch)) or
				(has_right_patch and then right_patch.has (patch)) or
				(has_above_patch and then above_patch.has (patch)) or
				(has_below_patch and then below_patch.has (patch))
		end

	no_patches: BOOLEAN is
			-- Are all of the patch elements either void or do they
			-- have a count of zero?
		do
			Result := not has_left_patch and not has_right_patch and
				not has_above_patch and not has_below_patch
		end

	all_patches: DS_ARRAYED_LIST [PATCH_OBJECT] is
			-- Have all the patches been busted?
		do
			create Result.make (1)
			if has_left_patch then
				Result.append_last (left_patch)
			end

			if has_right_patch then
				Result.append_last (right_patch)
			end

			if has_above_patch then
				Result.append_last (above_patch)
			end

			if has_below_patch then
				Result.append_last (below_patch)
			end
		ensure
			Result_not_void: Result /= Void
		end

feature -- Status setting

	trim is
			-- Trim the grid
		do
		end

	left_start (v: DOUBLE): INTEGER is
			-- Return the grid index that is closest but less than `v'
		local
			index: INTEGER
		do
			index := ((v - parent_element.geometry.v_minimum) / v_stride).truncated_to_integer + 1
			Result := grid_index (1, index)
		end

	right_start (v: DOUBLE): INTEGER is
			-- Set `right_index' to the start of the edge
		local
			index: INTEGER
		do
			index := ((v - parent_element.geometry.v_minimum) / v_stride).truncated_to_integer + 1
			Result := grid_index (width, index)
		end

	above_start (u: DOUBLE): INTEGER is
			-- Set `above_index' to the start of the edge
		local
			index: INTEGER
		do
			index := ((u - parent_element.geometry.u_minimum) / u_stride).truncated_to_integer + 1
			Result := grid_index (index, height)
		end

	below_start (u: DOUBLE): INTEGER is
			-- Set `below_index' to the start of the edge
		local
			index: INTEGER
		do
			index := ((u - parent_element.geometry.u_minimum) / u_stride).truncated_to_integer + 1
			Result := grid_index (index, 1)
		end

	left_stride: INTEGER is
			-- Move the cursor foward one grid cell along the left edge
		do
			Result := width
		end

	right_stride: INTEGER is
			-- Move the cursor foward one grid cell along the right edge
		do
			Result := width
		end

	above_stride: INTEGER is
			-- Move the cursor foward one grid cell along the above edge
		do
			Result := 1
		end

	below_stride: INTEGER is
			-- Move the cursor foward one grid cell along the below edge
		do
			Result := 1
		end

	left_after (v: DOUBLE): INTEGER is
			-- Is `left_index' beyond the last grid point
		local
			index: INTEGER
		do
			index := clamp_to_range ((v - parent_element.geometry.v_minimum) /v_stride+1,
									 1, height).truncated_to_integer
			Result := grid_index (1, index)
		end

	right_after (v: DOUBLE): INTEGER is
			-- Is `right_index' beyond the last grid point
		local
			index: INTEGER
		do
			index := clamp_to_range ((v - parent_element.geometry.v_minimum) /v_stride+1,
									 1, height).truncated_to_integer
			Result := grid_index (width, index)
		end

	above_after (u: DOUBLE): INTEGER is
			-- Is `above_index' beyond the last grid point
		local
			index: INTEGER
		do
			index := clamp_to_range ((u - parent_element.geometry.u_minimum) /u_stride+1,
									 1, width).truncated_to_integer
			Result := grid_index (index, height)
		end

	below_after (u: DOUBLE): INTEGER is
			-- Is `below_index' beyond the last grid point
		local
			index: INTEGER
		do
			index := clamp_to_range ((u - parent_element.geometry.u_minimum) /u_stride+1,
									 1, width).truncated_to_integer
			Result := grid_index (index, 1)
		end

feature -- Cursor movement
feature -- Element change

	put (pt: SHADING_DATA; i, j: INTEGER) is
			-- Update the value at location (`i', `j' ) with `pt'
		require
			i_in_range: i > 0 and i <= width
			j_in_range: j > 0 and j <= height
			pt_not_void: pt /= Void
		do
			grid.put (pt, (j - 1) * width + i)
			pt.add_reference (Current)
		ensure
			updated: item (i, j) = pt
		end

	reset is
			-- Clear the grid
		local
			i: INTEGER
		do
			from
				i := 1
			invariant
				i_at_least_one: i >= 1
				i_small_enough: i <= 1 + grid.count
			variant
				i_increasing: 1 + grid.count - i
			until
				i > grid.count
			loop
				grid.item (i).remove_reference (Current)
				i := i + 1
			end
			grid.discard_items
		end

	release is
			-- Free the memory used by the grid
		require
			--all_diced: all_diced
--			patches_released: no_patches
		do
			reset
			parent_element.release
			parent_element := Void
			released := true
		ensure
			released: released
		end

	set_left_patch (patch_points: DS_ARRAYED_LIST [SHADING_DATA]) is
			-- Add the points that make up the left patch
		require
			patch_points_not_void: patch_points /= Void
			patch_points_not_empty: patch_points.count > 0
		local
			patch: PATCH_OBJECT
		do
			create patch.make (patch_points)

			if left_patch = Void then
				create left_patch.make (1)
			end

			left_patch.force_last (patch)
			patch.add_grid (Current)
		ensure
			left_patch_not_void: left_patch /= Void and left_patch.count >= 1
		end

	attach_left_patch (other: PATCH_OBJECT) is
			-- Attach the left patch to a the right patch of another
			-- grid
		require
			other_not_void: other /= Void
		do
			if left_patch = Void then
				create left_patch.make (1)
			end

			if not other.has (Current) then
				other.add_grid (Current)
			end

			if not left_patch.has (other) then
				left_patch.force_last (other)
			end
		ensure
			left_set: left_patch /= Void
			only_one_occurrence: only_one_occurrence (left_patch)
		end

	set_right_patch (patch_points: DS_ARRAYED_LIST [SHADING_DATA]) is
			-- Add the points that make up the right patch
		require
			patch_points_not_void: patch_points /= Void
			patch_points_not_empty: patch_points.count > 0
		local
			patch: PATCH_OBJECT
		do
			create patch.make (patch_points)

			if right_patch = Void then
				create right_patch.make (1)
			end

			right_patch.force_last (patch)
			patch.add_grid (Current)
		ensure
			right_patch_not_void: right_patch /= Void and right_patch.count >= 1
		end

	attach_right_patch (other: PATCH_OBJECT) is
			-- Attach the right patch to a the left patch of another
			-- grid
		require
			other_not_void: other /= Void
		do
			if right_patch = Void then
				create right_patch.make (1)
			end

			if not other.has (Current) then
				other.add_grid (Current)
			end

			if not right_patch.has (other) then
				right_patch.force_last (other)
			end
		ensure
			right_set: right_patch /= Void
			only_one_occurrence: only_one_occurrence (right_patch)
		end

	set_above_patch (patch_points: DS_ARRAYED_LIST [SHADING_DATA]) is
			-- Add the points that make up the above patch
		require
			patch_points_not_void: patch_points /= Void
			patch_points_not_empty: patch_points.count > 0
		local
			patch: PATCH_OBJECT
		do
			create patch.make (patch_points)

			if above_patch = Void then
				create above_patch.make (1)
			end

			above_patch.force_last (patch)
			patch.add_grid (Current)
		ensure
			above_patch_not_void: above_patch /= Void and above_patch.count >= 1
		end

	attach_above_patch (other: PATCH_OBJECT) is
			-- Attach the above patch to a the below patch of another
			-- grid
		require
			other_not_void: other /= Void
		do
			if above_patch = Void then
				create above_patch.make (1)
			end

			if not other.has (Current) then
				other.add_grid (Current)
			end

			if not above_patch.has (other) then
				above_patch.force_last (other)
			end
		ensure
			only_one_occurrence: only_one_occurrence (above_patch)
			above_set: above_patch /= Void
		end

	set_below_patch (patch_points: DS_ARRAYED_LIST [SHADING_DATA]) is
			-- Add the points that make up the below patch
		require
			patch_points_not_void: patch_points /= Void
			patch_points_not_empty: patch_points.count > 0
		local
			patch: PATCH_OBJECT
		do
			create patch.make (patch_points)

			if below_patch = Void then
				create below_patch.make (1)
			end

			below_patch.force_last (patch)
			patch.add_grid (Current)
		ensure
			below_patch_not_void: below_patch /= Void and below_patch.count >= 1
		end

	attach_below_patch (other: PATCH_OBJECT) is
			-- Attach the below patch to a the above patch of another
			-- grid
		require
			other_not_void: other /= Void
		do
			if below_patch = Void then
				create below_patch.make (1)
			end

			if not other.has (Current) then
				other.add_grid (Current)
			end

			if not below_patch.has (other) then
				below_patch.force_last (other)
			end
		ensure
			below_set: below_patch /= Void
			only_one_occurrence: only_one_occurrence (below_patch)
		end

	set_parent (parent: SCENE_DRAWABLE) is
			-- Set `parent_element' to `parent'
		require
			parent_not_void: parent /= Void
		do
			parent_element := parent
			u_stride := (parent.geometry.u_maximum - parent.geometry.u_minimum) / width
			v_stride := (parent.geometry.v_maximum - parent.geometry.v_minimum) / height
		ensure
			parent_element_set: parent_element = parent
		end

feature -- Removal

	remove (patch: PATCH_OBJECT) is
			-- Remove `patch'
		require
			patch_not_void: patch /= Void
			somebody_has_patch: has_patch (patch)
		do
			if has_left_patch and then left_patch.has (patch) then
				remove_patch (left_patch, patch)
			elseif has_right_patch and then right_patch.has (patch) then
				remove_patch (right_patch, patch)
			elseif has_above_patch and then above_patch.has (patch) then
				remove_patch (above_patch, patch)
			elseif has_below_patch and then below_patch.has (patch) then
				remove_patch (below_patch, patch)
			end
		ensure
			removed: not has_patch (patch)
		end

feature -- Resizing
feature -- Transformation
feature -- Conversion
feature -- Duplication
feature -- Basic operations

	bust (element: SCENE_DRAWABLE_ELEMENT;
		camera: SCENE_CAMERA): DS_ARRAYED_LIST [MICROPOLYGON] is
			-- Break the grid into micropolygons
		do
			create Result.make (count)
			bust_grid (Result, element, camera)
			if left_patch /= Void then
				bust_patch (Result, left_patch, camera)
				if left_patch.count = 0 then
					left_patch := Void
				end
			end
			if right_patch /= Void then
				bust_patch (Result, right_patch, camera)
				if right_patch.count = 0 then
					right_patch := Void
				end
			end
			if above_patch /= Void then
				bust_patch (Result, above_patch, camera)
				if above_patch.count = 0 then
					above_patch := Void
				end
			end
			if below_patch /= Void then
				bust_patch (Result, below_patch, camera)
				if below_patch.count = 0 then
					below_patch := Void
				end
			end
		ensure
			Result_not_void: Result /= Void
			Result_count_in_range: Result.count >= 1
		end

	displace is
			-- Run a displacment shader on the grid
		do
		end

	shade (camera: SCENE_CAMERA) is
			-- Run a surface shader on the grid
		require
			not_shaded: not shaded
		local
			i, j: INTEGER
			pt: SHADING_DATA
			lights: SCENE_LIGHT_ELEMENT
			eye_point: POINT3
			eye_direction: VECTOR3
			n: NORMAL
		do
			lights ?= parent_element.elements.element_item ("SCENE_LIGHT_ELEMENT")
			if lights /= Void then
				create eye_point
				-- Use the default camera position of (0, 0, 0) to 
				-- find the world coordinate
				eye_point := camera.world_to_camera.inverse.apply (eye_point)
				create n
				-- For the first light, set the shaded color
				from
					lights.start
					j := 1
				invariant
					j_at_least_one: j >= 1
					j_small_enough: j <= 1 + height
				variant
					j_increasing: 1 + height - j
				until
					j > height
				loop
					from
						i := 1
					invariant
						i_at_least_one: i >= 1
						i_small_enough: i <= 1 + width
					variant
						i_increasing: 1 + width - i
					until
						i > width
					loop
						pt := item (i, j)
						n.copy (pt.geometric_normal)
						eye_direction := eye_point - pt.point
						eye_direction.normalize
						if n.dot_with_vector (eye_direction) < 0 then
							n.make (-n.x, -n.y, -n.z)
						end
						pt.set_shaded_color (pt.color * (lights.item.ambient + lights.item.diffuse (n, pt.point)))

						i := i + 1
					end
					j := j + 1
				end

				from
					-- For the remaining lights, accumulate the
					-- light color
				until
					lights.after
				loop
					from
						j := 1
					invariant
						j_at_least_one: j >= 1
						j_small_enough: j <= 1 + height
					variant
						j_increasing: 1 + height - j
					until
						j > height
					loop
						from
							i := 1
						invariant
							i_at_least_one: i >= 1
							i_small_enough: i <= 1 + width
						variant
							i_increasing: 1 + width - i
						until
							i > width
						loop
							pt := item (i, j)
							n.copy (pt.geometric_normal)
							eye_direction := eye_point - pt.point
							eye_direction.normalize
							if n.dot_with_vector (eye_direction) < 0 then
								n.make (-n.x, -n.y, -n.z)
							end
							pt.shaded_color.summation (pt.color * lights.item.diffuse (n, pt.point))

							i := i + 1
						end
						j := j + 1
					end

					lights.forth
				end
			else
				pt.set_shaded_color (pt.color.black)
			end

				shaded := true
		ensure
			shaded: shaded
		end

feature -- Miscellaneous

	print_neighbor_ids is
			-- Print the ids of neighboring grids
		do
			if has_left_patch then
				print_patch (left_patch, identifier)
			else
				io.put_string ("LL ")
			end
			if has_above_patch then
				print_patch (above_patch, identifier)
			else
				io.put_string ("AA ")
			end
			if has_right_patch then
				print_patch (right_patch, identifier)
			else
				io.put_string ("RR ")
			end
			if has_below_patch then
				print_patch (below_patch, identifier)
			else
				io.put_string ("BB ")
			end
		end

	print_active_grids is
			-- Print a list of active grids
		do
			shading_data_pool.print_active_grids
		end

feature -- Obsolete
feature -- Inapplicable
feature {SCENE_GEOMETRY} -- Implementation

	left_patch, right_patch, above_patch, below_patch: DS_ARRAYED_LIST [PATCH_OBJECT]
			-- The points that join this grid to its neighbor

feature {NONE} -- Implementation

	grid: ARRAY [SHADING_DATA]
			-- The actual grid of shading information

	u_stride, v_stride: DOUBLE
			-- The stride in uv-space between grid points

	grid_index (i, j: INTEGER): INTEGER is
			-- Get the index for location (`i', `j' )
		require
			i_in_range: i > 0 and i <= width
			j_in_range: j > 0 and j <= height
		do
			Result := (j - 1) * width + i
		ensure
			Result_in_range: Result > 0 and Result <= grid.count
		end

	bust_grid (mp_list: DS_ARRAYED_LIST [MICROPOLYGON];
			   element: SCENE_DRAWABLE_ELEMENT;
			   camera: SCENE_CAMERA) is
			-- Break the grid into micropolygons
		do
			if parent_element.geometry.two_sided then
				bust_grid_two_sided (mp_list, element, camera)
			else
				bust_grid_one_sided (mp_list, element, camera)
			end
		end

	bust_grid_two_sided (mp_list: DS_ARRAYED_LIST [MICROPOLYGON];
						 element: SCENE_DRAWABLE_ELEMENT;
						 camera: SCENE_CAMERA) is
			-- Break a two_sided  grid into micropolygons
		local
			mp: MICROPOLYGON
			i, j: INTEGER
			pt1, pt2, pt3, pt4: SHADING_DATA
			cpt, rpt: POINT3
			depth, dpt1, dpt2, dpt3, dpt4: DOUBLE
		do
			from
				j := 1
				create cpt
			invariant
				j_at_least_one: j >= 1
				j_small_enough: j <= 1 + height - 1
			variant
				j_increasing: 1 + height - 1 - j
			until
				j > height - 1
			loop
				from
					i := 1
					pt1 := item (i, j + 1)
					rpt := pt1.raster_point
					rpt.copy (pt1.point)
					rpt.apply (camera.world_to_raster)
					cpt.copy (pt1.point)
					cpt.apply (camera.world_to_camera)
					dpt1 := cpt.z
					pt4 := item (i, j)
					rpt := pt4.raster_point
					rpt.copy (pt4.point)
					rpt.apply (camera.world_to_raster)
					cpt.copy (pt4.point)
					cpt.apply (camera.world_to_camera)
					dpt4 := cpt.z
				invariant
					i_at_least_one: i >= 1
					i_small_enough: i <= 1 + width - 1
				variant
					i_increasing: 1 + width - 1 - i
				until
					i > width - 1
				loop
					pt2 := item (i + 1, j + 1)
					rpt := pt2.raster_point
					rpt.copy (pt2.point)
					rpt.apply (camera.world_to_raster)
					cpt.copy (pt2.point)
					cpt.apply (camera.world_to_camera)
					dpt2 := cpt.z
					pt3 := item (i + 1, j)
					rpt := pt3.raster_point
					rpt.copy (pt3.point)
					rpt.apply (camera.world_to_raster)
					cpt.copy (pt3.point)
					cpt.apply (camera.world_to_camera)
					dpt3 := cpt.z

					-- Compute the depth in camera space; Use max
					-- since we're looking down -Z.
					depth := dpt1.max (dpt2.max (dpt3.max (dpt4)))

					create mp.make (pt1, pt2, pt3, pt4)
					mp.set_depth (depth)

					mp_list.force_last (mp)

					pt1 := pt2
					pt4 := pt3
					dpt1 := dpt2
					dpt4 := dpt3

					i := i + 1
				end

				j := j + 1
			end
		end

	bust_grid_one_sided (mp_list: DS_ARRAYED_LIST [MICROPOLYGON];
						 element: SCENE_DRAWABLE_ELEMENT;
						 camera: SCENE_CAMERA) is
			-- Break a one-sided grid into micropolygons
		local
			mp: MICROPOLYGON
			i, j: INTEGER
			pt1, pt2, pt3, pt4: SHADING_DATA
			cpt, rpt: POINT3
			depth, dpt1, dpt2, dpt3, dpt4: DOUBLE
			n: NORMAL
			bf1, bf2, bf3, bf4: BOOLEAN
		do
			from
				j := 1
				create n
				create cpt
			invariant
				j_at_least_one: j >= 1
				j_small_enough: j <= 1 + height - 1
			variant
				j_increasing: 1 + height - 1 - j
			until
				j > height - 1
			loop
				from
					i := 1
					pt1 := item (i, j + 1)
					rpt := pt1.raster_point
					rpt.copy (pt1.point)
					rpt.apply (camera.world_to_raster)
					cpt.copy (pt1.point)
					cpt.apply (camera.world_to_camera)
					dpt1 := cpt.z
					n.copy (pt1.geometric_normal)
					n.apply (camera.world_to_camera)
					bf1 := n.dot_with_vector (camera.projection_direction) < 0
					pt4 := item (i, j)
					rpt := pt4.raster_point
					rpt.copy (pt4.point)
					rpt.apply (camera.world_to_raster)
					cpt.copy (pt4.point)
					cpt.apply (camera.world_to_camera)
					dpt4 := cpt.z
					n.copy (pt4.geometric_normal)
					n.apply (camera.world_to_camera)
					bf4 := n.dot_with_vector (camera.projection_direction) < 0
				invariant
					i_at_least_one: i >= 1
					i_small_enough: i <= 1 + width - 1
				variant
					i_increasing: 1 + width - 1 - i
				until
					i > width - 1
				loop
					pt2 := item (i + 1, j + 1)
					rpt := pt2.raster_point
					rpt.copy (pt2.point)
					rpt.apply (camera.world_to_raster)
					cpt.copy (pt2.point)
					cpt.apply (camera.world_to_camera)
					dpt2 := cpt.z
					n.copy (pt2.geometric_normal)
					n.apply (camera.world_to_camera)
					bf2 := n.dot_with_vector (camera.projection_direction) < 0
					pt3 := item (i + 1, j)
					rpt := pt3.raster_point
					rpt.copy (pt3.point)
					rpt.apply (camera.world_to_raster)
					cpt.copy (pt3.point)
					cpt.apply (camera.world_to_camera)
					dpt3 := cpt.z
					n.copy (pt3.geometric_normal)
					n.apply (camera.world_to_camera)
					bf3 := n.dot_with_vector (camera.projection_direction) < 0

					if not (bf1 and bf2 and bf3 and bf4) then
						-- Compute the depth in camera space; Use max
						-- since we're looking down -Z.
						depth := dpt1.max (dpt2.max (dpt3.max (dpt4)))

						create mp.make (pt1, pt2, pt3, pt4)
						mp.set_depth (depth)

						mp_list.force_last (mp)
					end

					pt1 := pt2
					pt4 := pt3
					bf1 := bf2
					bf4 := bf3
					dpt1 := dpt2
					dpt4 := dpt3

					i := i + 1
				end

				j := j + 1
			end
		end

	bust_patch (mp_list: DS_ARRAYED_LIST [MICROPOLYGON];
				patch_list: DS_ARRAYED_LIST [PATCH_OBJECT];
				camera: SCENE_CAMERA) is
			-- Break the patch into micropolygons
		local
			patch: PATCH_OBJECT
		do
			from
				patch_list.start
			until
				patch_list.after
			loop
				patch := patch_list.item_for_iteration

				if not patch.all_grids_shaded then
					-- Force all the neighbor grids to be shaded
					from
						patch.referers.start
					until
						patch.referers.after
					loop
						-- Ensure that we don't try to shade a neighbor that was released. (HACK! the referers item
						-- should have been released as well as the reference to it).
						if not patch.referers.item_for_iteration.released and not patch.referers.item_for_iteration.shaded then
							patch.referers.item_for_iteration.shade (camera)
						end
						patch.referers.forth
					end
				end

				-- Bust the patch if it hasn't already been done
				if not patch.busted then
					patch.bust (mp_list, camera.world_to_camera, camera.world_to_raster)
					patch_list.forth
				else
					patch.release
				end
			end
		end

	remove_patch (patch_list: DS_ARRAYED_LIST [PATCH_OBJECT];
				patch_to_remove: PATCH_OBJECT) is
			-- Remove all occurances of `patch_to_remove' from
			-- `patch_list'
		do
			from
				patch_list.start
				patch_list.search_forth (patch_to_remove)
			until
				patch_list.after
			loop
				patch_list.remove_at
				patch_list.search_forth (patch_to_remove)
			end
		ensure
			no_occurences: patch_list.occurrences (patch_to_remove) = 0
		end

	only_one_occurrence (patch_list: DS_ARRAYED_LIST [PATCH_OBJECT]): BOOLEAN is
			-- Check that each element of `patch_list' is only listed once
		do
			from
				patch_list.start
				Result := true
			until
				patch_list.after or not Result
			loop
				Result := patch_list.occurrences (patch_list.item_for_iteration) = 1
				patch_list.forth
			end
		end

	print_patch (patch_list: DS_ARRAYED_LIST [PATCH_OBJECT];
				current_id: INTEGER) is
			-- Print ids from `patch_list', excluding `current_id'
		do
			from
				patch_list.start
			until
				patch_list.after
			loop
				if patch_list.item_for_iteration.referers.count = 1 and
					patch_list.item_for_iteration.referers.item (1).identifier = current_id then
					io.put_string ("-- ")
				else
					from
						patch_list.item_for_iteration.referers.start
					until
						patch_list.item_for_iteration.referers.after
					loop
						if patch_list.item_for_iteration.referers.item_for_iteration.identifier /= current_id then
							io.put_string (patch_list.item_for_iteration.referers.item_for_iteration.identifier.out)
							io.put_string (" ")
						end

						patch_list.item_for_iteration.referers.forth
					end
				end

				patch_list.forth
			end
		end

	shading_data_pool: SHADING_DATA_POOL is
			-- Create the pool
		once
			create Result
		end

	object_id: OBJECT_IDENTIFIER is
			-- Create the id generator
		once
			create Result
		end

invariant
	grid_not_void: grid /= Void
	width_positive: width > 0
	height_positive: height > 0
	vertex_count_equals_grid_capacity: vertex_count = grid.capacity

end -- MICROPOLYGON_GRID

