

indexing
	description: "Pixel information"
	patterns: "none"
	status: "unreviewed"
	date: "$Date: 2006-12-01 11:29:53 -0500 (Fri, 01 Dec 2006) $"
	revision: "$Revision: 270 $"



class

	PIXEL

inherit

	SINGLETON_ACCESSOR
		rename
			singleton as statistics
		redefine
			statistics
		select
			statistics
		end

	SINGLETON_ACCESSOR
		rename
			singleton as sample_missed_only
		redefine
			sample_missed_only
		end

	BASIC_NUMERIC

creation
	make, make_with_extent

feature -- Initialization

	make (x, y: INTEGER) is
			--
		require
			x_positive: x >= 0
			y_positive: y >= 0
		do
			make_with_extent (x, y, 1, 1)
		end

	make_with_extent (x, y: INTEGER; width, height: DOUBLE) is
			--
		require
			x_positive: x >= 0
			y_positive: y >= 0
			width_positive: width > 0
			height_positive: height > 0
		local
			lower, upper: POINT3
		do
			create lower.make (x, y, 0)
			create upper.make (x + width, y + height, 0)
			create extent.make (lower, upper)

			create micro_polys.make (1)

			samples_in_x := 1
			samples_in_y := 1

			create colors.make (samples_in_x * samples_in_y)

			create test_point

			closest_depth := 0

			create intervals.make (2)
			create interval_1.make (0, 1, 1)
			intervals.force_last (interval_1)
			create interval_2.make (0, 1, 1)
			intervals.force_last (interval_2)
			create domain.make_from_intervals (intervals)

			create {STRATIFIED_SAMPLER_2D} s_sampler.make (domain)

			samples_need_regeneration := true
		end

feature -- Access

	extent: AXIS_ALIGNED_BOX
			-- The domain of this pixel

	color: RGBA_COLOR is
			-- The final color after shading
		local
			i, sample_count: INTEGER
		do
			if is_empty then
				create Result.make (0, 0, 0, 1)
			else
				sample_count := samples_in_x * samples_in_y

				from
					i := 2
					Result := flatten (colors.item (1))
				invariant
					i_at_least_one: i >= 2
					i_small_enough: i <= 2 + sample_count
				variant
					i_increasing: 2 + sample_count - i
				until
					i > sample_count
				loop
					Result := Result + flatten (colors.item (i))
					i := i + 1
				end

				Result := Result.scale (1.0 / sample_count)
			end
		end

feature -- Measurement

	count: INTEGER is
			-- How many items cover this pixel?
		do
			Result := micro_polys.count
		end

feature -- Comparison
feature -- Status report

	is_empty: BOOLEAN is
			-- Is this pixel covered by any micropolygon?
		do
			Result := micro_polys.count = 0
		end

	has (micro: MICROPOLYGON): BOOLEAN is
			-- Does this pixel already have `micro'?
		do
			Result := micro_polys.has (micro)
		end

	closest_depth: DOUBLE
			-- What's the nearest depth of the micropolygons

	samples_in_x, samples_in_y: INTEGER
			-- The number of samples in each dimension

feature -- Status setting

	set_sample_rate (x_samples, y_samples: INTEGER) is
		-- Set the number of samples in each dimension to
		-- (`x_samples', `y_samples')
		require
			strictly_positive: x_samples >= 1 and  y_samples >= 1
		do
			samples_in_x := x_samples
			samples_in_y := y_samples
			if samples_in_x * samples_in_y /= colors.count then
				colors.resize (samples_in_x * samples_in_y)
			end
			samples_need_regeneration := true
		ensure
			samples_set: samples_in_x = x_samples and  samples_in_y = y_samples
		end

	statistics_records_misses_only (on_or_off: BOOLEAN) is
			-- Should we record only misses in the statistics
		do
				sample_missed_only.set_flag (on_or_off)
		end

feature -- Cursor movement
feature -- Element change

	add_micropolygon (micro: MICROPOLYGON) is
			-- Add a micropolygon to this pixel
		require
			micro_not_void: micro /= Void
			overlaps: overlaps (micro)
			new_micro: not has (micro)
		do
			if micro.depth < closest_depth then
				closest_depth := micro.depth
			end

			micro_polys.force_last (micro)
			update_references (micro)
		ensure
			added: micro_polys.last = micro
		end

	set_origin (x, y: INTEGER; width, height: DOUBLE) is
			-- Set the origin
		require
			width_positive: width > 0
			height_positive: height > 0
		do
			extent.lower.make (x, y, 0)
			extent.upper.make (x + width, y + height, 0)
		end

feature -- Removal

	release is
			-- Release memory
		local
			i: INTEGER
		do
			from
				i := 1
			invariant
				i_at_least_one: i >= 1
				i_small_enough: i <= 1 + micro_polys.count
			variant
				i_increasing: 1 + micro_polys.count - i
			until
				i > micro_polys.count
			loop
				remove_references (micro_polys.item (i))

				i := i + 1
			end

			micro_polys.wipe_out
			colors.wipe_out
			closest_depth := 0
		ensure
			discarded: micro_polys.count = 0 and colors.count = 0
		end

feature -- Resizing
feature -- Transformation
feature -- Conversion
feature -- Duplication
feature -- Basic operations

	overlaps (micro: MICROPOLYGON): BOOLEAN is
			-- Does `micro' overlap this pixel?
		require
			micro_not_void: micro /= Void
		local
			m_extent: AXIS_ALIGNED_BOX
		do
			m_extent := micro.extent
			Result := not
						(m_extent.lower.x >= extent.upper.x or else
						m_extent.upper.x < extent.lower.x or else
						m_extent.lower.y >= extent.upper.y or else
						m_extent.upper.y < extent.lower.y)
		end

	sort is
			-- Sort the micropolygons by depth
		do
			micro_polys.sort
		ensure
			default: true
		end

	sample is
			-- Sample the pixel
		local
			point: POINT
			hit: BOOLEAN
			opaque: BOOLEAN
			max_depth: DOUBLE
			mp: MICROPOLYGON
			i: INTEGER
			color_list: DS_ARRAYED_LIST [RGBA_COLOR]
			p: PAIR [POINT, DS_ARRAYED_LIST [RGBA_COLOR]]
		do
			if not is_empty then
				opaque := true
				if samples_need_regeneration then
					samples_need_regeneration := false
					generate_samples
				end

				--micro_polys.sort

				from
					s_sampler.start
					s_sampler.generate_samples
				until
					s_sampler.after
				loop
					max_depth := -maximum_double
					point := s_sampler.item
					hit := false
					test_point.make (extent.lower.x + point.item (1),
									 extent.lower.y + point.item (2), 0)

					debug ("dump_samples")
						if not sample_missed_only.flag then
							statistics.samples.force_last (test_point.twin)
						end
					end

					create color_list.make (1)
					create p.make (point, color_list)

					from
						i := 1
					invariant
						i_at_least_one: i >= 1
						i_small_enough: i <= 1 + micro_polys.count
					variant
						i_increasing: 1 + micro_polys.count - i
					until
						i > micro_polys.count
					loop
						mp := micro_polys.item (i)

						-- Distances are negative
						if mp.depth >= max_depth then
							if mp.has (test_point) then
								-- Opacity accumulation will have to
								-- be computed at the point of shading
								hit := true
								if opaque then
									max_depth := mp.depth
								end
								color_list.force_last (mp.lower_left.shaded_color)
							end
						end

						i := i + 1
					end

					if not hit then
						color_list.force_last (black)
						debug ("dump_samples")
							if sample_missed_only.flag then
								statistics.samples.force_last (test_point.twin)
							end
						end
					end

					colors.force_last (p)

					s_sampler.forth
				end
			end
		end

feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {NONE} -- Implementation

	micro_polys: SORTABLE_ARRAYED_LIST [MICROPOLYGON]
	s_sampler: STRATIFIED_SAMPLER

	domain: DOMAIN_2D
	interval_1, interval_2: LINEAR_INTERVAL
	intervals: DS_ARRAYED_LIST [LINEAR_INTERVAL]
		-- Variables for `generate_samples'

	samples_need_regeneration: BOOLEAN
		-- Flag for determining if the samples need to be regenerated

	test_point: POINT3
			-- Temporary for sample testing

	colors: DS_ARRAYED_LIST [PAIR [POINT, DS_ARRAYED_LIST [RGBA_COLOR]]]
			-- The colors at a particular sample

	black: RGBA_COLOR is
			-- The color black
		once
			create Result.make (0, 0, 0, 1)
		ensure
			Result_not_void: Result /= Void
		end

	generate_samples is
			-- Create samples points
		do
			interval_1.make (0, 1, samples_in_x)
			intervals.replace (interval_1, 1)
			interval_2.make (0, 1, samples_in_y)
			intervals.replace (interval_2, 2)
			domain.make_from_intervals (intervals)

			s_sampler.make (domain)

			s_sampler.set_jittering (true)
		end

	flatten (the_sample: PAIR [POINT, DS_ARRAYED_LIST [RGBA_COLOR]]): RGBA_COLOR is
			-- Flatten the colors in `the_sample'
		require
			the_sample_not_void: the_sample /= Void
		do
			Result := the_sample.second.last
		end

	statistics: STATISTICS is
			-- The statistics
		once
			create Result
		end

	sample_missed_only: BOOLEAN_SINGLETON is
			-- Should we record only misses in the statistics
		once
			create Result
		end

	update_references (micro: MICROPOLYGON) is
			-- Add a reference to the micropolygon for this pixel
		do
			micro.lower_left.add_reference (Current)
			micro.lower_right.add_reference (Current)
			micro.upper_right.add_reference (Current)
			micro.upper_left.add_reference (Current)
		end

	remove_references (micro: MICROPOLYGON) is
			-- Free the memory used by the micropolygon for this pixel
		do
			micro.lower_left.remove_reference (Current)
			micro.lower_right.remove_reference (Current)
			micro.upper_right.remove_reference (Current)
			micro.upper_left.remove_reference (Current)
		end

invariant

	samples_strictly_positive: samples_in_x >= 1 and  samples_in_y >= 1
	micro_polys_not_void: micro_polys /= Void
--	sample_data_not_void: sample_data /= Void

end -- PIXEL

