

indexing
	description: "A container for pixels organized as a 2d bucket"
	patterns: "none"
	status: "unreviewed"



class

	PIXEL_BUCKET

inherit

	NUMERIC_CONVERSION
		export
			{NONE} all
		end

	SINGLETON_ACCESSOR
		rename
			singleton as statistics
		redefine
			statistics
		end

creation
	make, make_with_samples, make_with_extents

feature -- Initialization

	make (start_x, start_y: INTEGER) is
			-- Create a pixel bucket at (`start_x', `start_y')
			-- with dimensions (1, 1) and using samples (1, 1)
		require
			start_x_positive: start_x >= 0
			start_y_positive: start_y >= 0
		do
			make_with_extents (start_x, start_y, 1, 1, 1, 1)
		end

	make_with_samples (start_x, start_y, x_samples, y_samples: INTEGER) is
			-- Create a pixel bucket at (`start_x', `start_y')
			-- with dimensions (1, 1) and using
			-- (`x_samples', `y_samples')
		require
			start_x_positive: start_x >= 0
			start_y_positive: start_y >= 0
			x_samples_strictly_positive: x_samples >= 1
			y_samples_strictly_positive: y_samples >= 1
		do
			make_with_extents (start_x, start_y, 1, 1, x_samples, y_samples)
		end

	make_with_extents (start_x, start_y, new_width, new_height,
					   x_samples, y_samples: INTEGER) is
			-- Create a pixel bucket at (`start_x', `start_y')
			-- with dimensions (`new_width', `new_height') and using
			-- (`x_samples', `y_samples')
		require
			start_x_positive: start_x >= 0
			start_y_positive: start_y >= 0
			new_width_strictly_positive: new_width >= 1
			new_height_strictly_positive: new_height >= 1
			x_samples_strictly_positive: x_samples >= 1
			y_samples_strictly_positive: y_samples >= 1
		local
			i, j, p_index: INTEGER
			pixel: PIXEL
		do
			if origin /= Void then
				origin.make (start_x, start_y)
			else
				create origin.make (start_x, start_y)
			end

			if pixels /= Void then
				if width /= new_width or else height /= new_height then
					width := new_width
					height := new_height

					pixels.make (1, width * height)
				end
			else
				width := new_width
				height := new_height
				create pixels.make (1, width * height)
			end

			from
				j := start_y
			invariant
				j_at_least_one: j >= start_y
				j_small_enough: j <= start_y + height
			variant
				j_increasing: start_y + height - j
			until
				j >= start_y + height
			loop
				from
					i := start_x
				invariant
					i_at_least_one: i >= start_x
					i_small_enough: i <= start_x + width
				variant
					i_increasing: start_x + width - i
				until
					i >= start_x + width
				loop
					p_index := pixel_index (j, i)
					pixel := pixels.item (p_index)
					if pixel = Void then
						create pixel.make (i, j)
						pixels.put (pixel, p_index)
					else
						pixel.set_origin (i, j,
						pixel.extent.upper.x - pixel.extent.lower.x,
						pixel.extent.upper.y - pixel.extent.lower.y)
						pixel.release
					end

					if pixel.samples_in_x /= x_samples and then pixel.samples_in_y /= y_samples then
						pixel.set_sample_rate (x_samples, y_samples)
					end

					i := i + 1
				end
				j := j + 1
			end
		end

feature -- Access

	origin: POINT2
			-- The location of the lower left corner

	row, column: INTEGER
			-- The current pixel being processed

	width, height: INTEGER
			-- The dimensions

	item (the_row, the_column: INTEGER): PIXEL is
			-- The pixel at (`the_row', `the_column')
		require
			the_row_in_range: the_row >= origin.y and the_row < origin.y + height
			the_column_in_range: the_column >= origin.x and the_column < origin.x + width
		do
			Result := pixels.item (pixel_index (the_row, the_column))
		ensure
			Result_not_void: Result /= Void
			Result_is_pixel: Result = pixels.item (pixel_index (the_row, the_column))
		end

	item_for_iteration: PIXEL is
			-- The pixel at (`row', `column')
		do
			Result := pixels.item (pixel_index (row, column))
		ensure
			Result_not_void: Result /= Void
			Result_is_pixel: Result = pixels.item (pixel_index (row, column))
		end

feature -- Measurement
feature -- Comparison
feature -- Status report

	after: BOOLEAN is
			-- Are we after the last pixel?
		do
			Result := column = real_64_truncated_to_integer (origin.x) and
				row >= real_64_truncated_to_integer (origin.y + height)
		end

feature -- Status setting
feature -- Cursor movement

	start is
			-- Begin an iteration
		do
			row := real_64_truncated_to_integer (origin.y)
			column := real_64_truncated_to_integer (origin.x)
		end

	forth is
			-- Move forward a pixel
		do
			column := column + 1
			if column >= origin.x + width then
				column := real_64_truncated_to_integer (origin.x)
				row := row + 1
			end
		end

feature -- Element change

	add_micropolygon (mp: MICROPOLYGON) is
			-- Add a micropolygon to the pixels it overlaps
		local
			i, j: INTEGER
			lower_y, upper_y, lower_x, upper_x: INTEGER
		do
			lower_x := real_64_truncated_to_integer (mp.extent.lower.x.max (origin.x))
			lower_y := real_64_truncated_to_integer (mp.extent.lower.y.max (origin.y))
			upper_x := real_64_truncated_to_integer (mp.extent.upper.x.min (origin.x + width - 1))
			upper_y := real_64_truncated_to_integer (mp.extent.upper.y.min (origin.y + height - 1))

			from
				j := lower_y
			invariant
				j_at_least_one: j >= lower_y
				j_small_enough: j <= lower_y + upper_y + 1
			variant
				j_increasing: lower_y + upper_y +1 - j
			until
				j > upper_y
			loop
				from
					i := lower_x
				invariant
					i_at_least_one: i >= lower_x
					i_small_enough: i <= lower_x + upper_x + 1
				variant
					i_increasing: lower_x + upper_x + 1 - i
				until
					i > upper_x
				loop
					item (j, i).add_micropolygon (mp)
					check
						overlaps: item (j, i).overlaps (mp)
					end

					i := i + 1
				end
				j := j + 1
			end
		end

feature -- Removal
feature -- Resizing
feature -- Transformation
feature -- Conversion
feature -- Duplication
feature -- Basic operations

	sample is
			-- Sample the pixels to generate the appropriate color
		local
			depth_complexity: INTEGER
			p: PIXEL
		do
			from
				start
			until
				after
			loop
				p := item_for_iteration
				p.sample

				if not p.is_empty then
					depth_complexity := statistics.item (statistics.mp_per_pixel_min)
					if depth_complexity = 0 then
						statistics.put (p.count,
										statistics.mp_per_pixel_min)
						statistics.put (p.count,
										statistics.mp_per_pixel_max)
						statistics.put (p.count,
										statistics.mp_per_pixel_avg)
						statistics.increment (statistics.mp_pixel_count)
					else
						statistics.put (depth_complexity.min (p.count),
										statistics.mp_per_pixel_min)
						statistics.put (statistics.item (statistics.mp_per_pixel_max).max (p.count),
										statistics.mp_per_pixel_max)
						statistics.increase (p.count,
											 statistics.mp_per_pixel_avg)
						statistics.increment (statistics.mp_pixel_count)
					end
				end

				forth
			end
		end

feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {NONE} -- Implementation

	pixels: ARRAY [PIXEL]
			-- The actual pixel data

	pixel_index (the_row, the_column: INTEGER): INTEGER is
			-- The index into `pixels'
		require
			the_row_in_range: the_row >= origin.y and the_row < origin.y + height
			the_column_in_range: the_column >= origin.x and the_column < origin.x + width
		do
			Result := (the_row - real_64_truncated_to_integer_32 (origin.y)) * width +
				(the_column - real_64_truncated_to_integer_32 (origin.x)) + 1
		ensure
			Result_in_range: Result >= 1 and Result <= pixels.count
		end

	statistics: STATISTICS is
			-- The statistics
		once
			create Result
		end

invariant

	pixels_not_void: pixels /= Void

end -- PIXEL_BUCKET

