indexing
	description: "Interface to the OpenGL routines."
	note: "This only interfaces to the double variant of those %
			%routines that take multiple types. Converters are provided %
			%internally to those few routines that require reals"
	patterns: "none"
	date: "$Date: 2007-02-20 16:28:19 -0500 (Tue, 20 Feb 2007) $"
	revision: "$Revision: 283 $"
	status: "unreviewed"

class

	GL

inherit

	RENDERER
		redefine
			default_create
		end

	EWG_EXTERNAL_STRING_ROUTINES
		export
			{NONE} all
		redefine
			default_create
		end

	NUMERIC_CONVERSION
		redefine
			default_create
		end

	GL_IMPLEMENTATION
		redefine
			default_create
		end

creation
	default_create

feature -- Initialization

	default_create is
			-- Initialize the internals
		do
			initialize

			Precursor {GL_IMPLEMENTATION}

			-- matrix_array is 0-based
			create matrix_array.make (16)
			matrix_array.fill

			number_of_slices := 20
			number_of_stacks := 20

			check_for_context_valid := true
		end

	begin_statistics is
		do
			vertex_count := 0
			polygon_count := 0
			current_mode := 0
		end

	vertex_count, polygon_count: INTEGER

	initialize_extensions is
			-- Initialize the extensions
		do
			if glew_init = 0 then
				extensions_initialized := true
			else
				extensions_initialized := false
			end
		end
feature -- Access

	draw_style: INTEGER is
			-- What is the current drawing style?
		do
			Result := quadric_draw_style
		end

feature -- Status Setting

	set_number_of_slices (slice_count: INTEGER) is
			-- The number of polygons to use in u direction
		require
			slice_count_positive: slice_count >= 1
		do
			number_of_slices := slice_count
		ensure
			number_of_slices_set: number_of_slices = slice_count
		end

	set_number_of_stacks (stack_count: INTEGER) is
			-- The number of polygons to use in v direction
		require
			stack_count_positive: stack_count >= 1
		do
			number_of_stacks := stack_count
		ensure
			number_of_stacks_set: number_of_stacks = stack_count
		end

feature -- NURBS Primitives

	new_nurbs_renderer: NURBS_RENDERER is
			-- Create a NURBS renderer
		do
			create {GLU_NURBS_RENDERER} Result.make (Current)
		end

feature -- High Level Primitives

	cone (radius, height, sweep_angle,
		  u_minimum, u_maximum, v_minimum, v_maximum: DOUBLE) is
		local
			slices, stacks: INTEGER
		do
			slices := real_64_truncated_to_integer ((number_of_slices * (u_maximum - u_minimum)).rounded).max (4)
			stacks := real_64_truncated_to_integer ((number_of_stacks * (v_maximum - v_minimum)).rounded).max (3)
			gl_cone (radius, height, sweep_angle,
					 u_minimum, u_maximum, v_minimum, v_maximum,
					 slices, stacks)
		ensure then
			gl_error_free: get_error = false
		end

	cube (lower, upper: POINT3) is
		do
			gl_cube (lower, upper)
		ensure then
			gl_error_free: get_error = false
		end

	cylinder (base_radius, top_radius, z_minimum, z_maximum, sweep_angle,
			u_minimum, u_maximum, v_minimum, v_maximum: DOUBLE) is
		local
			slices, stacks: INTEGER
		do
			slices := real_64_truncated_to_integer ((number_of_slices * (u_maximum - u_minimum)).rounded).max (4)
			stacks := real_64_truncated_to_integer ((number_of_stacks * (v_maximum - v_minimum)).rounded).max (3)
			gl_cylinder (base_radius, top_radius, z_minimum, z_maximum, sweep_angle,
						 u_minimum, u_maximum, v_minimum, v_maximum,
						 slices, stacks)
		ensure then
			gl_error_free: get_error = false
		end

	disk (radius, height, sweep_angle: DOUBLE) is
		do
			gl_disk (radius, height, sweep_angle, 20, 20)
		ensure then
			gl_error_free: get_error = false
		end

	sphere (radius, z_minimum, z_maximum, sweep_angle,
			u_minimum, u_maximum, v_minimum, v_maximum: DOUBLE) is
		local
			slices, stacks: INTEGER
		do
			slices := real_64_truncated_to_integer ((number_of_slices * (u_maximum - u_minimum)).rounded).max (1)
			stacks := real_64_truncated_to_integer ((number_of_stacks * (v_maximum - v_minimum)).rounded).max (1)
			gl_sphere (radius, z_minimum, z_maximum, sweep_angle,
					   u_minimum, u_maximum, v_minimum, v_maximum,
					   slices, stacks)
		ensure then
			gl_error_free: get_error = false
		end

	torus (inner_radius, outer_radius, phi_min, phi_max, sweep_angle,
			u_minimum, u_maximum, v_minimum, v_maximum: DOUBLE) is
		local
			slices, stacks: INTEGER
		do
			slices := real_64_truncated_to_integer ((number_of_slices * (u_maximum - u_minimum)).rounded).max (1)
			stacks := real_64_truncated_to_integer ((number_of_stacks * (v_maximum - v_minimum)).rounded).max (1)
			gl_torus (inner_radius, outer_radius, phi_min, phi_max, sweep_angle,
					   u_minimum, u_maximum, v_minimum, v_maximum,
					  slices, stacks)
		ensure then
			gl_error_free: get_error = false
		end

feature -- Shading Language

	has_shading_language: BOOLEAN is
			-- Does this OpenGL implementation support GLslang
		do
			Result := has_fragment_shader_support and
				has_vertex_shader_support and
				has_shader_object_support and
				has_shading_language_support
		end

	create_vertex_shader: INTEGER is
			-- Create a vertex shader
		do
			Result := gl_create_shader_object (rd_vertex_shader)
		ensure then
			default: true
		end

	create_fragment_shader: INTEGER is
			-- Create a fragment shader
		do
			Result := gl_create_shader_object (rd_fragment_shader)
		ensure then
			default: true
		end

	set_shader_source (shader_id: INTEGER; buffer: STRING) is
			-- Set the source of a shader
		local
			ext_string: EWG_ZERO_TERMINATED_STRING
		do
			create ext_string.make_unshared_from_string (buffer)
			gl_shader_source (shader_id, 1, ext_string.item, default_pointer)
		end

	compile_shader (shader_id: INTEGER) is
			-- Compile the shader corresponding to `shader_id'
		do
			gl_compile_shader (shader_id)
			debug ("shader")
				io.put_string (shader_status (shader_id))
				io.put_string ("%N%N")
			end
		ensure then
			gl_error_free: get_error = false
			shader_compiled: shader_compiled (shader_id)
		end

	shader_compiled (shader_id: INTEGER): BOOLEAN is
			-- Has `shader_id' been compiled
		do
			Result := check_shading_status (shader_id, rd_compile_status)
		end

	program_linked (shader_id: INTEGER): BOOLEAN is
			-- Has `shader_id' been linked successfully?
		do
			Result := check_shading_status (shader_id, rd_link_status)
		end

	shader_status (shader_id: INTEGER): STRING is
			-- What is the status of `shader_id'?
		local
			compiled,status: EWG_INT_ARRAY
			ext_string: EWG_ZERO_TERMINATED_STRING
		do
			create compiled.make_new_unshared (1)
			gl_get_object_parameteriv (shader_id, rd_information_log_length,
									   compiled.array_address)
			if compiled.item (0) > 0 then
				create ext_string.make_new_unshared (compiled.item (0))
				create status.make_new_unshared (1)
				gl_get_info_log (shader_id, compiled.item (0),
								 status.array_address, ext_string.item)
			end

			Result := ext_string.string
		end

	create_shader_program: INTEGER is
			-- Create a shader program
		do
			Result := gl_create_program_object
		end

	attach_shader_object (program, object: INTEGER) is
			-- Attach a shader to the program
		do
			gl_attach_object (program, object)
		end

	link_shader_program (program_id: INTEGER) is
			-- Link the program defined by `program_id'?
		local
			s: STRING
		do
			gl_link_program (program_id)
			s := shader_status (program_id)
		ensure then
			gl_error_free: get_error = false
			shader_compiled: program_linked (program_id)
		end

	use_shader_program (program_id: INTEGER) is
			-- Load the program defined by `program_id'?
		do
			gl_use_program_object (program_id)
		ensure then
			gl_error_free: get_error = false
		end

	set_uniform_parameters (program_id: INTEGER; parameter: STRING;
						   values: REAL_ARRAYED_LIST) is
			-- Set the value for the uniform `parameter' in
			-- `program_id'
		local
			location: INTEGER
			ext_string: EWG_ZERO_TERMINATED_STRING
		do
			create ext_string.make_unshared_from_string (parameter)
			location := gl_get_uniform_location (program_id, ext_string.item)
			if location >= 0 then
				inspect
					values.count
				when 1 then
					gl_uniform1f (location, values.item (1))
				when 2 then
					gl_uniform2f (location, values.item (1), values.item (2))
				when 3 then
					gl_uniform3f (location, values.item (1), values.item (2),
								   values.item (3))
				when 4 then
					gl_uniform4f (location, values.item (1), values.item (2),
								   values.item (3), values.item (4))
				end
			end
		ensure then
			gl_error_free: get_error = false
		end

	get_uniform_variable_names (program_id: INTEGER): DS_ARRAYED_LIST [STRING] is
			-- What uniform variables are used by `program_id'?
		local
			compiled,status: EWG_INT_ARRAY
			ext_string: EWG_ZERO_TERMINATED_STRING
			i, count, length: INTEGER
		do
			create compiled.make_new_unshared (1)
			create status.make_new_unshared (1)

			gl_get_object_parameteriv (program_id, rd_object_active_uniforms,
									   compiled.array_address)
			if compiled.item (0) > 0 then
				count := compiled.item (0)
				create Result.make (count)
				gl_get_object_parameteriv (program_id, rd_object_active_uniform_max_length,
										   compiled.array_address)
				length := compiled.item (0)
				create ext_string.make_new_unshared (length)
				create status.make_new_unshared (1)

				from
					i := 0
				invariant
					i_at_least_one: i >= 0
					i_small_enough: i <= 0 + count
				variant
					i_increasing: 0 + count - i
				until
					i >= count
				loop
					gl_get_active_uniform (program_id, i, length,
										   default_pointer,
										   compiled.array_address,
										   status.array_address, ext_string.item)

					Result.force_last (ext_string.string)
					i := i + 1
				end
			end
		end

feature -- Primitives

	render_begin (mode: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("render_begin (")
				io.put_double (mode)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_render_begin (mode)

			current_mode := mode

			check_for_context_valid := false
		ensure then
			gl_error_free: get_error = false
		end

	render_end is
		do
			debug ("graphics")
				io.put_string ("render_end")
				io.put_string ("%N")
			end

			gl_render_end

			if current_mode = rd_triangle_strip or current_mode = rd_triangle_fan then
				polygon_count := polygon_count + vertex_count - 2
			elseif current_mode = rd_quad_strip then
				polygon_count := polygon_count + (vertex_count - 2) // 2
			elseif current_mode = rd_triangles then
				polygon_count := polygon_count + vertex_count // 3
			elseif current_mode = rd_polygon then
				polygon_count := polygon_count + 1
			end

			vertex_count := 0
			check_for_context_valid := true
		ensure then
			gl_error_free: get_error = false or error_code = gl_invalid_operation
		end

	rect (x1, y1, x2, y2: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("rect (x1, y1, x2, y2: DOUBLE)")
				io.put_string ("%N")
			end

			gl_rect (x1, y1, x2, y2)
		ensure then
			gl_error_free: get_error = false
		end

	vertex2 (x, y: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("vertex2 (x, y: DOUBLE)")
				io.put_string ("%N")
			end

			gl_vertex2 (x, y)
		ensure then
			gl_error_free: get_error = false
		end

	vertex3 (x, y, z: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("gl_vertex3 (")
				io.put_double (x)
				io.put_string (", ")
				io.put_double (y)
				io.put_string (", ")
				io.put_double (z)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_vertex3 (x, y, z)

			vertex_count := vertex_count + 1
		ensure then
			gl_error_free: get_error = false
		end

	vertex4 (x, y, z, w: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("gl_vertex4 (")
				io.put_double (x)
				io.put_string (", ")
				io.put_double (y)
				io.put_string (", ")
				io.put_double (z)
				io.put_string (", ")
				io.put_double (w)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_vertex4 (x, y, z, w)

			vertex_count := vertex_count + 1
		ensure then
			gl_error_free: get_error = false
		end

	vertex_pointer (p: POINTER) is
		do
			debug ("graphics")
				io.put_string ("vertex_pointer (p: POINTER)")
				io.put_string ("%N")
			end

			gl_vertex_pointer (3, gl_double, 0, p)
		ensure then
			gl_error_free: get_error = false
		end

feature -- Coordinate Transforms

	rotate (angle, x, y, z: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("rotate (")
				io.put_double (angle)
				io.put_string (", ")
				io.put_double (x)
				io.put_string (", ")
				io.put_double (y)
				io.put_string (", ")
				io.put_double (z)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_rotate (angle, x, y, z)
		ensure then
			gl_error_free: get_error = false
		end

	scale (x, y, z: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("scale (")
				io.put_double (x)
				io.put_string (", ")
				io.put_double (y)
				io.put_string (", ")
				io.put_double (z)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_scale (x, y, z)
		ensure then
			gl_error_free: get_error = false
		end

	translate (x, y, z: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("translate (")
				io.put_double (x)
				io.put_string (", ")
				io.put_double (y)
				io.put_string (", ")
				io.put_double (z)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_translate (x, y, z)
		ensure then
			gl_error_free: get_error = false
		end

	multiply_matrix (m: TRANSFORM) is
			-- Interface to glMultMatrix
		do
			m.transpose
			debug ("graphics")
				io.put_string ("gl_multiply_matrix (")
				io.put_string (m.out)
				io.put_string (")")
				io.put_string ("%N")
			end
			m.to_array_argument (matrix_array)
			gl_mult_matrix (matrix_array.storage.item)
			m.transpose
		ensure then
			gl_error_free: get_error = false
		end

	frustum (left, right, bottom, top, near, far: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("frustum (")
				io.put_double (left)
				io.put_string (", ")
				io.put_double (right)
				io.put_string (", ")
				io.put_double (bottom)
				io.put_string (", ")
				io.put_double (top)
				io.put_string (", ")
				io.put_double (near)
				io.put_string (", ")
				io.put_double (far)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_frustum (left, right, bottom, top, near, far)
		ensure then
			gl_error_free: get_error = false
		end

	ortho (left, right, bottom, top, near, far: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("ortho (")
				io.put_double (left)
				io.put_string (", ")
				io.put_double (right)
				io.put_string (", ")
				io.put_double (bottom)
				io.put_string (", ")
				io.put_double (top)
				io.put_string (", ")
				io.put_double (near)
				io.put_string (", ")
				io.put_double (far)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_ortho (left, right, bottom, top, near, far)
		ensure then
			gl_error_free: get_error = false
		end

	load_identity is
		do
			debug ("graphics")
				io.put_string ("load_identity")
				io.put_string ("%N")
			end

			gl_load_identity
		ensure then
			gl_error_free: get_error = false
		end

	load_matrix (m: TRANSFORM) is
		do
			m.transpose

			debug ("graphics")
				io.put_string ("gl_load_matrix (")
				io.put_string ("%N")
				io.put_string (m.out)
				io.put_string (")")
				io.put_string ("%N")
			end

			m.to_array_argument (matrix_array)
			gl_load_matrix (matrix_array.storage.item)
			m.transpose
		ensure then
			gl_error_free: get_error = false
		end

	matrix_mode (mode: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("matrix_mode (")
				io.put_integer (mode)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_matrix_mode (mode)
		ensure then
			gl_error_free: get_error = false
		end

	pop_matrix is
		do
			debug ("graphics")
				io.put_string ("pop_matrix")
				io.put_string ("%N")
			end

			gl_pop_matrix
		ensure then
			gl_error_free: get_error = false
		end

	push_matrix is
		do
			debug ("graphics")
				io.put_string ("push_matrix")
				io.put_string ("%N")
			end

			gl_push_matrix
		ensure then
			gl_error_free: get_error = false
		end

	viewport (left, right, bottom, top: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("viewport (")
				io.put_integer (left)
				io.put_string (", ")
				io.put_integer (right)
				io.put_string (", ")
				io.put_integer (bottom)
				io.put_string (", ")
				io.put_integer (top)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_viewport (left, right, bottom, top)
		ensure then
			gl_error_free: get_error = false
		end

feature -- Coloring and Lighting

	color_3 (red, green, blue: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("color_3 (")
				io.put_double (red)
				io.put_string (", ")
				io.put_double (green)
				io.put_string (", ")
				io.put_double (blue)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_color_3 (red, green, blue)
		ensure then
			gl_error_free: get_error = false
		end

	color_4 (red, green, blue, alpha: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("color_4 (")
				io.put_double (red)
				io.put_string (", ")
				io.put_double (green)
				io.put_string (", ")
				io.put_double (blue)
				io.put_string (", ")
				io.put_double (alpha)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_color_4 (red, green, blue, alpha)
		ensure then
			gl_error_free: get_error = false
		end

	color_material (face, mode: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("color_material (face, mode: INTEGER)")
				io.put_string ("%N")
			end

			gl_color_material (face, mode)
		ensure then
			gl_error_free: get_error = false
		end

	color_index (index: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("color_index (index: DOUBLE)")
				io.put_string ("%N")
			end

			gl_color_index_imp (index)
		ensure then
			gl_error_free: get_error = false
		end

	normal (nx, ny, nz: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("normal (")
				io.put_double (nx)
				io.put_string (", ")
				io.put_double (ny)
				io.put_string (", ")
				io.put_double (nz)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_normal (nx, ny, nz)
		ensure then
			gl_error_free: get_error = false
		end

	material (light_number, pname: INTEGER; parameter: ARRAY [DOUBLE]) is
		local
			vals: EWG_REAL_ARRAY
			i: INTEGER
		do
			debug ("graphics")
				io.put_string ("material (light_number, pname: INTEGER; parameter: ARRAY [DOUBLE])")
				io.put_string ("%N")
			end

			create vals.make_new_unshared (parameter.upper - parameter.lower + 1)

			from
				i := parameter.lower
			variant
				parameter.upper - i + 1
			until
				i > parameter.upper
			loop
				vals.put (real_64_truncated_to_real_32 (parameter.item (i)),
														i - parameter.lower)
				i := i + 1
			end
			gl_material_imp (light_number, pname, vals.array_address)
		ensure then
			gl_error_free: get_error = false
		end

	set_light_position (light_number: INTEGER; position: POINT4) is
		do
			debug ("graphics")
				io.put_string ("set_light_position (")
				io.put_integer (light_number)
				io.put_string (", ")
				io.put_string (position.out)
				io.put_string (")")
				io.put_string ("%N")
			end

			cached_array.put (real_64_truncated_to_real_32 (position.x), 0)
			cached_array.put (real_64_truncated_to_real_32 (position.y), 1)
			cached_array.put (real_64_truncated_to_real_32 (position.z), 2)
			cached_array.put (real_64_truncated_to_real_32 (position.w), 3)

			gl_light_imp (light_number + rd_light_0, rd_position, cached_array.array_address)
		ensure then
			gl_error_free: get_error = false
		end

	set_light_direction (light_number: INTEGER; direction: VECTOR4) is
		do
			debug ("graphics")
				io.put_string ("set_light_direction (")
				io.put_integer (light_number)
				io.put_string (", ")
				io.put_string (direction.out)
				io.put_string (")")
				io.put_string ("%N")
			end

			cached_array.put (real_64_truncated_to_real_32 (direction.x), 0)
			cached_array.put (real_64_truncated_to_real_32 (direction.y), 1)
			cached_array.put (real_64_truncated_to_real_32 (direction.z), 2)

			gl_light_imp (light_number + rd_light_0, rd_spot_direction, cached_array.array_address)
		ensure then
			gl_error_free: get_error = false
		end

	set_diffuse_light_color (light_number: INTEGER; new_color: RGBA_COLOR) is
		do
			debug ("graphics")
				io.put_string ("set_diffuse_light_color (")
				io.put_integer (light_number)
				io.put_string (", ")
				io.put_string (new_color.out)
				io.put_string (")")
				io.put_string ("%N")
			end

			cached_array.put (real_64_truncated_to_real_32 (new_color.red), 0)
			cached_array.put (real_64_truncated_to_real_32 (new_color.green), 1)
			cached_array.put (real_64_truncated_to_real_32 (new_color.blue), 2)
			cached_array.put (real_64_truncated_to_real_32 (new_color.alpha), 3)

			gl_light_imp (light_number + rd_light_0, rd_diffuse, cached_array.array_address)
		ensure then
			gl_error_free: get_error = false
		end

	set_ambient_light_color (light_number: INTEGER; new_color: RGBA_COLOR) is
		do
			debug ("graphics")
				io.put_string ("set_ambient_light_color (")
				io.put_integer (light_number)
				io.put_string (", ")
				io.put_string (new_color.out)
				io.put_string (")")
				io.put_string ("%N")
			end

			cached_array.put (real_64_truncated_to_real_32 (new_color.red), 0)
			cached_array.put (real_64_truncated_to_real_32 (new_color.green), 1)
			cached_array.put (real_64_truncated_to_real_32 (new_color.blue), 2)
			cached_array.put (real_64_truncated_to_real_32 (new_color.alpha), 3)

			gl_light_imp (light_number + rd_light_0, rd_ambient, cached_array.array_address)
		ensure then
			gl_error_free: get_error = false
		end

	set_specular_light_color (light_number: INTEGER; new_color: RGBA_COLOR) is
		do
			debug ("graphics")
				io.put_string ("set_specular_light_color (")
				io.put_integer (light_number)
				io.put_string (", ")
				io.put_string (new_color.out)
				io.put_string (")")
				io.put_string ("%N")
			end

			cached_array.put (real_64_truncated_to_real_32 (new_color.red), 0)
			cached_array.put (real_64_truncated_to_real_32 (new_color.green), 1)
			cached_array.put (real_64_truncated_to_real_32 (new_color.blue), 2)
			cached_array.put (real_64_truncated_to_real_32 (new_color.alpha), 3)

			gl_light_imp (light_number + rd_light_0, rd_specular, cached_array.array_address)
		ensure then
			gl_error_free: get_error = false
		end

	set_spotlight_exponent (light_number: INTEGER; exponent: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("set_spotlight_exponent (")
				io.put_integer (light_number)
				io.put_string (", ")
				io.put_double (exponent)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_lightf (light_number + rd_light_0, rd_spot_exponent,
					   real_64_truncated_to_real_32 (exponent))
		ensure then
			gl_error_free: get_error = false
		end

	set_spotlight_cutoff (light_number: INTEGER; cutoff: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("set_spotlight_cutoff (")
				io.put_integer (light_number)
				io.put_string (", ")
				io.put_double (cutoff)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_lightf (light_number + rd_light_0, rd_spot_cutoff,
					   real_64_truncated_to_real_32 (cutoff))
		ensure then
			gl_error_free: get_error = false
		end

	set_constant_attenuation (light_number: INTEGER; factor: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("set_constant_attenuation (")
				io.put_integer (light_number)
				io.put_string (", ")
				io.put_double (factor)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_lightf (light_number + rd_light_0, rd_constant_attenuation,
					   real_64_truncated_to_real_32 (factor))
		ensure then
			gl_error_free: get_error = false
		end

	set_linear_attenuation (light_number: INTEGER; factor: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("set_linear_attenuation (")
				io.put_integer (light_number)
				io.put_string (", ")
				io.put_double (factor)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_lightf (light_number + rd_light_0, rd_linear_attenuation,
					   real_64_truncated_to_real_32 (factor))
		ensure then
			gl_error_free: get_error = false
		end

	set_quadratic_attenuation (light_number: INTEGER; factor: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("set_quadratic_attenuation (")
				io.put_integer (light_number)
				io.put_string (", ")
				io.put_double (factor)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_lightf (light_number + rd_light_0, rd_quadratic_attenuation,
					   real_64_truncated_to_real_32 (factor))
		ensure then
			gl_error_free: get_error = false
		end

	light (light_number, pname: INTEGER; parameter: VECTOR) is
		local
			vals: EWG_REAL_ARRAY
			i: INTEGER
		do
			debug ("graphics")
				io.put_string ("light (light_number, pname: INTEGER; parameter: VECTOR)")
				io.put_integer (light_number)
				io.put_string (", ")
				io.put_integer (pname)
				io.put_string (", ")
				io.put_string (parameter.out)
				io.put_string (")")
				io.put_string ("%N")
			end

			create vals.make_new_shared (parameter.count)

			from
				i := 1
			variant
				parameter.count - i + 1
			until
				i > parameter.count
			loop
				vals.put (real_64_truncated_to_real_32 (parameter.item (i)), i - 1)
				i := i + 1
			end

			gl_light_imp (light_number + rd_light_0, pname, vals.array_address)
		ensure then
			gl_error_free: get_error = false
		end

	assign_light_index: INTEGER is
			-- Generate an index for use with `light'. Use
			-- `enable_light' to active the index
		local
			i: INTEGER
		do
			debug ("graphics")
				io.put_string ("assign_light_index: INTEGER")
				io.put_string ("%N")
			end

			from
				i := 0
			invariant
				i_at_least_one: i >= 0
				i_small_enough: i <= 0 + maximum_number_of_lights
			variant
				i_increasing: 0 + maximum_number_of_lights - i
			until
				not light_array.item (i) or i >= maximum_number_of_lights
			loop
				i := i + 1
				Result := i
			end

			if Result = maximum_number_of_lights then
				Result := -1
			else
				light_array.put (true, Result)
			end
		ensure then
			light_set: Result /= -1 implies light_array.item (Result)
		end

	enable_light (light_index: INTEGER) is
			-- Turn on the light at `light_index'
		do
			debug ("graphics")
				io.put_string ("enable_light (")
				io.put_integer (light_index)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_enable (light_index + rd_light_0)
		ensure then
			gl_error_free: get_error = false
		end

	disable_light (light_index: INTEGER) is
			-- Turn off the light at `light_index'
		do
			debug ("graphics")
				io.put_string ("disable_light (")
				io.put_integer (light_index)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_disable (light_index + rd_light_0)
			light_array.put (false, light_index)
		ensure then
			gl_error_free: get_error = false
		end

	maximum_number_of_lights: INTEGER is
			-- The maximum number of active lights allowed by the renderer
		once
			gl_get_integer (gl_max_lights, $tmp_integer)
			Result := tmp_integer
		ensure then
			gl_error_free: get_error = false
		end

	maximum_evaluation_order: INTEGER is
			-- The maximum order for evaluation of points in a 'map'
		once
			gl_get_integer (gl_max_eval_order, $tmp_integer)
			Result := tmp_integer
		ensure then
			gl_error_free: get_error = false
		end

	shade_model (mode: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("maximum_number_of_lights: INTEGER")
				io.put_string ("%N")
			end

			gl_shade_model (mode)
		ensure then
			gl_error_free: get_error = false
		end

	enable_two_sided_lighting is
		do
			debug ("graphics")
				io.put_string ("enable_two_sided_lighting")
				io.put_string ("%N")
			end

			gl_light_model_i (rd_light_model_local_viewer, 1)
			gl_light_model_i (rd_light_model_two_side, 1)
		ensure then
			gl_error_free: get_error = false
		end

	enable_one_sided_lighting is
		do
			debug ("graphics")
				io.put_string ("enable_one_sided_lighting")
				io.put_string ("%N")
			end

			gl_light_model_i (rd_light_model_two_side, 0)
		ensure then
			gl_error_free: get_error = false
		end

	set_scene_ambient_color (new_color: RGBA_COLOR) is
		do
			debug ("graphics")
				io.put_string ("set_scene_ambient_color (")
				io.put_string (new_color.out)
				io.put_string (")")
				io.put_string ("%N")
			end

			cached_array.put (real_64_truncated_to_real_32 (new_color.red), 0)
			cached_array.put (real_64_truncated_to_real_32 (new_color.green), 1)
			cached_array.put (real_64_truncated_to_real_32 (new_color.blue), 2)
			cached_array.put (real_64_truncated_to_real_32 (new_color.alpha), 3)

			gl_light_model_f (rd_light_model_ambient, cached_array.array_address)
		ensure then
			gl_error_free: get_error = false
		end

	front_face (mode: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("front_face (")
				if mode = gl_cw then
					io.put_string ("gl_cw [")
				else
					io.put_string ("gl_ccw [")
				end
				io.put_integer (mode)
				io.put_string ("])")
				io.put_string ("%N")
			end

			gl_front_face (mode)
		ensure then
			gl_error_free: get_error = false
		end

	color_pointer (p: POINTER) is
		do
			debug ("graphics")
				io.put_string ("color_pointer (p: POINTER)")
				io.put_string ("%N")
			end

			gl_color_pointer (4, gl_double, 0, p)
		ensure then
			gl_error_free: get_error = false
		end

feature -- Drawing

	begin_query (a1: INTEGER; a2: INTEGER) is
		do
		end

	bind_buffer (a1: INTEGER; a2: INTEGER) is
		do
		end

	buffer_data (a1: INTEGER; a2: POINTER; a3: POINTER; a4: INTEGER) is
		do
		end

	buffer_sub_data (a1: INTEGER; a2: POINTER; a3: POINTER; a4: POINTER) is
		do
		end

	delete_buffers (a1: INTEGER; a2: POINTER) is
		do
		end

	delete_queries (a1: INTEGER; a2: POINTER) is
		do
		end

	end_query (a1: INTEGER) is
		do
		end

	generate_buffers (a1: INTEGER): INTEGER is
		do
		end

	gen_queries (a1: INTEGER; a2: POINTER) is
		do
		end

	is_buffer (a1: INTEGER): INTEGER is
		do
		end

	is_query (a1: INTEGER): INTEGER is
		do
		end

	map_buffer (a1: INTEGER; a2: INTEGER): POINTER is
		do
		end

	unmap_buffer (a1: INTEGER): INTEGER is
		do
		end

feature -- Clipping

	clip_plane (which_plane: INTEGER; equation: EWG_DOUBLE_ARRAY) is
			-- Interface to the glClipPlane
		do
			debug ("graphics")
				io.put_string ("clip_plane (which_plane: INTEGER; equation: ARRAY [DOUBLE])")
				io.put_string ("%N")
			end

			gl_clip_plane_imp (which_plane, equation.array_address)

		ensure then
			gl_error_free: get_error = false
		end

feature -- Rasterization

	raster_position_2 (x, y: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("raster_position_2 (x, y: DOUBLE)")
				io.put_string ("%N")
			end

			gl_raster_position_2 (x, y)
		ensure then
			gl_error_free: get_error = false
		end

	raster_position_3 (x, y, z: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("raster_position_3 (x, y, z: DOUBLE)")
				io.put_string ("%N")
			end

			gl_raster_position_3 (x, y, z)
		ensure then
			gl_error_free: get_error = false
		end

	raster_position_4 (x, y, z, w: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("raster_position_4 (x, y, z, w: DOUBLE)")
				io.put_string ("%N")
			end

			gl_raster_position_4 (x, y, z, w)
		ensure then
			gl_error_free: get_error = false
		end

	point_size (size: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("point_size (size: DOUBLE)")
				io.put_string ("%N")
			end

			gl_point_size (real_64_truncated_to_real_32 (size))
		ensure then
			gl_error_free: get_error = false
		end

	line_width (width: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("line_width (width: DOUBLE)")
				io.put_string ("%N")
			end

			gl_line_width (real_64_truncated_to_real_32 (width))
		ensure then
			gl_error_free: get_error = false
		end

	polygon_mode (face, mode: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("polygon_mode (face, mode: INTEGER)")
				io.put_string ("%N")
			end

			gl_polygon_mode (face, mode)
		ensure then
			gl_error_free: get_error = false
		end

feature -- Pixel Operations

	read_buffer (mode: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("read_buffer (mode: INTEGER)")
				io.put_string ("%N")
			end

			gl_read_buffer (mode)
		ensure then
			gl_error_free: get_error = false
		end

	pixel_store (pname, param: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("pixel_store (pname, param: INTEGER)")
				io.put_string ("%N")
			end

			gl_pixel_store (pname, param)
		ensure then
			gl_error_free: get_error = false
		end

	pixel_transfer (pname:INTEGER; param: REAL_32) is
		do
			debug ("graphics")
				io.put_string ("pixel_transfer (pname: INTEGER; param: REAL_32)")
				io.put_string ("%N")
			end

			gl_pixel_transfer (pname, param)
		ensure then
			gl_error_free: get_error = false
		end

	draw_pixels (width, height, mode: INTEGER; pixels: POINTER) is
		do
			debug ("graphics")
				io.put_string ("draw_pixels (width, height, mode: INTEGER; pixels: POINTER")
				io.put_string ("%N")
			end

			gl_draw_pixels (width, height, mode, gl_float, pixels)
		ensure then
			gl_error_free: get_error = false
		end
feature -- Texture Mapping

	texture_coord_1 (s: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("texture_coord_1 (s: DOUBLE)")
				io.put_string ("%N")
			end

			gl_texture_coord_1 (s)
		ensure then
			gl_error_free: get_error = false
		end

	texture_coord_2 (s, t: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("texture_coord_2 (s, t: DOUBLE)")
				io.put_string ("%N")
			end

			gl_texture_coord_2 (s, t)
		ensure then
			gl_error_free: get_error = false
		end

	texture_coord_3 (s, t, r: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("texture_coord_3 (s, t, r: DOUBLE)")
				io.put_string ("%N")
			end

			gl_texture_coord_3 (s, t, r)
		ensure then
			gl_error_free: get_error = false
		end

	texture_coord_4 (s, t, r, q: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("texture_coord_4 (s, t, r, q: DOUBLE)")
				io.put_string ("%N")
			end

			gl_texture_coord_4 (s, t, r, q)
		ensure then
			gl_error_free: get_error = false
		end

	texture_env (target, pname, param: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("texture_env (target, pname, param: INTEGER)")
				io.put_string ("%N")
			end

			gl_texture_env (target, pname, param)
		ensure then
			gl_error_free: get_error = false
		end

	texture_gen (coord, pname: INTEGER; param: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("texture_gen (coord, pname: INTEGER; param: DOUBLE)")
				io.put_string ("%N")
			end

			gl_texture_gen (coord, pname, param)
		ensure then
			gl_error_free: get_error = false
		end

	texture_parameter (target, pname, param: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("texture_parameter (target, pname, param: INTEGER)")
				io.put_string ("%N")
			end

			gl_texture_parameter (target, pname, param)
		ensure then
			gl_error_free: get_error = false
		end

feature -- Fog
feature -- Frame Buffer Operations

	scissor (x, y, width, height: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("scissor (x, y, width, height: INTEGER)")
				io.put_string ("%N")
			end

			gl_scissor (x, y, width, height)
		ensure then
			gl_error_free: get_error = false
		end

	stencil_func (func, ref, mask: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("stencil_func (func, ref, mask: INTEGER)")
				io.put_string ("%N")
			end

			gl_stencil_func (func, ref, mask)
		ensure then
			gl_error_free: get_error = false
		end

	stencil_mask (mask: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("stencil_mask (mask: INTEGER)")
				io.put_string ("%N")
			end

			gl_stencil_mask (mask)
		ensure then
			gl_error_free: get_error = false
		end

	stencil_op (fail, zfail, zpass: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("stencil_op (fail, zfail, zpass: INTEGER)")
				io.put_string ("%N")
			end

			gl_stencil_op (fail, zfail, zpass)
		ensure then
			gl_error_free: get_error = false
		end

	clear (mask: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("clear (mask: INTEGER)")
				io.put_string ("%N")
			end

			gl_clear (mask)
		ensure then
			gl_error_free: get_error = false
		end

	clear_color (red, green, blue, alpha: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("clear_color (red, green, blue, alpha: DOUBLE)")
				io.put_string ("%N")
			end

			gl_clear_color (red, green, blue, alpha)
		ensure then
			gl_error_free: get_error = false
		end

	clear_depth (depth: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("clear_depth (depth: DOUBLE)")
				io.put_string ("%N")
			end

			gl_clear_depth (depth)
		ensure then
			gl_error_free: get_error = false
		end

	clear_index (index: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("clear_index (index: DOUBLE)")
				io.put_string ("%N")
			end

			gl_clear_index (index)
		ensure then
			gl_error_free: get_error = false
		end

	draw_buffer (mode: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("draw_buffer (mode: INTEGER)")
				io.put_string ("%N")
			end

			gl_draw_buffer (mode)
		ensure then
			gl_error_free: get_error = false
		end

	draw_arrays (mode, offset, count: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("draw_arrays (mode, offset, count: INTEGER)")
				io.put_string ("%N")
			end

			gl_draw_arrays (mode, offset, count)
		ensure then
			gl_error_free: get_error = false
		end

	draw_elements (mode, offset, count: INTEGER; indices: POINTER) is
		do
			debug ("graphics")
				io.put_string ("draw_elements (mode, offset, count: INTEGER; indices: POINTER)")
				io.put_string ("%N")
			end

			gl_draw_elements (mode, offset, count, indices)
		ensure then
			gl_error_free: get_error = false
		end

feature -- Evaluators
	
	map_1d (target: INTEGER; u1, u2: DOUBLE; stride, order: INTEGER; points: POINTER) is
		do
			debug ("graphics")
				io.put_string ("map_1d (")
				io.put_string (target.out)
				io.put_string (", ")
				io.put_string (u1.out)
				io.put_string (", ")
				io.put_string (u2.out)
				io.put_string (", ")
				io.put_string (stride.out)
				io.put_string (", ")
				io.put_string (order.out)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_map_1d (target, u1, u2, stride, order, points)
		ensure then
			gl_error_free: get_error = false
		end
	
	map_grid_1d (un: INTEGER; u1, u2: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("map_1d (")
				io.put_string (un.out)
				io.put_string (", ")
				io.put_string (u1.out)
				io.put_string (", ")
				io.put_string (u2.out)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_map_grid_1d (un, u1, u2)
		ensure then
			gl_error_free: get_error = false			
		end
	
	map_2d (target: INTEGER; u1, u2: DOUBLE; ustride, uorder: INTEGER;
	v1, v2: DOUBLE; vstride, vorder: INTEGER; points: POINTER) is
		do
				check
					uorder_less_than_max_eval_order: uorder < maximum_evaluation_order
					vorder_less_than_max_eval_order: vorder < maximum_evaluation_order
				end
			debug ("graphics")
				io.put_string ("map_2d (")
				io.put_string (target.out)
				io.put_string (", ")
				io.put_string (u1.out)
				io.put_string (", ")
				io.put_string (u2.out)
				io.put_string (", ")
				io.put_string (ustride.out)
				io.put_string (", ")
				io.put_string (uorder.out)
				io.put_string (", ")
				io.put_string (v1.out)
				io.put_string (", ")
				io.put_string (v2.out)
				io.put_string (", ")
				io.put_string (vstride.out)
				io.put_string (", ")
				io.put_string (vorder.out)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_map_2d (target, u1, u2, ustride, uorder, v1, v2,
			vstride, vorder, points)
		ensure then
			gl_error_free: get_error = false
		end

	map_grid_2d (un: INTEGER; u1, u2: DOUBLE; vn: INTEGER; v1, v2: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("map_1d (")
				io.put_string (un.out)
				io.put_string (", ")
				io.put_string (u1.out)
				io.put_string (", ")
				io.put_string (u2.out)
				io.put_string (", ")
				io.put_string (vn.out)
				io.put_string (", ")
				io.put_string (v1.out)
				io.put_string (", ")
				io.put_string (v2.out)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_map_grid_2d (un, u1, u2, vn, v1, v2)
		ensure then
			gl_error_free: get_error = false			
		end
	
	eval_mesh_1d (mode, i1, i2: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("eval_mesh_1d (")
				io.put_string (mode.out)
				io.put_string (", ")
				io.put_string (i1.out)
				io.put_string (", ")
				io.put_string (i2.out)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_eval_mesh_1d (mode, i1, i2)
		ensure then
			gl_error_free: get_error = false
		end
		
	eval_mesh_2d (mode, i1, i2, j1, j2: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("eval_mesh_2d (")
				io.put_string (mode.out)
				io.put_string (", ")
				io.put_string (i1.out)
				io.put_string (", ")
				io.put_string (i2.out)
				io.put_string (", ")
				io.put_string (j1.out)
				io.put_string (", ")
				io.put_string (j2.out)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_eval_mesh_2d (mode, i1, i2, j1, j2)
		ensure then
			gl_error_free: get_error = false
		end
		
feature -- Selection and Feedback

	render_mode (mode: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("render_mode (mode: INTEGER)")
				io.put_string ("%N")
			end

			gl_render_mode (mode)
		ensure then
			gl_error_free: get_error = false
		end

feature -- Display Lists

	generate_lists (count: INTEGER): INTEGER is
			-- Generate a contiguous set of empty display lists
		do
			debug ("graphics")
				io.put_string ("generate_lists (count: INTEGER)")
				io.put_string ("%N")
			end

			Result := gl_generate_lists (count)
		end

	delete_lists (count, base: INTEGER) is
			-- Delete a contiguous set of empty display lists
		do
			debug ("graphics")
				io.put_string ("delete_lists (count, base: INTEGER)")
				io.put_string ("%N")
			end

			gl_delete_lists (count, base)
		end

	list_base (offset: INTEGER) is
			-- Set the display list base for call_lists
		do
			debug ("graphics")
				io.put_string ("list_base (offset: INTEGER)")
				io.put_string ("%N")
			end

			gl_list_base (offset)
		end

	new_list (list, mode: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("new_list (list, mode: INTEGER)")
				io.put_string ("%N")
			end

			gl_new_list (list, mode)
		ensure then
			gl_error_free: get_error = false
		end

	end_list is
		do
			debug ("graphics")
				io.put_string ("end_list")
				io.put_string ("%N")
			end

			gl_end_list
		ensure then
			gl_error_free: get_error = false
		end

	call_list (list: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("call_list (list: INTEGER)")
				io.put_string ("%N")
			end

			gl_call_list (list)
		ensure then
			gl_error_free: get_error = false
		end

	call_lists (n, type: INTEGER; lists: POINTER) is
		do
			debug ("graphics")
				io.put_string ("call_lists (n, type: INTEGER; lists: POINTER)" )
				io.put_string ("%N")
			end

			gl_call_lists (n, type, lists)
		ensure then
			gl_error_free: get_error = false
		end

feature -- Modes and Execution

	disable (mode: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("gl_disable (")
				io.put_integer (mode)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_disable (mode)
		ensure then
			gl_error_free: get_error = false
		end

	enable (mode: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("gl_enable (")
				io.put_integer (mode)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_enable (mode)
		ensure then
			gl_error_free: get_error = false
		end

	enable_client_state (mode: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("gl_enable_client_state (")
				io.put_integer (mode)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_enable_client_state (mode)
		ensure then
			gl_error_free: get_error = false
		end

	disable_client_state (mode: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("gl_disable_client_state (")
				io.put_integer (mode)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_disable_client_state (mode)
		ensure then
			gl_error_free: get_error = false
		end

	finish is
		do
			debug ("graphics")
				io.put_string ("finish")
				io.put_string ("%N")
			end

			gl_finish
		ensure then
			gl_error_free: get_error = false
		end

	flush is
		do
			debug ("graphics")
				io.put_string ("flush")
				io.put_string ("%N")
			end

			gl_flush
		ensure then
			gl_error_free: get_error = false
		end

	load_font (font_ptr, offset, count, display_list_base: INTEGER) is
		do
			debug ("graphics")
				io.put_string ("load_font")
				io.put_string ("%N")
			end

			glx_use_x_font (font_ptr, offset, count, display_list_base)
		ensure then
			gl_error_free: get_error = false
		end

feature -- State Queries

	context_enabled: BOOLEAN is
			-- Is there a valid context for rendering
		local
			p: POINTER
		do
			if check_for_context_valid then
				p := glx_current_context
				Result := not (p = default_pointer)
			else
				Result := true
			end
		ensure then
			gl_error_free: get_error = false
		end

	get_error: BOOLEAN is
			-- Did OpenGL return an error?
		local
			code: INTEGER
		do
			if check_for_context_valid then
				error_code := gl_get_error
				Result := error_code /= 0
				if Result then
					from
						code := error_code
					until
						code = 0
					loop
						io.put_string ("Error Code = ")
						io.put_string (code.out)
						io.put_string ("%N")

						code := gl_get_error
					end
				end
			else
				Result := false
			end
		end

	push_attribute (mask: INTEGER) is
			-- Push the attribute stack
		do
			debug ("graphics")
				io.put_string ("push_attribute (mask: INTEGER)")
				io.put_string ("%N")
			end
			gl_push_attribute (mask)
		end

	pop_attribute is
			-- Pop the attribute stack
		do
			debug ("graphics")
				io.put_string ("pop_attribute")
				io.put_string ("%N")
			end
			gl_pop_attribute
		end

feature -- Implementation

	perspective (fovy, aspect, znear, zfar: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("gl_perspective (")
				io.put_double (fovy)
				io.put_string (", ")
				io.put_double (aspect)
				io.put_string ("," )
				io.put_double (znear)
				io.put_string ("," )
				io.put_double (zfar)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_perspective (fovy, aspect, znear, zfar)
		ensure then
			gl_error_free: get_error = false
		end

	look_at (eyex, eyey, eyez, centerx, centery, centerz, upx, upy, upz: DOUBLE) is
		do
			debug ("graphics")
				io.put_string ("look_at (")
				io.put_double (eyex)
				io.put_string (", ")
				io.put_double (eyey)
				io.put_string ("," )
				io.put_double (eyez)
				io.put_string ("," )
				io.put_double (centerx)
				io.put_string (", ")
				io.put_double (centery)
				io.put_string ("," )
				io.put_double (centerz)
				io.put_string ("," )
				io.put_double (upx)
				io.put_string (", ")
				io.put_double (upy)
				io.put_string ("," )
				io.put_double (upz)
				io.put_string (")")
				io.put_string ("%N")
			end

			gl_look_at (eyex, eyey, eyez, centerx, centery, centerz, upx, upy, upz)
		end


feature {NONE}-- Implementation

	number_of_slices, number_of_stacks: INTEGER
			-- The resolution for drawing primitives

	check_for_context_valid: BOOLEAN
			-- Only check for a valid context when this flag is set, i.e., not between a glBeing and a glEnd.

	light_array: ARRAY [BOOLEAN] is
			-- Create a temporary object
		once
			create Result.make (0, maximum_number_of_lights)
		ensure then
			gl_error_free: get_error = false
		end

	cached_array: EWG_REAL_ARRAY is
			-- Create a temporary object
		once
			create Result.make_new_unshared (4)
		ensure then
			result_not_void: Result /= Void
		end

feature {NONE}-- Implementation

	current_mode: INTEGER

	matrix_array: DOUBLE_ARRAYED_LIST
			-- Temporary array for matrix load and multiply

	tmp_integer: INTEGER

invariant

	matrix_array_not_void: matrix_array /= Void
	quadric_normals_in_range: ((quadric_normals = rd_none or
								quadric_normals = rd_flat or
								quadric_normals = rd_smooth))
	quadric_draw_style_is_valid:	((quadric_draw_style = rd_line or
									  quadric_draw_style = rd_point or
									  quadric_draw_style = rd_fill))
	quadric_orientation_in_range: ((quadric_orientation = rd_inside or
								quadric_orientation = rd_outside))

end -- class GL

