

indexing
	description: "Convert a shader in an AST to GLSL"
	names: "GLslang", "GLSL"
	representation: "fixed"
	access: "fixed"
	size: "fixed"
	contents: "STRING"
	patterns: "none"
	date: "$Date: 2006-06-23 10:34:45 -0400 (Fri, 23 Jun 2006) $"
	revision: "$Revision: 226 $"
	author: "Mark A. Bolstad"
	status: "unreviewed"




class 

	GLSL_OUTPUT
	
inherit

	SL_ACTION
		redefine
			default_create, pre_action, traversal_action, post_action
		end

creation
	default_create

feature -- Initialization
	
	default_create is
			-- Initialize the action
		do
			Precursor

			unroll_loops := true
			
			create hash_table.make (1000)
			create indentation_string.make (4)
		end

feature -- Access

	vertex_shader: STRING
	fragment_shader: STRING
			-- The GLSL shaders generated from traversing the AST

	user_variables: STRING
			-- Temporary string to hold user-defined variable 
			-- definitions

	unroll_loops: BOOLEAN
			-- Should we unroll loops (does the hardware support 
			-- looping)
	
	dag_node: DAG_NODE
	root_node: SHADER_PARSE_NODE
			-- Nodes for the root of the dag and AST for the shader 
			-- that we want to convert.

	glsl_normal_name: STRING is "normal_glsl"
			-- Name to be used inplace of the RSL normal, N
	
	glsl_eye_vector_name: STRING is "viewVec_glsl"
			-- Name to be used inplace of the RSL eye vector, I

	glsl_light_vector_name: STRING is "lightVec_glsl"
			-- Name to be used inplace of the RSL light vector, L

	glsl_half_vector_name: STRING is "halfVec_glsl"
			-- Name to be used inplace of the RSL half vector

	max_lights: INTEGER is 2
			-- The maximum number of lights to use in an unrolled loop
			
	indentation_string: STRING
			-- The string that has the correct indentation
	
feature -- Measurement
feature -- Comparison
feature -- Status report
	
	variable_defined (variable_name: STRING): BOOLEAN is
			-- Have we defined `variable_name' before
		require 
			variable_name_not_void: variable_name /= Void
		do
			Result := hash_table.has (variable_name) and then
			hash_table.item (variable_name)
		end 

feature -- Status setting
	
	push (node: PARSE_NODE) is
			-- Increase depth in tree
		do
		end 

	pop (node: PARSE_NODE) is
			-- Decrease depth in tree
		do
		end 
	
	indent is
			-- Increase depth in tree
		do
			indentation_string.append_string ("%T")
		end 

	dedent is
			-- Decrease depth in tree
		do
			indentation_string.remove_tail (1)
		end 
	
feature -- Cursor movement
feature -- Element change

	set_dag_node (dag: DAG_NODE) is
			-- Set the `dag_node' to `dag'
		require 
			dag_not_void: dag /= Void
		do
			dag_node := dag
		ensure 
			dag_node_set: dag_node = dag
		end

	set_root_node (node: SHADER_PARSE_NODE) is
			-- Set the `root_node' to `node'
		require 
			node_set: node /= Void
		do
			root_node := node
		ensure 
			root_node_set: root_node = node
		end 

	defining_variable (variable_name: STRING) is
			-- Indicate that the node is defining the variable `name'
		require 
			variable_name_not_void: variable_name /= Void
		do
			hash_table.force (true, variable_name)
		ensure 
			variable_defined: variable_defined (variable_name)
		end

	set_unrolling_of_loops (on_or_off: BOOLEAN) is
			-- Set `unroll_loops' to `on_or_off'
		do
			unroll_loops := on_or_off
		ensure 
			unroll_loops_set: unroll_loops = on_or_off
		end 

feature -- Removal
feature -- Resizing
feature -- Transformation
feature -- Conversion
feature -- Duplication
feature -- Basic operations

	begin_action is
			-- Begin the traversal of the DAG and shader. As opposed 
			-- to the other visitor actions, this object starts its 
			-- own traversal
		require 
			dag_not_void: dag_node /= Void
			root_node_not_void: root_node /= Void
		do
			create_vertex_shader
			create_fragment_shader
		ensure 
			vertex_shader_not_void: vertex_shader /= Void
			fragment_shader_not_void: fragment_shader /= Void
		end 

	pre_action (node: PARSE_NODE) is
			-- Apply this action to `node'
		do
			node.glsl_pre_action (Current)
		end 

	traversal_action (node: PARSE_NODE) is
			-- Apply this action during the traversal of children of `node'
		do
			node.glsl_traverse_action (Current)
		end 

	post_action (node: PARSE_NODE) is
			-- Apply this action after the traversal of children of `node'
		do
			node.glsl_post_action (Current)
		end
	
feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {NONE} -- Implementation

	hash_table: DS_HASH_TABLE [BOOLEAN, STRING]
			-- A hash table to store if a user-variable has already 
			-- been defined

	create_vertex_shader is
			-- Create a vertex shader from the shader
		do
			create vertex_shader.make (2048)
			vertex_shader.append_string ("// Vertex Shader created from:%
											 % ")
			vertex_shader.append_string (root_node.shader_name)
			vertex_shader.append_string ("%N%N")
			vertex_shader.append_string ("varying vec3 normal, ")
			vertex_shader.append_string (glsl_eye_vector_name)
			vertex_shader.append_string (";%N")
			
			vertex_shader.append_string ("%Nvoid main( void )%N{%N")
			vertex_shader.append_string ("%Tnormal = normalize(gl_NormalMatrix * gl_Normal);%N")
			
			vertex_shader.append_string ("%T")
			vertex_shader.append_string (glsl_eye_vector_name)
			vertex_shader.append_string (" = vec3( gl_ModelViewMatrix * gl_Vertex );%N")
			
			vertex_shader.append_string ("%Tgl_FrontColor = gl_Color;%N")
			vertex_shader.append_string ("%Tgl_BackColor = gl_Color;%N")
			vertex_shader.append_string ("%Tgl_TexCoord[0]	= gl_MultiTexCoord0;%N")
			vertex_shader.append_string ("%Tgl_Position = ftransform();%N")
			vertex_shader.append_string ("}%N")
		ensure 
			vertex_shader_not_void: vertex_shader /= Void
			vertex_shader_not_empty: vertex_shader.count > 30 -- made up number
		end 

	create_fragment_shader is
			-- Begin the traversal of the DAG and shader. As opposed 
			-- to the other visitor actions, this object starts its 
			-- own traversal
		require 
			dag_not_void: dag_node /= Void
			root_node_not_void: root_node /= Void
		do
			create fragment_shader.make (2048)
			fragment_shader.append_string ("// Fragment Shader created from:%
											 % ")
			fragment_shader.append_string (root_node.shader_name)
			fragment_shader.append_string ("%N%N")
			fragment_shader.append_string ("varying vec3 normal, ")
			fragment_shader.append_string (glsl_eye_vector_name)
			fragment_shader.append_string (";%N")
			fragment_shader.append_string ("%N")
			
			fragment_shader.append_string ("%N%N")
			fragment_shader.append_string (rsl_to_glsl_light_definitions)

			create user_variables.make (1000)
			
			dag_node.traverse (Current)

			fragment_shader.append_string ("%Nvoid main( void )%N{%N")
			fragment_shader.append_string (user_variables)
			fragment_shader.append_string ("%Tvec3 ")
			fragment_shader.append_string (glsl_normal_name)
			fragment_shader.append_string (";%N")
			-- Add Color and Opacity
			fragment_shader.append_string ("%Tvec4 Os;%N")
			fragment_shader.append_string ("%Tvec4 Oi;%N")
			fragment_shader.append_string ("%Tvec4 Cs;%N")
			fragment_shader.append_string ("%Tvec4 Ci;%N")
			fragment_shader.append_string ("%N")
			fragment_shader.append_string ("%TOs = vec4( 1.0, 1.0, 1.0, gl_Color[ 3 ] );%N")
			fragment_shader.append_string ("%TCs = gl_Color;%N%N")
			fragment_shader.append_string ("%T")
			fragment_shader.append_string (glsl_normal_name)
			fragment_shader.append_string (" = normalize( normal );%N")

			indent
			
			root_node.traverse (Current)
			
			fragment_shader.append_string ("%N%Tgl_FragColor = Ci;%N")
			fragment_shader.append_string ("}%N")
		ensure 
			fragment_shader_not_void: fragment_shader /= Void
			fragment_shader_not_empty: fragment_shader.count > 30 -- made up number
		end 

	rsl_to_glsl_light_definitions: STRING is
			-- Convert RSL lighting functions, ambient(), diffuse(), and 
			-- specular() to GLSL functions
		do
			Result := ambient_light_definition

			Result.append_string (diffuse_light_definition)
			Result.append_string (specular_light_definition)
		end

	ambient_light_definition: STRING is
			-- The string for RSL ambient function
		local
			i: INTEGER
		do
			create Result.make (100)
			Result.append_string ("vec4 ambient( void )%N")
			Result.append_string ("{%N")
			Result.append_string ("%Tvec4 amb;%N")
			Result.append_string ("%N")
			Result.append_string ("%Tamb = gl_LightModel.ambient;%N")
			if not unroll_loops then
				Result.append_string ("%Tfor( int _i = 0; _i < gl_MaxLights;%
									  % _i++ ) {%N")
				Result.append_string ("%T%Tamb += gl_LightSource[_i].ambient;%N")
				Result.append_string ("}%N")
			else
				from
					i := 0 
				invariant
					i_at_least_one: i >= 0
					i_small_enough: i <= 0 + max_lights
				variant
					i_increasing: 0 + max_lights - i
				until
					i >= max_lights
				loop
					Result.append_string ("%Tamb += gl_LightSource[")
					Result.append_string (i.out)
					Result.append_string ("].ambient;%N")
					i := i + 1
				end
			end
			Result.append_string ("%Treturn (amb);%N")
			Result.append_string ("}%N%N")
		end

	diffuse_light_definition: STRING is
			-- The string for RSL diffuse definition
		local
			i: INTEGER
		do
			create Result.make (100)
			
			Result.append_string ("vec4 diffuse( vec3 _Nf )%N")
			Result.append_string ("{%N")
			Result.append_string ("%Tvec3 diffuse = vec3(0.0);%N")
			Result.append_string ("%Tvec3 ")
			Result.append_string (glsl_light_vector_name)
			Result.append_string (";%N")
			Result.append_string ("%Tfloat NdotL, _dist;%N")
			Result.append_string ("	%N")
			
			if not unroll_loops then
				Result.append_string ("%Tfor( int _i = 0; _i < gl_MaxLights;%
									  % _i++ ) {%N")
				Result.append_string ("%T%T")
				Result.append_string (glsl_light_vector_name)
				Result.append_string (" = vec3(gl_LightSource[_i].position)-")
				Result.append_string (glsl_eye_vector_name)
				Result.append_string (";%N")
				
				Result.append_string ("%T_dist  = dot( ")
				Result.append_string (glsl_light_vector_name)
				Result.append_string (", ")
				Result.append_string (glsl_light_vector_name)
				Result.append_string (" );%N")
			
				Result.append_string ("%T")
				Result.append_string (glsl_light_vector_name)
				Result.append_string (" = normalize( ")
				Result.append_string (glsl_light_vector_name)
				Result.append_string (" );%N")

				Result.append_string ("%TNdotL = max(dot(_Nf, normalize( ")
				Result.append_string (glsl_light_vector_name)
				Result.append_string (" ) ), 0.0);%N")
				Result.append_string ("%Tdiffuse += NdotL * (gl_LightSource[_i].diffuse.rgb)/_dist;%N")
				Result.append_string ("}%N")
				Result.append_string ("%N")
			else
				from
					i := 0 
				invariant
					i_at_least_one: i >= 0
					i_small_enough: i <= 0 + max_lights
				variant
					i_increasing: 0 + max_lights - i
				until
					i >= max_lights
				loop
					Result.append_string (generate_diffuse_string (i))
					i := i + 1
				end
			end
			
			Result.append_string ("%Treturn (vec4(diffuse, 1));%N")
			Result.append_string ("}%N%N")
		end

	specular_light_definition: STRING is
			-- The string for RSL specular definition
		local
			i: INTEGER
		do
			create Result.make (100)

			Result.append_string ("vec4 specular( vec3 _Nf, vec3 _If, float _roughness )%N")
			Result.append_string ("{%N")
			Result.append_string ("%Tvec3 specular = 0;%N")
			Result.append_string ("%Tfloat NdotL, NdotHV;%N")
			
			Result.append_string ("%Tvec3 ")
			Result.append_string (glsl_light_vector_name)
			Result.append_string (", ")
			Result.append_string (glsl_half_vector_name)
			Result.append_string (";%N")
			
			Result.append_string ("	%N")
			if not unroll_loops then
				Result.append_string ("%Tfor( int _i = 0; _i < gl_MaxLights;%
									  % _i++ ) {%N")
				Result.append_string ("%T%T")
				Result.append_string (glsl_light_vector_name)
				Result.append_string (" = vec3(gl_LightSource[_i].position)-")
				Result.append_string (glsl_eye_vector_name)
				Result.append_string (";%N")
				
				Result.append_string ("%T")
				Result.append_string (glsl_light_vector_name)
				Result.append_string (" = normalize( ")
				Result.append_string (glsl_light_vector_name)
				Result.append_string (" );%N")

				Result.append_string ("%TNdotL = max(dot(_Nf, normalize( ")
				Result.append_string (glsl_light_vector_name)
				Result.append_string (" ) ), 0.0);%N")
				
				Result.append_string ("%T")
				Result.append_string (glsl_half_vector_name)
				Result.append_string (" = normalize( ")
				Result.append_string (glsl_light_vector_name)
				Result.append_string (" - ")
				Result.append_string (glsl_eye_vector_name)
				Result.append_string (");%N")
			
				Result.append_string ("%TNdotHV = max(dot(_Nf, normalize( ")
				Result.append_string (glsl_half_vector_name)
				Result.append_string (" )), 0.0);%N")
				Result.append_string ("%Tspecular += NdotL > 0.0 ?  2 * gl_LightSource[_i].specular * pow( NdotHV, (1.0/(_roughness/20.0)) ) : 0.0;%N")
				Result.append_string ("}%N")
			else
				from
					i := 0 
				invariant
					i_at_least_one: i >= 0
					i_small_enough: i <= 0 + max_lights
				variant
					i_increasing: 0 + max_lights - i
				until
					i >= max_lights
				loop
					Result.append_string (generate_specular_string (i))
					i := i + 1
				end
			end
			Result.append_string ("%N")
			Result.append_string ("%Treturn (vec4( specular, 1 ));%N")
			Result.append_string ("}%N%N")
		ensure
			Result_not_void: Result /= Void
		end

	generate_diffuse_string (index: INTEGER): STRING is
			-- Create the diffuse string for the fragment shader
		require 
			index_positive: index >= 0
		do
			create Result.make (150)
			Result.append_string ("%T")
			Result.append_string (glsl_light_vector_name)
			Result.append_string (" = vec3(gl_LightSource[")
			Result.append_string (index.out)
			Result.append_string ("].position)-")
			Result.append_string (glsl_eye_vector_name)
			Result.append_string (";%N")
			
			Result.append_string ("%T_dist  = dot( ")
			Result.append_string (glsl_light_vector_name)
			Result.append_string (", ")
			Result.append_string (glsl_light_vector_name)
			Result.append_string (" );%N")
			
			Result.append_string ("%T")
			Result.append_string (glsl_light_vector_name)
			Result.append_string (" = normalize( ")
			Result.append_string (glsl_light_vector_name)
			Result.append_string (" );%N")

			Result.append_string ("%TNdotL = max(dot(_Nf, normalize( ")
			Result.append_string (glsl_light_vector_name)
			Result.append_string (" ) ), 0.0);%N")
			Result.append_string ("%Tdiffuse += NdotL * (gl_LightSource[")
			Result.append_string (index.out)
			Result.append_string ("].diffuse.rgb)/_dist;%N")
			Result.append_string ("%N")
		ensure 
			Result_not_void: Result /= Void
			length_greater_than_100: Result.count > 100				 
		end

	
	generate_specular_string (index: INTEGER): STRING is
			-- Create the diffuse string for the fragment shader
		require 
			index_positive: index >= 0
		do
			create Result.make (150)
			Result.append_string ("%T")
			Result.append_string (glsl_light_vector_name)
			Result.append_string (" = vec3(gl_LightSource[")
			Result.append_string (index.out)
			Result.append_string ("].position)-")
			Result.append_string (glsl_eye_vector_name)
			Result.append_string (";%N")
			
			Result.append_string ("%T")
			Result.append_string (glsl_half_vector_name)
			Result.append_string (" = normalize( ")
			Result.append_string (glsl_light_vector_name)
			Result.append_string (" - ")
			Result.append_string (glsl_eye_vector_name)
			Result.append_string (");%N")
			
			Result.append_string ("%T")
			Result.append_string (glsl_light_vector_name)
			Result.append_string (" = normalize( ")
			Result.append_string (glsl_light_vector_name)
			Result.append_string (" );%N")

			Result.append_string ("%TNdotL = max(dot(_Nf, normalize( ")
			Result.append_string (glsl_light_vector_name)
			Result.append_string (" ) ), 0.0);%N")

			Result.append_string ("%TNdotHV = max(dot(_Nf, normalize( ")
			Result.append_string (glsl_half_vector_name)
			Result.append_string (" )), 0.0);%N")

			Result.append_string ("%Tspecular += NdotL > 0.0 ?  1.5 * gl_LightSource[")
			Result.append_string (index.out)
			Result.append_string ("].specular.rgb * pow( NdotHV, (1.0/(_roughness/5.0)) ) : 0.0;%N")
		ensure 
			Result_not_void: Result /= Void
			length_greater_than_100: Result.count > 100				 
		end

	
invariant
	hash_table_not_void: hash_table /= Void
	
end -- GLSL_OUTPUT

