

indexing
	description: "State element for shaders for OpenGL Shading Language"
	names: "shader", "shading", "OpenGL"
	representation: "fixed"
	access: "fixed"
	size: "fixed"
	contents: "INTEGER"
	patterns: "none"



class 

	SCENE_GLSL_SHADER_ELEMENT
	
inherit
	
	SCENE_STATE_ELEMENT
		redefine
			apply, print_features, default_create
		end

	SL_PARSER
		rename
			make as sl_parser_make
		redefine
			default_create
		end

	NUMERIC_CONVERSION
		redefine
			default_create
		end
	
creation
	default_create, make, make_from_shader
	
feature -- Initialization

	default_create is
			-- Initialize to double-sided
		do
			Precursor {SCENE_STATE_ELEMENT}
			Precursor {SL_PARSER}

			sl_parser_make
			
			shader_initialized := False
			
			create vertex_shaders.make (10)
			create fragment_shaders.make (10)
			create shader_table.make (100)
		ensure then
			vertex_shaders_not_void: vertex_shaders /= Void
			fragment_shaders_not_void: fragment_shaders /= Void
			shader_table_not_void: shader_table /= Void
		end

	make (new_vertex_shader, new_fragment_shader: STRING) is
			-- Initialize the shaders
		require
			at_least_one_not_void: new_vertex_shader /= Void or
								   new_fragment_shader /= Void
		do
			default_create
			add_vertex_shader (new_vertex_shader)
			add_fragment_shader (new_fragment_shader)
		end

	make_from_shader (new_shader: SCENE_SHADER_ELEMENT) is
			-- Initialize the shader from a shader element
		require
			new_shader_not_void: new_shader /= Void
			new_shader_has_searchpath: new_shader.parameters.has_tag ("shader")
		local
			start_index, index: INTEGER
			path: STRING
			shader_path: STRING
			shader_file: KL_TEXT_INPUT_FILE
			found_shader: BOOLEAN
			daggy: DAG_ACTION
			shader_node: SHADER_PARSE_NODE
			glsler: GLSL_OUTPUT
			gl_parameters: GLSL_PARAMETER_ACTION
			p: PAIR [SL_VARIABLE, REAL_ARRAYED_LIST]
			variable_item: SL_VARIABLE
		do
			default_create
			
			-- Find the search paths
			index := new_shader.parameters.index_of_tag ("shader")
			path := new_shader.parameters.item (index).second.first.first
				check
					path_not_void: path /= Void
					path_not_empty: path.count >= 1
				end

			-- Check where we can find the shader
			from
				start_index := 1
				found_shader := false
			until
				index > path.count	or found_shader
			loop
				index := path.substring_index (":", start_index)
				if index = 0 then
					index := path.count + 1
				end
				shader_path := path.substring (start_index, index - 1)
				shader_path.append ("/")
				shader_path.append (new_shader.shader)
				shader_path.append (".sl")
				create shader_file.make (shader_path)
				-- If the shader exists
				if shader_file.exists then
					-- and it's parsable
					if process_shader (shader_path) then
						-- and valid
						shader_node ?= root_node
						if shader_node /= Void then
							found_shader := true
							
							-- then create a DAG
							create daggy
							root_node.traverse (daggy)
							daggy.finalize
							daggy.build ("Ci")

							create glsler
							glsler.set_dag_node (daggy.root_node)
							glsler.set_root_node (shader_node)
							glsler.begin_action

							from
								daggy.parameters.start
							until
								daggy.parameters.after
							loop
								create gl_parameters
								
								variable_item := daggy.parameters.item_for_iteration
								variable_item.default_value.traverse (gl_parameters)
								create p.make (variable_item, gl_parameters.parameters)
								shader_table.force (p, variable_item.name)
								daggy.parameters.forth
							end

							override_default_parameters (new_shader)
							update_parameters
							
							io.put_string (glsler.vertex_shader)
							io.put_string ("%N%N")
							io.put_string (glsler.fragment_shader)
							io.put_string ("%N%N")
							
							vertex_shaders.force (glsler.vertex_shader.twin,
												  new_shader.shader)
							fragment_shaders.force (glsler.fragment_shader.twin,
													new_shader.shader)
						end
					end
				end
				start_index := index + 1
			end
		end

feature -- Access

	uniform_variables: DS_ARRAYED_LIST [STRING]
			-- Uniform variable names from the current program

	shader_initialized: BOOLEAN
			-- Has the shader been initialized
	
feature -- Measurement
	
	count: INTEGER_64 is
			-- The maximum number of shaders
		do
			if	vertex_shaders.count < fragment_shaders.count then
				Result := fragment_shaders.count
			else
				Result := vertex_shaders.count
			end
		end
	
feature -- Comparison
feature -- Status report
feature -- Comparison
feature -- Status report

	has_vertex_shader (the_shader: STRING): BOOLEAN is
			-- Is `the_shader' in`vertex_shaders'
		do
			Result := vertex_shaders.has (the_shader)
		end

	has_fragment_shader (the_shader: STRING): BOOLEAN is
			-- Is `the_shader' in`fragment_shaders'
		do
			Result := fragment_shaders.has (the_shader)
		end

	enabled: BOOLEAN
			-- Is the shader enabled?

	has_parameters: BOOLEAN is
		-- Have parameters for this shader been computed. This is 
		-- typically valid only through the `make_from_shader' 
		-- creation routine
		do
			Result := shader_table.count > 0
		end
	
feature -- Status setting

	enable_shader (on_or_off: BOOLEAN) is
			-- Set `enabled' to `on_or_off'
		do
			enabled := on_or_off
		ensure
			enabled_set: enabled = on_or_off
		end
	
feature -- Cursor movement
feature -- Element change
	
	add_vertex_shader (the_shader: STRING) is
			-- Add `the_shader' to `vertex_shaders'
		require 
			the_shader_is_valid: the_shader /= Void
		do
			read_shader (the_shader)
			vertex_shaders.force (shader_buffer.twin, the_shader)
		ensure
			vertex_shaders_increased: vertex_shaders.count = old vertex_shaders.count + 1
		end 

	add_fragment_shader (the_shader: STRING) is
			-- Add `the_shader' to `fragment_shaders'
		require 
			the_shader_is_valid: the_shader /= Void
		do
			read_shader (the_shader)
			fragment_shaders.force (shader_buffer.twin, the_shader)
		ensure
			fragment_shaders_increased: fragment_shaders.count = old fragment_shaders.count + 1
		end 

	set_parameter_value (renderer: RENDERER; parameter: STRING; values: REAL_ARRAYED_LIST) is
			-- Set `parameter' to `value'
		require
			shader_initialized: shader_initialized
			parameter_not_void: parameter /= Void
			has_parameter: uniform_variables.has (parameter)
			values_not_void: values /= Void
			renderer_not_void: renderer /= Void
		do
			renderer.set_uniform_parameters (current_shader_program, parameter, values)
		end
		
feature -- Removal

	remove_vertex_shader (the_shader: STRING) is
			-- Remove `the_shader' from `vertex_shaders'
		require 
			the_shader_is_valid: the_shader /= Void
			shader_has_the_shader: has_vertex_shader (the_shader)
		do
			vertex_shaders.remove (the_shader)
				check
					not_implemented: false
				end
		ensure
			vertex_shaders_decreased: vertex_shaders.count = old vertex_shaders.count - 1
		end 

	remove_fragment_shader (the_shader: STRING) is
			-- Remove `the_shader' from `fragment_shaders'
		require 
			the_shader_is_valid: the_shader /= Void
			shader_has_the_shader: has_fragment_shader (the_shader)
		do
			fragment_shaders.remove (the_shader)
				check
					not_implemented: false
				end
		ensure
			fragment_shaders_decreased: fragment_shaders.count = old fragment_shaders.count - 1
		end 
	
feature -- Resizing
feature -- Transformation
feature -- Conversion
feature -- Duplication
feature -- Basic operations
	
	initialize_shaders (renderer: RENDERER) is
			-- Initialize the shaders
		local
			tester: KL_EQUALITY_TESTER [STRING]
		do
				check
					context_enabled: renderer.context_enabled
					extensions_initialized: renderer.extensions_initialized
				end
			if renderer.has_shading_language then
				if not shader_initialized then
					begin_shader
					read_vertex_shaders (renderer)
					read_fragment_shaders (renderer)
					end_shader (renderer)
				end
				
				renderer.use_shader_program (current_shader_program)
				uniform_variables := renderer.get_uniform_variable_names (current_shader_program)
				create tester
				uniform_variables.set_equality_tester (tester)

				shader_initialized := true
			end
		end 

	create_parameters (renderer: RENDERER) is
			-- Create the parameters from internally computed values
		require 
			shader_initialized: shader_initialized
			has_parameters: has_parameters
			renderer_not_void: renderer /= Void
		local
			key: STRING
		do
			from
				shader_table.start
			until
				shader_table.after
			loop
				key := shader_table.key_for_iteration
				set_parameter_value (renderer, key, shader_table.item (key).second)
				shader_table.forth
			end
		end 

	apply (renderer: RENDERER) is
			-- apply the `renderer' to `Current'. Default action is no-op.
		do
			check
				context_enabled: renderer.context_enabled
			end
			
			if renderer.extensions_initialized then		
				if shader_initialized and enabled then
					renderer.use_shader_program (current_shader_program)
				else
					renderer.use_shader_program (0)
				end
			end
		end 

feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {NONE} -- Implementation
	
	shaders: INTEGER_ARRAYED_LIST
	vertex_shaders: DS_HASH_TABLE [STRING, STRING]
	fragment_shaders: DS_HASH_TABLE [STRING, STRING]
			-- Lists to hold the shaders
	
	shader_buffer: STRING
			-- Temporary buffer for holding shader code

	current_shader_program: INTEGER
	
	shader_table: DS_HASH_TABLE [PAIR [SL_VARIABLE, REAL_ARRAYED_LIST], STRING]
			-- Store the variable with it's default or overridden value

	override_default_parameters (new_shader: SCENE_SHADER_ELEMENT) is
			-- Update the default parameters with the values defined 
			-- in the shader element
		require 
			new_shader_not_void: new_shader /= Void
		local
			the_variable: SL_VARIABLE
			values: REAL_ARRAYED_LIST
			p_values: DOUBLE_ARRAYED_LIST
			name: STRING
		do
			from
				new_shader.parameters.start
			until
				new_shader.parameters.after
			loop
				name := new_shader.parameters.item_for_iteration.first
				if name.item (1) = '%"' then
					name.remove_head (1)
					name.remove_tail (1)					
				end
				if shader_table.has (name) then
					if new_shader.parameters.item_for_iteration.second.second /= Void then
						-- Overwrite the default values in the table 
						-- with the values specified in the scene graph
						the_variable := shader_table.item (name).first
						values := shader_table.item (name).second
						p_values := new_shader.parameters.item_for_iteration.second.second
							check
								p_values_count_less_than_type_count:
								p_values.count <= type_count (the_variable.type)
							end

						values.wipe_out
						from
							p_values.start
						until
							p_values.after
						loop
							values.force_last (real_64_truncated_to_real_32 (p_values.item_for_iteration))
							p_values.forth
						end
					end
				end
				new_shader.parameters.forth
			end
		end 

	update_parameters is
			-- Update the default parameters with the values defined 
			-- in the shader element
		do
			from
				shader_table.start
			until
				shader_table.after
			loop
				convert_parameter_to_correct_count (shader_table.key_for_iteration)
				shader_table.forth
			end
		end
	
	convert_parameter_to_correct_count (name: STRING) is
			-- Convert a parameter to the correct number of elements 
			-- for the type
		require 
			name_not_void: name /= Void
			name_in_table: shader_table.has (name)
		local
			the_variable: SL_VARIABLE
			values: REAL_ARRAYED_LIST
			p_value: REAL
			index: INTEGER
		do
			the_variable := shader_table.item (name).first
			values := shader_table.item (name).second
			if values.count < type_count (the_variable.type) then
				-- If the parameter count is smaller than 
				-- the number of elements in the type, 
				-- then the count should be 1.
					check
						values_count_is_one: values.count = 1
					end
				from
					index := 2
					p_value := values.item (1)
				invariant
					index_at_least_one: index >= 2
					index_small_enough: index <= 2 + type_count (the_variable.type)
				variant
					index_increasing: 2 + type_count (the_variable.type) - index
				until
					index > type_count (the_variable.type)
				loop
					values.force_last (real_64_to_real_32 (p_value))
					index := index + 1
				end
			end
			
			if the_variable.type = color_type then
				values.force_last (1)
			end
			
		ensure 
			count_correct: (shader_table.item (name).second.count =
							type_count (shader_table.item (name).first.type)) or
						   (shader_table.item (name).first.type = color_type and
							shader_table.item (name).second.count =
							type_count (shader_table.item (name).first.type)+1)
		end 

	print_features (action: PRINT_ACTION) is
			-- print the node's features
		do
			Precursor (action)
			
			action.io.put_string ("	 vertex shaders = ")
			from
				vertex_shaders.start
			until
				vertex_shaders.after
			loop
				action.io.put_string ("%N%T")
				action.io.put_string (vertex_shaders.key_for_iteration)
				vertex_shaders.forth
			end
			
			action.io.put_string ("%N	 fragment shaders = ")
			from
				fragment_shaders.start
			until
				fragment_shaders.after
			loop
				action.io.put_string ("%N%T")
				action.io.put_string (fragment_shaders.key_for_iteration)
				fragment_shaders.forth
			end
			
			action.io.put_new_line
		end 
	
	begin_shader is
			-- Start the build of a shader
		do
			shader_initialized := true
			create shaders.make (2)
		ensure 
			shaders_are_initialized: shaders /= Void
		end 

	end_shader (renderer: RENDERER) is
			-- Finalize the shader
		do
			current_shader_program := renderer.create_shader_program
			from
				shaders.start
			until
				shaders.after
			loop
				renderer.attach_shader_object (current_shader_program, shaders.item_for_iteration)
				shaders.forth
			end

			renderer.link_shader_program (current_shader_program)
		end
	
	read_vertex_shaders (renderer: RENDERER) is
			-- Reads a vertex shader
		local
			vertex_shader_buffer: STRING
			status: STRING
			vertex_shader: INTEGER
		do
			from
				vertex_shaders.start
			until
				vertex_shaders.after
			loop
				vertex_shader := renderer.create_vertex_shader
				vertex_shader_buffer := vertex_shaders.item (vertex_shaders.key_for_iteration)
				renderer.set_shader_source (vertex_shader, vertex_shader_buffer)
				renderer.compile_shader (vertex_shader)
				status := renderer.shader_status (vertex_shader)
				
				shaders.force_last (vertex_shader)
				
				io.put_string ("Vertex shader status:%N")
				io.put_string (status)
				io.put_string ("%N")
				
				vertex_shaders.forth
			end
			
		ensure 
			default: true
		end 

	read_fragment_shaders (renderer: RENDERER) is
			-- Reads a fragment shader
		local
			fragment_shader_buffer: STRING
			status: STRING
			fragment_shader: INTEGER
		do
			from
				fragment_shaders.start
			until
				fragment_shaders.after
			loop
				fragment_shader_buffer := fragment_shaders.item (fragment_shaders.key_for_iteration)
				fragment_shader := renderer.create_fragment_shader
				renderer.set_shader_source (fragment_shader, fragment_shader_buffer)
				renderer.compile_shader (fragment_shader)
				status := renderer.shader_status (fragment_shader)
				
				shaders.force_last (fragment_shader)
				
				io.put_string ("Fragment shader status:%N")
				io.put_string (status)
				io.put_string ("%N")
				
				fragment_shaders.forth
			end
		ensure 
			default: true
		end 

	read_shader (filename: STRING) is
			-- Read the shader in `filename' into a buffer
		require 
			filename_not_void: filename /= Void
		local
			input_file: KL_TEXT_INPUT_FILE
		do
			create input_file.make (filename)
			input_file.open_read
			create shader_buffer.make (1024)

			from
				input_file.read_string (1024)
				shader_buffer := input_file.last_string.twin
			until
				input_file.end_of_file
			loop
				input_file.read_string (1024)
				shader_buffer.append (input_file.last_string.twin)
			end
			input_file.close
		ensure 
			default: true
		end 

	process_shader (shader: STRING): BOOLEAN is
			-- Parse a RenderMan `shader'
		local
			tmp_file: STRING
			a_file: KL_TEXT_INPUT_FILE
		do
			Result := true
			tmp_file := preprocess_shader (shader)
			create a_file.make (tmp_file)
			a_file.open_read
			if a_file.is_open_read then
				reset
				set_input_buffer (new_file_buffer (a_file))
				parse
				if root_node = Void then
					Result := false
				else
					root_node.resolve_type (nil_type)
					if root_node.has_valid_type then
						Result := false
						io.put_string (root_node.error)
					end
				end
				a_file.close
			end
			a_file.delete
		end
	
	preprocess_shader (shader: STRING): STRING is
			-- Run the shader through a preprocessor
		local
			p: POINTER
			command_string: STRING
			cmd: KL_SHELL_COMMAND
		do
			p := c_tempnam (p, p)
			create Result.make_from_c (p)
--			p.memory_free
			
			create command_string.make_from_string ("cpp -DPI=3.141592654 -x c ")
			
			command_string.append_string (shader)
			
			command_string.append_string (" > ")
			command_string.append_string (Result)
			
			create cmd.make (command_string)
			cmd.execute
			
			io.put_string (command_string)
			io.put_string ("%N")
		ensure 
			Result_not_void: Result /= Void
		end 

	c_tempnam (d, n: POINTER): POINTER is
		external
			"C use <stdio.h>"
		alias
			"tempnam"
		end

invariant
	vertex_shaders_not_void: vertex_shaders /= Void
	fragment_shaders_not_void: fragment_shaders /= Void
	shader_table_not_void: shader_table /= Void
	
end -- SCENE_GLSL_SHADER_ELEMENT

