

indexing
	description: "A container for state elements"
	patterns: "none"
	note: "Detemination of containment is defined by class name comparison,%
		  % not by object type. Therefore, you are only allowed one %
		  %state element for each implemented class."
	status: "unreviewed"




class 

	SCENE_STATE_ELEMENT_TABLE
	
inherit
	
	MEMORY_REPORTER
		undefine
			default_create
		redefine
			detailed_memory_usage, default_create, is_equal, copy
		end

creation
	default_create, make_from_list

feature -- Initialization

	default_create is
			-- Initialize the table
		do
			invalidate
			
			if state_elements = Void then
				create state_elements.make (2)
			else
				state_elements.wipe_out
			end
		end

	make_from_list (elements: DS_ARRAYED_LIST [SCENE_STATE_ELEMENT]) is
			-- add items from `elements'. Clones the items
		require 
			elements_not_void: elements /= Void
		do
			default_create
			
			from
				elements.start
			until
				elements.after
			loop
				add_element (elements.item_for_iteration.twin)
				elements.forth
			end
		ensure
			all_items_added: element_count = elements.count
		end
	
feature -- Access
	
	element_item (s: STRING): SCENE_STATE_ELEMENT is
			-- Return the element corresponding to `s'
		require 
			string_not_void: s /= Void and then s.count > 6 -- SCENE_
--			valid_key: valid_key (s)
		do
			if valid_key (s) then
				Result := state_elements.item (s)				
			end
		end 

	has_element (v: SCENE_STATE_ELEMENT): BOOLEAN is
			-- Does `v' appear in array?
		do
			Result := state_elements.has (v.class_name)
		end
	
	item: SCENE_STATE_ELEMENT is
			-- Return the element at the cursor
		do
			Result := state_elements.item_for_iteration
		end 
	
feature -- Measurement
	
	element_count: INTEGER is
			-- The number of state_elements
		do
			Result := state_elements.count
		ensure 
			Result_is_positive: Result >= 0
		end
	
feature -- Comparison
	
	is_equal (other: like Current): BOOLEAN is
			-- Are we equal?
		do
			Result := state_elements.is_equal (other.state_elements)
		end
	
feature -- Status report
	
	valid_key (s: STRING): BOOLEAN is
			-- Is `s' a valid key?
		require 
			s_not_void: s /= Void
		do
			Result := state_elements.has (s)
		end 

	after: BOOLEAN is
			-- Is the cursor off the end?
		do
			Result := state_elements.after
		end
	
feature -- Status setting
feature -- Cursor movement
	
	start is
			-- Move to the beginning of the elements
		do
			state_elements.start
		end
	
	forth is
			-- Move to the	cursor one element
		do
			state_elements.forth
		end
	
feature -- Element change
	
	copy (other: like Current) is
			-- Make `Current' a copy of `other'
		do
			--old_copy (other)
			create state_elements.make (2)
			state_elements.copy (other.state_elements)
			
			invalidate
		end 

	add_element (element: SCENE_STATE_ELEMENT) is
			-- add `element' as a new element
		require 
			element_not_void: element /= Void
			not_already_a_element: not has_element (element)
		do
			state_elements.force (element, element.class_name)
			
			invalidate
		ensure 
			count_increased_by_one: element_count = old element_count + 1
			element_added: has_element (element)
		end
	
	replace_element (element: SCENE_STATE_ELEMENT) is
			-- replace the element with `element'. Will add `element' 
			-- if not already present.
		require 
			element_not_void: element /= Void
		do
			state_elements.force (element, element.class_name)
			
			invalidate
		ensure 
			element_added: has_element (element)
			count_unchanged: old has_element (element) implies element_count = old element_count
		end

	merge (new_elements: SCENE_STATE_ELEMENT_TABLE) is
			-- Update the current state by adding or replacing 
			-- elements from `new_elements' into `Current'
		require 
			new_elements_not_void: new_elements /= Void
		local
			se: SCENE_STATE_ELEMENT
		do
			from
				new_elements.start
			until
				new_elements.after
			loop
				se := new_elements.item
				if has_element (se) then
					-- If the elements are not the same, then update 
					-- with the new one
					if not se.is_equal (element_item (se.class_name)) then
						replace_element (se)
					end
				else
					add_element (se)
				end
				new_elements.forth
			end
			
			invalidate
		end 

feature -- Removal
	
	remove_element (element: SCENE_STATE_ELEMENT) is
			-- remove the element corresponding to `element'
		require 
			element_not_void: element /= Void
			has_element: has_element (element)
		do
			state_elements.remove (element.class_name)
			
			invalidate
		ensure 
			element_is_removed: not has_element (element)
		end
	
	remove_element_by_name (element_name: STRING) is
			-- remove the element corresponding to `element_name'
		require 
			string_not_void: element_name /= Void and then element_name.count > 6 -- SCENE_
			valid_key: valid_key (element_name)
		local
			element: SCENE_STATE_ELEMENT
		do
			element ?= element_item (element_name)
			check
				element_not_void: element /= Void
			end
			
			remove_element (element)
			
			invalidate
		end
	
	remove_all_elements is
			-- remove all elements
		do
			state_elements.wipe_out
			
			invalidate
		ensure 
			count_is_zero: element_count = 0
		end

feature -- Resizing
feature -- Transformation
feature -- Conversion
	
	action_print (action: PRINT_ACTION) is
			-- print the table
		do
			from
				state_elements.start
			until
				state_elements.after
			loop
				action.apply (state_elements.item_for_iteration)
				state_elements.forth
			end
		end 
	
feature -- Duplication
feature -- Basic operations
	
	apply (renderer: RENDERER) is
			-- apply the `renderer' to `Current'. 
		do
			from
				state_elements.start
			until
				state_elements.after
			loop
				state_elements.item_for_iteration.apply (renderer)
				state_elements.forth
			end
		end 

feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {SCENE_STATE_ELEMENT_TABLE} -- Implementation

	state_elements: DS_HASH_TABLE [SCENE_STATE_ELEMENT, STRING]
			-- The state elements
	
	
feature {NONE} -- Implementation

	detailed_memory_usage: INTEGER is
			-- Compute additional memory usage as required
		do
			from
				state_elements.start
			until
				state_elements.after
			loop
				Result := Result + state_elements.item_for_iteration.memory_usage
				state_elements.forth
			end
		end 

	invalidate is
			-- Elements have been changed, so invalidate state as 
			-- necessary
		do
		end
	
invariant
	
	state_elements_not_void: state_elements /= Void
	
end -- SCENE_STATE_ELEMENT_TABLE

