

indexing
	description: "Auxillary class for NURBS representing knot vectors"
	patterns: "none"
	date: "$Date: 2007-04-10 14:11:16 -0400 (Tue, 10 Apr 2007) $"
	revision: "$Revision: 288 $"
	status: "unreviewed"




class 

	KNOT_VECTOR
	
inherit
	
	BASIC_NUMERIC
		redefine
			default_create, out, copy, is_equal
		end

	NUMERIC_CONVERSION
		export
			{NONE} all
		undefine
			default_create, out, copy, is_equal
		end
	
creation
	default_create, make_from_array, make_from_arrayed_list
	
feature -- Initialization
	
	default_create is
			-- Initialize the knots
		do
			if knots = Void then
				create knots.make_from_array (<<0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0>>)
				create unique_knots.make (1)
			else
--				knots.make_from_array (<<0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0>>)
				unique_knots.wipe_out
			end
			
			order := 4
			
			insertion_count := 0
			
			update_unique
		end
	
	make_from_array (vals: ARRAY [DOUBLE]; new_order: INTEGER) is
			-- Create a new knot vector from `vals', normalizing if necessary
		require
			new_order_is_strictly_positive: new_order > 0
		do
			default_create
			
			create knots.make_from_array (vals)
			
			order := new_order
			
--			normalize
			
			update_unique
		ensure
			order_set: order = new_order
--			normalized: minimum = 0 and maximum = 1
		end
	
	make_from_arrayed_list (vals: DOUBLE_ARRAYED_LIST; new_order: INTEGER) is
			-- Create a new knot vector from `vals', normalizing if necessary
		require
			new_order_is_strictly_positive: new_order > 0
		do
			default_create
			
			knots := vals.twin
			
			order := new_order
			
--			normalize
			
			update_unique
		ensure
			order_set: order = new_order
--			normalized: minimum = 0 and maximum = 1
		end
	
feature -- Access
	
	item, infix "@" (i: INTEGER_64): DOUBLE is
			-- Knot at index `i'
		require
			i_in_range: i >= 1 and i <= count
		do
			Result := knots @ i
		end

	unique_item (interval: INTEGER_64): DOUBLE is
			-- Left knot value of parametric `interval'
		require
			interval_in_range: interval >= 1 and interval <= intervals
		do
			Result := knots @ (unique_knots @ interval)
		end
	
	insertion_count: INTEGER
			-- How many times was the knot inserted in `insert_knot'
	
	insertion_point: INTEGER
			-- What index was the knot inserted?
	
feature -- Measurement

	minimum: DOUBLE is
			-- The smallest knot value
		do
			Result := knots @ 1
		end
	
	maximum: DOUBLE is
			-- The largest knot value
		do
			Result := knots @  (knots.count)
		end
	
	order: INTEGER
			-- The `order' of the knot vector
	
feature -- Comparison

	is_equal (other: like Current): BOOLEAN is
			-- Is `other' attached to an object considered
			-- equal to current object?
		do
			Result := insertion_count = other.insertion_count and
				insertion_point = other.insertion_point and
				order = other.order and
				knots.is_equal (other.knots) and
				unique_knots.is_equal (other.unique_knots)
		end

feature -- Status report
	
	clamped: BOOLEAN is
			-- Is this knot vector nonperiodic, i.e., `clamped'
		do
			Result := multiplicity (1) = order and multiplicity (intervals) = order 
		end
	
	count: INTEGER is
			-- The number of knots
		do
			Result := knots.count.to_integer
		ensure
			Result_strictly_positive: Result > 0
		end
	
	intervals: INTEGER is
			-- The number of unique intervals in the knot sequence
		do
			Result := unique_knots.count.to_integer - 1
		ensure
			Result_strictly_positive: Result > 0
		end	
	
	multiplicity (interval: INTEGER_64): INTEGER is
			-- The multiplicity of `interval'
		require 
			interval_in_range: interval >= 1 and interval <= intervals
		do
			Result := integer_64_to_integer_32 (unique_knots @ (interval + 1) - unique_knots @ interval)
		ensure 
			Result_strictly_positive: Result >= 1
		end 

feature -- Status setting
feature -- Cursor movement
feature -- Element change
feature -- Removal
feature -- Resizing
feature -- Transformation
feature -- Conversion

	out: STRING is
			-- Yield a printable representation.
		do
			create Result.make (80)
			
			from
				knots.start
				Result.append_character (opening_out_delimiter)
			until
				knots.after
			loop
				Result.append_string (knots.item_for_iteration.out)
				
				knots.forth
				
				if not knots.after then
					Result.append_character (' ')
				end
			end
			
			Result.append_character (closing_out_delimiter)
		ensure then
			result_not_void: Result /= Void
		end

feature -- Duplication

	copy (other: like Current) is
			-- Update current object using fields of object attached
			-- to `other', so as to yield equal objects.
		do
			insertion_count := other.insertion_count
			insertion_point := other.insertion_point
			order := other.order
			knots.copy (other.knots)
			unique_knots.copy (other.unique_knots)
		end

feature -- Basic operations
	
	find_interval (val: DOUBLE): INTEGER is
			-- Find the interval [`Result', `Result' + 1] that contains`val'
		require 
			val_in_range: val >= minimum and val <= maximum
		local
			index, l, r: INTEGER
			flag: BOOLEAN
		do
			l := 1
			r := intervals
			
			from
			until
				flag
			loop
				index := (l + r) // 2
				
				if unique_item (index) <= val then
					if index = intervals or else unique_item (index + 1) > val then
						flag := true
					else
						l := index + 1
					end
				else
					r := index - 1
				end
			end
			
			Result := index
			
		ensure 
			Result_is_valid: Result >= 1 and Result <= count
			contains_val: Result < intervals implies (val >= unique_item (Result) and val <= unique_item (Result + 1))
		end 
	
	find_span (val: DOUBLE): INTEGER is
			-- Find the span [`Result', `Result' + 1] that contains`val'
		require 
			val_in_range: val >= minimum and val <= maximum
			is_clamped: clamped
		local
			mid, l, r: INTEGER
		do
			if val = maximum then
				Result := count - order
			elseif val = minimum then
				Result := order
			else
				l := order - 1
				r := count - order + 1
				mid := (l + r) // 2
				
				from
				until
					val >= knots @ mid and val < knots @ (mid + 1)
				loop
					if val < knots @ mid then
						r := mid
					else
						l := mid
					end
					
					mid := (l + r) // 2
				end
				
				Result := mid
			end
				
		ensure 
			Result_is_valid: Result >= 1 and Result <= count
			contains_val: val >= knots @ (Result) and val <= knots @ (Result + 1) 
		end 
	
	insert_knot (knot_val: DOUBLE; knot_multiplicity: INTEGER) is
			-- Insert `knot_val' into the vector `knot_multiplicity' times
		require 
			knot_val_in_range: knot_val >= minimum and knot_val <= maximum
			knot_multiplicity_strictly_positive: knot_multiplicity >= 1
		local
			i, p, s: INTEGER
			flag: BOOLEAN
		do
			insertion_count := knot_multiplicity
			insertion_point := count
			p := order - 1
			
			if knot_val < knots @ order or knot_val > knots @ (count - order + 1) then
				insertion_count := 0
			else
				from
					i := 1 
				invariant
					i_at_least_one: i >= 1
					i_small_enough: i <= 1 + count
				variant
					i_increasing: 1 + count - i
				until
					i > count or flag
				loop
					if knots @ i > knot_val then
						insertion_point := i - 1
						flag := true
					end
					
					i := i + 1
				end
				
				if knot_val <= knots @ insertion_point then
					s := 1
					
					from
						i := insertion_point
						flag := false
					invariant
						i_at_least_one: i >= 1
						i_small_enough: i <= insertion_point
					until
						i < 2 or flag
					loop
						if knots @ i <= knots @ (i - 1) then
							s := s + 1
						else
							flag := true
						end
						
						i := i - 1
					end
				else
					s := 0
				end
				
				if knot_multiplicity + s > order then
					insertion_count := order - s
					if insertion_count < 0 then
						insertion_count := 0
					end
				end
				
				from
					i := 1
					knots.go_i_th (insertion_point)
				invariant
					i_at_least_one: i >= 1
					i_small_enough: i <= 1 + insertion_count
				variant
					i_increasing: 1 + insertion_count - i
				until
					i > insertion_count
				loop
					knots.force_right (knot_val)
					i := i + 1
				end
				
				update_unique
			end
				
		ensure 
			knot_inserted: multiplicity (find_interval (knot_val)) >= 
						   knot_multiplicity
		end 
	
	refine (new_knots: DOUBLE_ARRAYED_LIST) is
			-- Insert `new_knots' into knot vector
		require 
			new_knots_not_void: new_knots /= Void
			new_knots_count_positive: new_knots.count >= 1
		local
			a, b, i, j, k: INTEGER
			degree: INTEGER
			tmp_knots: DOUBLE_ARRAYED_LIST
		do
			degree := order - 1
			
			a := find_span (new_knots.item (1))
			b := find_span (new_knots.item (new_knots.count))
			b := b + 1
			
			i := b + degree - 1
			k := b + degree + new_knots.count.to_integer - 1
			
			create tmp_knots.make (knots.count + new_knots.count)
			tmp_knots.fill
			from
				j := 1 
			invariant
				j_at_least_one: j >= 1
				j_small_enough: j <= 1 + a
			variant
				j_increasing: 1 + a - j
			until
				j > a
			loop
				tmp_knots.replace (knots @ j, j)
				j := j + 1
			end
			
			from
				j := b
			invariant
				j_at_least_one: j >= b
				j_small_enough: j <= b + knots.count
			variant
				j_increasing: b + knots.count - j
			until
				j > knots.count
			loop
				tmp_knots.replace (knots @ j, j + new_knots.count)
				j := j + 1
			end
			
			from
				j := new_knots.count.to_integer
			until
				j = 0
			loop
				from
				until
					not (new_knots.item (j) <= knots.item (i) and i >= a)
				loop
					tmp_knots.replace (knots @ i, k)
					i := i - 1
					k := k - 1
				end
				
				tmp_knots.replace (new_knots @ j, k)
				k := k - 1
				j := j - 1
			end
			
			knots.copy (tmp_knots)
			
			update_unique
		ensure 
			default: true
		end 

	normalize is
			-- Reset the knots to be in the range [0, 1]
		local
			min, scale: DOUBLE
		do
			from
				min := minimum
				scale := 1.0 / (maximum - min)
				knots.start
			until
				knots.after
			loop
				knots.replace_at ((knots.item_for_iteration - min) * scale)
				
				knots.forth
			end
		ensure 
			minimum_is_zero: minimum = 0
			maximum_is_one: maximum = 1
		end 
	
	trim (min_val, max_val: DOUBLE) is
			-- Trim the knot vector to `min' and `max'
		require
			min_val_in_range: min_val >= minimum and min_val <= maximum
			max_val_in_range: max_val >= minimum and max_val <= maximum
			ordered: min_val < max_val
		do
			if min_val > minimum then
				from
					knots.start
				until
					knots.item_for_iteration >= min_val
				loop
					knots.remove_at
				end
			end
			
			if max_val < maximum then
				-- Search for the first value greater than `max_val'
				from
					knots.start
				until
					knots.item_for_iteration > max_val
				loop
					knots.forth
				end
				
				-- Then remove items from here to the end
				from
				until
					knots.after
				loop
					knots.remove_at
				end
			end
			
			update_unique
		ensure 
			trimmed: minimum = min_val and maximum = max_val
		end 

feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {SCENE_NURBS, KNOT_VECTOR} -- Implementation
	
	knots: DOUBLE_ARRAYED_LIST
	unique_knots: INTEGER_64_ARRAYED_LIST
	
feature {NONE} -- Implementation
	
	
	opening_out_delimiter: CHARACTER is '<'
	closing_out_delimiter: CHARACTER is '>'
	
	update_unique is
			-- Update the `unique_knots' vector
		local
			i: INTEGER
		do
			from
				unique_knots.wipe_out
				unique_knots.force_last (1)
				i := 2
			invariant
				i_at_least_one: i >= 2
				i_small_enough: i <= 1 + knots.count
			variant
				i_increasing: 2 + knots.count - i
			until
				i > knots.count
			loop
				if knots @ i /= knots @ (i - 1) then
					unique_knots.force_last (i)
				end
				
				i := i + 1
			end
			
			unique_knots.force_last (knots.count + 1)
		end 

invariant
	
	minimum_less_than_maximum: minimum < maximum
	knots_not_void: knots /= Void
	unique_knots_not_void: unique_knots /= Void
	unique_knots_valid: (unique_knots.count > 0 implies
		(unique_knots.item (1) >= 1 and unique_knots.item (1) < knots.capacity and
			unique_knots.item (unique_knots.count) >= unique_knots.item (1) and
			unique_knots.item (unique_knots.count) <= knots.capacity))
	intervals_less_than_count: intervals <= count
	order_strictly_positive: order > 0
	clamped: clamped implies multiplicity (1) = order and multiplicity (intervals) = order
	
end -- KNOT_VECTOR

