

indexing
	description: "Abstract class for indexed geometry. All descendant classes are %
				% geometry that are constructed from verticies defined by indexing %
				% into the current coordinates"
	patterns: "none"
	date: "$Date: 2007-03-29 13:43:21 -0400 (Thu, 29 Mar 2007) $"
	revision: "$Revision: 286 $"
	status: "unreviewed"




deferred class 

	SCENE_INDEXED_GEOMETRY

inherit
	
	BASIC_NUMERIC
		export
			{ANY} Tolerance
		undefine
			copy, is_equal, default_create
		end
	
	SCENE_GEOMETRY
		rename
			insert_intersection_information as geometry_insert_intersection_information
		export
			{NONE} geometry_insert_intersection_information
		redefine
			default_create, create_color_element, print_features,
			invalidate, i_th_bound
		end

	SCENE_COORDINATES
		undefine
			copy, is_equal
		redefine
			default_create, make_with_coordinates, set_coordinates_and_indices, set_coordinates,
			add_indices
		end
	
feature -- Initialization

	default_create is
			-- Initialize the coordinate arrays
		do
			Precursor {SCENE_GEOMETRY}
			Precursor {SCENE_COORDINATES}
			
			create {SCENE_BOX_BOUND} extents
		end

	make_with_coordinates (new_coordinates: POINT3_ARRAYED_LIST; 
						   new_coordinate_indices: INTEGER_64_ARRAYED_LIST) is
			-- 
		do
			Precursor {SCENE_COORDINATES}  (new_coordinates, new_coordinate_indices)
			
			update_extents
		end

feature -- Access
	
	extents: SCENE_BOUND
			-- The `extents' or bounds of the volume.

	i_th_bound (index: INTEGER): AXIS_ALIGNED_BOX is
			-- Return a bound around the `index'th element
		local
			indices: INTEGER_64_ARRAYED_LIST
		do
			indices := coordinate_indices.item (index)
			
			create Result.make (coordinates.item (indices.item (1)),
								coordinates.item (indices.item (2)))
			from
				indices.start
				indices.forth
				indices.forth
			until
				indices.after
			loop
				Result.grow (coordinates.item (indices.item_for_iteration))
				indices.forth
			end		
		end 

feature -- Measurement
feature -- Comparison
feature -- Status report

	use_vertex_array: BOOLEAN
			-- Use vertex arrays for passing data to the renderer
	
feature -- Status setting

	vertex_array_usage (enabled: BOOLEAN) is
			-- Enable/disable the usage of vertex arrays
		do
			use_vertex_array := enabled
		ensure
			use_vertex_array_set: use_vertex_array = enabled
		end
	
feature -- Cursor movement
feature -- Element change
	
	compute_normals (how: INTEGER) is
			-- Computes normals associated with the object. 
			-- Destroys any existing normals
		require
			how_is_correct: how = per_object or how = per_element or 
							how = per_vertex
		do
			if how = per_object or how = per_element then
				compute_face_normals
			else
				compute_vertex_normals
			end
			
			normal_mapping := how
		ensure
			normals_in_state_table: get_indexed_normals /= Void
			each_object_has_a_normal: get_indexed_normals.count = count
									   
		end 

	compute_face_normals is
			-- Computes face normals associated with the object. 
			-- Destroys any existing normals
		local
			ine: SCENE_INDEXED_NORMAL_ELEMENT
			indices: INTEGER_64_ARRAYED_LIST
		do
			ine := get_indexed_normals
			if ine = Void then
				create ine
			else
				ine.remove_all_values
			end
			
			
			from
				coordinate_indices.start
			until
				coordinate_indices.after
			loop
				
				indices := coordinate_indices.item_for_iteration
				
				ine.add_item (face_normal (indices.item (1),
										   indices.item (2),
										   indices.item (3)))
								
				coordinate_indices.forth
				if not coordinate_indices.after then
					ine.create_strip
				end
			end
			
			ine.set_strip_index (1)
			replace_element (ine)
			
			update_display_lists
			
			set_normal_mapping (per_object)
		ensure
			count_correct: get_indexed_normals.count = count
		end
	
	compute_vertex_normals is
			-- Computes vertex normals associated with the object. 
			-- Destroys any existing normals
		do
			io.put_string ("Cannot compute vertex normals for triangles,%
						   % no connectivity information. Computing %
						   %face normals%N")
			compute_face_normals
		ensure
			count_correct: get_indexed_normals.count = mapping_count
		end
	
	set_coordinates_and_indices (new_coordinates: POINT3_ARRAYED_LIST; 
					 new_coordinate_indices: INTEGER_64_ARRAYED_LIST) is
			-- Update `coordinates' and `coordinate_indices' to 
			-- `new_coordinates' and `new_coordinate_indices'.
		do
			Precursor {SCENE_COORDINATES}  (new_coordinates, new_coordinate_indices)
			
			update_extents
			
			if valid_key ("SCENE_INDEXED_NORMAL_ELEMENT") then
				remove_element_by_name ("SCENE_INDEXED_NORMAL_ELEMENT")
				cached_indexed_normals := Void
			end
			
			update_display_lists
		end

	set_coordinates (new_coordinates: POINT3_ARRAYED_LIST) is
			-- Update `coordinates'	 to `new_coordinates'.
		do
			Precursor {SCENE_COORDINATES}  (new_coordinates)
			
			update_extents
			
			if valid_key ("SCENE_INDEXED_NORMAL_ELEMENT") then
				remove_element_by_name ("SCENE_INDEXED_NORMAL_ELEMENT")
				cached_indexed_normals := Void
			end
			
			update_display_lists
		end

	add_indices (new_coordinate_indices: INTEGER_64_ARRAYED_LIST) is
			-- Update `coordinate_indices' to `new_coordinate_indices'.
		do
			Precursor {SCENE_COORDINATES}  (new_coordinate_indices)
			
			if valid_key ("SCENE_INDEXED_NORMAL_ELEMENT") then
				remove_element_by_name ("SCENE_INDEXED_NORMAL_ELEMENT")
				cached_indexed_normals := Void
			end
			
			update_display_lists
		end

	update_extents is
			-- Compute new `extents' information
		local
			sbb: SCENE_BOX_BOUND
		do
			sbb ?= extents
			sbb.default_create
			sbb.expand_to_include_points (coordinates)
		end 
	
feature -- Removal
feature -- Resizing
feature -- Transformation
feature -- Conversion
feature -- Duplication
feature -- Basic operations
feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable

feature {SCENE_INDEXED_GEOMETRY} -- Implementation

	intersect_item (intersecter: INTERSECT_ACTION; index: INTEGER_64) is
			-- Test the intersection against the item at `index'
		require 
			intersecter_not_void: intersecter /= Void
			index_positive: index > 0			
		deferred
		end
	
feature {NONE} -- Implementation
	
	cached_indexed_colors: SCENE_INDEXED_COLOR_ELEMENT
	cached_indexed_normals: SCENE_INDEXED_NORMAL_ELEMENT
	
	invalidate is
			-- Elements have been changed, so invalidate state as 
			-- necessary
		do
			Precursor
			cached_indexed_colors := Void
			cached_indexed_normals := Void
		ensure then
			cache_cleared: cached_indexed_colors = Void and cached_indexed_normals = Void
		end
	
	get_indexed_colors: SCENE_INDEXED_COLOR_ELEMENT is
			-- Return the appropriate colors, possibly Void if 
			-- it doesn't exist
		do
			if cached_indexed_colors = Void then
				cached_indexed_colors ?= 
				element_item ("SCENE_INDEXED_COLOR_ELEMENT")
			end
			
			Result := cached_indexed_colors
		end 

	get_indexed_normals: SCENE_INDEXED_NORMAL_ELEMENT is
			-- Return the appropriate display list, possibly Void if 
			-- it doesn't exist
		do
			if cached_indexed_normals = Void then
				cached_indexed_normals ?= 
				element_item ("SCENE_INDEXED_NORMAL_ELEMENT")
			end
			
			Result := cached_indexed_normals
		end 

	face_normal (i1, i2, i3: INTEGER_64): NORMAL is
			-- Compute a normal to the three verticies `i1', `i2', `i3'
		require 
			all_positive: i1 > 0 and i2 > 0 and i3 > 0
			in_range: i1 <= coordinates.count and 
					  i2 <= coordinates.count and 
					  i3 <= coordinates.count
		local		
			v21, v32: VECTOR3
		do
			v21 := coordinates.item (i2) - coordinates.item (i1)
			v32 := coordinates.item (i3) - coordinates.item (i2)
			
			create Result
			
			Result.make_from_vector (v21.cross (v32))
			
		ensure 
			Result_not_void: Result /= Void
			Result_is_unit: Result.is_unit
		end 

	insert_intersection_information (action: INTERSECT_ACTION;
									 u, v, t: DOUBLE; 
									 ice: SCENE_INDEXED_COLOR_ELEMENT;
									 ine: SCENE_INDEXED_NORMAL_ELEMENT;
									 point: POINT3;
									 ray: RAY) is
			-- Insert `t', `point', and other computed information 
			-- into `intersection'
		require 
			t_positive: t >= 0
			point_not_void: point /= Void
			ray_not_void: ray /= Void
		deferred
		ensure 
			inserted: action.intersection.count = old action.intersection.count + 1
		end 

	create_color_element is
			-- Initialize common geometry features
		local
			sce: SCENE_COLOR_ELEMENT
		do
			create {SCENE_INDEXED_COLOR_ELEMENT} sce
			
			add_element (sce)
		end 
	
	print_features (action: PRINT_ACTION) is
			-- print the sphere's features
		do
			Precursor (action)
			
			action.io.put_string ("	 coordinate count = ")
			action.io.put_string (coordinates.count.out)
			action.io.put_new_line
			
			action.io.put_string ("	 geometry count = ")
			action.io.put_string (count.out)
			action.io.put_new_line
		end
	
	geometry_insert_intersection_information (action: INTERSECT_ACTION;
									 u, v, t: DOUBLE; point: POINT3;
									 ray: RAY) is
			-- Insert `t', `point', and other computed information 
			-- into `intersection'
		do
			check
				never_called: false
			end
		end

end -- SCENE_INDEXED_GEOMETRY

