

indexing
	description: "Class describing indexed individual triangles."
	patterns: "none"
	date: "$Date: 2007-03-29 13:43:21 -0400 (Thu, 29 Mar 2007) $"
	revision: "$Revision: 286 $"
	status: "unreviewed"




class 

	SCENE_INDEXED_TRIANGLES
	
inherit
	
	SCENE_INDEXED_GEOMETRY
		undefine
			i_th_bound, intersectable_intersect, intersect, default_create
		redefine
			compute_face_normals
		end
	
	TRIANGLE_INTERSECTOR
		undefine
			face_normal, compute_face_normals
		end

creation
	default_create, make_with_coordinates

feature -- Initialization
	
feature -- Access
feature -- Measurement
	
	count: INTEGER_64 is
			-- The number of items
		do
			Result := vertex_count // 3
		end
	

	vertex_count: INTEGER_64 is
			-- The number of verticies
		do
			from
				coordinate_indices.start
			until
				coordinate_indices.after
			loop
				Result := Result + coordinate_indices.item_for_iteration.count
				coordinate_indices.forth
			end
		ensure
			vertex_count_is_multiple_of_3: Result \\ 3 = 0
		end
	
	mapping_count: INTEGER_64 is
			-- The number of elements with respect to mapping colors, 
			-- normals, ...
		do
			inspect color_mapping
			when constant_color then
				Result := 1
			when per_object, per_element then
				Result := count
			when per_vertex then
				Result := vertex_count
			end
		end 

feature -- Comparison
feature -- Status report
	
	has_extents: BOOLEAN is
			-- Does this volume have extents?
		do
			Result := extents /= Void
		end		
	
feature -- Status setting

	set_intersection_status (status: BOOLEAN) is
			-- Set the intersection status
		do
			intersected := status
		end
	
feature -- Cursor movement
feature -- Element change

	compute_face_normals is
			-- Computes face normals associated with the object. 
			-- Destroys any existing normals
		local
			ine: SCENE_INDEXED_NORMAL_ELEMENT
			indices: INTEGER_64_ARRAYED_LIST
			i: INTEGER
			n: NORMAL
		do
			ine := get_indexed_normals
			if ine = Void then
				create ine
			else
				ine.remove_all_values
			end
			
			
			from
				coordinate_indices.start
			until
				coordinate_indices.after
			loop
				
				indices := coordinate_indices.item_for_iteration
				
				
				from
					i := 1 
				invariant
					i_at_least_one: i >= 1
					i_small_enough: i <= 1 + indices.count
				variant
					i_increasing: 1 + indices.count - i
				until
					i > indices.count
				loop
					n := face_normal (indices.item (i),
									  indices.item (i + 1),
									  indices.item (i + 2))
					
					ine.add_item (n)

					i := i + 3
				end								
				
				coordinate_indices.forth
				if not coordinate_indices.after then
					ine.create_strip
				end
			end
			
			ine.set_strip_index (1)
			replace_element (ine)
			
			update_display_lists
			
			normal_mapping := per_object
		ensure then
			count_correct: get_indexed_normals.count = count
		end
	
feature -- Removal
feature -- Resizing
feature -- Transformation
feature -- Conversion
feature -- Duplication
feature -- Basic operations
	
	evaluate (u, v: DOUBLE; points_only: BOOLEAN; result_data: SHADING_DATA) is
			-- What's the point at parametric coordinates (`u', `v')?
		do
			check
				implemented: false
			end
		end

	tesselate is
			-- Transform the object into a polygonal structure
		do
			check
				implemented: false
			end
		end

feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {NONE} -- Implementation

	check_color_count: BOOLEAN is
			-- Check if the mapping count and the count for each 
			-- strip match
		local
			ice: SCENE_INDEXED_COLOR_ELEMENT
		do
			ice := get_indexed_colors
			
			if color_mapping = constant_color then
				ice.strip_start
				if ice.count >= 1 then
					Result := true
				end
			else
				Result := ice.count >= mapping_count
			end
		end 

	check_normal_count: BOOLEAN is
			-- Check if the mapping count and the count for each 
			-- strip match
		local
			ine: SCENE_INDEXED_NORMAL_ELEMENT
			normal_count: INTEGER_64
		do
			ine := get_indexed_normals
			
			normal_count := ine.count
			
			if normal_mapping = per_object then
				Result := normal_count = count
			else
				Result := normal_count = vertex_count
			end
		end 

	draw_immediate (renderer: RENDERER) is
			-- Apply the renderer to this object
		local
			ice: SCENE_INDEXED_COLOR_ELEMENT
			ine: SCENE_INDEXED_NORMAL_ELEMENT
			indices: INTEGER_64_ARRAYED_LIST
			n_shared, c_shared: BOOLEAN
			i: INTEGER
		do
			ice := get_indexed_colors
			check
				color_element_not_void: ice /= Void
			end
			
			ine := get_indexed_normals
			if ine = Void then
				compute_face_normals
				ine := get_indexed_normals

			end
			
			check
				normal_element_not_void: ine /= Void
			end
				
			renderer.enable (renderer.rd_color_material_constant)			
			renderer.color_material (renderer.rd_front_and_back,
									 renderer.rd_ambient_and_diffuse)
			renderer.polygon_mode (renderer.rd_front_and_back, renderer.rd_fill)
			
			if is_left_handed then
				renderer.front_face (renderer.rd_ccw)
			else
				renderer.front_face (renderer.rd_cw)
			end

			if use_vertex_array then
				renderer.enable_client_state (renderer.rd_vertex_array)
				renderer.enable_client_state (renderer.rd_normal_array)
				renderer.vertex_pointer (coordinates.base_address)
				renderer.draw_elements (renderer.rd_triangles, 0, renderer.rd_unsigned_integer,
										coordinate_indices.item (1).base_address)
				renderer.disable_client_state (renderer.rd_normal_array)
				renderer.disable_client_state (renderer.rd_vertex_array)
			else
				-- loop over strips
				from
					coordinate_indices.start
					ice.strip_start
					ine.strip_start
				until
					coordinate_indices.after
				loop
					indices := coordinate_indices.item_for_iteration
					
					from
						indices.start
						c_shared := ice.shared_indices (indices)
						n_shared := ine.shared_indices (indices)
						ice.start
						ine.start
					until
						indices.after
					loop
						renderer.render_begin (renderer.rd_triangles)
						
						if normal_mapping = per_object then
							renderer.normal (ine.item.x, ine.item.y, ine.item.z) 
						end
						
						from
							i := 1 
						invariant
							i_at_least_one: i >= 1
							i_small_enough: i <= 4
						variant
							i_increasing: 4 - i
						until
							i > 3
						loop
							renderer.rgba_color (ice.item)
							if normal_mapping = per_vertex then
								renderer.normal (ine.item.x, ine.item.y, ine.item.z)
							end
							renderer.set_vertex (coordinates.item (indices.item_for_iteration))
							
							indices.forth
							if normal_mapping = per_vertex and not n_shared then
								ine.forth
							end
							if color_mapping = per_vertex and not c_shared then
								ice.forth
							end
							
							i := i + 1
						end
						
						if normal_mapping = per_object then
							ine.forth
						end
						
						if color_mapping = per_object or color_mapping = per_element then
							ice.forth
						end
						
						renderer.render_end
					end
					
					ine.strip_forth
					if not (color_mapping = constant_color) then
						ice.strip_forth
					end
					coordinate_indices.forth
				end
			end
			renderer.disable (renderer.rd_color_material_constant)			
			
			renderer.front_face (renderer.rd_cw)
		end
	
invariant
	
	color_count_correct: check_color_count
	normal_count_correct: (valid_key ("SCENE_INDEXED_NORMAL_ELEMENT")
						   implies
						   check_normal_count)
	vertex_count_is_multiple_of_3: vertex_count \\ 3 = 0
	
end -- SCENE_INDEXED_TRIANGLES

