

indexing
	description: "Bilinear and Bicubic Patch meshes"
	patterns: "none"
	status: "unreviewed"
	Date: "$Date: 2007-01-25 13:00:51 -0500 (Thu, 25 Jan 2007) $"
	Revision: "$Revision: 280 $"



class

	SCENE_PATCH_MESH

inherit

	SCENE_GEOMETRY
		redefine
			test_diceability, split
		end

creation
	make

feature -- Initialization

	make (nu: INTEGER; u_periodic: BOOLEAN;
		nv: INTEGER; v_periodic: BOOLEAN;
		points: POINT3_ARRAYED_LIST) is
			-- Create a patch
		require
			nu_positive: nu > 0
			nv_positive: nv > 0
			points_not_void: points /= Void
			points_count_not_empty: points.count > 0
		do
			default_create
			coordinates := points

			u_count := nu
			v_count := nv

			if u_periodic then
				patch_u_count := nu
			else
				patch_u_count := nu - 1
			end
			
			if v_periodic then
				patch_v_count := nv
			else
				patch_v_count := nv - 1
			end
			
			periodic_in_u := u_periodic
			periodic_in_v := v_periodic
			
			create {SCENE_BOX_BOUND} extents
			update_extents
		end

feature -- Access

	periodic_in_u, periodic_in_v: BOOLEAN
			-- Is the mesh periodic

	patch_u_count, patch_v_count: INTEGER
			-- The number of u and v patches
	
feature -- Measurement

	coordinates: POINT3_ARRAYED_LIST
			-- The control points of the patch

	count: INTEGER_64 is
			-- The number of items
		do
			Result := 1
		end

	extents: SCENE_BOUND
			-- The `extents' or bounds of the volume.

	mapping_count: INTEGER_64 is
			-- The number of elements with respect to mapping colors,
			-- normals, ...
		do
			inspect color_mapping
			when constant_color then
				Result := 1
			when per_object, per_element then
				Result := patch_u_count * patch_v_count
			when per_vertex then
				Result := coordinates.count
			end
		end

feature -- Comparison
feature -- Status report

	has_extents: BOOLEAN is
			-- Does this volume have extents?
		do
			Result := extents /= Void
		end

	is_linear: BOOLEAN is
			-- Is this patch linear (as opposed to cubic)?
		do
			Result := coordinates.count = 4
		end

feature -- Status setting

	test_diceability (the_renderer: MICROPOLYGON_RENDERER) is
			-- Is the resulting of dicing this object "small enough"?
		do
			diceable := false
		end
	
feature -- Cursor movement
feature -- Element change
feature -- Removal
feature -- Resizing
feature -- Transformation
feature -- Conversion
feature -- Duplication
feature -- Basic operations

	evaluate (u, v: DOUBLE; points_only: BOOLEAN; result_data: SHADING_DATA) is
			-- What's the point at parametric coordinates (`u', `v')?
		local
			left_pt, right_pt, pt: POINT3
			min_vec, max_vec, dpdu, dpdv: VECTOR3
			normal: NORMAL
			sce: SCENE_COLOR_ELEMENT
		do
			left_pt := coordinates.item (1) + (coordinates.item (3) - coordinates.item (1)).scale (v)
			right_pt := coordinates.item (2) + (coordinates.item (4) - coordinates.item (2)).scale (v)

			pt := result_data.point
			pt.copy (left_pt + (right_pt - left_pt).scale (u))
			result_data.set_two_sided (two_sided)

			if not points_only then
				dpdu := result_data.dpdu
				dpdv := result_data.dpdv
				min_vec := coordinates.item (2) - coordinates.item (1)
				max_vec := coordinates.item (4) - coordinates.item (3)
				dpdu.copy (min_vec + (max_vec - min_vec).scale (v))
				min_vec := coordinates.item (3) - coordinates.item (1)
				max_vec := coordinates.item (4) - coordinates.item (2)
				dpdv.copy (min_vec + (max_vec - min_vec).scale (u))

				normal := result_data.geometric_normal
				normal.make_from_vector (dpdu * dpdv)
				normal.apply (object_to_world)
				normal.normalize
				--result_data.set_geometric_normal (normal)

				sce := get_colors
					check
						color_element_not_void: sce /= Void
					end
				result_data.set_color (sce.i_th (1))
			end
			-- Transform to world space. Done here, since we need the
			-- point in object space to compute the normal
			pt.apply (object_to_world)
		end

	split is
			-- Split the object into smaller pieces
			-- This is the generic routine for when the object is 
			-- split into two pieces. Override if need more parts
		local
			i, j: INTEGER
			patch: SCENE_PATCH
			sce: SCENE_COLOR_ELEMENT
			sne: SCENE_NORMAL_ELEMENT
		do
			from
				i := 1 
			invariant
				i_at_least_one: i >= 1
				i_small_enough: i <= 1 + patch_u_count
			variant
				i_increasing: 1 + patch_u_count - i
			until
				i > patch_u_count
			loop
				from
					j := 1 
				invariant
					j_at_least_one: j >= 1
					j_small_enough: j <= 1 + patch_v_count
				variant
					j_increasing: 1 + patch_v_count - j
				until
					j > patch_v_count
				loop
					generate_subpatch (i, j)
					create patch.make (subpatch_points)

					sce ?= patch.element_item ("SCENE_COLOR_ELEMENT")
					if sce /= Void then
						sce.make_from_list (subpatch_colors)
					else
						create sce.make_from_list (subpatch_colors)
					end
					patch.replace_element (sce)
				
					sne ?= patch.element_item ("SCENE_NORMAL_ELEMENT")
					if sne /= Void then
						sne.make_from_list (subpatch_normals)
					else
						create sne.make_from_list (subpatch_normals)
					end
					patch.replace_element (sne)
					
					if j > 1 then
						patch.set_neighbor (split_items.last, below)
					end
					
					split_items.force_last (patch)

					if i > 1 then
						patch.set_neighbor (split_items.item (split_items.count - patch_v_count), left)
					end
					
					j := j + 1
				end
				
				i := i + 1
			end
		end
	
	tesselate is
			-- Transform the object into a polygonal structure
		do
				check
					implemented: false
				end
		end

feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {NONE} -- Implementation

	u_count, v_count: INTEGER
			-- The number of control points in u and v

	subpatch_points: POINT3_ARRAYED_LIST
	subpatch_normals: NORMAL_ARRAYED_LIST
	subpatch_colors: RGBA_COLOR_ARRAYED_LIST
			-- Data for busting the mesh to a subpatch

	generate_subpatch (i, j: INTEGER) is
			-- Generate the points, colors, and normals for the patch at 
			-- `i', `j'
		require 
			i_in_range: i >= 1 and i <= patch_u_count
			j_in_range: j >= 1 and j <= patch_v_count
		local
			index: INTEGER
			color_is_per_vertex: BOOLEAN
		do
			create subpatch_points.make (4)
			create subpatch_normals.make (4)
			create subpatch_colors.make (4)

			if get_colors.count = coordinates.count then
				color_is_per_vertex := true
			else
				color_is_per_vertex := false
				subpatch_colors.put_last (get_colors.i_th (1))
			end
			
			index := (j - 1) * u_count
			subpatch_points.put_last (coordinates.item (index + i))
			subpatch_normals.put_last (get_normals.i_th (index + i))
			if color_is_per_vertex then
				subpatch_colors.put_last (get_colors.i_th (index + i))
			end
			if i = u_count then
					check
						is_periodic_in_u: periodic_in_u
					end
					
					subpatch_points.put_last (coordinates.item (index))
					subpatch_normals.put_last (get_normals.i_th (index))
				if color_is_per_vertex then
					subpatch_colors.put_last (get_colors.i_th (index))
				end
			else
				subpatch_points.put_last (coordinates.item (index + i + 1))
				subpatch_normals.put_last (get_normals.i_th (index + i + 1))
				if color_is_per_vertex then
					subpatch_colors.put_last (get_colors.i_th (index + i + 1))
				end
			end

			-- Handle the periodic_in_v case
			if j = v_count then
						check
							is_periodic_in_v: periodic_in_v
						end
				index := 1
				subpatch_points.put_last (coordinates.item (index + i))
				subpatch_normals.put_last (get_normals.i_th (index + i))
				if color_is_per_vertex then
					subpatch_colors.put_last (get_colors.i_th (index + i))
				end
				if i = u_count then
						check
							is_periodic_in_u: periodic_in_u
						end
					
					subpatch_points.put_last (coordinates.item (index))
					subpatch_normals.put_last (get_normals.i_th (index))
					if color_is_per_vertex then
						subpatch_colors.put_last (get_colors.i_th (index))
					end
				else
					subpatch_points.put_last (coordinates.item (index + i + 1))
					subpatch_normals.put_last (get_normals.i_th (index + i + 1))
					if color_is_per_vertex then
						subpatch_colors.put_last (get_colors.i_th (index + i + 1))
					end
				end
			else
				index := j * u_count
				subpatch_points.put_last (coordinates.item (index + i))
				subpatch_normals.put_last (get_normals.i_th (index + i))
				if color_is_per_vertex then
					subpatch_colors.put_last (get_colors.i_th (index + i))
				end
				if i = u_count then
						check
							is_periodic_in_u: periodic_in_u
						end
					
					subpatch_points.put_last (coordinates.item (index))
					subpatch_normals.put_last (get_normals.i_th (index))
					if color_is_per_vertex then
						subpatch_colors.put_last (get_colors.i_th (index))
					end
				else
					subpatch_points.put_last (coordinates.item (index + i + 1))
					subpatch_normals.put_last (get_normals.i_th (index + i + 1))
					if color_is_per_vertex then
						subpatch_colors.put_last (get_colors.i_th (index + i + 1))
					end
				end
			end
		end 

	update_extents is
			-- Compute new `extents' information
		local
			sbb: SCENE_BOX_BOUND
			u_start, u_end, v_start, v_end: INTEGER
			pt1, pt2, pt3, pt4: INTEGER
		do
			create {SCENE_BOX_BOUND} extents
			sbb ?= extents
			sbb.default_create

			u_start := (u_minimum * u_count).truncated_to_integer + 1
			u_end := (u_maximum * u_count).ceiling
			v_start := (v_minimum * v_count).truncated_to_integer + 1
			v_end := (v_maximum * v_count).ceiling

			pt1 := (v_start - 1) * u_count + u_start
			pt2 := (v_start - 1) * u_count + u_end
			pt3 := (v_end - 1) * u_count + u_start
			pt4 := (v_end - 1) * u_count + u_end
			
			sbb.make (coordinates.item (pt1), coordinates.item (pt2))
			sbb.expand_to_include_point (coordinates.item (pt3))
			sbb.expand_to_include_point (coordinates.item (pt4))
		end

	draw_immediate (renderer: RENDERER) is
			-- Apply the renderer to this object.
		local
			sce: SCENE_COLOR_ELEMENT
			sne: SCENE_NORMAL_ELEMENT
			vals: POINT3_ARRAYED_LIST
			normals: NORMAL_ARRAYED_LIST
			i, j: INTEGER
			n, p: POINTER
		do

			if is_left_handed then
				renderer.front_face (renderer.rd_cw)
			else
				renderer.front_face (renderer.rd_ccw)
			end

			sce := get_colors
			check
				color_element_not_void: sce /= Void
			end

			sne := get_normals
			if sne = Void then
				--compute_normals
				sne := get_normals
			end

--			check
--				normal_element_not_void: sne /= Void
--			end

			renderer.color_material (renderer.rd_front_and_back,
							   renderer.rd_ambient_and_diffuse)
			renderer.enable (renderer.rd_color_material_constant)

			renderer.rgba_color (sce.i_th (1))
			renderer.enable (renderer.rd_map2_vertex_3)
			renderer.enable (renderer.rd_map2_normal)
			renderer.enable (renderer.rd_normalize)

			from
				i := 1
				create vals.make (4)
				vals.fill
				create normals.make (4)
				normals.fill
			invariant
				i_at_least_one: i >= 1
				i_small_enough: i <= 1 + v_count
			variant
				i_increasing: 1 + v_count - i
			until
				i >= v_count
			loop
				from
					j := 1
				invariant
					j_at_least_one: j >= 1
					j_small_enough: j <= 1 + u_count
				variant
					j_increasing: 1 + u_count - j
				until
					j >= u_count
				loop
					vals.replace (coordinates.item ((i - 1) * u_count + j), 1)
					normals.replace (sne.i_th ((i - 1) * u_count + j), 1)
					vals.replace (coordinates.item ((i - 1) * u_count + j + 1), 2)
					normals.replace (sne.i_th ((i - 1) * u_count + j + 1), 2)
					vals.replace (coordinates.item (i * u_count + j), 3)
					normals.replace (sne.i_th (i * u_count + j), 3)
					vals.replace (coordinates.item (i * u_count + j + 1), 4)
					normals.replace (sne.i_th (i * u_count + j + 1), 4)
					debug ("patch_mesh")
						io.put_string (((i - 1) * u_count + j).out)
						io.put_string (", ")
						io.put_string (((i - 1) * u_count + j + 1).out)
						io.put_string (", ")
						io.put_string ((i * u_count + j).out)
						io.put_string (", ")
						io.put_string ((i * u_count + j + 1).out)
						io.put_string ("%N")
					end
					p := vals.storage.item
					n := normals.storage.item
					renderer.map_2d (renderer.rd_map2_vertex_3, 0, 1, 3, 2, 0, 1, 6, 2, p)
					renderer.map_2d (renderer.rd_map2_normal, 0, 1, 3, 2, 0, 1, 6, 2, n)
					renderer.map_grid_2d (8, 0, 1, 8, 0, 1)
					renderer.eval_mesh_2d (renderer.rd_fill, 0, 8, 0, 8)
					
					j := j + 1
				end
				i := i + 1
			end

			-- Pick up the patches on the left edge
			if periodic_in_u then
				from
					i := 1
				invariant
					i_at_least_one: i >= 1
					i_small_enough: i <= 1 + v_count
				variant
					i_increasing: 1 + v_count - i
				until
					i >= v_count - 1
				loop
					vals.replace (coordinates.item (i * u_count), 1)
					normals.replace (sne.i_th (i * u_count), 1)
					vals.replace (coordinates.item ((i - 1) * u_count + 1), 2)
					normals.replace (sne.i_th ((i - 1) * u_count + 1), 2)
					vals.replace (coordinates.item ((i + 1) * u_count), 3)
					normals.replace (sne.i_th ((i + 1) * u_count), 3)
					vals.replace (coordinates.item (i * u_count + 1), 4)
					normals.replace (sne.i_th (i * u_count + 1), 4)
					
					p := vals.storage.item
					n := normals.storage.item
					renderer.map_2d (renderer.rd_map2_vertex_3, 0, 1, 3, 2, 0, 1, 6, 2, p)
					renderer.map_2d (renderer.rd_map2_normal, 0, 1, 3, 2, 0, 1, 6, 2, n)
					renderer.map_grid_2d (4, 0, 1, 4, 0, 1)
					renderer.eval_mesh_2d (renderer.rd_fill, 0, 4, 0, 4)

					i := i + 1
				end
			end

			if periodic_in_v then
			end
		end

	insert_intersection_information (action: INTERSECT_ACTION;
									 u, v, t: DOUBLE; point: POINT3;
									 ray: RAY) is
			-- Insert `t', `point', and other computed information
			-- into `intersection'
		do
				check
					implemented: false
				end
		end

	object_intersect (intersecter: INTERSECT_ACTION) is
			-- Intersect with `intersecter'.
			-- If `intersecter' does not intersect then `intersected' will be False
			-- else `intersected' will be True and `intersection' will yield
			-- the ordered parametric coordinates along `ray' of each
			-- successive point of intersection.
		do
				check
					implemented: false
				end
		end

end -- SCENE_PATCH

