

indexing
	description: "Abstract class for Quadric Primitives"
	patterns: "none"
	date: "$Date: 2007-02-09 16:45:54 -0500 (Fri, 09 Feb 2007) $"
	revision: "$Revision: 282 $"
	status: "unreviewed"




deferred class 

	SCENE_QUADRIC
	
inherit
	
	BASIC_NUMERIC
		export
			{ANY} Tolerance
		undefine
			copy, is_equal, default_create
		end
	
	SCENE_GEOMETRY
		export
			{NONE} set_color_mapping
		redefine
			default_create, print_features, add_attributes, process_u_split
		end
	
feature -- Initialization

	default_create is
			-- Create a full quadric of radius 1
		do
			Precursor {SCENE_GEOMETRY}
			
			radius := 1
			z_minimum := -1
			z_maximum := 1
			sweep_angle := 2 * pi
			
			if origin = Void then
				create origin
			else
				origin.default_create
			end
			
			compute_intermediates

			set_wrapping
		end

	make_by_parameters (r, zmin, zmax, theta: DOUBLE) is
			-- Create a quadric of radius `r', with bottom 
			-- `zmin', top	`zmax', and sweep angle `theta' 
			-- (closed quadric = 360)
		require
			r_positive: r > 0
		do
			default_create
			
			radius := r
			z_minimum := zmin
			z_maximum := zmax
			sweep_angle := theta * pi / 180.0
			
			compute_intermediates

			-- Set the wrap around
			set_wrapping
		ensure
			radius_set: radius = r
			z_minimum_set: z_minimum = zmin
			z_maximum_set: z_maximum = zmax
			sweep_angle_set: sweep_angle = theta * pi / 180.0			 
		end

feature -- Access
	
	extents: SCENE_BOUND
			-- The `extents' or bounds of the volume.
	
	radius: DOUBLE
			-- The radius of the sphere
	
	z_minimum, z_maximum: DOUBLE
			-- The top and bottom radius of the sphere. If these are 
			-- less than `radius' the sphere is flat-topped.
	
	sweep_angle: DOUBLE
			-- The angle a half-circle is swept through (2 pi radians 
			-- = closed sphere)
	
	origin: POINT3
			-- The center of the sphere
	
feature -- Measurement
	
	count: INTEGER_64 is 1
			-- There is only one sphere
	
	mapping_count: INTEGER_64 is 1
			-- There is only one element
	
feature -- Comparison
feature -- Status report

	has_extents: BOOLEAN is
			-- Does this volume have extents?
		do
			Result := extents /= Void
		end
	
	valid_radius (r: DOUBLE): BOOLEAN is
		-- Is `r' valid for `radius'?
		do
			Result := r > 0
		end
	
	valid_sweep_angle (new_sweep_angle: DOUBLE): BOOLEAN is
			-- Is `new_sweep_angle' valid
		do 
			Result := new_sweep_angle > 0.0 and new_sweep_angle <= 360.0
		end
	
	valid_sweep_angle_in_radians (new_sweep_angle: DOUBLE): BOOLEAN is
			-- Is `new_sweep_angle' valid
		do 
			Result := new_sweep_angle > 0.0 and new_sweep_angle <= 6.2831853
		end
	
	valid_z_minimum (z: DOUBLE): BOOLEAN is
		-- Is `z' valid for `z_minimum'?
		do
			Result := z <= z_maximum
		end
	
	valid_z_maximum (z: DOUBLE): BOOLEAN is
		-- Is `z' valid for `z_maximum'?
		do
			Result := z >= z_minimum
		end
	
feature -- Status setting
feature -- Cursor movement
feature -- Element change
	
	set_origin (new_origin: POINT3) is
			-- Set `origin' to `new_origin'
		require 
			new_origin_not_void: new_origin /= Void
		do
			origin.copy (new_origin)
			compute_intermediates		
		ensure 
			origin_set: origin.is_equal (new_origin)
		end 
	
	set_radius (new_radius: DOUBLE) is
			-- Set `radius' to `new_radius'
		require 
			new_radius_is_valid: new_radius > 0.0
		do
			radius := new_radius
			compute_intermediates		
		ensure 
			radius_set: radius = new_radius
		end 
	
	set_sweep_angle (new_sweep_angle: DOUBLE) is
			-- Set `sweep_angle' to `new_sweep_angle'
		require 
			new_sweep_angle_is_valid: new_sweep_angle > 0.0 and new_sweep_angle <= 360.0
		do
			sweep_angle := new_sweep_angle * pi / 180.0
			compute_intermediates		
		ensure 
			sweep_angle_set: sweep_angle = new_sweep_angle
		end
	
	set_sweep_angle_in_radians (new_sweep_angle: DOUBLE) is
			-- Set `sweep_angle' to `new_sweep_angle'
		require 
			new_sweep_angle_is_valid: new_sweep_angle > 0.0 and 
									  new_sweep_angle <= 6.2831853
		do
			sweep_angle := new_sweep_angle
			compute_intermediates		
		ensure 
			sweep_angle_set: sweep_angle = new_sweep_angle
		end 
	set_z_minimum (new_minimum: DOUBLE) is
			-- Set `z_minimum' to `new_minimum'
		require 
			new_minimum_is_valid: new_minimum <= z_maximum
		do
			z_minimum := new_minimum
			compute_intermediates		
		ensure 
			z_minimum_set: z_minimum = new_minimum
		end 

	set_z_maximum (new_maximum: DOUBLE) is
			-- Set `z_maximum' to `new_maximum'
		require 
			new_maximum_is_valid: new_maximum >= z_minimum
		do
			z_maximum := new_maximum
			compute_intermediates		
		ensure 
			z_mimum_set: z_maximum = new_maximum
		end 

feature -- Removal
feature -- Resizing
feature -- Transformation
feature -- Conversion
feature -- Duplication
feature -- Basic operations
feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {NONE} -- Implementation

	process_u_split is
			-- Set the neighbors for a split in u
		local
			first, second: SCENE_DRAWABLE_ELEMENT
		do
			if wrapped_in_u and then u_maximum - u_minimum > 0.5 then
				first := split_items.first
				second := split_items.last
				
				second.set_neighbor (first, right)
				second.set_neighbor (first, left)
				update_u_neighbors (first, left)
				update_v_neighbors (first, above)
				update_v_neighbors (first, below)
				
				update_u_neighbors (second, right)
				update_v_neighbors (second, above)
				update_v_neighbors (second, below)
			else
				Precursor
			end
		end

	set_wrapping is
			-- Set the wrapping parameter based on the `sweep_angle'
		do
			if about_equal (sweep_angle, 2 * pi)  then
				set_neighbor (current, right)
				wrapped_in_u := true
			else
				remove_neighbor (current, right)
				wrapped_in_u := false
			end
		end 

	compute_intermediates is
			-- Require intermediate computations
		do
			update_display_lists
		end

	sweep_for_bounds (rmin, rmax, tmin, tmax: DOUBLE) is
			-- 
		require 
			default: true
		do
			extents.make (polar (rmin, tmin), polar (rmin, tmin))
			extents.expand_to_include_point (polar (rmax, tmin))
			extents.expand_to_include_point (polar (rmin, tmax))
			extents.expand_to_include_point (polar (rmax, tmax))

			if tmin < pi_over_two and tmax > pi_over_two then
				extents.expand_to_include_point (polar (rmax, pi_over_two))
			end
			if tmin < pi and tmax > pi then
				extents.expand_to_include_point (polar (rmax, pi))
			end
			if tmin < pi_over_two + pi and tmax > pi_over_two + pi then
				extents.expand_to_include_point (polar (rmax, pi_over_two + pi))
			end
			
		ensure 
			default: true
		end 

	polar (rad, theta: DOUBLE): POINT3 is
			-- What is the cartesian coordinate for the polar 
			-- coordinate (`rad', `theta')?
		do
			create Result.make (rad * cosine (theta), rad * sine (theta), 0)
		end
	
	add_attributes (the_attributes: DS_ARRAYED_LIST [SCENE_ATTRIBUTE]) is
			-- Add attributes for the GUI_ACTION
		local
			attribute: SCENE_ATTRIBUTE_INFO [DOUBLE]
		do
			create attribute.make ("radius", index_of_field ("radius"), 
									agent valid_radius, agent set_radius)
			the_attributes.force_last (attribute)
			create attribute.make ("z_minimum", index_of_field ("z_minimum"), 
									agent valid_z_minimum, agent set_z_minimum)
			the_attributes.force_last (attribute)
			create attribute.make ("z_maximum", index_of_field ("z_maximum"), 
									agent valid_z_maximum, agent set_z_maximum)
			the_attributes.force_last (attribute)
			create attribute.make ("sweep_angle", index_of_field ("sweep_angle"), 
									agent valid_sweep_angle_in_radians, 
									agent set_sweep_angle_in_radians)
			the_attributes.force_last (attribute)
		end
	
	print_features (action: PRINT_ACTION) is
			-- print the sphere's features
		do
			Precursor (action)
			
			action.io.put_string ("	 radius = ")
			action.io.put_double (radius)
			action.io.put_new_line
			
			action.io.put_string ("	 z_minimum = ")
			action.io.put_double (z_minimum)
			action.io.put_new_line
			
			action.io.put_string ("	 z_maximum = ")
			action.io.put_double (z_maximum)
			action.io.put_new_line
			
			action.io.put_string ("	 sweep_angle = ")
			action.io.put_double (sweep_angle)
			action.io.put_new_line
			
			action.io.put_string ("	 origin = ")
			action.io.put_string (origin.out)
			action.io.put_new_line
		end 
	
invariant
	
	radius_not_negative: radius >= 0
	quadrics_are_only_constant_colored: color_mapping = constant_color
	origin_not_void: origin /= Void
	sweep_angle_in_range: sweep_angle > 0 and sweep_angle <= 2 * pi
	is_wrapped: about_equal (sweep_angle, 2 * pi) implies wrapped_in_u = true
	
end -- SCENE_QUADRIC

