

indexing
	description: "Abstract class representing lights"
	patterns: "none"
	date: "$Date: 2007-01-09 16:27:59 -0500 (Tue, 09 Jan 2007) $"
	revision: "$Revision: 278 $"
	status: "unreviewed"




deferred class

	LIGHT_SOURCE

inherit

	BASIC_NUMERIC
		undefine
			default_create, is_equal, copy
		end

	COMPARABLE
		undefine
			is_equal, default_create, copy
		end

	SCENE_STATE_ELEMENT_TABLE
		rename
			start as element_start,
			after as element_after,
			forth as element_forth
		redefine
			default_create, apply, invalidate
		end

	SCENE_OBJECT
		export
			{NONE} all
			{ANY} print_object
		undefine
			detailed_memory_usage, is_equal, copy
		redefine
			add_attributes, default_create, print_features
		end

	NUMERIC_CONVERSION
		export
			{NONE} all
		undefine
			default_create, is_equal, copy
		end

feature -- Initialization

	default_create is
			-- Initialize a light; By default it's white with
			-- constant attenuation
		local
			sce: SCENE_COLOR_ELEMENT
		do
			Precursor {SCENE_STATE_ELEMENT_TABLE}
			Precursor {SCENE_OBJECT}

			handle := 1
			dynamic_id := -1
			intensity := 1.0
			light_samples := 1
			sample_number := 1
			enabled := false

			create sce

			add_element (sce)

			create light_probe
			create light_ray
			create light_direction
			create illumination_at_point

			constant_attenuation := 1.0
			linear_attenuation := 0.0
			quadratic_attenuation := 0.0

			create light_to_world
		end

	make (the_handle: INTEGER; the_shader: STRING;
		  parameters: PARAMETER_LIST) is
			-- Initialize a light
		require
			handle_positive: the_handle > 0
			valid_shader: the_shader /= Void and then the_shader.count > 0
		local
			sce: SCENE_COLOR_ELEMENT
			c: RGBA_COLOR
		do
			default_create -- The descendant version is actually called

			handle := the_handle

			shader := the_shader.twin

			if parameters /= Void then
				from
					parameters.start
				until
					parameters.after
				loop
					if parameters.item_for_iteration.first.is_equal ("%"lightcolor%"") then
						create c.make (parameters.item_for_iteration.second.second @ 1,
								parameters.item_for_iteration.second.second @ 2,
								parameters.item_for_iteration.second.second @ 3,
								1.0)

						create sce.make (c)
						replace_element (sce)
					elseif parameters.item_for_iteration.first.is_equal ("%"intensity%"") then
						intensity := parameters.item_for_iteration.second.second @ 1
					end
					parameters.forth
				end
			end
		ensure
			handle_set: handle = the_handle
			shader_set: shader.is_equal (the_shader)
		end

feature -- Access

	enabled: BOOLEAN
			-- Has the light been enabled for drawing?

	shader: STRING
			-- The name of the light shader

	intensity: DOUBLE
			-- The `intensity' of the light source

	constant_attenuation: DOUBLE
			-- A constant factor used for determining the attenuation
			-- of the light source

	linear_attenuation: DOUBLE
			-- The attenuation of the light as a linear function of distance

	quadratic_attenuation: DOUBLE
			-- The attenuation of the light as quadratic function of
			-- distance. Computes standard 1/R^2 falloff

	light_color: SCENE_COLOR_ELEMENT is
			-- The light's color
		do
			if cached_color = Void then
				cached_color ?=
				element_item ("SCENE_COLOR_ELEMENT")
			end

			Result := cached_color
		end

	illumination_at_point: RGBA_COLOR
			-- What is the illumination at `origin' of `intersect'?

	light_direction: VECTOR3
			-- The direction of the light used at intersect

	was_occluded: BOOLEAN
			-- Was the light occluded

	handle: INTEGER
			-- The `handle' for refering to this light

	dynamic_id: INTEGER
			-- An `id' assigned by an external source (a renderer)
			-- for use when processing. Different from `handle' in
			-- that a system may allow infinte number of lights, but
			-- only a limitied set may be illuminating at any given
			-- moment.
			--
			-- -1 indicates that the id is currently unassigned

	light_samples: INTEGER
			-- The number of light samples to use when rendering

	emission (light_pt, geom_pt: POINT3): RGBA_COLOR is
			-- The emissive property of the light
		require
			light_pt_not_void: light_pt /= Void
			geom_pt_not_void: geom_pt /= Void
		do
			Result := light_color.i_th (1).scale (intensity)
		end

	light_to_world: TRANSFORM
			-- The transform to put the light in world space

feature -- Measurement
feature -- Comparison

	infix "<" (other: like Current): BOOLEAN is
			-- Is `Current' less than `other'?
		do
			Result := handle < other.handle
		end

feature -- Status report

	valid_shader (a_shader: STRING): BOOLEAN is
			-- Is `a_shader' valid?
		do
			Result := a_shader /=Void and then a_shader.count > 0
		end

	valid_intensity (new_intensity: DOUBLE): BOOLEAN is
			-- Is `new_intensity' valid?
		do
			Result := new_intensity > 0
		end

	valid_constant_attenuation (new_attenuation: DOUBLE): BOOLEAN is
			-- Is `new_attenuation' valid?
		do
			Result := new_attenuation >= 0.0
		end

	valid_linear_attenuation (new_attenuation: DOUBLE): BOOLEAN is
			-- Is `new_attenuation' valid?
		do
			Result := new_attenuation >= 0.0
		end

	valid_quadratic_attenuation (new_attenuation: DOUBLE): BOOLEAN is
			-- Is `new_attenuation' valid?
		do
			Result := new_attenuation >= 0.0
		end

feature -- Status setting

	enable_light is
			-- Request the light to be enabled
		do
			enable_requested := true
		ensure
			enable_requested_set: enable_requested
		end

	disable_light is
			-- Request the light to be disabled
		do
			enable_requested := false
		ensure
			enable_requested_set: not enable_requested
		end

	reset (renderer: RENDERER) is
			-- Force the light to be disabled
		do
			disable_light
			if enabled then
				renderer.disable_light (dynamic_id)
				release_dynamic_id
			end
			enabled := false
		ensure
			enabled_set: not enabled
		end

feature -- Cursor movement

	start is
			-- The first light sample
		do
			sample_number := 1
		ensure
			beginning: sample_number = 1
		end

	after: BOOLEAN is
			-- Are we at the last sample?
		do
			Result := sample_number > light_samples
		ensure
			correct: Result implies sample_number > light_samples
		end

	forth is
			-- The next light sample
		do
			sample_number := sample_number + 1
		ensure
			changed: old sample_number = sample_number - 1
		end

feature -- Element change

	set_dynamic_id (new_id: INTEGER) is
			-- Set `dynamic_id' to `new_id'
		require
			new_id_is_valid: new_id >= 0
		do
			dynamic_id := new_id
		ensure
			dynamic_id_set: dynamic_id = new_id
		end

	release_dynamic_id is
			-- Set `dynamic_id' to default
		do
			dynamic_id := -1
		ensure
			dynamic_id_set: dynamic_id = -1
		end

	set_shader (new_shader: STRING) is
			-- Set `shader' to `new_shader'
		require
			valid_shader: new_shader /= Void and then new_shader.count > 0
		do
			shader := new_shader.twin
		end

	set_intensity (new_intensity: DOUBLE) is
			-- Set `intensity' to `new_intensity'
		require
			new_intensity_strictly_positive: new_intensity > 0
		do
			intensity := new_intensity
		ensure
			intensity_set: intensity = new_intensity
		end

	set_constant_attenuation (new_attenuation: DOUBLE) is
			-- Set `constant_attenuation' to `new_attenuation'
		require
			new_attenuation_is_positive: new_attenuation >= 0.0
		do
			constant_attenuation := new_attenuation
		ensure
			constant_attenuation_set: constant_attenuation = new_attenuation
		end

	set_linear_attenuation (new_attenuation: DOUBLE) is
			-- Set `linear_attenuation' to `new_attenuation'
		require
			new_attenuation_is_positive: new_attenuation >= 0.0
		do
			linear_attenuation := new_attenuation
		ensure
			linear_attenuation_set: linear_attenuation = new_attenuation
		end

	set_quadratic_attenuation (new_attenuation: DOUBLE) is
			-- Set `quadratic_attenuation' to `new_attenuation'
		require
			new_attenuation_is_positive: new_attenuation >= 0.0
		do
			quadratic_attenuation := new_attenuation
		ensure
			quadratic_attenuation_set: quadratic_attenuation = new_attenuation
		end

	set_light_samples (how_many: INTEGER) is
			-- Set `light_samples' to `how_many'
		require
			how_many_strictly_positive: how_many > 0
		do
			light_samples := how_many
		ensure
			light_samples_set: light_samples = how_many
		end

	set_light_to_world_transform (new_transform: TRANSFORM) is
			-- Set `light_to_world' to `new_transform'
		require
			new_transform_not_void: new_transform /= Void
		do
			light_to_world.copy (new_transform)
		ensure
			light_to_world_set: light_to_world.is_equal (new_transform)
		end

feature -- Removal
feature -- Resizing
feature -- Transformation
feature -- Conversion
feature -- Duplication
feature -- Basic operations

	apply (renderer: RENDERER) is
			-- apply the `renderer' to `Current'.
		local
			id: INTEGER
		do
			if not enabled and enable_requested then
				-- the light is off, and we want it turned on
				renderer.push_matrix
				-- Need to ensure this tranform is referenced from
				-- the global transform not the local one
-- Removed: The viewing transform is now rolled in the camera' world_to_camera
-- matrix.
				renderer.load_viewing_transform
				renderer.multiply_matrix (light_to_world)
				id := renderer.assign_light_index
				if id >= 0 then
					set_dynamic_id (id)
					draw (renderer)
					renderer.enable_light (dynamic_id)
					renderer.pop_matrix
					enabled := true
				else
					release_dynamic_id
					enabled := false
				end
			elseif enabled and not enable_requested then
				-- the light is on, and we want it turned off
				renderer.disable_light (dynamic_id)
				release_dynamic_id
				enabled := false
			end
		end

	draw (renderer: RENDERER) is
			-- `Draw' the scene through the camera's lens
		require
			renderer_not_void: renderer /= Void
		local
			color: RGBA_COLOR
		do
			renderer.set_ambient_light_color (dynamic_id, black)
			renderer.set_specular_light_color (dynamic_id, grey_50)
			color := light_color.i_th (1).scale (intensity)
			renderer.set_diffuse_light_color (dynamic_id, color)
			renderer.set_constant_attenuation (dynamic_id,
											   real_64_truncated_to_real_32 (constant_attenuation))
			renderer.set_linear_attenuation (dynamic_id,
											 real_64_truncated_to_real_32 (linear_attenuation))
			renderer.set_quadratic_attenuation (dynamic_id,
												real_64_truncated_to_real_32 (quadratic_attenuation))
		end

	generate_sample_point: POINT3 is
			-- Generate a point on the light
		deferred
		ensure
			Result_not_void: Result /= Void
		end

	intersect (scene: SCENE_ACCELERATOR [SCENE_DRAWABLE]; origin: POINT3;
			   info: INTERSECTION_DATA) is
			-- Intersect the `scene' with a ray at `origin'
		require
			scene_not_void: scene /= Void
			origin_not_void: origin /= Void
			info_not_void: info /= Void
		deferred
		end

	compute_illumination (point: SHADING_DATA; camera: SCENE_CAMERA) is
			-- Compute the illumination at `point'
		require
			point_not_void: point /= Void
			camera_not_void: camera /= Void
		deferred
		end

	ambient: RGBA_COLOR is
			-- Compute the `ambient' illumination at `shading_point'
		do
			create Result.make (0, 0, 0, 0)
		ensure
			Result_not_void: Result /= Void
		end 

	diffuse (n: NORMAL; shading_point: POINT3): RGBA_COLOR is
			-- Compute the `diffuse' illumination at `shading_point'
		require 
			n_not_void: n /= Void
			shading_point_not_void: shading_point /= Void
		do
			create Result.make (0, 0, 0, 0)
		ensure
			Result_not_void: Result /= Void
		end 

feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {NONE} -- Implementation

	enable_requested: BOOLEAN
			-- Has this light been requested to be enabled?

	cached_color: SCENE_COLOR_ELEMENT

	light_probe: ALPHA_INTERSECT_ACTION
	light_ray: RAY

	sample_number: INTEGER
			-- which light sample

	invalidate is
			-- Elements have been changed, so invalidate state as
			-- necessary
		do
			cached_color := Void
		ensure then
			cache_cleared: cached_color = Void
		end

	occluded (probe: INTERSECT_ACTION; distance_to_light: DOUBLE): BOOLEAN is
			-- Test if a shadow ray is occluded
		require
			probe_intersected: probe.intersected
		local
			intersection: SORTABLE_ARRAYED_LIST [INTERSECTION_DATA]
			info: INTERSECTION_DATA
			done: BOOLEAN
		do
			intersection := probe.intersection
			intersection.sort

			if intersection.item (intersection.count).parametric_distance < Tolerance then
				-- self-intersecting
				Result := false
			else
				from
					intersection.start
					done := false
				until
					intersection.after or done
				loop
					info := intersection.item_for_iteration

					if info.parametric_distance > Tolerance and
						info.parametric_distance < distance_to_light then
						done := true
					end

					intersection.forth
				end

				if done then
					Result := true
				end
			end
		end

	white: RGBA_COLOR is
			-- white
		once
			create Result.make (1, 1, 1, 1)
		end

	black: RGBA_COLOR is
			-- black
		once
			create Result.make (0, 0, 0, 1)
		end

	grey_50: RGBA_COLOR is
			-- 50% grey
		once
			create Result.make (0.5, 0.5, 0.5, 1)
		end

	add_attributes (the_attributes: DS_ARRAYED_LIST [SCENE_ATTRIBUTE]) is
			-- Add public modifiable attributes to `attributes'
		local
			double_attribute: SCENE_ATTRIBUTE_INFO [DOUBLE]
			string_attribute: SCENE_ATTRIBUTE_INFO [STRING]
		do
			create string_attribute.make ("shader", index_of_field ("shader"),
										  agent valid_shader, agent set_shader)
			the_attributes.force_last (string_attribute)
			create double_attribute.make ("intensity", index_of_field ("intensity"),
										  agent valid_intensity, agent set_intensity)
			the_attributes.force_last (double_attribute)
			create double_attribute.make ("constant_attenuation",
										  index_of_field ("constant_attenuation"),
										  agent valid_constant_attenuation,
										  agent set_constant_attenuation)
			the_attributes.force_last (double_attribute)
			create double_attribute.make ("linear_attenuation",
										  index_of_field ("linear_attenuation"),
										  agent valid_linear_attenuation,
										  agent set_linear_attenuation)
			the_attributes.force_last (double_attribute)
			create double_attribute.make ("quadratic_attenuation",
										  index_of_field ("quadratic_attenuation"),
										  agent valid_quadratic_attenuation,
										  agent set_quadratic_attenuation)
			the_attributes.force_last (double_attribute)
		end

	print_features (action: PRINT_ACTION) is
			-- print the group's features
		do
			Precursor (action)

			action.io.put_string ("	 handle = ")
			action.io.put_integer (handle)
			action.io.put_new_line

			action.io.put_string ("	 light_to_world = ")
			action.io.push
			action.io.put_new_line
			light_to_world.print_using_io (action.io)
			action.io.pop
			action.io.put_new_line

			action.io.put_string ("	 shader = ")
			action.io.put_string (shader)
			action.io.put_new_line

			action.io.put_string ("	 intensity = ")
			action.io.put_double (intensity)
			action.io.put_new_line

			action.io.put_string ("	 light_color = ")
			action.io.put_string (light_color.i_th (1).out)
			action.io.put_new_line

			action.io.put_string ("	 dynamic_id = ")
			action.io.put_integer (dynamic_id)
			action.io.put_new_line
		end

invariant

	handle_positive: handle > 0
	shader_valid: shader /= Void and then shader.count > 0
	dynamic_id_is_valid: dynamic_id >= 0 or dynamic_id = -1
	intensity_is_positive: intensity >= 0
	light_color_not_void: light_color /= Void
	constant_attenuation_is_positive: constant_attenuation >= 0.0
	linear_attenuation_is_positive: linear_attenuation >= 0.0
	quadratic_attenuation_is_positive: quadratic_attenuation >= 0.0
	light_samples_strictly_positive: light_samples > 0
	light_to_world_not_void: light_to_world /= Void
	illumination_at_point_not_void: illumination_at_point /= Void
	sample_number_in_range: sample_number >= 1 and sample_number <= light_samples + 1

end -- LIGHT_SOURCE

