

indexing
	description: "Axis-aligned box bounding volumes"
	patterns: "none"
	date: "$Date: 2007-01-09 16:27:59 -0500 (Tue, 09 Jan 2007) $"
	revision: "$Revision: 278 $"
	status: "unreviewed"




class

	SCENE_BOX_BOUND

inherit

	SCENE_BOUND
		redefine
			default_create, intersect_ray
		end

	BASIC_NUMERIC
		undefine
			out
		redefine
			default_create
		end

creation
	default_create, make, make_infinite

feature -- Initialization

	default_create is
			-- Create an empty bounding box
		do
			if lower = Void then
				create lower
			else
				lower.default_create
			end

			if upper = Void then
				create upper
			else
				upper.default_create
			end

			Precursor {SCENE_BOUND}

			reset_intersections
		end

	make (first_corner, second_corner: POINT3) is
			-- Arbitrary-ordered vertices to be bounded.
		do
			default_create
			
			lower.make (first_corner.x.min (second_corner.x),
						first_corner.y.min (second_corner.y),
						first_corner.z.min (second_corner.z))
			upper.make (first_corner.x.max (second_corner.x),
						first_corner.y.max (second_corner.y),
						first_corner.z.max (second_corner.z))
		end

	make_infinite is
			-- Create an infinte bounding box
		do
			create lower.make (-huge_value, -huge_value, -huge_value)
			create upper.make (huge_value, huge_value, huge_value)

			reset_intersections
		ensure then
			is_empty: empty
		end

feature -- Access

	lower: POINT3
			-- Lower corner.

	upper: POINT3
			-- Upper corner.

	centroid: POINT3 is
			-- Where is the centroid of the box?
		do
			create Result.make ((upper.x + lower.x) * 0.5,
						(upper.y + lower.y) * 0.5,
						(upper.z + lower.z) * 0.5)
		end

	length: DOUBLE is
			-- The longest distance from the center to the outside of
			-- the bounds. A sphere of center `centroid` with radius
			-- `length` is guaranteed to encompass the bounds
		do
			Result := (upper - lower).magnitude / 2
		end

feature -- Measurement
feature -- Comparison
feature -- Status report

	contains (point: POINT3): BOOLEAN is
			-- Does this object contain `point'?
		do
			Result := within_range (point.x, lower.x, upper.x) and then
						within_range (point.y, lower.y, upper.y) and then
						within_range (point.z, lower.z, upper.z)
		end

	empty: BOOLEAN is
			-- Is the bounds empty?
		do
			Result := lower.is_equal (upper)
		ensure then
			result_correct: Result = lower.is_equal (upper)
		end

	infinite: BOOLEAN is
			-- Is the bounds infinite?
		do
			Result :=
				lower.x = -huge_value and
				lower.y = -huge_value and
				lower.z = -huge_value and
				upper.x = huge_value and
				upper.y = huge_value and
				upper.z = huge_value
		end

feature -- Status setting
feature -- Cursor movement
feature -- Element change

	expand_to_include (bound: SCENE_BOUND) is
			-- Adapt this bounds to include `bound'
		local
			box: AXIS_ALIGNED_BOX
		do
			box := bound.circumscribed_box

			expand_to_include_point (box.lower)
			expand_to_include_point (box.upper)
		end

	expand_to_include_point (point: POINT3) is
			-- Adapt this bounds to include `point'
		do
			lower.make (lower.x.min (point.x),
						lower.y.min (point.y),
						lower.z.min (point.z))
			upper.make (upper.x.max (point.x),
						upper.y.max (point.y),
						upper.z.max (point.z))
		end

	expand_to_include_points (points: POINT3_ARRAYED_LIST) is
		do
			from
				points.start
			until
				points.after
			loop
				expand_to_include_point (points.item_for_iteration)
				points.forth
			end
		end

	set_lower (new_lower: POINT3) is
			-- Set `lower'.
		require
			x_less_than_upper: new_lower.x <= upper.x
			y_less_than_upper: new_lower.y <= upper.y
			z_less_than_upper: new_lower.z <= upper.z
		do
			lower := new_lower
		ensure
			lower_set: lower = new_lower
		end

	set_upper (new_upper: POINT3) is
			-- Set `upper'.
		require
			x_greater_than_lower: lower.x <= new_upper.x
			y_greater_than_lower: lower.y <= new_upper.y
			z_greater_than_lower: lower.z <= new_upper.z
		do
			upper := new_upper
		ensure
			upper_set: upper = new_upper
		end

feature -- Removal
feature -- Resizing
feature -- Transformation
feature -- Conversion

	circumscribed_box: AXIS_ALIGNED_BOX is
			-- A box that circumscribes the bound
		do
			create Result.make (lower, upper)
		end

	out: STRING is
			-- Stringified version of `Current'
		do
			create Result.make (160)
			if infinite then
				Result.append ("infinite")
			elseif empty then
				Result.append ("empty")
			else
				Result.append ("lower = ")
				Result.append (lower.out)
				Result.append (", upper = ")
				Result.append (upper.out)
			end
		end

feature -- Duplication
feature -- Basic operations

	apply (transform: TRANSFORM): like Current is
			-- Apply `transform' to `Current'
		do
			create Result.make (transform.apply (lower), transform.apply (upper))
		end

	draw_immediate (renderer: RENDERER) is
			-- Apply the renderer to this object
		local
			draw_style: INTEGER
		do
			draw_style := renderer.draw_style
			renderer.set_draw_style (renderer.rd_line)
			renderer.disable (renderer.rd_lighting)

			renderer.enable (renderer.rd_color_material_constant)
			renderer.color_material (renderer.rd_front_and_back,
							   renderer.rd_ambient_and_diffuse)

			renderer.rgba_color (green)

			renderer.cube (lower, upper)

			renderer.enable (renderer.rd_lighting)
			renderer.set_draw_style (draw_style)
		end

	intersect_ray (ray: RAY) is
			-- Intersect with `ray'.
			-- If `ray' does not intersect then `intersected' will be False
			-- else `intersected' will be True and `intersection' will yield
			-- the ordered parametric coordinates along `ray' of each
			-- successive point of intersection.
		local
			t_near, t_far: DOUBLE
					-- Parametric coordinates along the ray of the near and far
					-- points of intersection with the box.

			t1, t2: DOUBLE
					-- Intersection distances to each pair of bounding planes.

			direction, origin: DOUBLE
					-- Ray's coordinate direction and origin.

			x_intersected, y_intersected, z_intersected: BOOLEAN
					-- Did the coordinate intersect?
		do
			Precursor (ray)
			reset_intersections

			if empty then
				outside := true
			elseif infinite then
				inside := true
			else
				outside := true
				t_near := -huge_value
				t_far := huge_value
				direction := ray.direction.x
				origin := ray.origin.x

				if direction = 0.0 then
					-- Ray direction is parallel to the bounding x-planes so
					-- it intersects only if it originates at a point with
					-- x coordinates that lie between the bounding x-planes.
					x_intersected := within_range (origin, lower.x, upper.x)
				else
					-- Ray is not parallel to the bounding x-planes so
					-- calculate near and far intersection distances of planes.
					direction := 1.0 / direction
					t1 := (lower.x - origin) * direction
					t2 := (upper.x - origin) * direction

					if t1 > t2 then
						t_near := t2
						t_far := t1
					else
						t_near := t1
						t_far := t2
					end

					check
						t_near_not_more_than_t_far: t_near <= t_far
					end
					x_intersected := t_far >= 0.0
				end

				if x_intersected then
					direction := ray.direction.y
					origin := ray.origin.y

					if direction = 0.0 then
						-- Ray direction is parallel to the bounding y-planes so
						-- it intersects only if it originates at a point with
						-- y coordinates that lie between the bounding y-planes.
						y_intersected := within_range (origin, lower.y, upper.y)
					else
						-- Ray is not parallel to the bounding y-planes so
						-- calculate near and far intersection distances of
						-- planes.
						direction := 1.0 / direction
						t1 := (lower.y - origin) * direction
						t2 := (upper.y - origin) * direction

						if t1 > t2 then

							if t2 > t_near then
								t_near := t2
							end

							if t1 < t_far then
								t_far := t1
							end
						else

							if t1 > t_near then
								t_near := t1
							end

							if t2 < t_far then
								t_far := t2
							end
						end

						y_intersected := t_near <= t_far and then t_far >= 0.0
					end

					if y_intersected then
						direction := ray.direction.z
						origin := ray.origin.z

						if direction = 0.0 then
							-- Ray direction is parallel to the bounding
							-- z-planes so it intersects only if it originates
							-- at a point with z coordinates that lie between
							-- the bounding z-planes.
							z_intersected := within_range (origin, lower.z, upper.z)
						else
							-- Ray is not parallel to the bounding z-planes so
							-- calculate near and far intersection distances of
							-- planes.
							direction := 1.0 / direction
							t1 := (lower.z - origin) * direction
							t2 := (upper.z - origin) * direction

							if t1 > t2 then

								if t2 > t_near then
									t_near := t2
								end

								if t1 < t_far then
									t_far := t1
								end
							else

								if t1 > t_near then
									t_near := t1
								end

								if t2 < t_far then
									t_far := t2
								end
							end

							z_intersected := t_near <= t_far and then t_far >= 0.0
						end

						if z_intersected then
							outside := false
							if t_near < 0.0 or else about_equal (t_near, t_far) then
								inside := true
							else
								intersected := true
							end
						end
					end
				end
			end

			if not outside then
				intersection_hit := intersection_hit + 1
			end
		end

	intersect_plane (plane: PLANE) is
			-- Intersect with `plane'.
			-- One and only one of the following will be true,
			-- `inside', `outside', or `intersected'. `inside' is
			-- defined by the normal direction
		do
			check
				implemented: false
			end
		end

	intersect_polyhedron (poly: POLYHEDRON) is
			-- Intersect with `plane'.
			-- One and only one of the following will be true,
			-- `inside', `outside', or `intersected'. `inside' is
			-- defined by the normal direction
		do
			check
				implemented: false
			end
		end

feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {NONE} -- Implementation

	reset_intersections is
			-- Reset the location of the intersection queries to false
		do
			intersected := false
			outside := false
			inside := false
		end

invariant

	lower_not_void: lower /= Void
	upper_not_void: upper /= Void
	x_ordered: lower.x <= upper.x
	y_ordered: lower.y <= upper.y
	z_ordered: lower.z <= upper.z

end -- SCENE_BOX_BOUND

