

indexing
	description: "Abstract Camera for viewing the scene graph"
	patterns: "none"
	status: "unreviewed"
	date: "$Date: 2006-09-15 16:35:58 -0400 (Fri, 15 Sep 2006) $"
	revision: "$Revision: 267 $"



deferred class

	SCENE_CAMERA

inherit

	RI_CONSTANTS

feature -- Initialization

	make is
			--
		do
			x_resolution := default_x_resolution
			y_resolution := default_y_resolution

			pixel_aspect_ratio := default_pixel_aspect_ratio
			frame_aspect_ratio := 4.0/3.0

			create screen_window.make(1, 4)
			update_screen_window (-frame_aspect_ratio, frame_aspect_ratio,
								-1.0, 1.0)

			create crop_window.make(1, 4)
			set_crop_window (0.0, 1.0, 0.0, 1.0)

			near_clipper := epsilon
			far_clipper := infinity

			f_stop := infinity

			shutter_open := 0
			shutter_close := 0

			create camera_focal_point
			create position
			create orientation

			create world_to_camera.make
			create world_to_raster.make
			create camera_to_screen.make
			create internal_world_to_camera.make
			create screen_to_raster.make
			create raster_to_camera.make
			create camera_to_raster.make
		end

feature -- Access

	x_resolution: INTEGER
			-- Display resolution in x

	y_resolution: INTEGER
			-- Display resolution in y

	default_x_resolution: INTEGER is 640
			-- Display resolution in x

	default_y_resolution: INTEGER is 480
			-- Display resolution in y

	pixel_aspect_ratio: DOUBLE
			-- Ratio of physical height to width of a pixel

	default_pixel_aspect_ratio: DOUBLE is 1.0
			-- Ratio of physical height to width of a pixel

	frame_aspect_ratio: DOUBLE
			-- Ratio of physical height to width of the image

	screen_window: ARRAY[DOUBLE]
			-- The coordinates defining the raster to screen mapping

	crop_window: ARRAY[DOUBLE]
			-- Coordinates for a subwindow for rendering

	crop_bound: AXIS_ALIGNED_BOX is
			-- The pixel coordinates of the crop_window
		local
			lower, upper: POINT3
		do
			create lower.make ((crop_window.item (1) * x_resolution).floor,
							   (crop_window.item (3) * y_resolution).floor,
							   0)
			create upper.make ((crop_window.item (2) * x_resolution).floor,
							   (crop_window.item (4) * y_resolution).floor,
							   0)
			create Result.make (lower, upper)
		end

	near_clipper, far_clipper: DOUBLE
			-- The near and far clipping planes

	f_stop, focal_length, focal_distance: DOUBLE
			-- Camera parameters for Depth of field

	focal_point: POINT3 is
			-- Camera `look at' position
		do
			--Result := world_to_camera.item.apply (camera_focal_point)
			Result := camera_focal_point
		end

	shutter_open, shutter_close: DOUBLE
			-- Camera parameters for motion blur

	position: POINT3
			-- The camera's current location

	orientation: QUATERNION
			-- The camera's current orientation

	view_volume: VIEW_VOLUME is
			-- The camera's view volume
		deferred
		ensure
			Result_not_void: Result /= Void
		end

	world_to_camera: TRANSFORM
			-- The world space to camera space transformation

	world_to_raster: TRANSFORM
			-- The world space to raster space transformation

	camera_to_screen: TRANSFORM
			-- The camera space to screen space transform

	screen_to_raster: TRANSFORM
			-- The screen space to raster space transformation

	raster_to_camera: TRANSFORM
			-- The raster space to camera space transformation

	camera_to_raster: TRANSFORM
			-- The camera space to raster space transformation

feature -- Measurement
feature -- Comparison
feature -- Status report

	projection_direction: VECTOR3 is
			-- The direction of projection
		do
			create Result.make (0, 0, 1)
			--Result := view_volume.projection_direction
		end

	valid_display_type (str: STRING): BOOLEAN is
			-- Check if the display type is either a "framebuffer" or "file"
		require
			void_string: str /= Void
		do
			Result := str.is_equal ("%"framebuffer%"") or
					  str.is_equal ("%"file%"")
		end -- valid_display_type

	valid_display_mode (str: STRING): BOOLEAN is
			-- Check if the display mode is in [rgb][a][z]
		require
			void_string: str /= Void
		do
			Result := str.is_equal ("%"a%"") or
					  str.is_equal ("%"z%"") or
					  str.is_equal ("%"az%"") or
					  str.is_equal ("%"rgb%"") or
					  str.is_equal ("%"rgba%"") or
					  str.is_equal ("%"rgbaz%"")
		end -- valid_display_mode

feature -- Status setting
feature -- Cursor movement
feature -- Element change

	set_position_xyz (x, y, z: DOUBLE) is
			-- Set the `position` to <`x`, `y`, `z`>
		do
			position.make (x, y, z)

			update_view_volume
		ensure
			position_set: position.x = x and position.y = y and position.z = z
		end

	set_position (pt: POINT3) is
			-- Set the `position` to `pt'
		require
			pt_not_void: pt /= Void
		do
			position.copy (pt)

			update_view_volume
		ensure
			position_set: position.is_equal (pt)
		end

	set_focal_point (pt: POINT3) is
			-- Set the `position` to `pt'
		require
			pt_not_void: pt /= Void
		do
			camera_focal_point.copy (pt)

			update_view_volume
		ensure
			focal_point_set: camera_focal_point.is_equal (pt)
		end

	display_format (xresolution, yresolution: INTEGER; pixelaspect: DOUBLE) is
			-- Set the horizontal and vertical resolution of the image
		require
			positive: xresolution > 0 and yresolution > 0
		do
			x_resolution := xresolution
			y_resolution := yresolution
			if pixelaspect > 0.0 then
				pixel_aspect_ratio := pixelaspect
			end
			if not frame_aspect_set then
				set_frame_aspect ((x_resolution*pixel_aspect_ratio)/y_resolution)
			end

			update_view_volume
		ensure
			parameters_set: x_resolution = xresolution and
							y_resolution = yresolution and
							pixelaspect > 0.0 implies
							pixel_aspect_ratio = pixelaspect
		end

	set_frame_aspect_ratio (ratio: DOUBLE) is
			-- Ratio of width to height of image
		require
			positive: ratio > 0.0
		do
			set_frame_aspect (ratio)
			frame_aspect_set := true

			update_view_volume
		ensure
			default: true
		end

	set_screen_window (left, right, bottom, top: DOUBLE) is
			-- Set the screen to raster mapping
		require
			default: true
		do
			update_screen_window (left, right, bottom, top)
			screen_window_set := true

			update_view_volume
		ensure
			default: true
		end

	set_crop_window (xmin, xmax, ymin, ymax: DOUBLE) is
			-- Define a subwindow for rendering
		require
			correct_horizontal_relation: xmin >= 0.0 and xmin <= xmax and xmax <= 1.0
			correct_vertical_relation: ymin >= 0.0 and ymin <= ymax and ymax <= 1.0
		do
			crop_window.put (xmin, 1)
			crop_window.put (xmax, 2)
			crop_window.put (ymin, 3)
			crop_window.put (ymax, 4)
		ensure
			default: true
		end

	set_projection (name: STRING; parameters: PARAMETER_LIST) is
			-- Define the mapping from camera to screen coordinates
		require
			valid_name: name /= Void and name.count > 0
		do

		ensure
			default: true
		end

	set_clipping (near, far: DOUBLE) is
			-- Set the clipping planes
		require
			proper_relation: near < far
		do
			near_clipper := near
			far_clipper := far

			update_view_volume
		ensure
			near_set: near_clipper = near
			far_set: far_clipper = far
		end

	set_depth_of_field (fstop, focallength, focaldistance: DOUBLE) is
			--
		require
			positive_distance: focaldistance >= 0.0
			positive_length: focallength >= 0.0
			positive_fstop: fstop > 0.0
		do
			f_stop := fstop
			focal_length := focallength
			focal_distance := focaldistance

			update_view_volume
		ensure
			f_stop_set: f_stop = fstop
			focal_length_set: focal_length = focallength
			focal_distance_set: focal_distance = focaldistance
		end

	set_shutter (min, max: DOUBLE) is
			-- Sets the times when the shutter opens and closes
		require
			valid_times: min <= max
		do
			shutter_open := min
			shutter_close := max
		ensure
			shutter_times_set: shutter_open = min and shutter_close = max
		end

	set_zoom (z: DOUBLE) is
			--
		do
			zoom_factor := z
		end

	set_world_to_camera (new_xform: TRANSFORM) is
			-- Set `world_to_camera' to `new_xform'; Updates
			-- `screen_to_raster' and `raster_to_camera'
		do
			internal_world_to_camera := new_xform.twin

			update_matrices
		ensure
			world_to_camera_set: internal_world_to_camera.is_equal (new_xform)
		end

	update_clipping_planes (node: SCENE_NODE) is
			-- Update the clipping planes
		deferred
		end

feature -- Removal
feature -- Resizing
feature -- Transformation
feature -- Conversion
feature -- Duplication

	copy_camera (other: SCENE_CAMERA) is
			--
		require
			other_not_void: other /= Void
		do
			x_resolution := other.x_resolution
			y_resolution := other.y_resolution
			pixel_aspect_ratio := other.pixel_aspect_ratio
			frame_aspect_ratio := other.frame_aspect_ratio
			screen_window := other.screen_window.deep_twin
			crop_window:= other.crop_window.deep_twin
			near_clipper := other.near_clipper
			far_clipper := other.far_clipper
			f_stop := other.f_stop
			focal_length := other.focal_length
			focal_distance := other.focal_distance
			camera_focal_point := other.camera_focal_point
			shutter_open := other.shutter_open
			shutter_close := other.shutter_close
			position := other.position
			orientation := other.orientation
			frame_aspect_set := other.frame_aspect_set
			screen_window_set := other.screen_window_set
			pixel_aspect_set := other.pixel_aspect_set
		ensure
			default: true
		end

feature -- Basic operations

	draw (pool: SCENE_DRAWABLE_POOL; renderer: RENDERER) is
			-- `Draw' the scene through the camera's lens
		require
			pool_not_void: pool /= Void
			renderer_not_void: renderer /= Void
		deferred
		end

	view_all (node: SCENE_NODE) is
			-- Center the scene graph in the viewer
		require
			node_not_void: node /= Void
		deferred
		end

	generate_ray (sample_point: POINT): RAY is
			-- Generate a ray from the eye through `sample_point'
		require
			sample_point_not_void: sample_point /= Void
		deferred
		ensure
			Result_not_void: Result /= Void
		end

	update_view_volume is
			-- Updates the `view_volume'
		do
			if screen_window @ 4 - screen_window @ 3 <
				screen_window @ 2 - screen_window @ 1 then
				screen_scale := (screen_window @ 4 - screen_window @ 3) / 2
			else
				screen_scale := (screen_window @ 2 - screen_window @ 1) / 2
			end
		end

feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {SCENE_CAMERA} -- Implementation

	frame_aspect_set, screen_window_set, pixel_aspect_set: BOOLEAN
			-- Flags for computation of inter-relationship between
			-- `frame_aspect_ratio', `pixel_aspect_ratio', and `screen_window'

	camera_focal_point: POINT3
			-- The focal_point in camera space

feature {NONE} -- Implementation

	internal_world_to_camera: TRANSFORM

	zoom_factor: DOUBLE

	screen_scale: DOUBLE
			-- Scale factor used for scaling based on screen window

	set_frame_aspect (ratio: DOUBLE) is
			-- The frame aspect ratio
		require
			is_positive: ratio > 0.0
		do
			frame_aspect_ratio := ratio
			if not screen_window_set then
				if ratio >= 1.0 then
					update_screen_window (-ratio, ratio, -1.0, 1.0)
				else
					update_screen_window (-1.0, 1.0, -1.0/ratio, 1.0/ratio)
				end
			end
		ensure
			aspect_set: frame_aspect_ratio = ratio
		end

	update_screen_window (left, right, bottom, top: DOUBLE) is
			-- Defines the raster to screen coordinate mapping
		require
			correct_relation: left < right and bottom < top
		do
			screen_window.put (left, 1)
			screen_window.put (right, 2)
			screen_window.put (bottom, 3)
			screen_window.put (top, 4)
		ensure
			default: true
		end

	update_matrices is
			-- Updates all the matrices
		local
			t1, t2, t3, t4: TRANSFORM
		do
			camera_to_screen := view_volume.projection_transformation

			create t1
			t1.scale (x_resolution, y_resolution, 1)
			create t2
			t2.scale (1, -1, 1)
			create t3
			t3.scale (1.0 / (screen_window @ 2 - screen_window @ 1),
					  1.0 / (screen_window @ 4 - screen_window @ 3),
					  1.0)
			create t4
			t4.translate (-(screen_window @ 1), -(screen_window @ 4), 0)
			screen_to_raster := t1 * t2 * t3 * t4
			raster_to_camera := camera_to_screen.inverse * screen_to_raster.inverse

			camera_to_raster := raster_to_camera.inverse
			world_to_camera :=	view_volume.affine_transformation * internal_world_to_camera
			world_to_raster :=	camera_to_raster * world_to_camera
		ensure
			default: true
		end


invariant
	positive_values: (x_resolution > 0 and y_resolution > 0 and
					  pixel_aspect_ratio > 0.0 and frame_aspect_ratio > 0.0)
	correct_size: screen_window.count = 4 and crop_window.count = 4
	crop_window_validity: ((crop_window.item (1) >= 0 and
							crop_window.item (1) <= crop_window.item (2) and
							crop_window.item (2) <= 1.0) and
						   (crop_window.item (3) >= 0 and
							crop_window.item (3) <= crop_window.item (4) and
							crop_window.item (4) <= 1.0))
	screen_window_validity: (screen_window.item (1) < screen_window.item (2) and
							 screen_window.item (3) < screen_window.item (4))
	frame_aspect_priority: not frame_aspect_set implies
	(frame_aspect_ratio-((x_resolution*pixel_aspect_ratio)/y_resolution)).abs < 0.000001
	depth_of_field_validity: (focal_distance >= 0.0 and focal_length >= 0.0 and
							  f_stop > 0.0)
	clipping_order: near_clipper < far_clipper
	shutter_parameter_validity: shutter_open >= 0 and
								shutter_open <= shutter_close
	world_to_camera_not_void: world_to_camera /= Void

end -- SCENE_CAMERA

