

indexing
	description: "Spherical Bounding Volumes"
	patterns: "none"
	date: "$Date: 2006-12-06 13:14:08 -0500 (Wed, 06 Dec 2006) $"
	revision: "$Revision: 271 $"
	status: "unreviewed"




class 

	SCENE_SPHERE_BOUND
	
inherit
	
	SCENE_BOUND
		redefine
			default_create, intersect_ray
		end
	
	BASIC_NUMERIC
		undefine
			out
		redefine
			default_create
		end

creation
	default_create, make, make_with_point_and_radius, make_infinite

feature -- Initialization

	default_create is
			-- Create an empty bounding sphere
		do
			create the_origin.default_create
			the_radius := 0.0
			
			reset_intersections		
			
			Precursor {SCENE_BOUND}
		end
	
	make (first_corner, second_corner: POINT3) is
			-- Arbitrary-ordered vertices to be bounded.
		do
			make_with_point_and_radius (first_corner.midpoint (second_corner),
										first_corner.distance (second_corner)/2)
		end

	make_with_point_and_radius (new_origin: POINT3; new_radius: DOUBLE) is
			-- Create a new bounding sphere with `new_radius' at 
			-- `new_origin'.
		require 
			new_radius_strictly_positive: new_radius > 0.0
		do
			default_create
			the_radius := new_radius
			the_origin.copy (new_origin)
		ensure 
			finite_bound: not empty and not infinite
			radius_set: about_equal (new_radius, radius)
			origin_set: origin.is_equal (new_origin)
		end 

	make_infinite is
			-- Create an infinite bounding sphere
		do
			default_create
			the_radius := -1.0
		ensure
			is_infinite: infinite
		end

feature -- Access
	
	length, radius: DOUBLE is
			-- The radius of the bounds
		do
			Result := the_radius
		end
	
	origin, centroid: POINT3 is
			-- The origin of the bounds
		do
			Result := the_origin
		end
	
feature -- Measurement
feature -- Comparison
feature -- Status report
	
	contains (point: POINT3): BOOLEAN is
			-- Does this object contain `point'?
		local
			v: VECTOR3
		do
			if infinite then
				Result := true
			elseif empty then
				-- only true if `point' and `origin' are the sme object
				Result := point.is_equal (the_origin)
			else
				v := origin - point
				Result := v.magnitude < the_radius
			end
		end 

	empty: BOOLEAN is
			-- Is the bounds empty?
		do
			Result := about_equal (the_radius, 0.0)
		end 

	infinite: BOOLEAN is
			-- Is the bounds infinite?
		do
			Result := the_radius < 0.0
		end 

feature -- Status setting
	
	set_radius (new_radius: DOUBLE) is
			-- Set `radius' to `new_radius'
		require 
			new_radius_strictly_positive: new_radius > 0.0
		do
			the_radius := new_radius
		ensure 
			finite_bound: not empty and not infinite
			radius_set: about_equal (new_radius, radius)
		end 

	set_origin (new_origin: POINT3) is
			-- Set `origin' to `new_origin'.
		require
			not_infinite: not infinite
		do
			the_origin := new_origin
		ensure 
			origin_set: origin.is_equal (new_origin)
		end 

feature -- Cursor movement
feature -- Element change
	
	expand_to_include (bound: SCENE_BOUND) is
			-- Adapt this bounds to include `bound'
		local
			new_radius, delta, dist: DOUBLE
			v: VECTOR3
		do
			-- if I'm infinite, I encompass everything, a NOP
			-- Check the other bound
			if bound.infinite then
				make_infinite
			elseif bound.empty then
				-- NOP
			elseif empty then
				if not bound.empty then
					make_with_point_and_radius	(bound.centroid, bound.length)
				end
			else
				v := centroid - bound.centroid
				dist := v.magnitude
				
				-- Does `bound' contain me?
				if bound.length > (dist + radius) then
					make_with_point_and_radius	(bound.centroid, bound.length)
				-- Do I contain `bound'?
				elseif radius > (dist + bound.length) then
					-- NOP
				else
					if dist > Tolerance then
						new_radius := 0.5 * (radius + dist + bound.length)
						delta := new_radius - bound.length
						
						the_origin := bound.centroid + v.scale (delta/dist)
						the_radius := new_radius + Tolerance
					else
						if radius > bound.length then
							the_radius := radius + Tolerance
						else
							the_radius := bound.length + Tolerance
						end
					end
				end
			end
		end 

	expand_to_include_point (point: POINT3) is
			-- Adapt this bounds to include `point'
		local
			dist: DOUBLE
			v: VECTOR3
			aabb: AXIS_ALIGNED_BOX
		do
			-- if I'm infinite, I encompass everything, a NOP
			if empty then
				if not point.is_equal (the_origin) then
					v := origin - point
					dist := v.magnitude
					the_origin := the_origin.midpoint (point)
					the_radius := dist/2.0 + Tolerance
				end
			else
				-- Check if the point is outside the sphere
				if not contains (point) then
					create v.make (1, 1, 1)
					v.normalize
					-- Circumscribe a box; radius of box 2*R/sqrt(3)
					create aabb.make (the_origin + v.scale (-1.1547006 * the_radius),
									  the_origin + v.scale (1.1547006 * the_radius))
					aabb.grow (point)
					the_origin := aabb.centroid
					the_radius := 0.5 * ((aabb.upper - aabb.lower).magnitude) + Tolerance
				end
			end
		end
	
	expand_to_include_points (points: POINT3_ARRAYED_LIST) is
			-- Adapt this bounds to include `point'
		local
			dist: DOUBLE
			v: VECTOR3
			aabb: AXIS_ALIGNED_BOX
		do
			-- if I'm infinite, I encompass everything, a NOP
			if not infinite then
				if empty then
					v := origin - points.item (1)
					dist := v.magnitude
					the_origin := the_origin.midpoint (points.item (1))
					the_radius := dist/2.0 + Tolerance
				end
				
				aabb := circumscribed_box
				
				from
					points.start
				until
					points.after
				loop
					
					if not aabb.has (points.item_for_iteration) then
						aabb.grow (points.item_for_iteration)
					end
					
					points.forth
				end
				
				the_origin := aabb.centroid
				the_radius := 0.5 * ((aabb.upper - aabb.lower).magnitude) + Tolerance
			end
		end
	
feature -- Removal
feature -- Resizing
feature -- Transformation
feature -- Conversion
	
	circumscribed_box: AXIS_ALIGNED_BOX is
			-- A box that circumscribes the sphere, i.e., the sphere is 
			-- inscribed in the box; radius of box 2*R/sqrt(3)
		local
			v: VECTOR3
		do
			create v.make (1, 1, 1)
--			v.normalize
			
			create Result.make (the_origin + v.scale (-the_radius),
							  the_origin + v.scale (the_radius))
		end
	
	out: STRING is
			-- Stringified version of `Current'
		do
			create Result.make (30)
			if infinite then
				Result.append ("infinite")
			elseif empty then
				Result.append ("empty")
			else
				Result.append ("origin = ")
				Result.append (origin.out)
				Result.append (", radius = ")
				Result.append_string (radius.out)
			end
		end 

feature -- Duplication
feature -- Basic operations
	
	apply (transform: TRANSFORM): like Current is
			-- Apply `transform' to `Current'
		local
			v: VECTOR3
			r: DOUBLE
		do
			if empty then
				create Result
				Result.set_origin (transform.apply (origin))
			elseif infinite then
				create Result.make_infinite
			else
				create v.make (radius, 0, 0)
				r := transform.apply_as_direction (v).magnitude_squared
				v.make (0, radius, 0)
				r := r.max (transform.apply_as_direction (v).magnitude_squared)
				v.make (0, 0, radius)
				r := r.max (transform.apply_as_direction (v).magnitude_squared)
				
				create Result.make_with_point_and_radius  (transform.apply (origin), square_root (r))
			end 
		end

	draw_immediate (renderer: RENDERER) is
			-- Apply the renderer to this object
		local
			draw_style: INTEGER
		do
			draw_style := renderer.draw_style
			renderer.set_draw_style (renderer.rd_line)
			renderer.disable (renderer.rd_lighting)
			
			renderer.push_matrix
			renderer.translate (origin.x, origin.y, origin.z)
			
			renderer.enable (renderer.rd_color_material_constant)
			renderer.color_material (renderer.rd_front_and_back,
							   renderer.rd_ambient_and_diffuse)
			
			renderer.rgba_color (green)
			renderer.sphere (radius, -radius, radius, 2 * pi,
											0, 1, 0, 1)
			renderer.pop_matrix
			
			renderer.enable (renderer.rd_lighting)
			renderer.set_draw_style (draw_style)
		end
	
	intersect_ray (ray: RAY) is
			-- Intersect with `ray'.
			-- If `ray' does not intersect then `intersected' will be False
			-- else `intersected' will be True and `intersection' will yield
			-- the ordered parametric coordinates along `ray' of each
			-- successive point of intersection.
		local
			origin_to_center: VECTOR3
			dist_to_center, radius_squared, closest_approach: DOUBLE 
			half_chord_squared: DOUBLE
		do
			Precursor (ray)
			reset_intersections
			
			if empty then
				outside := true
			elseif infinite then
				inside := true
			else
				origin_to_center := origin - ray.origin
				dist_to_center := origin_to_center # origin_to_center
				radius_squared := radius * radius
				
				if dist_to_center < radius_squared then
					inside := true
				else
					closest_approach := origin_to_center # ray.direction
					if closest_approach < 0.0 then
						outside := true
					else
						half_chord_squared := radius_squared - dist_to_center + 
							closest_approach * closest_approach
						if half_chord_squared < 0.0 then
							outside := true
						else
							intersected := true
						end
					end
				end
			end
			
			if not outside then
				intersection_hit := intersection_hit + 1
			end
		ensure then
			only_one_true: (inside implies (not outside and not intersected)) or
						   (outside implies (not inside and not intersected)) or
						   (intersected implies (not outside and not inside))
		end

	intersect_plane (plane: PLANE) is
			-- Intersect with `plane'.
			-- One and only one of the following will be true, 
			-- `inside', `outside', or `intersected'. `inside' is 
			-- defined by the normal direction
		local
			p: DOUBLE
		do
			reset_intersections
			
			if infinite then
				intersected := true
			elseif empty then
				outside := true
			else
				p := plane.distance_from_point (origin)
				if p - radius > 0.0 then
					inside := true
				elseif p + radius < 0.0 then
					outside := true
				else
					intersected := true
				end
			end
		end

	intersect_polyhedron (poly: POLYHEDRON) is
			-- Intersect with `plane'.
			-- One and only one of the following will be true, 
			-- `inside', `outside', or `intersected'. `inside' is 
			-- defined by the normal direction
		do
			check
				implemented: false
			end
		end

feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {NONE} -- Implementation
	
	the_radius: DOUBLE
	the_origin: POINT3
	
	reset_intersections is
			-- Reset the location of the intersection queries to false
		do
			intersected := false
			outside := false
			inside := false
		end
	
invariant
	
	empty_implies_zero_radius: empty implies the_radius <= Tolerance
	empty_implies_not_infinite: empty implies not infinite
	infinite_implies_not_empty: infinite implies not empty
	finite_bounds: radius > 0.0 implies not empty and not infinite
	infinite_implies_origin: infinite implies origin.is_zero
	
end -- SCENE_SPHERE_BOUND

