

indexing
	description: "Defines a 3D view volume."
	patterns: "none"
	date: "$Date: 2006-12-06 13:14:08 -0500 (Wed, 06 Dec 2006) $"
	revision: "$Revision: 271 $"
	status: "unreviewed"




class 

	VIEW_VOLUME
	
inherit
	
	BASIC_NUMERIC
		redefine
			default_create
		end

creation
	default_create, make_as_orthographic, make_as_perspective

feature -- Initialization
	
	default_create is
		do
			make_as_orthographic (-1, 1, -1, 1, 0.1, 1000)
		end
	
	make_as_orthographic (left, right, bottom, top, near, far: DOUBLE) is
			-- Sets up an orthographic view volume with the given sides.
		require 
			proper_order_for_left_right: left < right
			proper_order_for_top_bottom: bottom < top					 
			proper_order_for_near_far:	 near < far
		do
			is_orthographic := true
			create projection_point.make (0, 0, 0)
			create projection_direction.make (0, 0, -1)
			create llf0.make (left, bottom, -near)
			create lrf0.make (right, bottom, -near)
			create ulf0.make (left, top, -near)
			
			llf := llf0
			lrf := lrf0
			ulf := ulf0
			
			near_clipping_plane := near
			depth := far - near

			needs_updating := true
		ensure 
			is_orthographic: is_orthographic
		end 
	
	make_as_perspective (y_fov, aspect, near, far: DOUBLE) is
			-- Sets up a perspective view volume with the given field of view
			-- and aspect ratio.
		require 
			y_fov_is_positive: y_fov > 0
			aspect_is_positive: aspect > 0
			proper_order_for_near_far:	 near < far
		local
			tan_fov: DOUBLE
		do
			tan_fov := tangent (y_fov / 2.0)
			
			is_orthographic := false

			create projection_point.make (0, 0, 0)
			create projection_direction.make (0, 0, -1)
			
			create ulf0.make (aspect * -near * tan_fov, near * tan_fov, -near)
			create llf0.make (aspect * -near * tan_fov, -ulf0.y, -near)
			create lrf0.make (-llf0.x, -ulf0.y, -near)
			
			llf := llf0
			lrf := lrf0
			ulf := ulf0
			
			near_clipping_plane := near
			depth := far - near

			needs_updating := true
		ensure 
			is_perspective: not is_orthographic
		end 

feature -- Access

	frustum: DS_ARRAYED_LIST [POINT3] is
			-- The points of the frustum
		local
			tmp_vec: VECTOR3
			tmp_pt: POINT3
			scale_factor: DOUBLE
		do
			create Result.make (8)

			create tmp_pt.make (ulf0.x, ulf0.y, ulf0.z)
			Result.force_last (tmp_pt)
			create tmp_pt.make (llf0.x, llf0.y, llf0.z)
			Result.force_last (tmp_pt)
			create tmp_pt.make (lrf0.x, lrf0.y, lrf0.z)
			Result.force_last (tmp_pt)
			create tmp_pt.make (lrf0.x, ulf0.y, llf0.z)
			Result.force_last (tmp_pt)
			scale_factor := (near_clipping_plane + depth) / near_clipping_plane
			tmp_vec := ulf0.scale (scale_factor)
			create tmp_pt.make (tmp_vec.x, tmp_vec.y, tmp_vec.z)
			Result.force_last (tmp_pt)
			tmp_vec := llf0.scale (scale_factor)
			create tmp_pt.make (tmp_vec.x, tmp_vec.y, tmp_vec.z)
			Result.force_last (tmp_pt)
			tmp_vec := lrf0.scale (scale_factor)
			create tmp_pt.make (tmp_vec.x, tmp_vec.y, tmp_vec.z)
			Result.force_last (tmp_pt)
			create tmp_vec.make (lrf0.x, ulf0.y, llf0.z)
			tmp_vec := tmp_vec.scale (scale_factor)
			create tmp_pt.make (tmp_vec.x, tmp_vec.y, tmp_vec.z)
			Result.force_last (tmp_pt)
		ensure 
			Result_not_void: Result /= Void
		end
		
	affine_transformation: TRANSFORM is
			-- an affine transformation corresponding to the view volume
		do
			if needs_updating then
				needs_updating := false
				compute_affine_transformation
				compute_projection_transformation
			end
			Result := affine_matrix
		ensure 
			Result_not_void: Result /= Void
		end
	
	
	projection_transformation: TRANSFORM is
			-- a projection transformation useful for a RENDERER
		do
			if needs_updating then
				needs_updating := false
				compute_affine_transformation
				compute_projection_transformation
			end
			Result := projection_matrix
		ensure 
			Result_not_void: Result /= Void
		end
	
	camera_space_transform: TRANSFORM is
			-- transforms the view volume into camera space: 
			-- it translates the view volume so the view point is at
			-- the origin, and rotates it so the view direction is along the
			-- negative z axis.
		local
			t: TRANSFORM
			q: QUATERNION
			v: VECTOR3
		do
			create t
			t.translate (-projection_point.x, -projection_point.y, -projection_point.z)
			create v.make (0, 0, -1)
			create q.make_from_vectors (projection_direction, v)
			Result := t * q.rotation_matrix
		ensure 
			Result_not_void: Result /= Void
		end

feature -- Measurement
feature -- Comparison
feature -- Status report
	
	is_orthographic: BOOLEAN
			-- Is this an orthographic view volume?
	
	projection_point: POINT3
			-- The center of the projection
	
	projection_direction: VECTOR3
			-- The direction of projection
	
	near_clipping_plane: DOUBLE
			-- distance from projection point to near plane
	
	width: DOUBLE is
			-- width of the viewing frustrum
		do
			Result := (lrf0-llf0).magnitude
		end
	
	height: DOUBLE is
			-- height of the viewing frustrum
		do
			Result := (ulf0-llf0).magnitude
		end
	
	depth: DOUBLE
			-- depth of the viewing frustrum
	

feature -- Status setting
feature -- Cursor movement
feature -- Element change
	
	narrow_volume_in_z (near, far: DOUBLE) is
			-- narrow view volume which contains as tightly as
			-- possible the given interval on the z axis (in eye space). 
		require 
			default: true
		do

		ensure 
			default: true
		end 
	
	scale (factor: DOUBLE) is
			-- Scales width and height of view volume by given factor
		require 
			factor_is_positive: factor > 0.0
		local
			diff: DOUBLE
			width_vec, height_vec: VECTOR3
			tmp: POINT3
		do
			diff := (1.0 - factor)/2.0
			
			width_vec := (lrf0 - llf0).scale (diff)
			height_vec := (ulf0 - llf0).scale (diff)
			
			llf0 := llf0 + height_vec + width_vec
			lrf0 := lrf0 + height_vec - width_vec
			ulf0 := ulf0 - height_vec + width_vec
			
			tmp := projection_point + llf0
			llf.make (tmp.x, tmp.y, tmp.z)
			tmp := projection_point + lrf0
			lrf.make (tmp.x, tmp.y, tmp.z)
			tmp := projection_point + ulf0
			ulf.make (tmp.x, tmp.y, tmp.z)

			needs_updating := true
		ensure 
			needs_updating: needs_updating
		end 

	scale_width (ratio: DOUBLE) is
			-- Scales view volume to be the given ratio of its current width
			-- or height, leaving the resulting view volume centered about the
			-- same point (in the near plane) as the current one.
		require 
			ratio_is_positive: ratio > 0.0
		local
			diff: DOUBLE
			width_vec: VECTOR3
			tmp: POINT3
		do
			diff := (1.0 - ratio)/2.0
			
			width_vec := (lrf0 - llf0).scale (diff)
			
			llf0 := llf0 + width_vec
			lrf0 := lrf0 + width_vec
			ulf0 := ulf0 - width_vec
			
			tmp := projection_point + llf0
			llf.make (tmp.x, tmp.y, tmp.z)
			tmp := projection_point + lrf0
			lrf.make (tmp.x, tmp.y, tmp.z)
			tmp := projection_point + ulf0
			ulf.make (tmp.x, tmp.y, tmp.z)

			needs_updating := true
		ensure 
			needs_updating: needs_updating
		end 

	scale_height (ratio: DOUBLE) is
			-- Scales view volume to be the given ratio of its current width
			-- or height, leaving the resulting view volume centered about the
			-- same point (in the near plane) as the current one.
		require 
			ratio_is_positive: ratio > 0.0
		local
			diff: DOUBLE
			height_vec: VECTOR3
			tmp: POINT3
		do
			diff := (1.0 - ratio)/2.0
			
			height_vec := (ulf0 - llf0).scale (diff)
			
			llf0 := llf0 + height_vec
			lrf0 := lrf0 + height_vec
			ulf0 := ulf0 - height_vec
			
			tmp := projection_point + llf0
			llf.make (tmp.x, tmp.y, tmp.z)
			tmp := projection_point + lrf0
			lrf.make (tmp.x, tmp.y, tmp.z)
			tmp := projection_point + ulf0
			ulf.make (tmp.x, tmp.y, tmp.z)

			needs_updating := true
		ensure 
			needs_updating: needs_updating
		end 

feature -- Removal
feature -- Resizing
feature -- Transformation
feature -- Conversion
feature -- Duplication
feature -- Basic operations
	
	project_point_to_line (pt: POINT2): RAY is
			-- Maps a 2d point (in 0 <= x,y <= 1) to a 3d ray (line).
		require 
			pt_is_unit_in_x: pt.x >= 0.0 and pt.x <= 1.0
			pt_is_unit_in_y: pt.y >= 0.0 and pt.y <= 1.0
		local
			pt_x, pt_y: DOUBLE
			x, y, z, w: DOUBLE
			m: TRANSFORM
			pt1, pt2: POINT3
			dir: VECTOR3
		do
			pt_x := 2.0 * pt.x - 1.0
			pt_y := 2.0 * pt.y - 1.0
			m := affine_transformation * projection_transformation
			m.invert
			
			check
				invert_failed: m.inverted
			end
			
			x := pt_x * m.a11 + pt_y * m.a12 - m.a13 + m.a14
			y := pt_x * m.a21 + pt_y * m.a22 - m.a23 + m.a24
			z := pt_x * m.a31 + pt_y * m.a32 - m.a33 + m.a34
			w := pt_x * m.a41 + pt_y * m.a42 - m.a43 + m.a44
			pt1.make (x/w, y/w, z/w)
			
			x := x + 2 * m.a13
			y := y + 2 * m.a23
			z := z + 2 * m.a33
			w := w + 2 * m.a43
			pt2.make (x/w, y/w, z/w)
			
			dir := pt2 - pt1
			dir.normalize
			Result.make (pt1, dir)
		ensure 
			Result_not_void: Result /= Void
		end 
	
	project_to_screen (src: POINT3): POINT2 is
			-- Maps the 3d point in world coordinates to a 2d point in
			-- normalized screen coordinates (0 <= x,y,z <= 1, 0 <= z <= 1).
			-- The z-screen coordinate represents the homogonized z coordinate
			-- which goes (non-linearly) from 0 at the near clipping plane to
			-- 1 at the far clipping plane.
		local
			mat: TRANSFORM
			tmp: POINT3
		do
			mat := affine_transformation * projection_transformation
			tmp := mat.apply (src)
			
			create Result.make ((1.0 + tmp.x)/2.0, (1 + tmp.y)/2.0)
		ensure 
			Result_is_unit_in_x: Result.x >= 0.0 and Result.x <= 1.0
			Result_is_unit_in_y: Result.y >= 0.0 and Result.y <= 1.0
		end 
	
	sight_point (distance_from_eye: DOUBLE): POINT3 is
			-- the point along the line of sight at the given distance
			-- from the projection point (eye)
		require 
			default: true
		local
			r: RAY
		do
			r.make (projection_point, projection_direction)
			Result := r.project (distance_from_eye)
		ensure 
			Result_not_void: Result /= Void
		end 
	
	aligning_transformation: TRANSFORM is
			-- a rotation (transformation) that would align a viewed object 
			-- so that its positive x-axis (of its object space) is to the 
			-- right in the view and it's positive y-axis is up.
		local
			vec, y_axis, up, right, new_right: VECTOR3
			q, q1: QUATERNION
			rot: TRANSFORM
		do
			up := ulf0 - llf0
			right := lrf0 - llf0
			up.normalize
			right.normalize
			
			q.make_from_vectors (y_axis, up)
			rot := q.rotation_matrix
			vec.make (1.0, 0.0, 0.0)
			new_right := rot.apply_as_direction (vec)
			q1.make_from_vectors (new_right, right)
			
			Result := (q + q1).rotation_matrix
		ensure 
			Result_not_void: Result /= Void
		end 
	
	world_to_screen_scale (world_center: POINT3;
						   normalized_radius: DOUBLE): DOUBLE is
			-- a scale factor that would scale a unit sphere centered
			-- at worldCenter so that it would appear to have the given radius
			-- in normalized screen coordinates when projected onto the near 
			-- plane
		require 
			default: true
		do

		ensure 
			default: true
		end 
	
	rotate_camera (xform: TRANSFORM) is
			-- Rotate the camera view direction.
		require
			xform_not_void: xform /= Void
		local
			tmp_pt: POINT3
		do
			projection_direction := xform.apply_as_direction (projection_direction)
		
			llf0 := xform.apply_as_direction (llf0)
			lrf0 := xform.apply_as_direction (lrf0)
			ulf0 := xform.apply_as_direction (ulf0)
			
			tmp_pt := projection_point + llf0
			llf.make (tmp_pt.x, tmp_pt.y, tmp_pt.z)
			tmp_pt := projection_point + lrf0
			lrf.make (tmp_pt.x, tmp_pt.y, tmp_pt.z)
			tmp_pt := projection_point + ulf0
			ulf.make (tmp_pt.x, tmp_pt.y, tmp_pt.z)

			needs_updating := true
		ensure 
			needs_updating: needs_updating
		end 
	
	translate_camera (vec: VECTOR3) is
			-- Translate the camera viewpoint.
		require 
			vec_not_void: vec /= Void
		do
			projection_point := projection_point + vec
			llf := llf + vec
			lrf := lrf + vec
			ulf := ulf + vec

			needs_updating := true
		ensure 
			needs_updating: needs_updating
		end 
	
	z_axis: VECTOR3 is
			-- The positive z axis in eye space.
		local
			p: PLANE
			pt1, pt2, pt3: POINT3
		do
			pt1.make (llf0.x, llf0.y, llf0.z)
			pt2.make (ulf0.x, ulf0.y, ulf0.z)
			pt3.make (lrf0.x, lrf0.y, lrf0.z)
			
			p.make_from_points (pt1, pt2, pt3)
			
			Result := -p.normal
		end 

feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {NONE} -- Implementation
	
	affine_matrix, projection_matrix: TRANSFORM
			-- Precomputed transformations

	needs_updating: BOOLEAN
			-- Do the matricies need updating
	
	transform_volume (xform: TRANSFORM) is
			-- Transforms the view volume by the given matrix
		do
			
		ensure 
			default: true
		end 

	llf0: VECTOR3
			-- Points on the near clipping plane.
			-- x = -w, y = -w, z = -w
	
	lrf0: VECTOR3
			-- Points on the near clipping plane.
			-- x =	w, y = -w, z = -w
	
	ulf0: VECTOR3
			-- Points on the near clipping plane.
			-- x = -w, y =	w, z = -w
	
	llf: VECTOR3
	lrf: VECTOR3
	ulf: VECTOR3
	
	compute_affine_transformation is
			-- compute an affine transformation corresponding to the view volume
		local
			skew_mat, skew_mat_inverse, tmp_mat: TRANSFORM
			right, up: VECTOR3
			eye, tmp: POINT3
		do
			right := lrf0 - llf0
			up := ulf0 - llf0
			
			right.normalize
			up.normalize
			
			create skew_mat.make_with_elements (right.x, right.y, right.z, 0,
										 up.x, up.y, up.z, 0,
										 -projection_direction.x, -projection_direction.y, -projection_direction.z, 0,
										  0, 0, 0, 1)
			skew_mat.transpose
			skew_mat_inverse := skew_mat.inverse
			
			check
				inverse_failed: skew_mat_inverse.inverted
			end
			
			tmp := -(projection_point + llf0)
			create affine_matrix
			affine_matrix.translate (tmp.x, tmp.y, tmp.z)
			affine_matrix := skew_mat_inverse * affine_matrix
			
			eye := affine_matrix.apply (projection_point)
			create tmp_mat
			tmp_mat.translate (-eye.x, -eye.y, -eye.z)
			affine_matrix := tmp_mat * affine_matrix
		end
	
	
	compute_projection_transformation is
			-- compute a projection transformation useful for a RENDERER
		local
			skew_mat, skew_mat_inverse: TRANSFORM
			right, up: VECTOR3
			llf_eye, lrf_eye, ulf_eye: VECTOR3
			w, h: DOUBLE
			right_minus_left, right_plus_left: DOUBLE
			top_minus_bottom, top_plus_bottom: DOUBLE
			far, far_plus_near, far_minus_near: DOUBLE
		do
			right := lrf0 - llf0
			up := ulf0 - llf0
			
			w := right.magnitude
			h := up.magnitude
			
			create skew_mat
			skew_mat.make_with_elements (right.x/w, right.y/w, right.z/w, 0,
										 up.x/h, up.y/h, up.z/h, 0,
										 -projection_direction.x,
										  -projection_direction.y,
										   -projection_direction.z, 0,
											0, 0, 0, 1)
			skew_mat.transpose
			skew_mat_inverse := skew_mat.inverse
			
			check
				inverse_failed: skew_mat_inverse.inverted
			end
			
			llf_eye := skew_mat_inverse.apply_as_direction (llf0)
			lrf_eye := skew_mat_inverse.apply_as_direction (lrf0)
			ulf_eye := skew_mat_inverse.apply_as_direction (ulf0)
			
			create projection_matrix.make
			right_minus_left := lrf_eye.x - llf_eye.x
			right_plus_left := lrf_eye.x + llf_eye.x
			
			top_minus_bottom := ulf_eye.y - llf_eye.y
			top_plus_bottom := ulf_eye.y + llf_eye.y
			
			far_minus_near := depth
			far := near_clipping_plane + depth
			far_plus_near := far + near_clipping_plane
			
			debug ("graphics")
				io.put_string ("Debug - perspective %N")
				io.put_double (right_minus_left)
				io.put_string ("%N")
				io.put_double (right_plus_left)
				io.put_string ("%N")
				io.put_double (top_minus_bottom)
				io.put_string ("%N")
				io.put_double (top_plus_bottom)
				io.put_string ("%N")
				io.put_double (far_minus_near)
				io.put_string ("%N")
				io.put_double (far_plus_near)
				io.put_string ("%N")
			end
			
			if is_orthographic then
				create projection_matrix.make_with_elements ( 2.0/right_minus_left, 0, 0, 0,
											0, 2.0/top_minus_bottom, 0, 0,
											0, 0, -2.0/far_minus_near, 0, 
				
											-right_plus_left / right_minus_left, 
											 -top_plus_bottom / top_minus_bottom, 
											  -far_plus_near / far_minus_near, 1)
				projection_matrix.transpose
			else
				create projection_matrix.make_with_elements (2 * near_clipping_plane / right_minus_left, 0, 0, 0,
										   0, 2 * near_clipping_plane / top_minus_bottom, 0, 0,
										   right_plus_left / right_minus_left, top_plus_bottom / top_minus_bottom, 
										   -far_plus_near / far_minus_near, -1.0, 
										   0, 0, -2 * near_clipping_plane * far / far_minus_near, 0)
				projection_matrix.transpose
			end
		end
	
invariant
	
	projection_point_not_void: projection_point /= Void
	projection_direction_not_void: projection_direction /= Void
	llf0_not_void: llf0 /= Void
	lrf0_not_void: lrf0 /= Void
	ulf0_not_void: ulf0 /= Void
	temporaries_not_void: llf /= Void and ulf /= Void and lrf /= Void
	
end -- VIEW_VOLUME

