

indexing
	description: "Samplers that generate samples by taking a single sample in each strata of a DOMAIN"
	names: "sampler", "stratification"
	representation: "list"
	access: "index", "cursor"
	size: "resizable"
	contents: "POINT"
	patterns: "none"
	date: "$Date: 2006-06-23 10:34:45 -0400 (Fri, 23 Jun 2006) $"
	revision: "$Revision: 226 $"
	author: "Mark A Bolstad"
	status: "unreviewed"




deferred class 

	STRATIFIED_SAMPLER

inherit
	
	BASIC_NUMERIC
		undefine
			default_create
		end
	
	SAMPLER

feature -- Initialization

	make (d: DOMAIN) is
			-- Create a `SAMPLER' object on the domain
		deferred
		ensure 
			domain_set: domain /= Void
		end

feature -- Access

	sample_number: INTEGER is
			-- The sample index 
		do
			Result := (index + 1) // domain.dimensions
		ensure
			in_range: Result >= 1 and Result <= sample_points.count // domain.dimensions
		end
	
	item: POINT is
			-- The current sample point
		local
			i, offset: INTEGER
		do
			from
				offset := 1
			invariant
				offset_at_least_one: offset >= 1
				offset_small_enough: offset <= 1 + domain.dimensions
			variant
				offset_increasing: 1 + domain.dimensions - offset
			until
				offset >= domain.dimensions
			loop
				current_point.put (sample_points.item (index + offset - 1), offset)
				offset := offset + 1
			end

			Result := current_point
		end 

feature -- Measurement
feature -- Comparison
feature -- Status report

	after: BOOLEAN is
			-- Is the cursor off the end?
		do
			Result := index > sample_points.count
		end
	
	end_of_reporting_interval: BOOLEAN is
			-- Is this sample at the end of the first dimension in `domain'?
		do
			Result := (index // domain.dimensions) \\ reporting_interval = 0
		end 

	using_hilbert_access: BOOLEAN
			-- Are we accessing the samples using a hilbert 
			-- space-filling curve?

	use_jittered_samples: BOOLEAN
			-- Should each sample be jittered within a strata?
	
feature -- Status setting
	
	set_hilbert_access (on_or_off: BOOLEAN) is
			-- Set whether or not to use hilbert sampling patterns
		do
			using_hilbert_access := on_or_off
		ensure 
			using_hilbert_access_set: using_hilbert_access = on_or_off
		end 

	set_jittering (on_or_off: BOOLEAN) is
			-- Jitter the samples
		do
			use_jittered_samples := on_or_off
		ensure
			use_jittered_samples_set: use_jittered_samples = on_or_off
		end
	
feature -- Cursor movement
	
	start is
			-- Move to the beginning of the samples
		do
			index := 1
		end
	
	forth is
			-- Move to the next sample
		do
			index := index + domain.dimensions
		end
	
feature -- Element change
feature -- Removal
feature -- Resizing
feature -- Transformation
feature -- Conversion
feature -- Duplication
feature -- Basic operations

	generate_samples is
			-- Generate the initial sampling pattern
		do
			index := 1
			if using_hilbert_access then
				generate_hilbert_samples
			else
				generate_uniform_samples
			end
		ensure 
			index_set: index = 1
		end 

feature -- Miscellaneous
feature -- Obsolete
feature -- Inapplicable
feature {NONE} -- Implementation

	sample_points: ARRAY [DOUBLE]
	index: INTEGER

	current_point: POINT

	random: RANDOM_GENERATOR
	
	generate_uniform_samples is
			-- Generate the initial samples
		deferred
		ensure 
			default: true
		end 

	generate_hilbert_samples is
			-- Generate the initial samples using a hilbert 
			-- space-filling curve as the access pattern
		deferred
		ensure 
			default: true
		end 

	hilbert_coordinates (dims, index1, index2: INTEGER): INTEGER is
		external
			"C macro use <hilbert.h>"
		alias
			"hilbert_coordinates"
		end
	
	hilbert_index (dims: INTEGER; coords: POINTER): INTEGER is
		external
			"C macro use <hilbert.h>"
		alias
			"hilbert_index"
		end

invariant

	samples_not_void: sample_points /= Void
	valid_generator: random /= Void and random.is_unit_distribution

end -- STRATIFIED_SAMPLER

