

indexing
	description: "Viewer class for observing geometry and bucketing behavior"
	names: "viewer", "3d", "bucket"
	representation: "none"
	access: "fixed"
	size: "resizable"
	contents: "SCENE_*"
	patterns: "none"
	date: "$Date: 2007-04-07 10:37:11 -0400 (Sat, 07 Apr 2007) $"
	revision: "$Revision: 287 $"
	author: "Mark A. Bolstad"
	status: "unreviewed"




class

	BUCKET_VIEWER

inherit

	SCENE_VIEWER
		rename
			make as viewer_make
		redefine
			on_expose_event, on_key_press_event
		end

creation
	make

feature -- Initialization

	make (the_micropoly_renderer: MICROPOLYGON_RENDERER) is
			-- Create a viewer.
		do
			micropoly_renderer := the_micropoly_renderer
			update_draw_pool := true
			make_with_camera (micropoly_renderer.camera)
			shading_rate := 1
		end

feature -- Status Report

	show_bounds: BOOLEAN

feature -- Status Setting

	set_show_bounds (on_or_off: BOOLEAN) is
			-- Set `show_bounds' to `on_or_off'
		do
			show_bounds := on_or_off
		ensure
			show_bounds_set: show_bounds = on_or_off
		end

feature -- Basic operations

	split is
			-- Split the primitives
		do
			split_objects := true
			update_draw_pool := true
		end

	on_expose_event (a_widget: GTK_WIDGET; a_event: GDK_EXPOSE_EVENT): BOOLEAN is
			-- Called when we need to redraw the GTKGL_widget
		local
			the_time: DOUBLE
			geometry: SCENE_DRAWABLE_ELEMENT
			the_extent: SCENE_BOUND
			count_string: STRING
			i: INTEGER
		do
			if update_draw_pool then
				culler.default_create
				culler.set_camera (camera)
				culler.set_culling (true)

				root_node.traverse (culler)
				draw_pool ?= culler.pool
				culled_draw_pool ?= culler.culled_pool
			end

			gl_drawer.begin_statistics

			if gl_begin then
				gl_drawer.enable_two_sided_lighting
				gl_drawer.set_scene_ambient_color (scene_ambient_color)
				gl_drawer.enable (gl_drawer.rd_normalize)
				gl_drawer.disable (gl_drawer.rd_cull_face)
				gl_drawer.enable (gl_drawer.rd_depth_test)
				gl_drawer.enable (gl_drawer.rd_lighting)
				gl_drawer.clear (gl_drawer.integer_or (gl_drawer.rd_color_buffer_bit,
													   gl_drawer.rd_depth_buffer_bit))

				gl_drawer.set_global_lights (culler.global_lights)
				gl_drawer.global_lights.reset (gl_drawer)

					check
						pool_not_void: draw_pool /= Void
					end

				if split_objects then
					process_split (draw_pool)
					split_objects := false
					update_draw_pool := false
				end

				camera.draw (draw_pool, gl_drawer)

				timer.stop_timer

				create count_string.make (30)
				count_string.append_string ("Viewable: ")
				count_string.append_string (draw_pool.count.out)
				count_string.append_string (", Culled: ")
				count_string.append_string (culled_draw_pool.count.out)
				count_string.append_string (", Shading Rate: ")
				output_information_string (count_string, shading_rate)
				if show_information then
					the_time := 1.0 / timer.elapsed_time
					if show_as_frames_per_sec then
						output_information_string ("Frames per sec", the_time)
					else
						the_time := the_time * gl_drawer.polygon_count
						output_information_string ("Polygons per sec", the_time)
					end
				end

				if show_bounds then
					from
						draw_pool.start
					until
						draw_pool.after
					loop
						geometry := draw_pool.item_for_iteration.geometry
						if not geometry.is_annotation then
							the_extent := geometry.extents
							draw_bounds (the_extent, draw_pool.item_for_iteration.transform)
						end
						draw_pool.forth
					end
				end

				timer.start_timer

				swap_buffers

				debug ("graphics")
					io.put_string ("Polygon count = ")
					io.put_integer (gl_drawer.polygon_count)
					io.put_string ("%N")
				end

				gl_end
			end
		end

	draw_bounds (the_extent: SCENE_BOUND; transform: TRANSFORM) is
			-- Draw a bounding box
		require
			the_extent_not_void: the_extent /= Void
			transform_not_void: transform /= Void
		local
			bound: AXIS_ALIGNED_BOX
		do
			bound := the_extent.circumscribed_box
			bound.transform (camera.world_to_raster * transform)

			draw_in_screenspace (gl_drawer, bound)
		end

feature -- Miscellaneous

	on_key_press_event (a_widget: GTK_WIDGET; a_event: GDK_KEY_EVENT): BOOLEAN is
			-- Handle key press events
		do
			Result := Precursor (a_widget, a_event)

			if a_event.keyval = key_lower_s.code then
				split
				if shading_rate > 1 then
					shading_rate := shading_rate / 2.0
				end
				a_widget.queue_draw
			elseif a_event.keyval = key_upper_s.code then
				split
				shading_rate := shading_rate * 2.0
				a_widget.queue_draw
			elseif a_event.keyval = key_upper_r.code then
				split
				a_widget.queue_draw
			elseif a_event.keyval = key_lower_b.code then
				set_show_bounds (not show_bounds)
				a_widget.queue_draw
			elseif a_event.keyval = key_lower_r.code then
				update_draw_pool := true
				a_widget.queue_draw
			elseif a_event.keyval = key_lower_n.code then
				view_normals := not view_normals
				split
				a_widget.queue_draw
			end

			Result := true
		end

feature -- Obsolete
feature -- Inapplicable
feature {NONE} -- Implementation

	micropoly_renderer: MICROPOLYGON_RENDERER

	split_objects: BOOLEAN
			-- Should I split the objects?

	shading_rate: DOUBLE
			-- What should I use for a shading_rate in the
			-- SCENE_GEOMETRY?

	update_draw_pool: BOOLEAN
			-- Should we force the culler to update the draw_pool

	view_normals: BOOLEAN
			-- Should we draw normals at the dice points?

	draw_pool: SCENE_DRAWABLE_POOL
	culled_draw_pool: SCENE_DRAWABLE_POOL
			-- The pool of drawable elements

	process_split (the_draw_pool: SCENE_DRAWABLE_POOL) is
			-- Process the objects in `the_draw_pool' for splitting
		require
			the_draw_pool_not_void: the_draw_pool /= Void
		local
			queue: DS_LINKED_QUEUE [SCENE_DRAWABLE_ELEMENT]
			cull_queue: DS_LINKED_QUEUE [INTEGER]
			collection: SCENE_GEOMETRIC_COLLECTION
			element: SCENE_DRAWABLE_ELEMENT
			cull_state, split_culled: INTEGER
			test_culler: CULL_ACTION
			normal_pool: SCENE_DRAWABLE_POOL
			sd: SCENE_DRAWABLE
			-- Debug related variables
			left: SCENE_GEOMETRY
			i: INTEGER
			colors: SCENE_COLOR_ELEMENT
			rnd: UNIFORM_RANDOM_GENERATOR
			r, g, b, a: DOUBLE
			color: RGBA_COLOR
		do
			create queue.make
			create cull_queue.make
			create rnd.make
			create test_culler.make (camera)

			if view_normals then
				create normal_pool.make (10)
			end

			split_culled := 0
			from
				the_draw_pool.start
			until
				the_draw_pool.after
			loop
					check
						queue_is_empty: queue.count = 0
					end

				if not the_draw_pool.item_for_iteration.geometry.is_annotation then
					create collection
					if the_draw_pool.item_for_iteration.geometry.is_split then
						the_draw_pool.item_for_iteration.geometry.release
					end
					
					queue.put (the_draw_pool.item_for_iteration.geometry)
					cull_queue.put (test_culler.partial)
					queue.item.set_split_direction (false)
					queue.item.set_shading_rate (shading_rate)

					from
					until
						queue.is_empty
					loop
						element := queue.item
						queue.remove
						element.test_diceability (micropoly_renderer)
						if not element.diceable then
							element.split
							if cull_queue.item = test_culler.inside then
								from
									element.split_items.start
								until
									element.split_items.after
								loop
									queue.put (element.split_items.item_for_iteration)
									cull_queue.put (test_culler.inside)

									element.split_items.forth
								end
							else
								test_culler.push_transform (element.object_to_world)
								from
									element.split_items.start
								until
									element.split_items.after
								loop
									cull_state := test_culler.cull_state (element.split_items.item_for_iteration.extents.circumscribed_box)
									if cull_state /= test_culler.outside then
										queue.put (element.split_items.item_for_iteration)
										cull_queue.put (cull_state)
									else
										split_culled := split_culled + 1
									end

									element.split_items.forth
								end
								test_culler.pop_transform
							end
						else
							if view_normals then
								sd := the_draw_pool.item_for_iteration.twin
								sd.replace_first (generate_normals (element))
								normal_pool.force_last (sd)
							end
							collection.add_child (element)
						end
						cull_queue.remove

							check
								queue_same_size: queue.count = cull_queue.count
							end
					end

					from
						i := 1
						a := 1.0
					invariant
						i_at_least_one: i >= 1
						i_small_enough: i <= 1 + collection.count
					variant
						i_increasing: 1 + collection.count - i
					until
						i >= collection.count
					loop
						left ?= collection.item (i)

						create colors
						r := rnd.item
						rnd.forth
						g := rnd.item
						rnd.forth
						b := rnd.item
						create color.make (r, g, b, a)
						colors.make (color)
						left.replace_element (colors)

						i := i + 1
					end

					io.put_string (collection.count.out)
					io.put_string (" splits generated, ")
					io.put_string (split_culled.out)
					io.put_string (" culled.%N")


					the_draw_pool.item_for_iteration.replace_first (collection)
				end
				the_draw_pool.forth
			end

			if view_normals then
				the_draw_pool.append_last (normal_pool)
			end
		ensure
			default: true
		end

	generate_normals (element: SCENE_DRAWABLE_ELEMENT): SCENE_LINES is
			-- Generate a set of normals for `element'
		require
			element_not_void: element /= Void
		local
			grid: MICROPOLYGON_GRID
			v: VECTOR3
			i, j: INTEGER
			sd: SHADING_DATA
			p: POLYLINE
			vals: DS_ARRAYED_LIST [POLYLINE]
		do
			create grid.make (element.requested_grid_x_dim, element.requested_grid_y_dim)
			element.dice (grid, false)
			create vals.make (grid.width * grid.height)

			from
				i := 1
			invariant
				i_at_least_one: i >= 1
				i_small_enough: i <= 1 + grid.width
			variant
				i_increasing: 1 + grid.width - i
			until
				i > grid.width
			loop
				from
					j := 1
				invariant
					j_at_least_one: j >= 1
					j_small_enough: j <= 1 + grid.height
				variant
					j_increasing: 1 + grid.height - j
				until
					j > grid.height
				loop
					sd := grid.item (i, j)
					create p.make_simple (sd.point, sd.point + sd.geometric_normal.scale (0.1))
					vals.force_last (p)
					j := j + 1
				end
				i := i + 1
			end

			create Result.make (vals)

		ensure
			Result_not_void: Result /= Void
		end

	draw_in_screenspace (renderer: RENDERER; bound: AXIS_ALIGNED_BOX) is
			-- Output the bounding box in screen-space
		do
			renderer.push_matrix
			renderer.disable (renderer.rd_depth_test)
			renderer.matrix_mode (renderer.rd_projection)
			renderer.push_matrix
			renderer.load_identity
			renderer.ortho (-0.25, renderer.viewing_camera.x_resolution-0.25,
					   renderer.viewing_camera.y_resolution-0.25, -0.25, -1.0, 1.0)
			renderer.matrix_mode (renderer.rd_modelview)
			renderer.load_identity
			renderer.enable (renderer.rd_blend)

			renderer.disable (renderer.rd_lighting)
			renderer.color_3 (1, 1, 0)
			renderer.render_begin (renderer.rd_line_strip)
			renderer.vertex2 (bound.lower.x, bound.lower.y)
			renderer.vertex2 (bound.lower.x, bound.upper.y)
			renderer.vertex2 (bound.upper.x, bound.upper.y)
			renderer.vertex2 (bound.upper.x, bound.lower.y)
			renderer.vertex2 (bound.lower.x, bound.lower.y)
			renderer.render_end
			renderer.enable (renderer.rd_lighting)

			renderer.disable (renderer.rd_blend)
			renderer.matrix_mode (renderer.rd_projection)
			renderer.pop_matrix
			renderer.matrix_mode (renderer.rd_modelview)
			renderer.enable (renderer.rd_depth_test)
			renderer.pop_matrix

		end

invariant

	shading_rate_positive: shading_rate >= 1

end -- BUCKET_VIEWER

